(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    198999,       6005]*)
(*NotebookOutlinePosition[    200430,       6053]*)
(*  CellTagsIndexPosition[    200284,       6045]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["\<\
The AlpHard syntax using
the Meta package\
\>", "Title",
  FontColor->RGBColor[1, 0, 0],
  CellTags->"firstcell"],

Cell["The Alpha Team", "Subsubtitle"],

Cell["Irisa, Campus de Beaulieu, 35042, Rennes Cedex, France", "SmallText"],

Cell[CellGroupData[{

Cell["Dec. 1998, Revised Dec. 27, 2002. Checked Dec. 2007", "Subsubtitle"],

Cell[CellGroupData[{

Cell["Cr\[EAcute]dits", "Subsubsection",
  FontColor->RGBColor[0, 1, 0]],

Cell["\<\
      Zbigniew Chamski 
      Catherine Dezan 
      Florent Dupont de Dinechin 
      Anne-Claire Guilloux
      Patricia Le Mo\[EDoubleDot]nner 
      Herv\[EAcute] Le Verge 
      Christophe Mauras 
      Patrice Quinton 
      Fabien Quillere
      Sanjay Rajopadhye 
      Sophie Robert 
      Tanguy Risset 
      Doran Wilde \
\>", "SmallText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Open and execute this cell to go in the right directory", \
"Subsubsection",
  InitializationCell->True,
  FontColor->RGBColor[1, 0, 0]],

Cell[BoxData[
    \(\(setMMADir[{"\<doc\>", "\<packages\>", "\<CheckAlpHard\>"}];\)\)], \
"Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["1. Introduction", "Section",
  FontColor->RGBColor[0, 0, 1],
  CellTags->"Introduction"],

Cell[TextData[{
  "This notebook describe ",
  StyleBox["CheckAlpHard", "Input"],
  " packages. Three packages were created using the meta translater: ",
  StyleBox["CheckCell", "Input"],
  ", ",
  StyleBox["CheckModule", "Input"],
  ", and ",
  StyleBox["CheckController", "Input"],
  ". These packages contains fonctions to check if an Alpha program is \
respectively a Cell, a Module, or a Controller of the subset of Alpha called \
AlpHard. "
}], "Text",
  TextAlignment->Left,
  TextJustification->1],

Cell[TextData[{
  "Recall that an ",
  StyleBox["AlpHard", "Input"],
  " is a subset of Alpha which can be translated into Vhdl (this is described \
in Patricia Le Moenner's PhD thesis). However, there exist no special parser \
for AlpHard: we use the parser of Alpha, then we can check that a program is \
an AlpHard program using the functions checkCell, checkModule and \
checkController. "
}], "Text",
  TextAlignment->Left,
  TextJustification->1],

Cell[TextData[{
  "Normally, AlpHard programs are generated automatically, and therefore, \
programs produced this way should be correct. However, there may be some bugs \
in the translator, and moreover, one may want to write AlpHard programs \
directly, or simply, modify AlpHard programs. Then checking the result is \
mandatory: the ",
  Cell[BoxData[
      ButtonBox["\<\"Vhdl\"\>",
        ButtonFunction:>NotebookOpen[ 
            StringJoin[ $demoDirectory, "NOTEBOOKS/Vhdl/Vhdl.nb"]],
        ButtonEvaluator->Automatic,
        Active->True,
        ButtonStyle->"Alpha`Private`Hyperlink"]]],
  " translater does not check that its input is an AlpHard program. "
}], "Text",
  TextAlignment->Left,
  TextJustification->1],

Cell[TextData[{
  "To modify one of these packages, say, CheckCell.m, it is necessary to \
modify the CheckCell.meta file, possibly the CheckCell.sem file, and \
recompile these file using the meta function. See the ",
  ButtonBox["Meta",
    ButtonData:>{
      FrontEnd`FileName[ {
        ParentDirectory[ ], "Meta"}, "Meta.nb", CharacterEncoding -> 
        "MacintoshRoman"], None},
    ButtonStyle->"Hyperlink"],
  " package. "
}], "Text",
  TextAlignment->Left,
  TextJustification->1]
}, Closed]],

Cell[CellGroupData[{

Cell["2. Syntax and semantic of the AlpHard elements", "Section",
  FontColor->RGBColor[0, 0, 1],
  CellTags->"Introduction"],

Cell[TextData[ButtonBox["Back to summary",
  ButtonData:>"Summary",
  ButtonStyle->"Hyperlink"]], "SmallText"],

Cell["\<\
See the notebook of the Meta package for more information on \
it.\
\>", "Text",
  TextJustification->1],

Cell[TextData[{
  "This notebook presents the meta-rule of the 4 AlpHard elements : Cells, \
controller, modules and interfaces. These meta-rules are contained in files \n\
- ",
  StyleBox["CheckCell.meta", "Input"],
  ", \n- ",
  StyleBox["CheckController.meta", "Input"],
  ", \n- ",
  StyleBox["CheckModule.meta", "Input"],
  " and \n- ",
  StyleBox["CheckInterface.meta", "Input"],
  ", respectively."
}], "Text",
  TextJustification->1],

Cell["\<\
First, write a file containing meta-rules, with suffix \".meta\", \
and a file containing the semantic functions, suffixed by \".sem\". Second, \
translate the source file :\
\>", "Text"],

Cell["\<\
There is still some  problems with the context so we need the \
following commands :\
\>", "Text",
  TextJustification->1],

Cell["\<\
Now we can translate the differents elements with the \
meta-traductor by using this commands\
\>", "Text",
  TextJustification->1],

Cell["Finally, load the translated file.", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["3. Examples of cells", "Section",
  FontColor->RGBColor[0, 0, 1],
  CellTags->"How to define"],

Cell[CellGroupData[{

Cell["To compile rules for CheckCell.meta", "Subsubsection"],

Cell["\<\
These commands are necessary only it you do a modification to CheckCell.meta. \
Don't worry about the error messages... Before you try CheckCell in this \
notebook, do not forget to set the directory (re-execute the red cell at the \
beginning of this notebook).\
\>", "Text",
  TextAlignment->Left,
  TextJustification->1],

Cell[CellGroupData[{

Cell[BoxData[
    \(setMMADir[{"\<lib/Packages/Alpha\>"}]\)], "Input"],

Cell[BoxData[
    \("/Users/quinton/mmalpha/lib/Packages/Alpha"\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(<< Meta.m\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(meta["\<CheckCell\>"];\)\)], "Input"],

Cell[BoxData[
    RowBox[{\(Remove::"rmnsm"\), \(\(:\)\(\ \)\), "\<\"There are no symbols \
matching \\\"\\!\\(\\\"Alpha`Work`*\\\"\\)\\\". \\!\\(\\*ButtonBox[\\\"More\
\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"Remove::rmnsm\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(Remove::"rmnsm"\), \(\(:\)\(\ \)\), "\<\"There are no symbols \
matching \\\"\\!\\(\\\"Alpha`Work`Private`*\\\"\\)\\\". \\!\\(\\*ButtonBox[\\\
\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\\\", \
ButtonFrame->None, ButtonData:>\\\"Remove::rmnsm\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(Remove::"rmnsm"\), \(\(:\)\(\ \)\), "\<\"There are no symbols \
matching \\\"\\!\\(\\\"Alpha`CheckCell`Private`*\\\"\\)\\\". \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Remove::rmnsm\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(Remove :: \\\"rmnsm\\\"\\) will be suppressed during this calculation. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(General::"spell1"\), \(\(:\)\(\ \)\), "\<\"Possible spelling \
error: new symbol name \\\"\\!\\(systemName\\)\\\" is similar to existing \
symbol \\\"\\!\\(systemNames\\)\\\". \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\
\", ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"General::spell1\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(General::"spell1"\), \(\(:\)\(\ \)\), "\<\"Possible spelling \
error: new symbol name \\\"\\!\\(varType\\)\\\" is similar to existing symbol \
\\\"\\!\\(varTypes\\)\\\". \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"General::spell1\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(General::"spell1"\), \(\(:\)\(\ \)\), "\<\"Possible spelling \
error: new symbol name \\\"\\!\\(minus\\)\\\" is similar to existing symbol \
\\\"\\!\\(Minus\\)\\\". \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"General::spell1\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(General :: \\\"spell1\\\"\\) will be suppressed during this \
calculation. \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    \("CheckCell.m"\)], "Print"],

Cell[BoxData[
    \("Meta translator, version 0.9 -- October 1998"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("The new package is created in file: "\
\[InvisibleSpace]"CheckCell.m"\),
      SequenceForm[ "The new package is created in file: ", "CheckCell.m"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("List of new functions :\n\
"\[InvisibleSpace]{"CheckCellTranslatelist", 
          "CheckCellTranslateSYSTEMDECLARATION", 
          "CheckCellTranslateDECLARATION", "CheckCellTranslateDOMAINPARAM", 
          "CheckCellTranslateDOMAIN", "CheckCellTranslatePOLYHEDRON", 
          "CheckCellTranslateCONSTRAINT", "CheckCellTranslateGENERATOR", 
          "CheckCellTranslateEQUATION", "CheckCellTranslateASSIGNMENT", 
          "CheckCellTranslateELEMENTS", "CheckCellTranslateAFFEXP", 
          "CheckCellTranslateMATRIX", "CheckCellTranslateMATRIXNUM", 
          "CheckCellTranslateBINEXP", "CheckCellTranslateUNEXP", 
          "CheckCellTranslateVAREXPORCONST", "CheckCellTranslateVAREXPE", 
          "CheckCellTranslateCONSTEXPE", "CheckCellTranslateVAREXP", 
          "CheckCellTranslateCONSTEXP", "CheckCellTranslateSUBEXPRESSION", 
          "CheckCellTranslateCASEMUX", "CheckCellTranslateRESTEXP", 
          "CheckCellTranslateIFEXP", "CheckCellTranslateCONTROLEXPRESSION", 
          "CheckCellTranslateIFMUX", "CheckCellTranslateSUBMUX2", 
          "CheckCellTranslateCASEXP", "CheckCellTranslateRESTEXPMUX2", 
          "CheckCellTranslateUSESTATEMENT", "CheckCellTranslateCALLEXP"}\),
      SequenceForm[ 
      "List of new functions :\n", {"CheckCellTranslatelist", 
        "CheckCellTranslateSYSTEMDECLARATION", 
        "CheckCellTranslateDECLARATION", "CheckCellTranslateDOMAINPARAM", 
        "CheckCellTranslateDOMAIN", "CheckCellTranslatePOLYHEDRON", 
        "CheckCellTranslateCONSTRAINT", "CheckCellTranslateGENERATOR", 
        "CheckCellTranslateEQUATION", "CheckCellTranslateASSIGNMENT", 
        "CheckCellTranslateELEMENTS", "CheckCellTranslateAFFEXP", 
        "CheckCellTranslateMATRIX", "CheckCellTranslateMATRIXNUM", 
        "CheckCellTranslateBINEXP", "CheckCellTranslateUNEXP", 
        "CheckCellTranslateVAREXPORCONST", "CheckCellTranslateVAREXPE", 
        "CheckCellTranslateCONSTEXPE", "CheckCellTranslateVAREXP", 
        "CheckCellTranslateCONSTEXP", "CheckCellTranslateSUBEXPRESSION", 
        "CheckCellTranslateCASEMUX", "CheckCellTranslateRESTEXP", 
        "CheckCellTranslateIFEXP", "CheckCellTranslateCONTROLEXPRESSION", 
        "CheckCellTranslateIFMUX", "CheckCellTranslateSUBMUX2", 
        "CheckCellTranslateCASEXP", "CheckCellTranslateRESTEXPMUX2", 
        "CheckCellTranslateUSESTATEMENT", "CheckCellTranslateCALLEXP"}],
      Editable->False]], "Print"]
}, Open  ]],

Cell["Then load the result. ", "Text"],

Cell[BoxData[
    \(<< CheckCell.m\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example 1", "Subsubsection"],

Cell["This first example contains various assignments. ", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(load["\<cell1.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \("[cell1]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \("system cell1 : {p | 2<=p<=10}\n               (A : {t | 1<=t} of \
integer)\n       returns (S : {t | 2<=t} of integer);\nvar\n  a : {t | 1<=t} \
of integer;\n  b : {t | 1<=t} of integer;\n  e : {t | 1<=t} of integer;\n  c \
: {t | 2<=t} of integer;\n  d : {t | 2<=t} of integer;\nlet\n  a[t] = 0[] + \
A[t];\n  b[t] = a[t];\n  c[t] = b[t-1];\n  d[t] = a[t-1] + b[t-1] - c[t];\n  \
e[t] = -a[t];\n  S[t] = c[t];\ntel;"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(schematics[]\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .48571 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0 0.285714 0 0.285714 [
[ 0 0 0 0 ]
[ 1 .48571 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .48571 L
0 .48571 L
closepath
clip
newpath
0 g
gsave
.29286 .14286 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.5 Mabswid
[ ] 0 setdash
.29286 .14286 m
.42857 .14286 L
s
gsave
.42571 .14857 -71 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
gsave
.48571 .14286 -70.5625 -15 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 30 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
%%IncludeResource: font Mathematica4Mono
%%IncludeFont: Mathematica4Mono
/Mathematica4Mono findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
63.000 20.000 moveto
(#) show
78.125 20.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.42857 .14286 m
.57143 .14286 L
s
gsave
.54286 .15429 -61 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(a1) show
75.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.54286 .14286 m
.57143 .14286 L
s
1 0 0 r
.35 .2 m
.35 .0574 L
.57143 .0574 L
.57143 .08571 L
s
gsave
.57143 .08571 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.35 .2 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.35 .2 m
.42857 .2 L
s
gsave
.42571 .20571 -71 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(a) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.42857 .2 m
.57143 .2 L
s
gsave
.54286 .21143 -61 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(b) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.54286 .2 m
.57143 .2 L
s
1 0 0 r
.62857 .25714 m
.62857 .22961 L
.57143 .22961 L
.57143 .25714 L
s
gsave
.57143 .25714 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.62857 .25714 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.62857 .25714 m
.71429 .25714 L
s
gsave
.71143 .26286 -71 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(c) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.71429 .25714 m
.85714 .25714 L
s
gsave
.82857 .26857 -61 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(S) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.82857 .25714 m
.85714 .25714 L
s
0 1 0 r
.84571 .25143 m
.85143 .25143 L
.85714 .25714 L
.85143 .26286 L
.84571 .26286 L
.84571 .25714 L
F
0 g
.42857 .06286 m
.54286 .06286 L
.54286 .10857 L
.42857 .10857 L
.42857 .06286 L
s
1 0 0 r
.3 .07429 m
.3 .11481 L
.57143 .11481 L
.57143 .14286 L
s
gsave
.57143 .14286 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.3 .07429 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.3 .07429 m
.42857 .07429 L
s
gsave
.42571 .08 -77 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(a1) show
75.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.4 .09714 m
.42857 .09714 L
s
0 0 1 r
.4 .09143 m
.40571 .09143 L
.41143 .09714 L
.40571 .10286 L
.4 .10286 L
.4 .09714 L
F
0 g
gsave
.42571 .10286 -71.0625 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(A) show
69.062 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
1 0 0 r
gsave
.48571 .08571 -69 -15 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 30 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 20.000 moveto
%%IncludeResource: font Mathematica1Mono
%%IncludeFont: Mathematica1Mono
/Mathematica1Mono findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(+) show
75.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.54286 .09714 -61 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(a) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.54286 .08571 m
.57143 .08571 L
s
.42857 .23429 m
.54286 .23429 L
.54286 .28 L
.42857 .28 L
.42857 .23429 L
s
1 0 0 r
.30714 .25714 m
.30714 .17221 L
.57143 .17221 L
.57143 .2 L
s
gsave
.57143 .2 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.30714 .25714 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.30714 .25714 m
.42857 .25714 L
s
gsave
.42571 .26286 -71 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(b) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
1 0 0 r
gsave
.48571 .25714 -81 -15 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 30 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(reg) show
99.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.54286 .26857 -61 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(c) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.54286 .25714 m
.57143 .25714 L
s
.42857 .29143 m
.54286 .29143 L
.54286 .33714 L
.42857 .33714 L
.42857 .29143 L
s
1 0 0 r
.32143 .30286 m
.32143 .34442 L
.57143 .34442 L
.57143 .37143 L
s
gsave
.57143 .37143 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.32143 .30286 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.32143 .30286 m
.42857 .30286 L
s
gsave
.42571 .30857 -77 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(d2) show
75.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
1 0 0 r
.31429 .32571 m
.31429 .22961 L
.57143 .22961 L
.57143 .25714 L
s
gsave
.57143 .25714 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.31429 .32571 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.31429 .32571 m
.42857 .32571 L
s
gsave
.42571 .33143 -71 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(c) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
1 0 0 r
gsave
.48571 .31429 -69 -15 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 30 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 20.000 moveto
%%IncludeResource: font Mathematica1Mono
%%IncludeFont: Mathematica1Mono
/Mathematica1Mono findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(-) show
75.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.54286 .32571 -61 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(d) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.54286 .31429 m
.57143 .31429 L
s
.42857 .34857 m
.54286 .34857 L
.54286 .39429 L
.42857 .39429 L
.42857 .34857 L
s
1 0 0 r
.32857 .36 m
.32857 .05896 L
.85714 .05896 L
.85714 .08571 L
s
gsave
.85714 .08571 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.32857 .36 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.32857 .36 m
.42857 .36 L
s
gsave
.42571 .36571 -83 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(d23) show
81.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
1 0 0 r
.33571 .38286 m
.33571 .11636 L
.85714 .11636 L
.85714 .14286 L
s
gsave
.85714 .14286 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.33571 .38286 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.33571 .38286 m
.42857 .38286 L
s
gsave
.42571 .38857 -83 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(d24) show
81.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
1 0 0 r
gsave
.48571 .37143 -69 -15 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 30 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 20.000 moveto
%%IncludeResource: font Mathematica1Mono
%%IncludeFont: Mathematica1Mono
/Mathematica1Mono findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(+) show
75.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.54286 .38286 -61 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(d2) show
75.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.54286 .37143 m
.57143 .37143 L
s
.71429 .06286 m
.82857 .06286 L
.82857 .10857 L
.71429 .10857 L
.71429 .06286 L
s
1 0 0 r
.58571 .08571 m
.58571 .0574 L
.57143 .0574 L
.57143 .08571 L
s
gsave
.57143 .08571 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.58571 .08571 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.58571 .08571 m
.71429 .08571 L
s
gsave
.71143 .09143 -71 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(a) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
1 0 0 r
gsave
.77143 .08571 -81 -15 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 30 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(reg) show
99.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.82857 .09714 -61 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(d23) show
81.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.82857 .08571 m
.85714 .08571 L
s
.71429 .12 m
.82857 .12 L
.82857 .16571 L
.71429 .16571 L
.71429 .12 L
s
1 0 0 r
.6 .14286 m
.6 .17221 L
.57143 .17221 L
.57143 .2 L
s
gsave
.57143 .2 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.6 .14286 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.6 .14286 m
.71429 .14286 L
s
gsave
.71143 .14857 -71 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(b) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
1 0 0 r
gsave
.77143 .14286 -81 -15 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 30 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(reg) show
99.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.82857 .15429 -61 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(d24) show
81.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.82857 .14286 m
.85714 .14286 L
s
.71429 .17714 m
.82857 .17714 L
.82857 .22286 L
.71429 .22286 L
.71429 .17714 L
s
1 0 0 r
.58571 .2 m
.58571 .0574 L
.57143 .0574 L
.57143 .08571 L
s
gsave
.57143 .08571 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.58571 .2 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.58571 .2 m
.71429 .2 L
s
gsave
.71143 .20571 -71 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(a) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
1 0 0 r
gsave
.77143 .2 -69 -15 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 30 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 20.000 moveto
%%IncludeResource: font Mathematica1Mono
%%IncludeFont: Mathematica1Mono
/Mathematica1Mono findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(-) show
75.000 20.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 20.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.82857 .21143 -61 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(e) show
69.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.82857 .2 m
.85714 .2 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{387.312, 188.062},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40009400016R000`400?l00000o`00003ooooooooo
ooooodKooooo003oooooooooooooA_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo
003oooooooooooooA_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo003ooooooooo
ooooA_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo003oooooooooooooA_ooool0
0?oooooooooooom6ooooo`00ooooooooooooodKooooo003oooooooooooooA_ooool00?oooooooooo
oom6ooooo`00ooooooooooooodKooooo003oooooooooooooA_ooool00?oooooooooooom6ooooo`00
ooooooooooooodKooooo003oooooooooooooA_ooool00?oooooooooooom6ooooo`00oooooooooooo
odKooooo003oooooooooooooA_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo003o
ooooooooooooA_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo003ooooooooooooo
A_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo003oooooooooooooA_ooool00?oo
oooooooooom6ooooo`00ooooooooooooodKooooo003;ooooohWoo`00l?ooool00;kooooooooo000d
ool005?ooooo002nooooo`03ool00?oooooooooo00[ooooo00?oo`00ooooooooool0OOooool00ooo
003oooooooooo`05ooooo`03ool00?oooooooooo09[ooooo00?oo`00ooooooooool0DOooool00;ko
oooo00?oo`00ooooooooool02_ooool00ooo003oooooooooo`0Zoooood?o00004?ooool00ooo003o
ooooooooo`05ooooo`03ool00?oooooooooo04OoooooA?l0000?ooooo`03ool00?oooooooooo057o
oooo002nooooo`03ool00?oooooooooo00[ooooo00?oo`00ooooooooool0:_ooool00ol0003ooooo
ooooo`0oooooo`03o`000?oooooooooo00kooooo00?oo`00ooooooooool01Oooool00ooo003ooooo
ooooo`17ooooo`03o`000?oooooooooo02gooooo00?oofIVool00?oo00000_oo00000ooo<S;oohN7
ooo^kP0;ooooo`03o`000?oooooooooo00gooooo00?oo`00ooooooooool0DOooool00;kooooo00?o
o`00ooooooooool02_ooool00ooo003oooooooooo`0Zooooo`03o`000?oooooooooo03oooooo00?o
0000ooooooooool03_ooool00ooo003oooooooooo`05ooooo`03ool00?oooooooooo04Oooooo00?o
0000ooooooooool0;Oooool00ooo<S;oo`00ool00005ool00003oonYZOoooooooooo00Sooooo00?o
0000ooooooooool03Oooool00ooo003oooooooooo`1Aooooo`00__ooool00ooo003oooooooooo`0:
ooooo`03ool00?oooooooooo02[ooooo00?o0000ooooooooool0?oooool00ol0003oooooooooo`0>
ooooo`03ool00?oooooooooo00Gooooo00?oo`00ooooooooool0Aoooool00ol0003oooooooooo`0b
ooooo`05ooo<c?ooA4Coo`00ool00?oogMd02Oooool00ol0003oooooooooo`0=ooooo`03ool00?oo
oooooooo057ooooo002[ooooo`06onk^k_mEEEGo0000o`000?mVIVKok^k^3Oooool00ooo003ooooo
ooooo`0:ooooo`03ool00?oooooooooo02[ooooo00?o0000ooooooooool0?oooool00ol0003ooooo
ooooo`0>ooooo`03ool00?oooooooooo00Gooooo00?oo`00ooooooooool0Aoooool00ol0003ooooo
ooooo`0dooooo`03oolb<_oo003ooeEE00Wooooo00?o0000ooooooooool03Oooool00ooo003ooooo
ooooo`1Aooooo`00Zoooool00omEEEGo0000o`000002o`000003ofIVI_oooooooooo00_ooooo00?o
o`00ooooooooool02_ooool00ooo003oooooooooo`0Zooooo`03o`000?oooooooooo03oooooo00?o
0000ooooooooool03_ooool00ooo003oooooooooo`05ooooo`03ool00?oooooooooo04Oooooo00?o
0000ooooooooool0=?ooool00oooZJWoo`00ool@4009ooooo`03o`000?oooooooooo00gooooo00?o
o`00ooooooooool0DOooool00:_ooooo4ol000000ooo003o0000o`00000ho`00047ooooo00?o0000
ooooooooool03_ooool00ooo003oooooooooo`05ooooo`03ool00?oooooooooo04Oooooo00?o0000
ooooooooool0=?ooool00oook^koo`00ool00009ooooo`03o`000?oooooooooo00gooooo00?oo`00
ooooooooool0DOooool00:_ooooo1_l0000=ooooo`03ool00?oooooooooo00[ooooo00?oo`00oooo
ooooool05?ooool01_l0003o^[ZjoooooooMgMgoMgMgo`0000Kooooo00?o0000ooooooooool01ooo
ool00ol0003oooooooooo`0Mooooo`;oo`008?ooool00ol0003oooooooooo`0>ooooo`03ool00?oo
oooooooo00Gooooo00?oo`00ooooooooool0Aoooool00ol0003oooooooooo`05ooooo`03oolb<_oo
003oo`0000Soo`0000?oob4Qooooooooool01oooool00ooogMgoofIVoolQ8@03ool00004ool@4?oo
<S;oogMgoonj^Q7ooooo0_oo0009ooooo`03o`000?oooooooooo00gooooo00?oo`00ooooooooool0
DOooool00:_ooooo00?oEEEEo`000?l000000_l000000omVIVKoooooooooo`0;ooooo`03ool00?oo
oooooooo00[ooooo00?oo`00ooooooooool05?ooool01_l@413oQhN7onk^k_oooooogMgMo`0000Ko
oooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0Mooooo`;oo`008?ooool00ol0003o
ooooooooo`0>ooooo`03ool00?oooooooooo00Gooooo00?oo`00ooooooooool0Aoooool00ol0003o
ooooooooo`05ooooo`03oolb<_oo003oo`0000Soo`0000?oob4Qooooooooool01_ooool00oooMgOo
o`00ool00009ool00003oon7Qooooooooooo00Gooooo00?ooiRHoom4A?oo00000_oo00000oooA4Co
ojVYooooo`02ool000Wooooo00?o0000ooooooooool03Oooool00ooo003oooooooooo`1Aooooo`00
Zoooool01_o^k^koEEEEo`000?l0003oIVIVonk^kPgooooo00?oo`00ooooooooool02_ooool00ooo
003oooooooooo`0Dooooo`06olc<c?m4A4Co0000o`000?l@413o00001_ooool00ol0003ooooooooo
o`07ooooo`03o`000?oooooooooo01gooooo0_oo000Pooooo`03o`000?oooooooooo00kooooo00?o
o`00ooooooooool01Oooool00ooo003oooooooooo`17ooooo`03o`000?oooooooooo00Wooooo0_oo
000=ooooo`05oomVI_oo003oo`00oomgMoooc<`00oooool01_ooc<coojVYoomVI_oo<S;oo`00oolQ
8@Gooooo00?oonk^oolb<_oo00001_oo00000oooEEGoo`00ool00009ooooo`03o`000?oooooooooo
00gooooo00?oo`00ooooooooool0DOooool00:kooooo00?oo`00ooooooooool03Oooool00ooo003o
ooooooooo`0:ooooo`03ool00?oooooooooo01Wooooo00?o0000ooooooooool01?ooool00ol0003o
ooooooooo`07ooooo`03o`000?oooooooooo01gooooo0_oo000Pooooo`03o`000?oooooooooo00ko
oooo00?oo`00ooooooooool01Oooool00ooo003oooooooooo`17ooooo`03o`000?oooooooooo00Wo
oooo0_oo000<ooooo`04oonj^_oo003ooa0@ooo^kQ3ooooo00Sooc8bool00?oo003ooiRHooo^k_oo
ooooonk^oonHV0Coo`002Oooool00ol0003oooooooooo`0=ooooo`03ool00?oooooooooo057ooooo
002^ooooo`03ool00?oooooooooo00gooooo00?oo`00ooooooooool02_ooool00ooo003ooooooooo
o`0Hooooo`03ohN7Qolb<S;ooooo00Gooooo00?o0000ooooooooool01oooool00ol0003ooooooooo
o`0Mooooo`;oo`008?ooool00ol0003oooooooooo`0;ooooo`06ooo^k_ooEEGoo`00ool00?ooIVKo
onk^0oooool01_o^k^koEEEEool00?l0003oIVIVonk^kTKooooo00?o0000ooooooooool02Oooool2
ool000cooooo00?ooeEEool00?ooZJT04?ooool01?ooV9Soo`00ool00?ooc<`5ooooo`04ooo<c?oo
003oo`00ool000Wooooo00?o0000ooooooooool02oooool01_ook^kooeEEool00?oo003oofIVooo^
kU3ooooo002^ooooo`03ool00?oooooooooo00gooooo00?oo`00ooooooooool02_ooool00ooo003o
ooooooooo`0Eooooo`05odA4A?l0003o0000oa0@4?o<c<`01_ooool00ol0003oooooooooo`07oooo
o`03o`000?oooooooooo01Sooooo3?oo000Kooooo`03o`000?oooooooooo00_ooooo00?ooeEEool0
0?oo00000_oo00001oooIVKoooooooooooooooooEEEEo`000?oo00000_l000000omVIVKooooooooo
o`14ooooo`03o`000?oooooooooo00Wooooo0_oo000<ooooo`03ool@4?oo003ooooo013ooooo00?o
odA4ool00?ooV9P01oooool00oooV9Soo`00ool00009ooooo`03o`000?oooooooooo00_ooooo00?o
oeEEool00?oo00000_oo00000oooIVKoooooooooo`1>ooooo`00[_ooool00ooo003oooooooooo`0=
ooooo`03ool00?oooooooooo00[ooooo00?oo`00ooooooooool07_ooool00ol0003oA4A4o`000009
ooooo`03o`000?oooooooooo01Sooooo3?oo000Kooooo`ko00001_oo0003ooooo`;o000000?oo`00
o`000?l00000B?l0000;ooooo`;oo`003?ooool>ool00003oolQ8Ooooooooooo00;ooooo0_oo0000
0oook^koooooooooo`05ooooo`03ooo^k_oo003oo`0000Wooooo4Ol00003ool0053ooooo002^oooo
o`03ool00?oooooooooo00gooooo00?oo`00ooooooooool02_ooool00ooo003oooooooooo`0Noooo
o`03olc<c?mEEEGo000000Wooooo00?o0000ooooooooool07Oooool2ool0023ooooo00?o0000oooo
ooooool02oooool6ool000?ooooo0_l000001?oo003o0000o`000?l0000jooooo`06o`000?nj^[[o
ooooomgMgOmgMgOo00001_ooool00ol0003oooooooooo`09ooooo`;oo`003?ooool?ool000Cooooo
0_oo0009ooooo`;oo`002Oooool00ol0003oooooooooo`0;ooooo`Koo`00D?ooool00:kooooo00?o
o`00ooooooooool03Oooool00ooo003oooooooooo`0:ooooo`03ool00?oooooooooo02[ooooo00?o
0000ooooooooool07Oooool2ool0023ooooo00?o0000ooooooooool02oooool00oooEEGoo`00ool0
0002ool00007oomVI_oooooooooooooooomEEEGo0000ool00002o`000003ofIVI_oooooooooo03So
oooo00Ko410@ohN7Qoo^k^koooooomgMgOl00006ooooo`03o`000?oooooooooo00Wooooo0_oo0000
0oooc<coooooooooo`09ooooo`03oolQ8Ooo003oomgM00Wooooo00?oomgMool00?oo00001?ooool2
ool00003ooo^k_oooooooooo00Gooooo00?oonk^ool00?oo00002Oooool00ol0003oooooooooo`0;
ooooo`03oomEEOoo003oo`0000;oo`0000?oofIVooooooooool0C_ooool00:kooooo00?oo`00oooo
ooooool03Oooool00ooo003oooooooooo`0:ooooo`03ool00?oooooooooo02[ooooo00?o0000oooo
ooooool07Oooool2ool0023ooooo00_o0000ooooooooooooooooojVYZOl@413o0000ob4Q8OnHV9So
410@o`000003ooooo`06ooo^k_ooEEGoo`00ool00?ooIVKoonk^0oooool01_o^k^koEEEEool00?l0
003oIVIVonk^kS[ooooo00Koc<c<odA4A?l0003o0000oa0@4?l00006ooooo`03o`000?oooooooooo
00Wooooo0_oo00000ooo413oonk^ooooo`09ooooo`03oomVI_oo003oodA400Wooooo00?oogMgool0
0?ooEED01?ooool00oooA4Coo`00oonHV007ooooo`03oonHV?oo003oo`0000Wooooo00?o0000oooo
ooooool00_ooool02?n7QhOo410@o`000?lQ8B7oZJVYo`000?l@413ooooo1_l000001Oook^kooooo
oooooonYZJWo8B4Q00;o000000?oEEEEonk^k_ooool0Aoooool00:kooooo00?oo`00ooooooooool0
3Oooool00ooo003oooooooooo`0:ooooo`03ool00?oooooooooo01[ooooo0ol00?l=ooooo`03o`00
0?oooooooooo01gooooo0_oo000Pooooo`0:o`000?ooooooooooooooool0003o^[ZjoooooooMgMgo
MgMgo`0000oooooo00?oo`00ooooooooool0@?ooool00ol0003oooooooooo`04ooooo`03o`000?oo
oooooooo00Wooooo0ooo00000ooo8B7oolc<ooooo`08ooooo`04oooMgOoo003oo`00oon7Q`Oooooo
00CookZjool00?oo003ookZj1?ooool01?oo^[[oo`00ool00?ooc<`5ooooo`04ooo<c?oo003oo`00
ool000Wooooo00_o0000ooooooooooooooooohN7QolQ8B7oc<c<ooooooo<c<co8B4Qo`000002oooo
o`03oiRHV?l0003oc<c<00;ooooo00Ko410@oooooooooooooooooa0@4?nYZJT2ooooo`03ogMgMolb
<S;ooooo04Oooooo002^ooooo`03ool00?oooooooooo00gooooo00?oo`00ooooooooool02_ooool0
0ooo003oooooooooo`0Jooooo`Co003o3?ooool00ol0003oooooooooo`0Mooooo`;oo`008?ooool0
2_l0003ooooooooooooooooo410@ohN7Qoo^k^koooooomgMgOl0000?ooooo`03ool00?oooooooooo
03oooooo00?oQhN7oc8b<_ooool01Oooool00ol0003oooooooooo`09ooooo`;oo`0000CooiRHool0
0?oo003oogMg0_ook^h00oooMgOoo`00oomEE@04ooooo`06oonHV?oo003oo`00oolb<_ooV9Soonk^
0_ooool01Ooo^[[ooeEEool00?oo003oogMg00Kooooo00SooeEEool00?oo003ooiRHooo^k_oooooo
onk^oonHV0Coo`002Oooool01_l0003ooooooooooooooooo410@olc<c0?ooooo00?o^[Zjo`000?oo
ool00_ooool00onHV9So0000olc<c00:ooooo`03o`000?oooooooooo04Kooooo002^ooooo`03ool0
0?oooooooooo00gooooo00?oo`00ooooooooool02_ooool00ooo003oooooooooo`0Jooooo`;o003o
00Ko000@o`000?l0003o410@ooooool@4102o`000003oa0@4?oooooooooo00?ooooo00?o0000oooo
ooooool0?oooool01_l0003oooooooooooooooooc<c<odA4A0;o000000?o410@o`000?ooool03_oo
ool00ooo003oooooooooo`0looooo`05odA4A?l0003o0000oa0@4?o<c<`01_ooool00ol0003ooooo
ooooo`06ooooo`03oolQ8Ooo003oo`0000;oo`0000?ooooooonYZOoo00001_oo0005ooooo`03oonY
ZOoo003oo`0000Ooo`0000?oohN7ooooooooool01Oooool00oook^kooeEEool00006ool00005oolb
<_oo003oo`00ool00?oo8B401oooool01Ol0003ooooooooooooooooo000000Gooooo00?o0000oooo
ooooool00_ooool00onYZJWo410@olc<c008ooooo`03ogMgMolQ8B7ooooo04Oooooo002^ooooo`03
ool00?oooooooooo00gooooo00?oo`00ooooooooool02_ooool00ooo003oooooooooo`0Jooooo`?o
003o00?o0000o`00gOl00?l02ol00011ooooo`03o`000?oooooooooo00Kooooo00?o0000oooooooo
ool03Oooool00ooo003oooooooooo`17ooooo`03o`000?oooooooooo00Kooooo00?oob4Qool00?oo
00000_oo0002ooooo`03ooo^k_ooMgOoob4Q00;oo`0000?oob4QoonYZOooool01Oooool00oook^ko
ohN7oolb<P03ool00003oolQ8OooIVKoomgM00[ooooo00?ookZjoom4A?oo00000_oo00000oooA4Co
okZjooooo`03ool00003oolQ8Ooooooooooo00Gooooo00Ko0000oooooooooooooooooa0@4?o<c<`3
ooooo`03olc<c?l0003ooooo00Cooooo00?oc<c<oa0@4?o<c<`01Oooool01?lQ8B7o0000o`000?nj
^[Y8ooooo`00[_ooool00ooo003oooooooooo`0=ooooo`03ool00?oooooooooo00[ooooo00?oo`00
ooooooooool06_ooool3o`00o`03o`00EOl008Oo003o00;ooooo00?oQhN7oeEEEOooool01Oooool0
0ol0003oooooooooo`0oooooo`03o`000?oooooooooo00Gooooo00?oQhN7oc8b<_ooool03_ooool0
0ooo003oooooooooo`17ooooo`03o`000?oooooooooo043ooooo00_o0000ooooooooooooooooojVY
ZOlQ8B7oc<c<ooooooo<c<co8B4Qo`000006ooooo`03ojVYZOlb<S;ooooo00Gooooo00?ok^k^ogMg
Molb<S80B?ooool00:kooooo00?oo`00ooooooooool03Oooool00ooo003oooooooooo`0:ooooo`03
ool00?oooooooooo01[ooooo0ol00?l00ol00<co0000o`000003o`000003okZj^_oooooooooo00Co
oooo00?o0000ooooooooool0?oooool00ol0003oooooooooo`02ooooo`05odA4A?l0003o0000oa0@
4?o<c<`03oooool00ooo003oooooooooo`17ooooo`03o`000?oooooooooo043ooooo00?o0000oooo
ooooool00_ooool01_nYZJWo410@o`000?l@413oQhN7o`0000Oooooo00?o0000ooooooooool01_oo
ool00ol0003oooooooooo`16ooooo`00[_ooool00ooo003oooooooooo`0=ooooo`03ool00?oooooo
oooo00[ooooo00?oo`00ooooooooool06_ooool4o`00o`05oa0@4?o<c<coooooolc<c?l@41001ooo
ool00ol0003oooooooooo`0oooooo`03o`000?oooooooooo01Kooooo00?oo`00ooooooooool0Aooo
ool00ol0003oooooooooo`10ooooo`03o`000?oooooooooo00Oooooo00Wo0000oooooooooooo0000
ohN7Qoo^k^koooooogMgMolb<S800oooool01_lQ8B7oQhN7ooooooooooooIVIVoc8b<TSooooo002^
ooooo`03ool00?oooooooooo00gooooo00?oo`00ooooooooool02_ooool00ooo003oooooooooo`0N
ooooo`;oIVIV00?oooooofIVI_mgMgL01oooool00ol0003oooooooooo`0oooooo`03o`000?oooooo
oooo01Kooooo00?oo`00ooooooooool0Aoooool00ol0003oooooooooo`10ooooo`03o`000?oooooo
oooo00Oooooo00Go0000ooooooooooooZJVYoc8b<P02o`000003oeEEEOo^k^kooooo00;ooooo00Ko
^[ZjodA4A?l0003o0000odA4A?o^k^i8ooooo`00[_ooool00ooo003oooooooooo`0=ooooo`03ool0
0?oooooooooo00[ooooo00?oo`00ooooooooool07_ooool01Oo<c<co410@ooooool0003ogMgM00Oo
oooo00?o0000ooooooooool0?oooool00ol0003oooooooooo`0Fooooo`03ool00?oooooooooo04Oo
oooo00?o0000ooooooooool0@?ooool00ol0003oooooooooo`06ooooo`;o0000FOooool00:kooooo
00?oo`00ooooooooool03Oooool00ooo003oooooooooo`0:ooooo`03ool00?oooooooooo01oooooo
00?o410@ofIVI_lb<S802?oooom3o`0001Sooooo00?oo`00ooooooooool0Aooooom4o`0006?ooooo
002^ooooo`03ool00?oooooooooo00gooooo00?oo`00ooooooooool02_ooool00ooo003ooooooooo
o`0Nooooo`?o000000?oZJVYooooooooool0HOooool00ooo003oooooooooo`3^ooooo`00[_ooool0
0ooo003oooooooooo`0=ooooo`03ool00?oooooooooo00[ooooo00?oo`00ooooooooool0QOooool0
0ooo003oooooooooo`3^ooooo`00[_ooool00ooo003oooooooooo`0=ooooo`03ool00?oooooooooo
00[ooooo00?oo`00ooooooooool0QOooool00ooo003oooooooooo`3^ooooo`00[_oooonNool000Oo
oooo00?oo`00ooooooooool0k_ooool00;kooooo00Coo`00oooooooooooooooooooo000`ool005?o
oooo002nooooo`05ool00?oooooooooooooooooo00002?ooool00ooo003oooooooooo`1mooooo`03
ool00?oooooooooo00Gooooo00?oo`00ooooooooool0V_ooool00ooo003oooooooooo`1Aooooo`00
__ooool01Ooo003oooooooooooooooooo`0000Sooooo00?oo`00ooooooooool0OOooool00ooo003o
ooooooooo`05ooooo`03ool00?oooooooooo04OoooooA?l0000?ooooo`03ool00?oooooooooo057o
oooo002nooooo`05ool00?oooooooooooooooooo00002?ooool00ooo003oooooooooo`1mooooo`03
ool00?oooooooooo00Gooooo00?oo`00ooooooooool0Aoooool00ol0003oooooooooo`0]ooooo`03
oomVI_oo003oo`0000;oo`0000?ooc8boon7Qoook^h02oooool00ol0003oooooooooo`0=ooooo`03
ool00?oooooooooo057ooooo002nooooo`05ool00?oooooooooooooooooo00002?ooool00ooo003o
ooooooooo`1mooooo`03ool00?oooooooooo00Gooooo00?oo`00ooooooooool0Aoooool00ol0003o
ooooooooo`0]ooooo`03oolb<_oo003oo`0000Goo`0000?oojVYooooooooool02?ooool00ol0003o
ooooooooo`0=ooooo`03ool00?oooooooooo057ooooo002nooooo`05ool00?oooooooooooooooooo
00002?ooool00ooo003oooooooooo`1mooooo`03ool00?oooooooooo00Gooooo00?oo`00oooooooo
ool0Aoooool00ol0003oooooooooo`0booooo`05ooo<c?ooA4Coo`00ool00?oogMd02Oooool00ol0
003oooooooooo`0=ooooo`03ool00?oooooooooo057ooooo002nooooo`05ool00?oooooooooooooo
oooo00002?ooool00ooo003oooooooooo`1mooooo`03ool00?oooooooooo00Gooooo00?oo`00oooo
ooooool0Aoooool00ol0003oooooooooo`0dooooo`03oolb<_oo003ooeEE00Wooooo00?o0000oooo
ooooool03Oooool00ooo003oooooooooo`1Aooooo`00__ooool01Ooo003oooooooooooooooooo`00
00Sooooo00?oo`00ooooooooool0OOooool00ooo003oooooooooo`05ooooo`03ool00?oooooooooo
04Oooooo00?o0000ooooooooool0=?ooool00oooZJWoo`00ool@4009ooooo`03o`000?oooooooooo
00gooooo00?oo`00ooooooooool0DOooool00;kooooo00Goo`00ooooooooooooooooool00008oooo
o`03ool00?oooooooooo04kooooo00?oV9RHob4Q8Oo<c<`0;?ooool00ooo003oooooooooo`05oooo
o`03ool00?oooooooooo04Oooooo00?o0000ooooooooool0=?ooool00oook^koo`00ool00009oooo
o`03o`000?oooooooooo00gooooo00?oo`00ooooooooool0DOooool00;kooooo00Goo`00oooooooo
ooooooooool00008ooooo`03ool00?oooooooooo047ooooo0_l0000;ooooo`04oeEEEOl0003o0000
onk^kR_ooooo00?oo`00ooooooooool01Oooool00ooo003oooooooooo`17ooooo`03o`000?oooooo
oooo00Gooooo00?ooc8bool00?oo00002?oo00000ooo8B7oooooooooo`07ooooo`03oooMgOooIVKo
ob4Q00?oo`0000Cooa0@oolb<_ooMgOookZj4Oooool2ool000Wooooo00?o0000ooooooooool03Ooo
ool00ooo003oooooooooo`1Aooooo`00__ooool01Ooo003oooooooooooooooooo`0000Sooooo00?o
o`00ooooooooool0@Oooool2o`0000cooooo00CoEEEEo`000?lQ8B7ok^k^:_ooool00ooo003ooooo
ooooo`05ooooo`03ool00?oooooooooo04Oooooo00?o0000ooooooooool01Oooool00ooo<S;oo`00
ool00008ool00003oolQ8Ooooooooooo00Kooooo00?oogMgool00?oo00002Ooo00000oooQhOooooo
ooooo`05ooooo`03oonHV?ooA4Coo`0000;oo`0000?oodA4oonYZOooool00_oo0009ooooo`03o`00
0?oooooooooo00gooooo00?oo`00ooooooooool0DOooool00;kooooo00Goo`00oooooooooooooooo
ool00008ooooo`03ool00?oooooooooo047ooooo0_l0000=ooooo`04oeEEEOl0003o8B4Qonk^kRWo
oooo00?oo`00ooooooooool01Oooool00ooo003oooooooooo`17ooooo`03o`000?oooooooooo00Wo
oooo0_oo000=ooooo`05oomVI_oo003oo`00oomgMoooc<`00oooool01_ooc<coojVYoomVI_oo<S;o
o`00oolQ8@Gooooo00?oonk^oolb<_oo00001_oo00000oooEEGoo`00ool00009ooooo`03o`000?oo
oooooooo00gooooo00?oo`00ooooooooool0DOooool00;kooooo00Goo`00ooooooooooooooooool0
0008ooooo`03ool00?oooooooooo047ooooo0_l0000>ooooo`03oeEEEOl0003oEEEE02Wooooo00?o
o`00ooooooooool01Oooool00ooo003oooooooooo`17ooooo`03o`000?oooooooooo00Wooooo0_oo
000<ooooo`04oonj^_oo003ooa0@ooo^kQ3ooooo00Sooc8bool00?oo003ooiRHooo^k_ooooooonk^
oonHV0Coo`002Oooool00ol0003oooooooooo`0=ooooo`03ool00?oooooooooo057ooooo002Woooo
o`06onk^k_mEEEGo0000o`000?mVIVKok^k^4Oooool01Ooo003oooooooooooooooooo`0000Sooooo
00?oo`00ooooooooool0@Oooool2o`0000oooooo00?oEEEEo`000?mEEED09Oooool01_ook^kooeEE
ool00?oo003oofIVooo^kPGooooo00?oo`00ooooooooool00oooool01_o^k^koEEEEo`000?l0003o
IVIVonk^kSkooooo00?o0000ooooooooool02Oooool2ool000cooooo00?ooeEEool00?ooZJT04?oo
ool01?ooV9Soo`00ool00?ooc<`5ooooo`04ooo<c?oo003oo`00ool000Wooooo00?o0000oooooooo
ool02oooool01_ook^kooeEEool00?oo003oofIVooo^kU3ooooo002Wooooo`03oeEEEOl0003o0000
00;o000000?oIVIVooooooooool03oooool01Ooo003oooooooooooooooooo`0000Sooooo00?oo`00
ooooooooool0@OoooolDo`000003ogMgMooooooooooo02;ooooo00?ooeEEool00?oo00000_oo0000
0oooIVKoooooooooo`03ooooo`03ool00?oooooooooo00?ooooo00?oEEEEo`000?l000000_l00000
0omVIVKoooooooooo`0looooo`03o`000?oooooooooo00Wooooo0_oo000<ooooo`03ool@4?oo003o
oooo013ooooo00?oodA4ool00?ooV9P01oooool00oooV9Soo`00ool00009ooooo`03o`000?oooooo
oooo00_ooooo00?ooeEEool00?oo00000_oo00000oooIVKoooooooooo`1>ooooo`00YooooolGo`00
0005ool00?l0003o0000o`000?oo00002?l000000ooo003o0000o`00001jo`0000Koo`001Oooool0
0ooo003oooooooooo`03ooooodGo00002oooool2ool000cooooo3_oo00000ooo8B7oooooooooo`02
ooooo`;oo`0000?oonk^ooooooooool01Oooool00oook^koo`00ool00009oooooa7o00000ooo001@
ooooo`00Yoooool6o`00017ooooo00Goo`00ooooooooooooooooool00008ooooo`03ool00?oooooo
oooo01gooooo00Koc<c<oc8b<_o^k^kok^k^ob4Q8OnYZJTNooooo`;o00003oooool00omEEEGo0000
oeEEE@0Uooooo`Koo`001Oooool00ooo003oooooooooo`03ooooo`Ko0000<oooool01ol0003o8B4Q
olc<c?ooooooc<c<ob4Q8OnYZJT01?ooool00ol0003oooooooooo`09ooooo`;oo`003?ooool?ool0
00Cooooo0_oo0009ooooo`;oo`002Oooool00ol0003oooooooooo`0;ooooo`Koo`00D?ooool00:Oo
oooo00?oEEEEo`000?l000000_l000000omVIVKoooooooooo`0?ooooo`05ool00?oooooooooooooo
oooo00002?ooool00ooo003oooooooooo`0Mooooo`06odA4A?nYZJWooooooooooonYZJWo<S8b7_oo
ool2o`0000kooooo00?oEEEEo`000?lQ8B409_ooool00oooEEGoo`00ool00002ool00003oomVI_oo
oooooooo00?ooooo00?oo`00ooooooooool00oooool00omEEEGo0000o`000002o`000003ofIVI_oo
oooooooo037ooooo00?o0000olc<c?ooool00_ooool00oo<c<co410@ooooo`03ooooo`03o`000?oo
oooooooo00Wooooo0_oo00000oooc<coooooooooo`09ooooo`03oolQ8Ooo003oomgM00Wooooo00?o
omgMool00?oo00001?ooool2ool00003ooo^k_oooooooooo00Gooooo00?oonk^ool00?oo00002Ooo
ool00ol0003oooooooooo`0;ooooo`03oomEEOoo003oo`0000;oo`0000?oofIVooooooooool0C_oo
ool00:Oooooo00Kok^k^oeEEEOl0003o0000ofIVI_o^k^hAooooo`05ool00?oooooooooooooooooo
00002?ooool00ooo003oooooooooo`0Mooooo`03o`000?oooooooooo00;ooooo00?o0000oooooooo
ool07?ooool2o`0000gooooo00CoEEEEo`000?lQ8B7ok^k^7?ooool01onYZJWo410@o`000?lQ8B7o
V9RHoa0@4?l000000oooool01_l@003oZ@00ool00?oo003oW`H6ooo^kPGooooo00?oo`00oooooooo
ool00oooool01_o^k^koEEEEo`000?l0003oIVIVonk^kS?ooooo00?o0000ooooooooool00oooool0
0ol0003oooooooooo`02ooooo`03o`000?oooooooooo00Wooooo0_oo00000ooo413oonk^ooooo`09
ooooo`03oomVI_oo003oodA400Wooooo00?oogMgool00?ooEED01?ooool00oooA4Coo`00oonHV007
ooooo`03oonHV?oo003oo`0000Wooooo00?o0000ooooooooool00_ooool02?n7QhOo410@o`000?lQ
8B7oZJVYo`000?l@413ooooo1_l000000oook^koooooooooo`03ooooo`;o000000?o410@oooooooo
ool0A_ooool00;kooooo00Goo`00ooooooooooooooooool00008ooooo`03ool00?oooooooooo01go
oooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0Looooo`;o00003?ooool01?mEEEGo
0000ob4Q8Oo^k^hMooooo`06o`000?nj^[[oooooomgMgOmgMgOo00001_ooool00ol0003ooooooooo
o`06ooooo`03ool00?oooooooooo00Gooooo00?oo`00ooooooooool0=?ooool00ol0003oc<c<oooo
o`02ooooo`03olc<c?l@413ooooo00?ooooo00?o0000ooooooooool02Oooool3ool00003oolQ8Ooo
c<cooooo00Sooooo00CoomgMool00?oo003oohN71oooool01?oo^[[oo`00ool00?oo^[X4ooooo`04
oonj^_oo003oo`00ooo<c0Gooooo00Coolc<ool00?oo003oo`002Oooool02ol0003ooooooooooooo
ooooQhN7ob4Q8Oo<c<coooooolc<c?lQ8B7o000000;ooooo00?oV9RHo`000?o<c<`00_ooool00ol@
413oooooooooo`05ooooo`03o`000?oooooooooo04Oooooo002nooooo`05ool00?oooooooooooooo
oooo00002?ooool00ooo003oooooooooo`0Mooooo`03o`000?oooooooooo00;ooooo00?o0000oooo
ooooool07?ooool2o`0000_ooooo00CoEEEEo`000?l0003ogMgM7_ooool01_l@413oQhN7onk^k_oo
oooogMgMo`0000Kooooo00?o0000ooooooooool01_ooool00ooo003oooooooooo`05ooooo`03ool0
0?oooooooooo03Cooooo00Oo0000ob4Q8Oo<c<coooooolc<c?lQ8B7oV9RH00Cooooo00?o0000oooo
ooooool02Oooool2ool00004oonHV?oo003oo`00oomgM`;oonk^00?oogMgool00?ooEED01?ooool0
1_ooV9Soo`00ool00?oo<S;ooiRHooo^kP;ooooo00GookZjoomEEOoo003oo`00oomgM`06ooooo`08
oomEEOoo003oo`00oonHV?ook^koooooooo^k_ooV9P4ool000Wooooo00Ko0000oooooooooooooooo
oa0@4?o<c<`3ooooo`03okZj^_l0003ooooo00;ooooo00?oV9RHo`000?o<c<`01Oooool6o`0004So
oooo002nooooo`05ool00?oooooooooooooooooo00002?ooool00ooo003oooooooooo`0Mooooo`06
odA4A?nYZJWooooooooooonYZJWoA4A4:oooool00onHV9So8B4Qolc<c00Oooooo`06olc<c?m4A4Co
0000o`000?l@413o00001_ooool00ol0003oooooooooo`06ooooo`03ool00?oooooooooo00Gooooo
00?oo`00ooooooooool0=?ooool01_l0003oQhN7oa0@4?l0003o410@oiRHV0Gooooo00?o0000oooo
ooooool01_ooool00ooo8B7oo`00ool00002ool00003ooooooooZJWoo`0000Koo`001Oooool00ooo
ZJWoo`00ool00007ool00003oon7Qooooooooooo00Gooooo00?oonk^oomEEOoo00001_oo00001Ooo
<S;oo`00ool00?oo003oob4Q00Oooooo00Go0000ooooooooooooooooo`000005ooooo`03o`000?oo
oooooooo00;ooooo00?oZJVYoa0@4?o<c<`01?ooool01OmVIVKoZJVYoooooooooooo000004Wooooo
002nooooo`05ool00?oooooooooooooooooo00002?ooool00ooo003oooooooooo`0Mooooo`06olc<
c?lb<S;ok^k^onk^k_lb<S;oc<c<D_ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo
00Kooooo00?oo`00ooooooooool01Oooool00ooo003oooooooooo`0dooooo`03o`000?oooooooooo
00Sooooo00?o0000ooooooooool01_ooool00ooo8B7oo`00ool00002ool000;ooooo00?oonk^oomg
Mooo8B400_oo00000ooo8B7oojVYooooo`05ooooo`03ooo^k_ooQhOooc8b00?oo`0000?oob4QoomV
I_oogMd02_ooool00ooo^[[oodA4ool00002ool00003oom4A?oo^[[ooooo00?oo`0000?oob4Qoooo
ooooool01Oooool01_l0003ooooooooooooooooo410@olc<c0?ooooo00?oc<c<o`000?ooool01?oo
ool00oo<c<co410@olc<c003ooooo`05onk^k_lQ8B7ok^k^ooooool00000BOooool00;kooooo00Go
o`00ooooooooooooooooool00008ooooo`03ool00?oooooooooo01kooooo00Co^[Zjoa0@4?l@413o
c<c<D_ooool00on7QhOo<S8booooo`05ooooo`03o`000?oooooooooo00Kooooo00?oo`00oooooooo
ool01Oooool00ooo003oooooooooo`0dooooo`03o`000?oooooooooo00Sooooo00?o0000oooooooo
ool0@?ooool02ol0003oooooooooooooooooZJVYob4Q8Oo<c<coooooolc<c?lQ8B7o000000Kooooo
00?oZJVYoc8b<_ooool00oooool01?nYZJWoIVIVooooool00019ooooo`00__ooool01Ooo003ooooo
ooooooooooooo`0000Sooooo00?oo`00ooooooooool0LOooool01Om4A4Co0000o`000?l@413oc<c<
00Kooooo00?o0000ooooooooool01_ooool00ooo003oooooooooo`05ooooo`03ool00?oooooooooo
03?ooooo00?o410@o`000?ooool02Oooool00ol0003oooooooooo`10ooooo`03o`000?oooooooooo
00;ooooo00KoZJVYoa0@4?l0003o410@ohN7Qol00007ooooo`03o`000?oooooooooo00?ooooo00?o
A4A4olc<c?l00000BOooool00;kooooo00Goo`00ooooooooooooooooool00008ooooo`03ool00?oo
oooooooo07[ooooo00?o0000odA4A?l000002?ooool00ooo003oooooooooo`05ooooo`03ool00?oo
oooooooo03oooooo00?o0000ooooooooool0@?ooool00ol0003oooooooooo`07ooooo`09o`000?oo
ooooooooo`000?n7QhOok^k^oooooomgMgOo<S8b00Gooooo00?ogMgModA4A?l00000BOooool00;ko
oooo00Goo`00ooooooooooooooooool00008ooooo`03ool00?oooooooooo07[ooooo00?oc<c<oeEE
EOl000002?ooool00ooo003oooooooooo`05ooooo`03ool00?oooooooooo03oooooo00?o0000oooo
ooooool0@?ooool00ol0003oooooooooo`07ooooo`05o`000?ooooooooooojVYZOlb<S800_l00000
0omEEEGok^k^ooooo`05ooooo`03ohN7Qol0003ooooo04Sooooo002nooooo`05ool00?oooooooooo
oooooooo00002?ooool00ooo003oooooooooo`25ooooo`03ool00?oooooooooo00Gooooo00?oo`00
ooooooooool0?oooool00ol0003oooooooooo`10ooooo`03o`000?oooooooooo00Kooooo0_l0001I
ooooo`00__ooool01Ooo003oooooooooooooooooo`0000Sooooo00?oo`00ooooooooool0QOooool0
0ooo003oooooooooo`05ooooo`03ool00?oooooooooo03ooooooA?l0001Sooooo`00__ooool01Ooo
003oooooooooooooooooo`0000Sooooo00?oo`00ooooooooool0QOooool00ooo003oooooooooo`05
ooooo`03ool00?oooooooooo0>Kooooo002nooooo`05ool00?oooooooooooooooooo00002?ooool0
0ooo003oooooooooo`25ooooo`03ool00?oooooooooo00Gooooo00?oo`00ooooooooool0i_ooool0
0;kooooo00Goo`00ooooooooooooooooool00008ooooo`03ool00?oooooooooo08Gooooo00?oo`00
ooooooooool01Oooool00ooo003oooooooooo`3Vooooo`00/_oooonZool00>Sooooo002booooo`03
ool00?oooooooooo00Wooooo00Goo`00ooooooooooooooooool00008ooooo`03ool00?oooooooooo
07gooooo00?oo`00ooooooooool01Oooool00ooo003oooooooooo`3^ooooo`00/_ooool00ooo003o
ooooooooo`09ooooo`05ool00?oooooooooooooooooo00002?ooool00ooo003oooooooooo`1moooo
o`03ool00?oooooooooo00Gooooo00?oo`00ooooooooool0k_ooool00;;ooooo00?oo`00oooooooo
ool02Oooool01Ooo003oooooooooooooooooo`0000Sooooo00?oo`00ooooooooool0OOooool00ooo
003oooooooooo`05ooooo`03ool00?oooooooooo04OoooooA?l0001Sooooo`00/_ooool00ooo003o
ooooooooo`09ooooo`05ool00?oooooooooooooooooo00002?ooool00ooo003oooooooooo`1moooo
o`03ool00?oooooooooo00Gooooo00?oo`00ooooooooool0Aoooool00ol0003oooooooooo`10oooo
o`03o`000?oooooooooo067ooooo002booooo`03ool00?oooooooooo00Wooooo00Goo`00oooooooo
ooooooooool00008ooooo`03ool00?oooooooooo07gooooo00?oo`00ooooooooool01Oooool00ooo
003oooooooooo`17ooooo`03o`000?oooooooooo043ooooo00?o0000ooooooooool0HOooool00;;o
oooo00?oo`00ooooooooool02Oooool01Ooo003oooooooooooooooooo`0000Sooooo00?oo`00oooo
ooooool0OOooool00ooo003oooooooooo`05ooooo`03ool00?oooooooooo04Oooooo00?o0000oooo
ooooool0@?ooool00ol0003oooooooooo`1Qooooo`00/_ooool00ooo003oooooooooo`09ooooo`05
ool00?oooooooooooooooooo00002?ooool00ooo003oooooooooo`1mooooo`03ool00?oooooooooo
00Gooooo00?oo`00ooooooooool0Aoooool00ol0003oooooooooo`10ooooo`03o`000?oooooooooo
067ooooo002booooo`03ool00?oooooooooo00Wooooo00Goo`00ooooooooooooooooool00008oooo
o`03ool00?oooooooooo07gooooo00?oo`00ooooooooool01Oooool00ooo003oooooooooo`17oooo
o`03o`000?oooooooooo043ooooo00?o0000ooooooooool0HOooool00;;ooooo00?oo`00oooooooo
ool02Oooool01Ooo003oooooooooooooooooo`0000Sooooo00?oo`00ooooooooool0OOooool00ooo
003oooooooooo`05ooooo`03ool00?oooooooooo04Oooooo00?o0000ooooooooool0@?ooool00ol0
003oooooooooo`1Qooooo`00/_ooool00ooo003oooooooooo`09ooooo`05ool00?oooooooooooooo
oooo00002?ooool00ooo003oooooooooo`1mooooo`03ool00?oooooooooo00Gooooo00?oo`00oooo
ooooool0Aoooool00ol0003oooooooooo`10ooooo`03o`000?oooooooooo067ooooo002booooo`03
ool00?oooooooooo00Wooooo00Goo`00ooooooooooooooooool00008ooooo`03ool00?oooooooooo
07gooooo00?oo`00ooooooooool01Oooool00ooo003oooooooooo`17ooooo`03o`000?oooooooooo
043ooooo00?o0000ooooooooool0HOooool00;;ooooo00?oo`00ooooooooool02Oooool01Ooo003o
ooooooooooooooooo`0000Sooooo00?oo`00ooooooooool0OOooool00ooo003oooooooooo`05oooo
o`03ool00?oooooooooo04Oooooo00?o0000ooooooooool0@?ooool00ol0003oooooooooo`1Qoooo
o`00/_ooool00ooo003oooooooooo`09ooooo`05ool00?oooooooooooooooooo00002?ooool00ooo
003oooooooooo`1mooooo`03ool00?oooooooooo00Gooooo00?oo`00ooooooooool0Aoooool00ol0
003oooooooooo`10ooooo`03o`000?oooooooooo067ooooo002booooo`03ool00?oooooooooo00Wo
oooo00Goo`00ooooooooooooooooool00005ooooo`06onk^k_mEEEGo0000o`000?mVIVKok^k^N_oo
ool01_ook^kooeEEool00?oo003oofIVooo^kP?ooooo00Kok^k^oeEEEOl0003o0000ofIVI_o^k^i6
ooooo`03o`000?oooooooooo043ooooo00?o0000ooooooooool0HOooool00;;ooooo00?oo`00oooo
ooooool02Oooool01Ooo003oooooooooooooooooo`0000Gooooo00?oEEEEo`000?l000000_l00000
0omVIVKoooooooooo`1hooooo`03oomEEOoo003oo`0000;oo`0000GoofIVoooooooooooooooooeEE
E@04o`000003ofIVI_oooooooooo04Cooooo00?o0000ooooooooool06Oooool<ool001_ooooo00?o
0000ooooooooool0HOooool00;;ooooo00?oo`00ooooooooool02Oooool01Ooo003ooooooooooooo
ooooo`0000GoooooP?l00006ool000?oooooCOl0000Kooooo`coo`006ooooolAo`0005?ooooo002b
ooooo`03ool00?oooooooooo00Wooooo00Goo`00ooooooooooooooooool00005ooooo`Ko00007Ooo
ool01_l0003o^[ZjoooooooMgMgoMgMgo`0005Oooooo1_oo0003ooooo`Ko0000>_ooool01_l0003o
^[ZjoooooooMgMgoMgMgo`0000Kooooo00?o0000ooooooooool0@?ooool00ol0003oooooooooo`1Q
ooooo`00/_ooool00ooo003oooooooooo`09ooooo`05ool00?oooooooooooooooooo00001Oooool0
0omEEEGo0000o`000002o`000003ofIVI_oooooooooo01_ooooo00Ko410@ohN7Qoo^k^koooooomgM
gOl0001Gooooo`03oomEEOoo003oo`0000;oo`0000GoofIVoooooooooooooooooeEEE@04o`000003
ofIVI_oooooooooo03Sooooo00Ko410@ohN7Qoo^k^koooooomgMgOl00006ooooo`03o`000?oooooo
oooo043ooooo00?o0000ooooooooool0HOooool00;;ooooo00?oo`00ooooooooool02Oooool01Ooo
003oooooooooooooooooo`0000Gooooo00Kok^k^oeEEEOl0003o0000ofIVI_o^k^hMooooo`06olc<
c?m4A4Co0000o`000?l@413o0000COooool01ol@413o0000oiRHV?l@413o0000oa0@4?nYZJT00ooo
ool01_ook^kooeEEool00?oo003oofIVooo^kP?ooooo00Kok^k^oeEEEOl0003o0000ofIVI_o^k^hj
ooooo`06olc<c?m4A4Co0000o`000?l@413o00001_ooool00ol0003oooooooooo`10ooooo`03o`00
0?oooooooooo00;ooooo00KoMgMgoa0@4?l0003o410@oeEEEOoMgMeIooooo`00/_ooool00ooo003o
ooooooooo`09ooooo`05ool00?oooooooooooooooooo0000;Oooool00ol0003oooooooooo`1<oooo
o`07o`000?lQ8B7oc<c<ooooooo<c<co8B4QojVYZ@1@ooooo`03o`000?oooooooooo00Cooooo00?o
0000ooooooooool0@?ooool02ol0003oooooooooooooooooMgMgodA4A?oMgMgoooooomgMgOn7QhOo
8B4Q05Wooooo002booooo`03ool00?oooooooooo00Wooooo00Goo`00ooooooooooooooooool0000/
ooooo`03ohN7Qolb<S;ooooo04gooooo00?o0000olc<c?ooool00_ooool00oo<c<co410@ooooo`1>
ooooo`03ohN7Qolb<S;ooooo00Gooooo00?o0000ooooooooool0@?ooool01_l0003ooooooooooooo
oooo410@onk^kUkooooo002booooo`03ool00?oooooooooo00Wooooo00Goo`00oooooooooooooooo
ool0000Yooooo`05odA4A?l0003o0000oa0@4?o<c<`0C_ooool00ol0003oooooooooo`03ooooo`03
o`000?oooooooooo04[ooooo00GoA4A4o`000?l0003o410@olc<c006ooooo`03o`000?oooooooooo
043ooooo00Co0000oooooooooooooooo1ol0001Iooooo`00/_ooool00ooo003oooooooooo`09oooo
o`05ool00?oooooooooooooooooo0000O?ooool00ol0003oc<c<ooooo`02ooooo`03olc<c?l@413o
oooo05Kooooo00?o0000ooooooooool0@?ooool01_l0003ooooooooooooooooo410@omgMg@?ooooo
00?ogMgMo`000?ooool0F?ooool00;;ooooo00?oo`00ooooooooool02Oooool01Ooo003ooooooooo
ooooooooo`0007cooooo00Oo0000ob4Q8Oo<c<coooooolc<c?lQ8B7oV9RH05Oooooo00?o0000oooo
ooooool0@?ooool02ol0003oooooooooooooooooQhN7ob4Q8Onj^[[oooooomgMgOlQ8B7oMgMg05Wo
oooo002booooo`03ool00?oooooooooo00Wooooo00Goo`00ooooooooooooooooool0001looooo`06
o`000?n7QhOo410@o`000?l@413oV9RHF?ooool00ol0003oooooooooo`10ooooo`03o`000?oooooo
oooo00;ooooo00GoQhN7oa0@4?l0003o410@ogMgM`1Jooooo`00/_ooool00ooo003oooooooooo`09
ooooo`05ool00?oooooooooooooooooo0000O?ooool00ol0003oooooooooo`1Kooooo`03o`000?oo
oooooooo043ooooo00?o0000ooooooooool0HOooool00;;ooooo00?oo`00ooooooooool02Oooool0
1Ooo003oooooooooooooooooo`0007cooooo00?o0000ooooooooool0Foooool00ol0003ooooooooo
o`10ooooo`03o`000?oooooooooo067ooooo002booooo`03ool00?oooooooooo00Wooooo00Goo`00
ooooooooooooooooool0001kooooo`03oa0@4?l0003ooooo05cooooo00?o0000ooooooooool0@?oo
ool00ol0003oooooooooo`1Qooooo`00/_ooool00ooo003oooooooooo`09ooooo`05ool00?oooooo
oooooooooooo0000f_oooom4o`0006?ooooo002booooo`03ool00?oooooooooo00Wooooo00Goo`00
ooooooooooooooooool0003ooooooh;ooooo002booooo`03ool00?oooooooooo00Wooooo00Goo`00
ooooooooooooooooool0003ooooooh;ooooo002booooo`03ool00?oooooooooo00Wooooo00Goo`00
ooooooooooooooooool0003ooooooh;ooooo002booooo`04ool00?oooooooooooooookOoo`00eooo
ool00;;ooooo00Goo`00ooooooooooooooooool00007ooooo`05ool00?oooooooooooooooooo0000
R?ooool00ooo003oooooooooo`0Nooooo`03ool00?oooooooooo0=Gooooo002booooo`05ool00?oo
oooooooooooooooo00001oooool01Ooo003oooooooooooooooooo`0008Sooooo00?oo`00oooooooo
ool07_ooool00ooo003oooooooooo`3Eooooo`00/_ooool01Ooo003oooooooooooooooooo`0000Oo
oooo00Goo`00ooooooooooooooooool0000eoooood?o00004?ooool00ooo003oooooooooo`0Noooo
o`03ool00?oooooooooo0=Gooooo002booooo`05ool00?oooooooooooooooooo00001oooool01Ooo
003oooooooooooooooooo`0003Gooooo00?o0000ooooooooool0;?ooool00oooIVKoo`00ool00002
ool00003oolb<_ooQhOoonk^00_ooooo00?o0000ooooooooool03_ooool00ooo003oooooooooo`0N
ooooo`03ool00?oooooooooo0=Gooooo002booooo`05ool00?oooooooooooooooooo00001oooool0
1Ooo003oooooooooooooooooo`0003Gooooo00?o0000ooooooooool0;?ooool00ooo<S;oo`00ool0
0005ool00003oonYZOoooooooooo00Sooooo00?o0000ooooooooool03_ooool00ooo003ooooooooo
o`0Nooooo`03ool00?oooooooooo0=Gooooo002booooo`05ool00?oooooooooooooooooo00001ooo
ool01Ooo003oooooooooooooooooo`0003Gooooo00?o0000ooooooooool0<Oooool01Oooc<coodA4
ool00?oo003oomgM00Wooooo00?o0000ooooooooool03_ooool00ooo003oooooooooo`0Nooooo`03
ool00?oooooooooo0=Gooooo002booooo`05ool00?oooooooooooooooooo00001oooool01Ooo003o
ooooooooooooooooo`0003Gooooo00?o0000ooooooooool0<oooool00ooo<S;oo`00oomEE@09oooo
o`03o`000?oooooooooo00kooooo00?oo`00ooooooooool07_ooool00ooo003oooooooooo`3Eoooo
o`00/_ooool01Ooo003oooooooooooooooooo`0000Oooooo00Goo`00ooooooooooooooooool0000e
ooooo`03o`000?oooooooooo03?ooooo00?oojVYool00?oo41002Oooool00ol0003oooooooooo`0>
ooooo`03ool00?oooooooooo01kooooo00?oo`00ooooooooool0eOooool00;;ooooo00Goo`00oooo
ooooooooooooool00007ooooo`05ool00?oooooooooooooooooo0000=Oooool00ol0003ooooooooo
o`0cooooo`03ooo^k_oo003oo`0000Wooooo00?o0000ooooooooool03_ooool00ooo003ooooooooo
o`0Nooooo`03ool00?oooooooooo0=Gooooo002booooo`05ool00?oooooooooooooooooo00001ooo
ool01Ooo003oooooooooooooooooo`0003Gooooo00?o0000ooooooooool01?ooool00ooo<S;oo`00
ool00008ool00003oolQ8Ooooooooooo00Oooooo00?oomgMoomVI_oo8B400ooo00001?oo413ooc8b
oomgMooo^[XAooooo`;oo`002Oooool00ol0003oooooooooo`0>ooooo`03ool00?oooooooooo01ko
oooo00?oo`00ooooooooool0eOooool00;;ooooo00Goo`00ooooooooooooooooool00007ooooo`05
ool00?oooooooooooooooooo0000=Oooool00ol0003oooooooooo`04ooooo`03oolb<_oo003oo`00
00Soo`0000?oob4Qooooooooool01_ooool00oooMgOoo`00ool00009ool00003oon7Qooooooooooo
00Gooooo00?ooiRHoom4A?oo00000_oo00000oooA4CoojVYooooo`02ool000Wooooo00?o0000oooo
ooooool03_ooool00ooo003oooooooooo`0Nooooo`03ool00?oooooooooo0=Gooooo002booooo`05
ool00?oooooooooooooooooo00001oooool01Ooo003oooooooooooooooooo`0003Gooooo00?o0000
ooooooooool02?ooool2ool000gooooo00GoofIVool00?oo003oogMgooo<c003ooooo`06ooo<c?oo
ZJWoofIVoolb<_oo003oob4Q1Oooool00oook^kooc8bool00006ool00003oomEEOoo003oo`0000Wo
oooo00?o0000ooooooooool03_ooool00ooo003oooooooooo`0Nooooo`03ool00?oooooooooo0=Go
oooo002booooo`05ool00?oooooooooooooooooo00001oooool01Ooo003oooooooooooooooooo`00
03Gooooo00?o0000ooooooooool02?ooool2ool000cooooo00CookZjool00?oo413oonk^4?ooool0
2?oo<S;oo`00ool00?ooV9Soonk^ooooooook^kooiRH1?oo0009ooooo`03o`000?oooooooooo00ko
oooo00?oo`00ooooooooool07_ooool00ooo003oooooooooo`3Eooooo`00[oooool02?o^k^koEEEE
o`000?l0003oIVIVonk^k_ooooooo`001oooool01Ooo003oooooooooooooooooo`0003Gooooo00?o
0000ooooooooool02?ooool2ool000cooooo00?ooeEEool00?ooZJT04?ooool01?ooV9Soo`00ool0
0?ooc<`5ooooo`04ooo<c?oo003oo`00ool000Wooooo00?o0000ooooooooool02oooool01_ook^ko
oeEEool00?oo003oofIVooo^kQ_ooooo00Kok^k^oeEEEOl0003o0000ofIVI_o^k^ikooooo`Co0?l0
E_ooool00:oooooo00?oEEEEo`000?l000000_l000000omVIVKoooooool00007ooooo`05ool00?oo
oooooooooooooooo0000=Oooool00ol0003oooooooooo`08ooooo`;oo`003?ooool00ooo413oo`00
ooooo`0@ooooo`03oom4A?oo003ooiRH00Oooooo00?ooiRHool00?oo00002Oooool00ol0003ooooo
ooooo`0;ooooo`03oomEEOoo003oo`0000;oo`0000?oofIVooooooooool06Oooool00omEEEGo0000
o`000002o`000003ofIVI_oooooooooo07Wooooo1Ol0o`1Eooooo`00[oooool7o`000003ool00?l0
003o000000Go000000Goo`00o`000?l0003o0000ool0000fo`0000[ooooo0_oo000<ooooo`koo`00
00?oob4Qooooooooool00_ooool2ool00003ooo^k_oooooooooo00Gooooo00?oonk^ool00?oo0000
2Oooool>o`0000Koo`006ooooon1o`0000Ko0?l000?o0000ooooooooool0DOooool00:oooooo1_l0
00000oooooooo`00ooooo`06ooooo`05ool00?oooooooooooooooooo0000:Oooool01ol0003o8B4Q
olc<c?ooooooc<c<ob4Q8OnYZJT01Oooool00ol0003oooooooooo`08ooooo`;oo`003?ooool?ool0
00Cooooo0_oo0009ooooo`;oo`002Oooool00ol0003oooooooooo`0;ooooo`Koo`006oooool6o`00
02;ooooo00?oZJVYob4Q8Onj^[X00_ooool00onj^[[o8B4Qooooo`1Aooooo`Ko0?l0E?ooool00:oo
oooo00?oEEEEo`000?l000000_l000000omVIVKoooooool00007ooooo`05ool00?oooooooooooooo
oooo0000:Oooool00ol0003oc<c<ooooo`02ooooo`03olc<c?l@413ooooo00Cooooo00?o0000oooo
ooooool02?ooool2ool00003ooo<c?oooooooooo00Wooooo00?oob4Qool00?oogMd02Oooool00ooo
gMgoo`00ool00004ooooo`;oo`0000?oonk^ooooooooool01Oooool00oook^koo`00ool00009oooo
o`03o`000?oooooooooo00_ooooo00?ooeEEool00?oo00000_oo00000oooIVKoooooooooo`0Ioooo
o`03oeEEEOl0003o000000;o000000?oIVIVooooooooool08?ooool00ol@413o^[Zjooooo`1Foooo
o`Ko0?l0E?ooool00:oooooo00Sok^k^oeEEEOl0003o0000ofIVI_o^k^koooooool000Oooooo00Go
o`00ooooooooooooooooool0000Yooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0
0oooool00ol0003oooooooooo`08ooooo`;oo`0000?ooa0@ooo^k_ooool02Oooool00oooIVKoo`00
oom4A009ooooo`03oomgMooo003ooeEE00Cooooo00?oodA4ool00?ooV9P01oooool00oooV9Soo`00
ool00009ooooo`03o`000?oooooooooo00;ooooo00KoZJVYoa0@4?l0003o0000oeEEEOoMgMd3oooo
o`06ooo^k_ooEEGoo`00ool00?ooIVKoonk^6oooool01_o^k^koEEEEo`000?l0003oIVIVonk^kR;o
oooo00?o0000ooooooooool0D?ooool01_l@413oA4A4o`000?l0003o410@ogMgM`Go0?l0EOooool0
0;Kooooo00?oo`00ooooooooool01Oooool01Ooo003oooooooooooooooooo`0002Wooooo00?o0000
olc<c?ooool00_ooool00oo<c<co410@ooooo`04ooooo`03o`000?oooooooooo00Sooooo0ooo0000
0ooo8B7oolc<ooooo`08ooooo`04oooMgOoo003oo`00oon7Q`Oooooo00CookZjool00?oo003ookZj
1?ooool01?oo^[[oo`00ool00?ooc<`5ooooo`04ooo<c?oo003oo`00ool000Wooooo00Oo0000oooo
ooooooooooooojVYZOlQ8B7o^[Zj00;ooooo00?o^[Zjob4Q8Oooool0Boooool00ol@413o^[Zjoooo
o`1@ooooo`03o`000?nHV9Sok^k^00;ooooo00?oMgMgo`3o0?l0o`000_l0o`1Fooooo`00]_ooool0
0ooo003oooooooooo`05ooooo`05ool00?oooooooooooooooooo0000:Oooool01ol0003o8B4Qolc<
c?ooooooc<c<ob4Q8OnHV9P01Oooool00ol0003oooooooooo`08ooooo`;oo`0000CooiRHool00?oo
003oogMg0_ook^h00oooMgOoo`00oomEE@04ooooo`06oonHV?oo003oo`00oolb<_ooV9Soonk^0_oo
ool01Ooo^[[ooeEEool00?oo003oogMg00Kooooo00SooeEEool00?oo003ooiRHooo^k_ooooooonk^
oonHV0Coo`002Oooool01_l0003ooooooooooooooooo410@okZj^U7ooooo00OoZJVYob4Q8Onj^[[o
ooooonk^k_n7QhOo000004cooooo00?o8B4Qooooooooool00oooool00ol0003oooooooooo`1Goooo
o`00]_ooool00ooo003oooooooooo`05ooooo`05ool00?oooooooooooooooooo0000:Oooool01_l0
003oQhN7oa0@4?l0003o410@oiRHV0Kooooo00?o0000ooooooooool01Oooool00ooo8B7oo`00ool0
0002ool00003ooooooooZJWoo`0000Koo`001Oooool00oooZJWoo`00ool00007ool00003oon7Qooo
oooooooo00Gooooo00?oonk^oomEEOoo00001_oo00001Ooo<S;oo`00ool00?oo003oob4Q00Oooooo
00Go0000ooooooooooooooooo`00001Cooooo`06ojVYZOlQ8B7o0000o`000?m4A4Co0000D?ooool0
0ooMgMgoV9RHoa0@401Iooooo`00]_ooool00ooo003oooooooooo`05ooooo`05ool00?oooooooooo
oooooooo0000:Oooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00Gooooo00?oob4Q
ool00?oo00000_oo0002ooooo`03ooo^k_ooMgOoob4Q00;oo`0000?oob4QoonYZOooool01Oooool0
0oook^koohN7oolb<P03ool00003oolQ8OooIVKoomgM00[ooooo00?ookZjoom4A?oo00000_oo0000
0oooA4CookZjooooo`03ool00003oolQ8Ooooooooooo00Gooooo00Ko0000oooooooooooooooooa0@
4?nj^[ZTooooo`03onk^k_mEEEGo410@00;o000000?o<S8bolc<c?ooool0F?ooool00;Kooooo00?o
o`00ooooooooool01Oooool01Ooo003oooooooooooooooooo`0002Wooooo00?o0000ooooooooool0
2Oooool00ol0003oooooooooo`0oooooo`0;o`000?oooooooooooooooonYZJWo8B4QokZj^_oooooo
k^k^ohN7Qol00000Woooool01?lb<S;oMgMgokZj^_oMgMeLooooo`00]_ooool00ooo003ooooooooo
o`05ooooo`05ool00?oooooooooooooooooo0000:?ooool00ol@413o0000ooooo`0:ooooo`03o`00
0?oooooooooo03oooooo00?o0000ooooooooool00_ooool01_nYZJWo8B4Qo`000?l0003oA4A4o`00
09oooooo00?o0000ooooooooool00oooool00olQ8B7oooooooooo`1Gooooo`00]_ooool00ooo003o
ooooooooo`05ooooo`05ool00?oooooooooooooooooo0000=Oooool00ol0003oooooooooo`0ooooo
o`03o`000?oooooooooo0:Oooooo00Oo<S8bogMgMoo^k^koooooonk^k_mgMgOo000005Wooooo002f
ooooo`03ool00?oooooooooo00Gooooo00Goo`00ooooooooooooooooool0000eooooo`03o`000?oo
oooooooo03oooooo00?o0000ooooooooool0Yoooool01oo^k^koA4A4o`000?l0003o410@oc8b<_l@
4100FOooool00;Kooooo00?oo`00ooooooooool01Oooool01Ooo003oooooooooooooooooo`0003Go
oooo00?o0000ooooooooool0?oooool00ol0003oooooooooo`3oooooo`Sooooo002fooooo`03ool0
0?oooooooooo00Gooooo00Goo`00ooooooooooooooooool0000eoooood?o0000ooooool:ooooo`00
]_ooool00ooo003oooooooooo`05ooooo`05ool00?oooooooooooooooooo0000oooooon2ooooo`00
]_ooool00ooo003oooooooooo`05ooooo`05ool00?oooooooooooooooooo0000oooooon2ooooo`00
]_ooool00ooo003oooooooooo`05ooooo`05ool00?oooooooooooooooooo0000oooooon2ooooo`00
]_ooool00ooo003oooooooooo`05ooooo`05ool00?oooooooooooooooooo0000oooooon2ooooo`00
]_ooool00ooo003oooooooooo`05ooooo`05ool00?oooooooooooooooooo0000oooooon2ooooo`00
]_ooool00ooo003oooooooooo`05ooooo`05ool00?oooooooooooooooooo0000oooooon2ooooo`00
]_ooool00ooo003oooooooooo`05ooooo`05ool00?oooooooooooooooooo0000=Ooooom3o`000?oo
oooo2_ooool00;Kooooo00?oo`00ooooooooool01Oooool01Ooo003oooooooooooooooooo`0003Go
oooo00?o0000ooooooooool0?oooool00ol0003oooooooooo`3oooooo`Sooooo002fooooo`03ool0
0?oooooooooo00Gooooo00Goo`00ooooooooooooooooool0000eooooo`03o`000?oooooooooo03oo
oooo00?o0000ooooooooool0ooooool8ooooo`00]_ooool00ooo003oooooooooo`05ooooo`05ool0
0?oooooooooooooooooo0000=Oooool00ol0003oooooooooo`0oooooo`03o`000?oooooooooo0?oo
oooo2?ooool00;Kooooo00Coo`00ooooooo^k^koEEEE0_l000000omVIVKok^k^ool00003ooooo`03
ool00?oooooooooo03?ooooo00?o0000ooooooooool0?oooool00ol0003oooooooooo`3oooooo`So
oooo002fooooo`03ool00?ooooooEEEE00Co000000?oIVIVool00?ooool00_ooool00ooo003ooooo
ooooo`0cooooo`03o`000?oooooooooo03oooooo00?o0000ooooooooool0ooooool8ooooo`00]_oo
ool00ooo003oooooo`000005o`000005ool00?oooooooooooooooooo00008?ooool02?n7QhOo410@
o`000?lQ8B7oZJVYo`000?l@413ooooo1_l00007ooooo`03o`000?oooooooooo03oooooo00?o0000
ooooooooool0ooooool8ooooo`00]_ooool00ooo003oooooo`000005o`000005ool00?l0003o0000
o`000?oo0000=_l00011ooooo`03o`000?oooooooooo0?oooooo2?ooool00;Kooooo00?oo`00oooo
oomEEED01?l000000omVIVKoo`00ooooo`02ooooo`03ool00?oooooooooo01gooooo00?o410@olc<
c?ooool00_ooool00onj^[[o0000ooooo`02ooooo`03oiRHV?l0003oc<c<00Wooooo00?o0000oooo
ooooool0?oooool00ol0003oooooooooo`3oooooo`Sooooo002fooooo`04ool00?ooooook^k^oeEE
E@;o000000?oIVIVonk^k_oo00000oooool00ooo003oooooooooo`0Mooooo`03o`000?oooooooooo
00?ooooo00?o0000ooooooooool00_ooool00onYZJWo410@olc<c008ooooo`03o`000?oooooooooo
03oooooo00?o0000ooooooooool0ooooool8ooooo`00]_ooool01Ooo003oooooooooooooooooo`00
00?ooooo00Goo`00ooooooooooooooooool0000Oooooo`03oa0@4?o<c<cooooo00;ooooo00?oc<c<
o`000?ooool01?ooool00oo<c<co410@olc<c007ooooo`03o`000?oooooooooo03oooooo00?o0000
ooooooooool0ooooool8ooooo`00]_ooool01Ooo003oooooooooooooooooo`0000?ooooo00Goo`00
ooooooooooooooooool0000Oooooo`07ojVYZOlQ8B7oc<c<ooooooo<c<co8B4Qo`000006ooooo`03
ojVYZOlb<S;ooooo00Kooooo00?o0000ooooooooool0?oooool00ol0003oooooooooo`3oooooo`So
oooo002fooooo`05ool00?oooooooooooooooooo00000oooool01Ooo003oooooooooooooooooo`00
023ooooo00KoZJVYoa0@4?l0003o410@ohN7Qol00007ooooo`03o`000?oooooooooo00Gooooo00?o
0000ooooooooool06?ooool<ool001_ooooo00?o0000ooooooooool0ooooool8ooooo`00]_ooool0
1Ooo003oooooooooooooooooo`0000?ooooo00Goo`00ooooooooooooooooool0000Uooooo`09o`00
0?ooooooooooo`000?n7QhOok^k^oooooomgMgOo<S8b00Oooooo00?o0000ooooooooool06?ooool<
ool001_ooooo4_l0003hooooo`00]_ooool01Ooo003oooooooooooooooooo`0000?ooooo00Goo`00
ooooooooooooooooool0000Uooooo`05o`000?ooooooooooojVYZOlb<S800_l000000omEEEGok^k^
ooooo`06ooooo`03o`000?oooooooooo03oooooo00?o0000ooooooooool0ooooool8ooooo`00]_oo
ool01Ooo003oooooooooooooooooo`0000?ooooo00Goo`00ooooooooooooooooool0000Tooooo`;o
00003oooool00ol0003oooooooooo`0oooooo`03o`000?oooooooooo0?oooooo2?ooool00;Kooooo
00Goo`00ooooooooooooooooool00003ooooo`05ool00?oooooooooooooooooo0000=Oooool00ol0
003oooooooooo`0oooooo`03o`000?oooooooooo00;ooooo00OoQhN7oa0@4?l0003o8B4QojVYZOl0
003o410@0?kooooo002dooooo`07onk^k_mEEEGo0000o`000?mVIVKok^k^ool00003ooooo`05ool0
0?oooooooooooooooooo0000=Oooool00ol0003oooooooooo`0oooooo`0;o`000?oooooooooooooo
oon7QhOo8B4Qolc<c?ooooooc<c<ob4Q8Ol00000ooooool00;Cooooo00?oEEEEo`000?l000000_l0
00000omVIVKoo`00ooooo`02ooooo`05ool00?oooooooooooooooooo0000=Oooool00ol0003ooooo
ooooo`0oooooo`06o`000?ooooooooooooooool@413oc<c<0oooool00onj^[[o0000ooooo`3noooo
o`00]?ooool6o`000005ool00?oooooooooooooooooo00000oooool00ooo003oooooooooo`0coooo
o`03o`000?oooooooooo03oooooo00Go0000ooooooooooooooooo`000005ooooo`03o`000?oooooo
oooo0?gooooo002dooooo`[o000000Goo`00o`000?l0003o0000ool0000fo`00047ooooo00Ko0000
oooooooooooooooooa0@4?o<c<`3ooooo`03olc<c?l0003ooooo0?kooooo002dooooo`03oeEEEOl0
003o000000;o000000?oIVIVool00?ooool00_ooool01Ooo003oooooooooooooooooo`0002Sooooo
00?oZJVYob4Q8Onj^[X00_ooool00onj^[[o8B4Qooooo`05ooooo`03o`000?oooooooooo03oooooo
00_o0000ooooooooooooooooojVYZOlQ8B7oc<c<ooooooo<c<co8B4Qo`00003oooooo`00]?ooool0
1oo^k^koEEEEo`000?l0003oIVIVonk^k_oo00000oooool01Ooo003oooooooooooooooooo`0002So
oooo00?o410@okZj^_ooool02_ooool00ol0003oooooooooo`0oooooo`03o`000?oooooooooo00;o
oooo00KoZJVYoa0@4?l0003o410@ohN7Qol0003oooooo`00^_ooool01Ooo003ooooooooooooooooo
o`0000?ooooo00?oo`00ooooooooool09_ooool00ol0003oooooooooo`0:ooooo`03o`000?oooooo
oooo03oooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`3mooooo`00^_ooool01Ooo
003oooooooooooooooooo`0000?ooooo00?oo`00ooooooooool09_ooool00ol@413o^[Zjooooo`0:
ooooo`03o`000?oooooooooo03oooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`3m
ooooo`00^_ooool01Ooo003oooooooooooooooooo`0000?ooooo00?oo`00ooooooooool09_ooool0
1onYZJWo8B4QokZj^_ooooook^k^ohN7Qol000001_ooool00ol0003oooooooooo`0oooooo`03o`00
0?oooooooooo00Kooooo0_l0003oooooo`00^_ooool01Ooo003oooooooooooooooooo`0000?ooooo
00?oo`00ooooooooool09oooool01_nYZJWo8B4Qo`000?l0003oA4A4o`0000Kooooo00?o0000oooo
ooooool0?oooool00ol0003oooooooooo`3oooooo`Sooooo002jooooo`05ool00?oooooooooooooo
oooo00000oooool00ooo003oooooooooo`0coooood?o0000ooooool:ooooo`00^_ooool01Ooo003o
ooooooooooooooooo`0000?ooooo00?oo`00ooooooooool0oooooon0ooooo`00^_ooool01Ooo003o
ooooooooooooooooo`0000?ooooo00?oo`00ooooooooool0oooooon0ooooo`00^_ooool01Ooo003o
ooooooooooooooooo`0000?ooooo00?oo`00ooooooooool0oooooon0ooooo`00^_oooonBool00?So
oooo002nooooo`05ool00?oooooooooooooooooo0000R?ooool00ooo003oooooooooo`3fooooo`00
__ooool01Ooo003oooooooooooooooooo`0003Gooooo@ol0000@ooooo`03ool00?oooooooooo0?Ko
oooo002nooooo`05ool00?oooooooooooooooooo0000=Oooool00ol0003oooooooooo`0oooooo`03
o`000?oooooooooo00kooooo00?oo`00ooooooooool0m_ooool00;kooooo00Goo`00oooooooooooo
ooooool0000eooooo`03o`000?oooooooooo03oooooo00?o0000ooooooooool03_ooool00ooo003o
ooooooooo`3fooooo`00__ooool01Ooo003oooooooooooooooooo`0003Gooooo00?o0000oooooooo
ool0?oooool00ol0003oooooooooo`0>ooooo`03ool00?oooooooooo0?Kooooo002looooo`07onk^
k_mEEEGo0000o`000?mVIVKok^k^ool0000eooooo`03o`000?oooooooooo03oooooo00?o0000oooo
ooooool03_ooool00ooo003oooooooooo`3fooooo`00_?ooool00omEEEGo0000o`000002o`000003
ofIVI_oo003ooooo03Cooooo00?o0000ooooooooool0?oooool00ol0003oooooooooo`0>ooooo`03
ool00?oooooooooo0?Kooooo002looooo`Ko000000?oo`00ooooooooool0<oooool00ol0003ooooo
ooooo`0oooooo`03o`000?oooooooooo00kooooo00?oo`00ooooooooool0m_ooool00;cooooo1_l0
00000ooo003o0000o`00000do`0001oooooo0_oo000Pooooo`03o`000?oooooooooo00kooooo00?o
o`00ooooooooool0m_ooool00;cooooo00?oEEEEo`000?l000000_l000000omVIVKoo`00ooooo`0E
ooooo`07ohN7QolQ8B7oc<c<ooooooo<c<co8B4Qo`000002ooooo`03oiRHV?l0003oc<c<00;ooooo
00Ko410@oooooooooooooooooa0@4?nYZJT2ooooo`03ogMgMolb<S;ooooo00Kooooo00?o0000oooo
ooooool07Oooool2ool0023ooooo00?o0000ooooooooool03_ooool00ooo003oooooooooo`3foooo
o`00_?ooool01oo^k^koEEEEo`000?l0003oIVIVonk^k_oo00005_ooool00ol@413oc<c<ooooo`02
ooooo`03okZj^_l0003ooooo00;ooooo00?oV9RHo`000?o<c<`02_ooool00ol0003oooooooooo`05
ooooo`03o`000?oooooooooo01gooooo0_oo000Pooooo`03o`000?oooooooooo00kooooo00?oo`00
ooooooooool0m_ooool00<;ooooo00?oo`00ooooooooool05?ooool00ol0003oooooooooo`03oooo
o`03o`000?oooooooooo00;ooooo00?oZJVYoa0@4?o<c<`02?ooool00omgMgOo8B4Qooooo`06oooo
o`03o`000?oooooooooo01gooooo0_oo000Pooooo`03o`000?oooooooooo00kooooo00?oo`00oooo
ooooool0m_ooool00<;ooooo00?oo`00ooooooooool05?ooool00ol@413oc<c<ooooo`02ooooo`03
olc<c?l0003ooooo00Cooooo00?oc<c<oa0@4?o<c<`01Oooool01?lQ8B7o0000o`000?nj^[X7oooo
o`03o`000?oooooooooo01gooooo0_oo000Pooooo`03o`000?oooooooooo00_ooooo00Koonk^oomE
EOoo003oo`00oomVI_ook^kfooooo`00`_ooool00ooo003oooooooooo`0Dooooo`07ojVYZOlQ8B7o
c<c<ooooooo<c<co8B4Qo`000006ooooo`03ojVYZOlb<S;ooooo00Gooooo00?ok^k^ogMgMolb<S80
1oooool00ol0003oooooooooo`0Hooooo`coo`006oooool00ol0003oooooooooo`0;ooooo`03oomE
EOoo003oo`0000;oo`0000?oofIVooooooooool0m?ooool00<;ooooo00?oo`00ooooooooool05Ooo
ool01_nYZJWo410@o`000?l@413oQhN7o`0000Oooooo00?o0000ooooooooool01_ooool00ol0003o
ooooooooo`05ooooo`03o`000?oooooooooo01Sooooo3?oo000Koooooa;o00000_oo003fooooo`00
`_ooool00ooo003oooooooooo`0Jooooo`09o`000?ooooooooooo`000?n7QhOok^k^oooooomgMgOo
<S8b00?ooooo00Ko8B4QohN7QoooooooooooofIVI_lb<S87ooooo`03o`000?oooooooooo01gooooo
0_oo000Pooooo`03o`000?oooooooooo00_ooooo1_oo003fooooo`00`_ooool00ooo003ooooooooo
o`0Jooooo`05o`000?ooooooooooojVYZOlb<S800_l000000omEEEGok^k^ooooo`02ooooo`06okZj
^_m4A4Co0000o`000?m4A4Cok^k^1oooool00ol0003oooooooooo`0Mooooo`;oo`008?ooool00ol0
003oooooooooo`0;ooooo`03oomEEOoo003oo`0000;oo`0000?oofIVooooooooool0m?ooool00<;o
oooo00?oo`00ooooooooool06Oooool2o`0001Sooooo00?o0000ooooooooool07Oooool2ool0023o
oooo00?o0000ooooooooool00_ooool02?n7QhOo410@o`000?lQ8B7oZJVYo`000?l@413ooooo1_l0
00000oook^koooooooooo`3dooooo`00`?ooool01_o^k^koEEEEo`000?l0003oIVIVonk^kS;ooooo
00?o0000ooooooooool07Oooool2ool0023ooooo00_o0000ooooooooooooooooohN7QolQ8B7oc<c<
ooooooo<c<co8B4Qo`000002ooooo`03oiRHV?l0003oc<c<00;ooooo00?o410@ooooooooool0mOoo
ool00<3ooooo00?oEEEEo`000?l000000_l000000omVIVKoooooooooo`0`ooooo`03o`000?oooooo
oooo01gooooo0_oo000Pooooo`06o`000?ooooooooooooooool@413oc<c<0oooool00onj^[[o0000
ooooo`02ooooo`03oiRHV?l0003oc<c<0?Wooooo0030ooooo`Ko0000<_ooool00ol0003ooooooooo
o`0oooooo`05o`000?ooooooooooooooool000001Oooool00ol0003oooooooooo`02ooooo`03ojVY
ZOl@413oc<c<0?Sooooo0030ooooocWo0000@Oooool01_l0003ooooooooooooooooo410@olc<c0?o
oooo00?oc<c<o`000?ooool01?ooool00oo<c<co410@olc<c03gooooo`00`?ooool00omEEEGo0000
o`000002o`000003ofIVI_oooooooooo017ooooo00OoQhN7ob4Q8Oo<c<coooooolc<c?lQ8B7o0000
00;ooooo00?oV9RHo`000?o<c<`00_ooool00ol@413oooooooooo`05ooooo`03o`000?oooooooooo
00Kooooo00?o0000ooooooooool0?oooool02ol0003oooooooooooooooooZJVYob4Q8Oo<c<cooooo
olc<c?lQ8B7o000000Kooooo00?oZJVYoc8b<_ooool0m_ooool00<3ooooo00Kok^k^oeEEEOl0003o
0000ofIVI_o^k^hCooooo`03oa0@4?o<c<cooooo00;ooooo00?o^[Zjo`000?ooool00_ooool00onH
V9So0000olc<c005ooooo`Ko00001oooool00ol0003oooooooooo`0oooooo`03o`000?oooooooooo
00;ooooo00KoZJVYoa0@4?l0003o410@ohN7Qol00007ooooo`03o`000?oooooooooo0?Gooooo003I
ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00_ooool00onYZJWo410@olc<c004
ooooo`05ofIVI_nYZJWoooooooooool000002?ooool00ol0003oooooooooo`0oooooo`03o`000?oo
oooooooo00Oooooo00Wo0000oooooooooooo0000ohN7Qoo^k^koooooogMgMolb<S80moooool00=Wo
oooo00?o410@olc<c?ooool00_ooool00oo<c<co0000ooooo`04ooooo`03olc<c?l@413oc<c<00?o
oooo00Gok^k^ob4Q8Oo^k^koooooo`000008ooooo`03o`000?oooooooooo03oooooo00?o0000oooo
ooooool01oooool01Ol0003ooooooooooonYZJWo<S8b00;o000000?oEEEEonk^k_ooool0m_ooool0
0=Wooooo00OoZJVYob4Q8Oo<c<coooooolc<c?lQ8B7o000000Kooooo00?oZJVYoc8b<_ooool00ooo
ool01?nYZJWoIVIVooooool00008ooooo`03o`000?oooooooooo03oooooo00?o0000ooooooooool0
1_ooool2o`000?oooooo003Jooooo`06ojVYZOl@413o0000oa0@4?n7QhOo00001oooool00ol0003o
ooooooooo`03ooooo`03odA4A?o<c<co000000Sooooo00?o0000ooooooooool0?oooool00ol0003o
ooooooooo`3oooooo`Sooooo003Oooooo`09o`000?ooooooooooo`000?n7QhOok^k^oooooomgMgOo
<S8b00Gooooo00?ogMgModA4A?l000002?oooom3o`000?oooooo2_ooool00=oooooo00Go0000oooo
ooooooooZJVYoc8b<P02o`000003oeEEEOo^k^kooooo00Gooooo00?oQhN7o`000?ooool0oooooomD
ooooo`00g_ooool2o`000?ooooooIOooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo
003oooooooooooooA_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo003ooooooooo
ooooA_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo003oooooooooooooA_ooool0
0?oooooooooooom6ooooo`00ooooooooooooodKooooo003oooooooooooooA_ooool00?oooooooooo
oom6ooooo`00ooooooooooooodKooooo003oooooooooooooA_ooool00?oooooooooooom6ooooo`00
ooooooooooooodKooooo003oooooooooooooA_ooool00?oooooooooooom6ooooo`00oooooooooooo
odKooooo003oooooooooooooA_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo003o
ooooooooooooA_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo003ooooooooooooo
A_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo003oooooooooooooA_ooool00?oo
oooooooooom6ooooo`00ooooooooooooodKooooo003oooooooooooooA_ooool00?oooooooooooom6
ooooo`00ooooooooooooodKooooo003oooooooooooooA_ooool00?oooooooooooom6ooooo`00oooo
ooooooooodKooooo003oooooooooooooA_ooool00?oooooooooooom6ooooo`00ooooooooooooodKo
oooo003oooooooooooooA_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo003ooooo
ooooooooA_ooool00?oooooooooooom6ooooo`00ooooooooooooodKooooo003oooooooooooooA_oo
ool00?oooooooooooom6ooooo`00ooooooooooooodKooooo0000\
\>"],
  ImageRangeCache->{{{0, 579.938}, {281.062, 0}} -> {-0.00387131, \
-8.50002*^-6, 0.00907273, 0.00907273}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(analyze[]\)], "Input"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Static Analysis of system \
"\[InvisibleSpace]"cell1"\),
      SequenceForm[ "Static Analysis of system ", "cell1"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--Checking declaration of variables."\)], "Print"],

Cell[BoxData[
    \("--Checking single assignment rule."\)], "Print"],

Cell[BoxData[
    \("--Checking definitions of output/local variables."\)], "Print"],

Cell[BoxData[
    \("--Checking definition of input variables."\)], "Print"],

Cell[BoxData[
    \("--Checking that input/local variables are used."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("WARNING: Input/local variable "\[InvisibleSpace]"d"\
\[InvisibleSpace]" does not appear at the RHS of any equation\n"\),
      SequenceForm[ 
      "WARNING: Input/local variable ", "d", 
        " does not appear at the RHS of any equation\n"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("WARNING: Input/local variable "\[InvisibleSpace]"e"\
\[InvisibleSpace]" does not appear at the RHS of any equation\n"\),
      SequenceForm[ 
      "WARNING: Input/local variable ", "e", 
        " does not appear at the RHS of any equation\n"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--Checking type and domain consistency in the equations:"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("----equation defining "\[InvisibleSpace]"a"\),
      SequenceForm[ "----equation defining ", "a"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("----equation defining "\[InvisibleSpace]"b"\),
      SequenceForm[ "----equation defining ", "b"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("----equation defining "\[InvisibleSpace]"c"\),
      SequenceForm[ "----equation defining ", "c"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("----equation defining "\[InvisibleSpace]"d"\),
      SequenceForm[ "----equation defining ", "d"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("----equation defining "\[InvisibleSpace]"e"\),
      SequenceForm[ "----equation defining ", "e"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("----equation defining "\[InvisibleSpace]"S"\),
      SequenceForm[ "----equation defining ", "S"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--Checking regularity of output equations"\)], "Print"],

Cell[BoxData[
    \("\nAnalysis Successful..."\)], "Print"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkCell[]\)], "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkCell[debug -> True]\)], "Input"],

Cell[BoxData[
    \("This program seems to be an AlpHard Cell..."\)], "Print"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(a2v[$result]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Vhdl files will be generated in \
"\[InvisibleSpace]"/Users/quinton/mmalpha/doc/Packages/CheckAlpHard/"\),
      SequenceForm[ 
      "Vhdl files will be generated in ", 
        "/Users/quinton/mmalpha/doc/Packages/CheckAlpHard/"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("\nPlease wait...\n"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("cell1"\[InvisibleSpace]" was recognized as a \
Cell."\),
      SequenceForm[ "cell1", " was recognized as a Cell."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("WARNING: Some variable identifiers may be confused by Vhdl which is \
not case sensitive. Please, check and modify your Alpha program."\)], "Print"],

Cell[BoxData[
    \("All modules are known..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Vhdl code was written in file \
"\[InvisibleSpace]"cell1.vhd"\),
      SequenceForm[ "Vhdl code was written in file ", "cell1.vhd"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Component description :"\)], "Print"],

Cell[BoxData[
    \("\nCOMPONENT cell1\nPORT(\n  clk: IN STD_LOGIC;\n  CE : IN STD_LOGIC;\n \
 Rst : IN STD_LOGIC;\n  A : IN  SIGNED (15 DOWNTO 0);\n  S : OUT  SIGNED (15 \
DOWNTO 0)\n);\nEND COMPONENT;\n"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(\(0.03189199999999914`\ Second\)\[InvisibleSpace]" \
needed to translate "\[InvisibleSpace]"cell1"\),
      SequenceForm[ 
        Times[ 0.031891999999999143, Second], " needed to translate ", 
        "cell1"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("End of translation."\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Example 2",
  FontColor->GrayLevel[0]]], "Subsubsection"],

Cell["Error 1:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(load["\<cell1e1.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \("[cell1e1]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"]
}, Open  ]],

Cell["\<\
The error is related to the input A, which is not 1-dimensional. \
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \("system cell1e1 :{p | 2<=p<=10}\n               (A : {t,j | 1<=t} of \
integer)\n       returns (S : {t | 2<=t} of integer);\nvar\n  a : {t | 1<=t} \
of integer;\n  b : {t | 1<=t} of integer;\n  e : {t | 1<=t} of integer;\n  c \
: {t | 2<=t} of integer;\n  d : {t | 2<=t} of integer;\nlet\n  a[t] = 0[] + \
A[t,t-1];\n  b[t] = a[t];\n  c[t] = b[t-1];\n  d[t] = a[t-1] + b[t-1] - c[t];\
\n  e[t] = -a[t];\n  S[t] = c[t];\ntel;"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkCell[]\)], "Input"],

Cell[BoxData[
    \(False\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkCell[debug -> True]\)], "Input"],

Cell[BoxData[
    \(checkCell::"decldim" \(\(:\)\(\ \)\) 
      "In :\n \!\(\"A : {t,j,p | 1<=t} of integer\"\)\nthe dimension should \
be one plus the number of parameters"\)], "Message"],

Cell[BoxData[
    \(checkCell::"affexp1" \(\(:\)\(\ \)\) 
      "In \!\(\"(t,p->t,t-1,p)\"\), the matrix should be a translation \
matrix"\)], "Message"],

Cell[BoxData[
    \(checkCell::"assignment" \(\(:\)\(\ \)\) 
      "Error in equation\n \!\(\"a = 0.(t,p->) + A.(t,p->t,t-1,p);\"\)\n\
(check dimensions of variable or constant)."\)], "Message"],

Cell[BoxData[
    \("There is a Mistake somewhere..."\)], "Print"],

Cell[BoxData[
    \(False\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(a2v[$result]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Vhdl files will be generated in \
"\[InvisibleSpace]"/Users/quinton/mmalpha/doc/Packages/CheckAlpHard/"\),
      SequenceForm[ 
      "Vhdl files will be generated in ", 
        "/Users/quinton/mmalpha/doc/Packages/CheckAlpHard/"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("\nPlease wait...\n"\)], "Print"],

Cell[BoxData[
    \(a2v::"unknsys" \(\(:\)\(\ \)\) 
      "System \!\(\"cell1e1\"\) is neither recognized as a cell, a module or \
a controller. To find out the problem, use the following sequence\n \
getSystem[\"\!\(\"cell1e1\"\)\"];checkCell[debug->True];\n \
getSystem[\"\!\(\"cell1e1\"\)\"];checkController[debug->True];\n getSystem[\"\
\!\(\"cell1e1\"\)\"];checkModule[debug->True];\n One of these command might \
explain why your system cannot  be translated to Vhdl. Also, remember that a \
system cannot be tranlsated to Vhdl if the parameter values haven't been  set \
using the fixParameter function. "\)], "Message"],

Cell[BoxData[
    InterpretationBox[\(\(0.032569999999999766`\ Second\)\[InvisibleSpace]" \
needed to translate "\[InvisibleSpace]"cell1e1"\),
      SequenceForm[ 
        Times[ 0.032569999999999766, Second], " needed to translate ", 
        "cell1e1"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("End of translation."\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example 3", "Subsubsection"],

Cell["Error 1:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(load["\<cell1e2.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \("[cell1e2]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \("system cell1e2 :{p | 2<=p<=10}\n               (A : {t | 1<=t} of \
integer)\n       returns (S : {t | 2<=t} of integer);\nvar\n  a : {t | 1<=t} \
of integer;\n  b : {t | 1<=t} of integer;\n  e : {t | 1<=t} of integer;\n  c \
: {t | 2<=t} of integer;\n  d : {t | 2<=t} of integer;\nlet\n  a[t] = 0[] + \
A[t];\n  b[t] = a[t];\n  c[t] = b[t+1];\n  d[t] = a[t-1] + b[t-1] - c[t];\n  \
e[t] = -a[t];\n  S[t] = c[t];\ntel;"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkCell[]\)], "Input"],

Cell[BoxData[
    \(False\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(<< Alpha/CheckCell.sem\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkCell[debug -> True]\)], "Input"],

Cell[BoxData[
    \(checkCell::"affexp3" \(\(:\)\(\ \)\) 
      "In \!\(\"(t,p->t+1,p)\"\), the translation should not be positive"\)], \
"Message"],

Cell[BoxData[
    \(checkCell::"assignment" \(\(:\)\(\ \)\) 
      "Error in equation\n \!\(\"c = b.(t,p->t+1,p);\"\)\n(check dimensions \
of variable or constant)."\)], "Message"],

Cell[BoxData[
    \("There is a Mistake somewhere..."\)], "Print"],

Cell[BoxData[
    \(False\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(a2v[$result]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Vhdl files will be generated in \
"\[InvisibleSpace]"/Users/quinton/mmalpha/doc/Packages/CheckAlpHard/"\),
      SequenceForm[ 
      "Vhdl files will be generated in ", 
        "/Users/quinton/mmalpha/doc/Packages/CheckAlpHard/"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("\nPlease wait...\n"\)], "Print"],

Cell[BoxData[
    \(a2v::"unknsys" \(\(:\)\(\ \)\) 
      "System \!\(\"cell1e2\"\) is neither recognized as a cell, a module or \
a controller. To find out the problem, use the following sequence\n \
getSystem[\"\!\(\"cell1e2\"\)\"];checkCell[debug->True];\n \
getSystem[\"\!\(\"cell1e2\"\)\"];checkController[debug->True];\n getSystem[\"\
\!\(\"cell1e2\"\)\"];checkModule[debug->True];\n One of these command might \
explain why your system cannot  be translated to Vhdl. Also, remember that a \
system cannot be tranlsated to Vhdl if the parameter values haven't been  set \
using the fixParameter function. "\)], "Message"],

Cell[BoxData[
    InterpretationBox[\(\(0.032197999999997506`\ Second\)\[InvisibleSpace]" \
needed to translate "\[InvisibleSpace]"cell1e2"\),
      SequenceForm[ 
        Times[ 0.032197999999997506, Second], " needed to translate ", 
        "cell1e2"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("End of translation."\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example 4", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(load["\<cell2.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \("[cell2]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \("system cell2 : {p | 2<=p<=10}\n               (c : {t | 1<=t} of \
boolean; \n                B : {t | 1<=t} of integer)\n       returns (D : {t \
| 1<=t} of integer);\nvar\n  Cloc : {t | 1<=t} of integer;\nlet\n  Cloc[t] = \
\n      case\n        { | t=p} : if (c[t]) then 0[] else 0[];\n        { | \
p+1<=t} : if (c[t]) then 0[] else B[t] * B[t-1];\n      esac;\ntel;"\)], \
"Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkCell[]\)], "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkCell[debug -> True]\)], "Input"],

Cell[BoxData[
    \("This program seems to be an AlpHard Cell..."\)], "Print"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(a2v[$result]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Vhdl files will be generated in \
"\[InvisibleSpace]"/Users/quinton/mmalpha/doc/Packages/CheckAlpHard/"\),
      SequenceForm[ 
      "Vhdl files will be generated in ", 
        "/Users/quinton/mmalpha/doc/Packages/CheckAlpHard/"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("\nPlease wait...\n"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("cell2"\[InvisibleSpace]" was recognized as a \
Cell."\),
      SequenceForm[ "cell2", " was recognized as a Cell."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("All modules are known..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Vhdl code was written in file \
"\[InvisibleSpace]"cell2.vhd"\),
      SequenceForm[ "Vhdl code was written in file ", "cell2.vhd"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Component description :"\)], "Print"],

Cell[BoxData[
    \("\nCOMPONENT cell2\nPORT(\n  clk: IN STD_LOGIC;\n  CE : IN STD_LOGIC;\n \
 Rst : IN STD_LOGIC;\n  c : IN  STD_LOGIC;\n  B : IN  SIGNED (15 DOWNTO 0);\n \
 D : OUT  SIGNED (15 DOWNTO 0)\n);\nEND COMPONENT;\n"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(\(0.025572000000000372`\ Second\)\[InvisibleSpace]" \
needed to translate "\[InvisibleSpace]"cell2"\),
      SequenceForm[ 
        Times[ 0.025572000000000372, Second], " needed to translate ", 
        "cell2"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("End of translation."\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example 5", "Subsubsection"],

Cell["This example has an error", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(load["\<cell2e1.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \("[cell2e1]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \("system cell2e1 :{p | 2<=p<=10}\n               (c : {t | 1<=t} of \
boolean; \n                B : {t | 1<=t} of integer)\n       returns (D : {t \
| 1<=t} of integer);\nvar\n  Cloc : {t | 1<=t} of integer;\nlet\n  Cloc[t] = \
\n      case\n        { | t=p} : if (c[t]) then 0[] else 0[];\n        { | \
p+1<=t} : if (c[t]) then 0[] else B[t] * B[t-1];\n        { | t<=p-1} : if \
(c[t]) then 0[] else B[t] * B[t-1];\n      esac;\ntel;"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkCell[]\)], "Input"],

Cell[BoxData[
    \(False\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkCell[debug -> True]\)], "Input"],

Cell[BoxData[
    \(checkCell::"case" \(\(:\)\(\ \)\) 
      "In a case expression, the number of branches should be 2."\)], \
"Message"],

Cell[BoxData[
    \(checkCell::"assignment" \(\(:\)\(\ \)\) 
      "Error in equation\n \!\(\"Cloc = \\n    case\\n      {t,p | t=p} \"\
\[Ellipsis]\"p->t,p) * B.(t,p->t-1,p);\\n    esac;\"\)\n(check dimensions of \
variable or constant)."\)], "Message"],

Cell[BoxData[
    \("There is a Mistake somewhere..."\)], "Print"],

Cell[BoxData[
    \(False\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(a2v[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Vhdl files will be generated in \
"\[InvisibleSpace]"/Users/quinton/mmalpha/doc/Packages/CheckAlpHard/"\),
      SequenceForm[ 
      "Vhdl files will be generated in ", 
        "/Users/quinton/mmalpha/doc/Packages/CheckAlpHard/"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("\nPlease wait...\n"\)], "Print"],

Cell[BoxData[
    \(a2v::"unknsys" \(\(:\)\(\ \)\) 
      "System \!\(\"cell2e1\"\) is neither recognized as a cell, a module or \
a controller. To find out the problem, use the following sequence\n \
getSystem[\"\!\(\"cell2e1\"\)\"];checkCell[debug->True];\n \
getSystem[\"\!\(\"cell2e1\"\)\"];checkController[debug->True];\n getSystem[\"\
\!\(\"cell2e1\"\)\"];checkModule[debug->True];\n One of these command might \
explain why your system cannot  be translated to Vhdl. Also, remember that a \
system cannot be tranlsated to Vhdl if the parameter values haven't been  set \
using the fixParameter function. "\)], "Message"],

Cell[BoxData[
    InterpretationBox[\(\(0.02992900000000276`\ Second\)\[InvisibleSpace]" \
needed to translate "\[InvisibleSpace]"cell2e1"\),
      SequenceForm[ 
        Times[ 0.029929000000002759, Second], " needed to translate ", 
        "cell2e1"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("End of translation."\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example 6", "Subsubsection"],

Cell["This example has another error", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(load["\<cell2e2.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \("[cell2e1]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \("system cell2e1 :{p | 2<=p<=10}\n               (c : {t | 1<=t} of \
boolean; \n                B : {t | 1<=t} of integer)\n       returns (D : {t \
| 1<=t} of integer);\nvar\n  Cloc : {t | 1<=t} of integer;\nlet\n  Cloc[t] = \
\n      case\n        { | t=p} : if (B[t]) then 0[] else 0[];\n        { | \
p+1<=t} : if (c[t]) then 0[] else B[t] * B[t-1];\n      esac;\ntel;"\)], \
"Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(schematics[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("While processing element: "\[InvisibleSpace]if[
          affine[var["B"], 
            matrix[3, 3, {"t", "p"}, {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}]], 
          affine[const[0], matrix[1, 3, {"t", "p"}, {{0, 0, 1}}]], 
          affine[const[0], matrix[1, 3, {"t", "p"}, {{0, 0, 1}}]]]\),
      SequenceForm[ "While processing element: ", 
        if[ 
          affine[ 
            var[ "B"], 
            matrix[ 3, 3, {"t", "p"}, {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}]], 
          affine[ 
            const[ 0], 
            matrix[ 1, 3, {"t", "p"}, {{0, 0, 1}}]], 
          affine[ 
            const[ 0], 
            matrix[ 1, 3, {"t", "p"}, {{0, 0, 1}}]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \(skeleton::"params" \(\(:\)\(\ \)\) "wrong paramaters"\)], "Message"],

Cell[BoxData[
    InterpretationBox[\("While processing element: "\[InvisibleSpace]if[
          affine[var["c"], 
            matrix[3, 3, {"t", "p"}, {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}]], 
          affine[const[0], matrix[1, 3, {"t", "p"}, {{0, 0, 1}}]], 
          binop[mul, 
            affine[var["B"], 
              matrix[3, 3, {"t", "p"}, {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}]], 
            affine[var["B"], 
              matrix[3, 
                3, {"t", "p"}, {{1, 0, \(-1\)}, {0, 1, 0}, {0, 0, 1}}]]]]\),
      SequenceForm[ "While processing element: ", 
        if[ 
          affine[ 
            var[ "c"], 
            matrix[ 3, 3, {"t", "p"}, {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}]], 
          affine[ 
            const[ 0], 
            matrix[ 1, 3, {"t", "p"}, {{0, 0, 1}}]], 
          binop[ mul, 
            affine[ 
              var[ "B"], 
              matrix[ 3, 3, {"t", "p"}, {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}]], 
            affine[ 
              var[ "B"], 
              matrix[ 
              3, 3, {"t", "p"}, {{1, 0, -1}, {0, 1, 0}, {0, 0, 1}}]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \(skeleton::"params" \(\(:\)\(\ \)\) "wrong paramaters"\)], "Message"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .28 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
5.55112e-17 0.4 0 0.4 [
[ 0 0 0 0 ]
[ 1 .28 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .28 L
0 .28 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.6 .088 m
.76 .088 L
.76 .152 L
.6 .152 L
.6 .088 L
s
gsave
.45 .104 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.45 .104 m
.6 .104 L
s
gsave
.596 .112 -89 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Null) show
1.000 setlinewidth
grestore
gsave
.45 .136 -66 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding MacintoshRomanEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Mathematica3Mono
%%IncludeFont: Mathematica3Mono
/Mathematica3Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\350) show
69.000 13.000 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.45 .136 m
.6 .136 L
s
gsave
.596 .144 -89 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Null) show
1.000 setlinewidth
grestore
0 0 1 r
gsave
.68 .12 -81 -12.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 16.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 15.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 16.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 15.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(case) show
99.000 16.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 15.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 g
gsave
.76 .136 -61 -10 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Cloc) show
87.000 13.000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.76 .12 m
.8 .12 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 80.625},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000N2000`400?l00000o`00003ooooooona
ooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooona
ooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooona
ooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooona
ooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooona
ooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooona
ooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooona
ooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooonaooooo`00oooooona
ooooo`00oooooonaooooo`00ooooool3ooooodGo0000JOooool00?oooooo0oooool00ol0003ooooo
ooooo`11ooooo`03o`000?oooooooooo06Oooooo003oooooo`?ooooo00?o0000ooooooooool0@Ooo
ool00ol0003oooooooooo`1Wooooo`00ooooool3ooooo`03o`000?oooooooooo047ooooo00?o0000
ooooooooool0Ioooool00?oooooo0oooool00ol0003oooooooooo`11ooooo`03o`000?oooooooooo
06Oooooo002oooooo`06onk^k_mEEEGo0000o`000?mVIVKok^k^?Oooool00ol0003oooooooooo`11
ooooo`03o`000?oooooooooo06Oooooo002oooooo`03oeEEEOl0003o000000;o000000?oIVIVoooo
ooooool0>oooool00ol0003oooooooooo`11ooooo`03o`000?oooooooooo06Oooooo002oooooodCo
0000@oooool00ol0003oooooooooo`1Wooooo`00_oooool6o`0001Kooooo00Ko0000oooooooooooo
oooooa0@4?l00003ooooo`06oa0@4?mgMgOoooooonk^k_mEEEGo00001Oooool00ol0003ooooooooo
o`06ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool0@Oooool00ol0003ooooooooo
o`1Wooooo`00_oooool00omEEEGo0000o`000002o`000003ofIVI_oooooooooo01Cooooo00Ko0000
ooooooooooooV9RHoeEEEOl00003ooooo`06o`000?ooooooooooooooooo^k^ko00001Oooool00ol0
003oooooooooo`06ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool0@Oooool00ol0
003oooooooooo`1Wooooo`00_oooool01_o^k^koEEEEo`000?l0003oIVIVonk^kQKooooo00Ko0000
oooooooooooo410@omgMgOl00003ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool0
0oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool0
1oooool00oo<c?ooEEGoo`00o`02o`00o`03oc8boonHV?oooooo00Kooooo00SoMgOooa0@ool00?oo
003oodA4oonj^_ooooooofIVo`;o003o00So8B7ooooooooooooo8B7ooc8boonHV?ooA4Cooa0@o`;o
003o00?o<S;oojVYooooool01Oooool00onj^_ooEEGooa0@o`03o`00o`04oc8boomEEOooV9Soonk^
o`Kooooo00?o0000ooooooooool0Ioooool00=_ooooo00Ko0000oooooon7QhOoMgMgooooool00003
ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`06
ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool01_ooool00onHV?oo003oo`00o`05
o`00o`03ob4Qooo^k_oooooo00?ooooo00?oIVKoo`00ool00?l01?l00?l01OmVI_oo8B7oo`00ool0
0?oo8B7o00;ooooo2Ol00?l00omgMoooooooooooo`02ooooo`03ofIVool00?oo003o00Oo003o00?o
413oooooooooool01?ooool00ol0003oooooooooo`1Wooooo`00foooool01_l0003ok^k^oa0@4?o^
k^koooooo`0000?ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`03ooooo`03o`00
0?oooooooooo00Kooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`05ooooo`0;ojVY
ool00?oo003oogMgoooMgOooooooonk^oonHV?oo413oo`00oolQ8Ol01?ooool2o`00o`06ojVYoooo
ooooooooonk^oonYZOooA4Co0_l00?l4ooooo`;o003o00Ko<S;oojVYooooooooooooonk^oomgMol2
o`00o`?ooooo00GoZJWoo`00ool@4?ooV9SoomgMo`0=ooooo`03o`000?oooooooooo06Oooooo003K
ooooo`03o`000?mgMgOoMgMg00;ooooo00Go0000oooooooooooo410@o`000002ooooo`03oa0@4?l0
003o000000Gooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`07ooooo`03o`000?oo
oooooooo00Gooooo00?oA4Coo`00oon7Qol01Oooool00oo^k_ooA4Cooa0@o`04ooooo`;o003o00?o
MgOoolc<ooooool00_ooool00ooMgOoo003oo`00o`04ooooo`03o`00oomEEOoooooo00Cooooo00?o
k^koo`00ool00?l00oooool00olQ8Ooo003oomgMo`0?ooooo`03o`000?oooooooooo06Oooooo003K
ooooo`03o`000?l@413ok^k^00;ooooo00?o0000ooooooooool03?ooool00ol0003oooooooooo`06
ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool01Oooool2o`00o`03onk^oooooooo
oooo00[ooooo00?oV9Soo`00ool00?l01ol00?l5ooooo`03omgMoon7QoooEEGo00?o<S;o0_l00?l0
1?mVI_ooooooooooooooool;o`00o`03ob4Qoooooooooooo00Cooooo4ol0001Gooooo`00f_ooool0
1Ol@413o0000ohN7Qooooooo410@00;o000000?o410@ooooooooool02oooool00ol0003ooooooooo
o`06ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool01Oooool2o`00o`03onk^oooo
oooooooo00_ooooo00?o^[[ooeEEool@4?l00ol00?l00om4A?oo003oo`00o`04ooooo`03oiRHool0
0?oo003o00Co003o00?o<S;ooiRHooooool00oooool<o`00o`Kooooo00?o0000ooooooooool0Iooo
ool00>gooooo00?o410@o`000?l000001_ooool00ol@413o0000o`000009ooooo`03o`000?oooooo
oooo00Gooooo00?o<S;oo`00oon7Qol01Oooool00oo<c?oo413oo`00o`0;ooooo`03omgMool00?oo
003o00Cooooo0_l00?l00onYZOooooooooooo`02ooooo`03oiRHool00?oooooo00?ooooo00?oA4Co
o`00oonYZOl01_ooool00oo<c?oo003ooc8bo`06ooooo`03o`000?oooooooooo06Oooooo003ooooo
o`?ooooo00?o0000ooooooooool01Oooool01OnYZOoo003oo`00oomgMooogMgo00;ooooo00CoV9So
o`00ool00?oo003o1Oooool2ob4Qo`07oiRHooo^k_ooooooomgMoolb<_oo003ooc8bo`04ooooo`;o
003o00GoQhOoonk^oooooooogMgoodA4o`02o`00o`Cooooo00Go^[[oo`00ool00?ooMgOoolc<o`02
ooooo`05omgMoonHV?oo003oo`00oonj^_l01_ooool00ol0003oooooooooo`02ooooo`06olc<c?m4
A4Co0000o`000?m4A4CogMgM0_ooool00ol@413o0000o`000004o`0000?ooooo00GoQhN7oa0@4?l0
003o410@ohN7Q`04ooooo`06ojVYZOl@413o0000o`000?mEEEGogMgMA?ooool00;oooooo00Kok^k^
oeEEEOl0003o0000ofIVI_o^k^hmooooo`03o`000?oooooooooo00Kooooo00?oV9Soo`00ool00?l0
1ol00?l5ooooo`03oa0@ool00?oo003o00Go003o00?oZJWoooooooooool00_ooool00omVI_oo003o
o`00o`06o`00o`Gooooo00?oV9Soo`00ool00?l01_l00?l00onHV?ooooooooooo`05ooooo`0;o`00
0?oooooooooooooooooMgMgo410@oiRHV?ooooook^k^ogMgMolQ8B401Oooool00ol0003ooooooooo
o`03ooooo`07ohN7QolQ8B7oc<c<ooooooo<c<co8B4QohN7Q`02ooooo`03ojVYZOlQ8B7o^[Zj00;o
oooo00?o^[Zjob4Q8Oooool0@oooool00;oooooo00?oEEEEo`000?l000000_l000000omVIVKooooo
ooooo`0kooooo`03o`000?oooooooooo00Oooooo00?o^[[oodA4ool00?l00_l00?l01?m4A?oo^[[o
oa0@oolQ8Ol5ooooo`03onk^oomgMooo<S;o00?o003o00?oA4CookZjooooool01Oooool02?nHV?oo
<S;oo`00ool00?oo413oofIVoomEEOoo8B7o1_ooool00ooMgOooIVKoob4Qo`02o`00o`03oa0@oomE
EOoogMgo00Sooooo00Ko0000oooooooooooooooooeEEEOn7QhL:ooooo`03o`000?oooooooooo00?o
oooo00?o410@olc<c?ooool00_ooool01_o<c<co410@oooooooooooo410@okZj^TWooooo002ooooo
o`Ko0000?Oooool00ol0003oooooooooo`11ooooo`06o`000?ooooooooooooooool0003ok^k^2_oo
ool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00?ooooo00Co0000oooooooooooo0000
B_ooool00;ooooooA?l00013ooooo`05o`000?ooooooooooooooool000002oooool00ol0003ooooo
ooooo`03ooooo`03oa0@4?nj^[[ooooo00;ooooo00Ko^[Zjoa0@4?oooooooooooa0@4?nj^[Y9oooo
o`00_oooool00omEEEGo0000o`000002o`000003ofIVI_oooooooooo01Cooooo00Ko0000oooooooo
oooooooooa0@4?l00003ooooo`06oa0@4?mgMgOoooooonk^k_mEEEGo00001Oooool00ol0003ooooo
ooooo`06ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool0@Oooool01_l0003ooooo
oooooooooooo0000onk^kP[ooooo00?o0000ooooooooool00oooool01on7QhOo8B4Qolc<c?oooooo
c<c<ob4Q8On7QhL00_ooool01onYZJWo8B4QokZj^_ooooook^k^ohN7Qol00000A?ooool00;oooooo
00Kok^k^oeEEEOl0003o0000ofIVI_o^k^hFooooo`06o`000?oooooooooooiRHV?mEEEGo00000ooo
ool01_l0003oooooooooooooooook^k^o`0000Gooooo00?o0000ooooooooool01_ooool00ol0003o
ooooooooo`07ooooo`03o`000?oooooooooo047ooooo00Ko0000oooooooooooooooooeEEEOmgMgL3
ooooo`03onk^k_l0003ooooo00Cooooo00?o0000ooooooooool01?ooool01On7QhOo410@o`000?l@
413oQhN700Cooooo00KoZJVYob4Q8Ol0003o0000odA4A?l00014ooooo`00foooool01_l0003ooooo
ooooool@413ogMgMo`0000?ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`03oooo
o`03o`000?oooooooooo00Kooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`11oooo
o`07o`000?ooooooooooooooooo^k^ko410@oiRHV002ooooo`03ohN7Qol0003ooooo00Cooooo00?o
0000ooooooooool0Eoooool00=_ooooo00Ko0000oooooon7QhOoMgMgooooool00003ooooo`03o`00
0?oooooooooo00;ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`06ooooo`03o`00
0?oooooooooo00Oooooo00?o0000ooooooooool0@Oooool00ol0003oooooooooo`02ooooo`06omgM
gOm4A4Co0000o`000?lb<S;o00001Oooool00ol0003oooooooooo`1Gooooo`00foooool01_l0003o
k^k^oa0@4?o^k^koooooo`0000?ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`03
ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`11
ooooo`03o`000?oooooooooo00_ooooo00?o410@o`000?l00000FOooool00=_ooooo00?o0000ogMg
MomgMgL00_ooool01Ol0003oooooooooool@413o000000;ooooo00?o410@o`000?l000001Oooool0
0ol0003oooooooooo`06ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool0@Oooool0
0ol0003oooooooooo`1Wooooo`00foooool00ol0003o410@onk^kP02ooooo`03o`000?oooooooooo
00cooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`07ooooodGo0000JOooool00=[o
oooo00Go410@o`000?n7QhOooooooa0@4002o`000003oa0@4?oooooooooo00_ooooo00?o0000oooo
ooooool01_ooool00ol0003oooooooooo`2eooooo`00kOooool00ol@413o0000o`000006ooooo`03
oa0@4?l0003o00000;Oooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooo
ok7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooo
ok7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooo
ok7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooo
ok7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooo
ok7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooo
ok7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooo
ok7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooo
ok7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooo
ok7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo003ooooo
ok7ooooo003ooooook7ooooo003ooooook7ooooo003ooooook7ooooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {119.938, 0}} -> {-0.00775098, -3.50004*^-6, \
0.00875465, 0.00875465}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkCell[debug -> True]\)], "Input"],

Cell[BoxData[
    \("This program seems to be an AlpHard Cell..."\)], "Print"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(a2v[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Vhdl files will be generated in \
"\[InvisibleSpace]"/Users/quinton/mmalpha/doc/Packages/CheckAlpHard/"\),
      SequenceForm[ 
      "Vhdl files will be generated in ", 
        "/Users/quinton/mmalpha/doc/Packages/CheckAlpHard/"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("\nPlease wait...\n"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("cell2e1"\[InvisibleSpace]" was recognized as a \
Cell."\),
      SequenceForm[ "cell2e1", " was recognized as a Cell."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("All modules are known..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Vhdl code was written in file \
"\[InvisibleSpace]"cell2e1.vhd"\),
      SequenceForm[ "Vhdl code was written in file ", "cell2e1.vhd"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Component description :"\)], "Print"],

Cell[BoxData[
    \("\nCOMPONENT cell2e1\nPORT(\n  clk: IN STD_LOGIC;\n  CE : IN STD_LOGIC;\
\n  Rst : IN STD_LOGIC;\n  c : IN  STD_LOGIC;\n  B : IN  SIGNED (15 DOWNTO \
0);\n  D : OUT  SIGNED (15 DOWNTO 0)\n);\nEND COMPONENT;\n"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(\(0.025470999999999577`\ Second\)\[InvisibleSpace]" \
needed to translate "\[InvisibleSpace]"cell2e1"\),
      SequenceForm[ 
        Times[ 0.025470999999999577, Second], " needed to translate ", 
        "cell2e1"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("End of translation."\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example 7", "Subsubsection"],

Cell["Load the final result of the Fir example:", "Text"],

Cell[BoxData[
    \(\(load["\<firHard48man.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(systemNames[]\)], "Input"],

Cell[BoxData[
    \(\(getSystem["\<cellfirModule1\>"];\)\)], "Input"],

Cell[BoxData[
    \(checkCell[debug -> True]\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \(\(getSystem["\<cellfirModule3\>"];\)\)], "Input"],

Cell[BoxData[
    \(checkCell[debug -> True]\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \(Options[schematics]\)], "Input"],

Cell[BoxData[
    \(schematics[sPositions -> {columns -> 2}, offsetX -> 0.6, 
      offsetY -> 0.2, partShown -> {{0. , 1. }, {0. , 1. }}]\)], "Input"],

Cell[BoxData[
    \(\(getSystem["\<cellfirModule4\>"];\)\)], "Input"],

Cell[BoxData[
    \(checkCell[debug -> True]\)], "Input"],

Cell[BoxData[
    \(\(getSystem["\<firModule\>"];\)\)], "Input"],

Cell[BoxData[
    \(checkCell[]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Example 8", "Subsubsection"],

Cell["Load the final result of the MatVect example:", "Text"],

Cell[BoxData[
    \(\(load[$rootDirectory <> \
"\</demos/NOTEBOOKS/Matvect/prodVectHard10man.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(\(getSystem["\<cellprodVectModule1\>"];\)\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \(checkCell[]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Example 9 (error)",
  FontColor->RGBColor[1, 0, 0]]], "Subsubsection"],

Cell[BoxData[
    \(\(load["\<lyon.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(systemNames[]\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \(checkCell[debug -> True]\)], "Input"]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["4. Examples of modules", "Section",
  FontColor->RGBColor[0, 0, 1],
  CellTags->"How to define"],

Cell["\<\
These commands are necessary only it you do a modification to CheckCell.meta. \
Don't worry about the error messages... Before you try CheckCell in this \
notebook, do not forget to set the directory (re-execute the red cell at the \
beginning of this notebook).\
\>", "Text",
  TextAlignment->Left,
  TextJustification->1],

Cell[CellGroupData[{

Cell["To compile rules for CheckModule.meta", "Subsubsection"],

Cell["\<\
These commands are necessary only it you do a modification to \
CheckModuke.meta. Don't worry about the error messages... Before you try \
CheckModule in this notebook, do not forget to set the directory (re-execute \
the red cell at the beginning of this notebook).\
\>", "Text",
  TextAlignment->Left,
  TextJustification->1],

Cell[BoxData[
    \(setMMADir[{"\<lib/Packages/Alpha\>"}]\)], "Input"],

Cell["\<\
First, compile the meta file. Don't worry about the error \
messages...\
\>", "Text"],

Cell[BoxData[
    \(\(meta["\<CheckModule\>"]; \)\)], "Input"],

Cell["Then load the result. ", "Text"],

Cell[BoxData[
    \(<< CheckModule.m\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example 1", "Subsubsection"],

Cell[BoxData[
    \(\(load["\<firHard48man.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(flattenSkeleton[]\)], "Input"],

Cell[BoxData[
    \(g = 
      schematics[sPositions -> {columns -> 3}, 
        partShown -> {{0. , 1. }, {0. , 1. }}, fSize -> 10]\)], "Input"],

Cell[BoxData[
    \(\(getSystem["\<firModule\>"];\)\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \(schematics[fSize -> 10]\)], "Input"],

Cell[BoxData[
    \(Show[g, PlotRange -> {{1, 2}, {0, 2}}]\)], "Input"],

Cell[BoxData[
    \(checkModule[debug -> True]\)], "Input"],

Cell[BoxData[
    \(\(getSystem["\<cellfirModule1\>"];\)\)], "Input"],

Cell[BoxData[
    \(checkModule[]\)], "Input"],

Cell[BoxData[
    \(checkModule[debug -> True]\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Example 2", "Subsubsection"],

Cell["Load the final result of the MatVect example:", "Text"],

Cell[BoxData[
    \(\(load[$rootDirectory <> \
"\</demos/NOTEBOOKS/Matvect/prodVectHard10man.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(systemNames[]\)], "Input"],

Cell[BoxData[
    \(\(getSystem["\<prodVectModule\>"];\)\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \(checkModule[]\)], "Input"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["5. Examples of controllers", "Section",
  FontColor->RGBColor[0, 0, 1],
  CellTags->"How to define"],

Cell[CellGroupData[{

Cell["To compile rules for CheckController.meta", "Subsubsection"],

Cell["\<\
These commands are necessary only it you do a modification to \
CheckController.meta. Don't worry about the error messages... Before you try \
CheckController in this notebook, do not forget to set the directory \
(re-execute the red cell at the beginning of this notebook).\
\>", "Text",
  TextAlignment->Left,
  TextJustification->1],

Cell[BoxData[
    \(setMMADir[{"\<lib/Packages/Alpha\>"}]\)], "Input"],

Cell["\<\
First, compile the meta file. Don't worry about the error \
messages...\
\>", "Text"],

Cell[BoxData[
    \(\(meta["\<CheckController\>"];\)\)], "Input"],

Cell["Then load the result. ", "Text"],

Cell[BoxData[
    \(<< CheckController.m\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example 1", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(load["\<firHard48man.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \("[ControlfirModule][cellfirModule1][cellfirModule3][cellfirModule4][\
firModule]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"]
}, Open  ]],

Cell[BoxData[
    \(\(getSystem["\<ControlfirModule\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkController[debug -> True]\)], "Input"],

Cell[BoxData[
    \("This program seems to be an AlpHard Controller..."\)], "Print"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example 2", "Subsubsection"],

Cell["Load the final result of the MatVect example:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(load[$rootDirectory <> \
"\</demos/NOTEBOOKS/Matvect/prodVectHard10man.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(load::"notFound" \(\(:\)\(\ \)\) 
      " File \!\(\"/Users/quinton/mmalpha/demos/NOTEBOOKS/Matvect/\
prodVectHard10man.alpha\"\) not found "\)], "Message"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(systemNames[]\)], "Input"],

Cell[BoxData[
    \({"ControlfirModule", "cellfirModule1", "cellfirModule3", 
      "cellfirModule4", "firModule"}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(getSystem["\<ControlprodVectModule\>"];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("ControlprodVectModule"\[InvisibleSpace]" : not found \
in the library."\),
      SequenceForm[ "ControlprodVectModule", " : not found in the library."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkController[]\)], "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(getSystem["\<cellprodVectModule1\>"];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("cellprodVectModule1"\[InvisibleSpace]" : not found \
in the library."\),
      SequenceForm[ "cellprodVectModule1", " : not found in the library."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \("system ControlfirModule ( )\n       returns          (HPipe_ctl1_Init \
: {t | t=1} | {t | 2<=t<=5} of boolean; \n                         \
xPipe_ctl2_Init : {t | t=2} | {t | 3<=t<=6} of boolean);\nlet\n  \
HPipe_ctl1_Init[t] = \n      case\n        { | t=1} : True[];\n        { | \
2<=t<=5} : False[];\n      esac;\n  xPipe_ctl2_Init[t] = \n      case\n       \
 { | t=2} : True[];\n        { | 3<=t<=6} : False[];\n      esac;\ntel;"\)], \
"Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(checkController[]\)], "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell["This is the end of this notebook...", "Subsubtitle",
  FontColor->RGBColor[1, 0, 0]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Macintosh",
ScreenRectangle->{{0, 1440}, {0, 835}},
AutoGeneratedPackage->None,
WindowToolbars->"EditBar",
WindowSize->{1016, 664},
WindowMargins->{{1, Automatic}, {Automatic, 40}},
PrintingCopies->1,
PrintingPageRange->{1, 1},
IndexCreationOptions->{"Format"->"Hyperlinks"},
Magnification->1.5
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "firstcell"->{
    Cell[1776, 53, 123, 5, 204, "Title",
      CellTags->"firstcell"]},
  "Introduction"->{
    Cell[2922, 105, 94, 2, 57, "Section",
      CellTags->"Introduction"],
    Cell[5246, 173, 125, 2, 57, "Section",
      CellTags->"Introduction"]},
  "How to define"->{
    Cell[6614, 225, 100, 2, 57, "Section",
      CellTags->"How to define"],
    Cell[192571, 5734, 102, 2, 104, "Section",
      CellTags->"How to define"],
    Cell[195157, 5846, 106, 2, 57, "Section",
      CellTags->"How to define"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"firstcell", 199755, 6026},
  {"Introduction", 199847, 6029},
  {"How to define", 200020, 6034}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 123, 5, 204, "Title",
  CellTags->"firstcell"],
Cell[1902, 60, 37, 0, 45, "Subsubtitle"],
Cell[1942, 62, 75, 0, 40, "SmallText"],

Cell[CellGroupData[{
Cell[2042, 66, 74, 0, 45, "Subsubtitle"],

Cell[CellGroupData[{
Cell[2141, 70, 72, 1, 41, "Subsubsection"],
Cell[2216, 73, 361, 14, 304, "SmallText"]
}, Closed]],

Cell[CellGroupData[{
Cell[2614, 92, 142, 3, 30, "Subsubsection",
  InitializationCell->True],
Cell[2759, 97, 126, 3, 42, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[2922, 105, 94, 2, 57, "Section",
  CellTags->"Introduction"],
Cell[3019, 109, 505, 14, 99, "Text"],
Cell[3527, 125, 452, 10, 99, "Text"],
Cell[3982, 137, 732, 16, 126, "Text"],
Cell[4717, 155, 492, 13, 73, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[5246, 173, 125, 2, 57, "Section",
  CellTags->"Introduction"],
Cell[5374, 177, 110, 2, 40, "SmallText"],
Cell[5487, 181, 114, 4, 47, "Text"],
Cell[5604, 187, 441, 13, 177, "Text"],
Cell[6048, 202, 197, 4, 73, "Text"],
Cell[6248, 208, 132, 4, 47, "Text"],
Cell[6383, 214, 141, 4, 47, "Text"],
Cell[6527, 220, 50, 0, 47, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[6614, 225, 100, 2, 57, "Section",
  CellTags->"How to define"],

Cell[CellGroupData[{
Cell[6739, 231, 60, 0, 41, "Subsubsection"],
Cell[6802, 233, 333, 7, 99, "Text"],

Cell[CellGroupData[{
Cell[7160, 244, 70, 1, 39, "Input"],
Cell[7233, 247, 77, 1, 40, "Output"]
}, Open  ]],
Cell[7325, 251, 42, 1, 39, "Input"],

Cell[CellGroupData[{
Cell[7392, 256, 59, 1, 39, "Input"],
Cell[7454, 259, 303, 4, 30, "Message"],
Cell[7760, 265, 311, 4, 30, "Message"],
Cell[8074, 271, 318, 5, 30, "Message"],
Cell[8395, 278, 331, 5, 49, "Message"],
Cell[8729, 285, 369, 5, 49, "Message"],
Cell[9101, 292, 363, 5, 49, "Message"],
Cell[9467, 299, 358, 5, 49, "Message"],
Cell[9828, 306, 333, 5, 49, "Message"],
Cell[10164, 313, 46, 1, 34, "Print"],
Cell[10213, 316, 79, 1, 34, "Print"],
Cell[10295, 319, 221, 4, 34, "Print"],
Cell[10519, 325, 2460, 37, 418, "Print"]
}, Open  ]],
Cell[12994, 365, 38, 0, 47, "Text"],
Cell[13035, 367, 47, 1, 39, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[13119, 373, 34, 0, 41, "Subsubsection"],
Cell[13156, 375, 65, 0, 47, "Text"],

Cell[CellGroupData[{
Cell[13246, 379, 61, 1, 39, "Input"],
Cell[13310, 382, 42, 1, 34, "Print"],
Cell[13355, 385, 50, 1, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[13442, 391, 40, 1, 39, "Input"],
Cell[13485, 394, 452, 6, 418, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[13974, 405, 45, 1, 39, "Input"],
Cell[14022, 408, 136941, 3995, 295, 85033, 3350, "GraphicsData", \
"PostScript", "Graphics"],
Cell[150966, 4405, 130, 3, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[151133, 4413, 42, 1, 39, "Input"],
Cell[151178, 4416, 36, 1, 34, "Print"],
Cell[151217, 4419, 189, 4, 34, "Print"],
Cell[151409, 4425, 71, 1, 34, "Print"],
Cell[151483, 4428, 69, 1, 34, "Print"],
Cell[151555, 4431, 84, 1, 34, "Print"],
Cell[151642, 4434, 76, 1, 34, "Print"],
Cell[151721, 4437, 82, 1, 34, "Print"],
Cell[151806, 4440, 318, 6, 58, "Print"],
Cell[152127, 4448, 318, 6, 58, "Print"],
Cell[152448, 4456, 91, 1, 34, "Print"],
Cell[152542, 4459, 171, 3, 34, "Print"],
Cell[152716, 4464, 171, 3, 34, "Print"],
Cell[152890, 4469, 171, 3, 34, "Print"],
Cell[153064, 4474, 171, 3, 34, "Print"],
Cell[153238, 4479, 171, 3, 34, "Print"],
Cell[153412, 4484, 171, 3, 34, "Print"],
Cell[153586, 4489, 76, 1, 34, "Print"],
Cell[153665, 4492, 59, 1, 58, "Print"],
Cell[153727, 4495, 38, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[153802, 4501, 44, 1, 39, "Input"],
Cell[153849, 4504, 38, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[153924, 4510, 57, 1, 39, "Input"],
Cell[153984, 4513, 78, 1, 34, "Print"],
Cell[154065, 4516, 38, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[154140, 4522, 45, 1, 39, "Input"],
Cell[154188, 4525, 305, 6, 58, "Print"],
Cell[154496, 4533, 53, 1, 82, "Print"],
Cell[154552, 4536, 189, 4, 34, "Print"],
Cell[154744, 4542, 169, 2, 58, "Print"],
Cell[154916, 4546, 59, 1, 34, "Print"],
Cell[154978, 4549, 205, 4, 34, "Print"],
Cell[155186, 4555, 58, 1, 34, "Print"],
Cell[155247, 4558, 218, 3, 274, "Print"],
Cell[155468, 4563, 284, 6, 34, "Print"],
Cell[155755, 4571, 54, 1, 34, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[155858, 4578, 81, 1, 41, "Subsubsection"],
Cell[155942, 4581, 24, 0, 47, "Text"],

Cell[CellGroupData[{
Cell[155991, 4585, 63, 1, 39, "Input"],
Cell[156057, 4588, 44, 1, 34, "Print"],
Cell[156104, 4591, 50, 1, 34, "Print"]
}, Open  ]],
Cell[156169, 4595, 89, 2, 47, "Text"],

Cell[CellGroupData[{
Cell[156283, 4601, 40, 1, 39, "Input"],
Cell[156326, 4604, 459, 6, 418, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[156822, 4615, 44, 1, 39, "Input"],
Cell[156869, 4618, 39, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[156945, 4624, 57, 1, 39, "Input"],
Cell[157005, 4627, 188, 3, 68, "Message"],
Cell[157196, 4632, 153, 3, 30, "Message"],
Cell[157352, 4637, 194, 3, 68, "Message"],
Cell[157549, 4642, 66, 1, 34, "Print"],
Cell[157618, 4645, 39, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[157694, 4651, 45, 1, 39, "Input"],
Cell[157742, 4654, 305, 6, 58, "Print"],
Cell[158050, 4662, 53, 1, 82, "Print"],
Cell[158106, 4665, 627, 9, 163, "Message"],
Cell[158736, 4676, 289, 6, 34, "Print"],
Cell[159028, 4684, 54, 1, 34, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[159131, 4691, 34, 0, 41, "Subsubsection"],
Cell[159168, 4693, 24, 0, 47, "Text"],

Cell[CellGroupData[{
Cell[159217, 4697, 63, 1, 39, "Input"],
Cell[159283, 4700, 44, 1, 34, "Print"],
Cell[159330, 4703, 50, 1, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[159417, 4709, 40, 1, 39, "Input"],
Cell[159460, 4712, 453, 6, 418, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[159950, 4723, 44, 1, 39, "Input"],
Cell[159997, 4726, 39, 1, 40, "Output"]
}, Open  ]],
Cell[160051, 4730, 55, 1, 39, "Input"],

Cell[CellGroupData[{
Cell[160131, 4735, 57, 1, 39, "Input"],
Cell[160191, 4738, 148, 3, 30, "Message"],
Cell[160342, 4743, 180, 3, 68, "Message"],
Cell[160525, 4748, 66, 1, 34, "Print"],
Cell[160594, 4751, 39, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[160670, 4757, 45, 1, 39, "Input"],
Cell[160718, 4760, 305, 6, 58, "Print"],
Cell[161026, 4768, 53, 1, 82, "Print"],
Cell[161082, 4771, 627, 9, 163, "Message"],
Cell[161712, 4782, 289, 6, 34, "Print"],
Cell[162004, 4790, 54, 1, 34, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[162107, 4797, 34, 0, 41, "Subsubsection"],

Cell[CellGroupData[{
Cell[162166, 4801, 61, 1, 39, "Input"],
Cell[162230, 4804, 42, 1, 34, "Print"],
Cell[162275, 4807, 50, 1, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[162362, 4813, 40, 1, 39, "Input"],
Cell[162405, 4816, 407, 6, 322, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[162849, 4827, 44, 1, 39, "Input"],
Cell[162896, 4830, 38, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[162971, 4836, 57, 1, 39, "Input"],
Cell[163031, 4839, 78, 1, 34, "Print"],
Cell[163112, 4842, 38, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[163187, 4848, 45, 1, 39, "Input"],
Cell[163235, 4851, 305, 6, 58, "Print"],
Cell[163543, 4859, 53, 1, 82, "Print"],
Cell[163599, 4862, 189, 4, 34, "Print"],
Cell[163791, 4868, 59, 1, 34, "Print"],
Cell[163853, 4871, 205, 4, 34, "Print"],
Cell[164061, 4877, 58, 1, 34, "Print"],
Cell[164122, 4880, 240, 3, 298, "Print"],
Cell[164365, 4885, 285, 6, 34, "Print"],
Cell[164653, 4893, 54, 1, 34, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[164756, 4900, 34, 0, 41, "Subsubsection"],
Cell[164793, 4902, 41, 0, 47, "Text"],

Cell[CellGroupData[{
Cell[164859, 4906, 63, 1, 39, "Input"],
Cell[164925, 4909, 44, 1, 34, "Print"],
Cell[164972, 4912, 50, 1, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[165059, 4918, 40, 1, 39, "Input"],
Cell[165102, 4921, 470, 6, 346, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[165609, 4932, 44, 1, 39, "Input"],
Cell[165656, 4935, 39, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[165732, 4941, 57, 1, 39, "Input"],
Cell[165792, 4944, 137, 3, 30, "Message"],
Cell[165932, 4949, 254, 4, 125, "Message"],
Cell[166189, 4955, 66, 1, 34, "Print"],
Cell[166258, 4958, 39, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[166334, 4964, 38, 1, 39, "Input"],
Cell[166375, 4967, 305, 6, 58, "Print"],
Cell[166683, 4975, 53, 1, 82, "Print"],
Cell[166739, 4978, 627, 9, 163, "Message"],
Cell[167369, 4989, 288, 6, 34, "Print"],
Cell[167660, 4997, 54, 1, 34, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[167763, 5004, 34, 0, 41, "Subsubsection"],
Cell[167800, 5006, 46, 0, 47, "Text"],

Cell[CellGroupData[{
Cell[167871, 5010, 63, 1, 39, "Input"],
Cell[167937, 5013, 44, 1, 34, "Print"],
Cell[167984, 5016, 50, 1, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[168071, 5022, 40, 1, 39, "Input"],
Cell[168114, 5025, 408, 6, 322, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[168559, 5036, 45, 1, 39, "Input"],
Cell[168607, 5039, 739, 17, 106, "Print"],
Cell[169349, 5058, 88, 1, 30, "Message"],
Cell[169440, 5061, 1123, 27, 130, "Print"],
Cell[170566, 5090, 88, 1, 30, "Message"],
Cell[170657, 5093, 18115, 478, 133, 8733, 358, "GraphicsData", "PostScript", \
"Graphics"],
Cell[188775, 5573, 130, 3, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[188942, 5581, 57, 1, 39, "Input"],
Cell[189002, 5584, 78, 1, 34, "Print"],
Cell[189083, 5587, 38, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[189158, 5593, 38, 1, 39, "Input"],
Cell[189199, 5596, 305, 6, 58, "Print"],
Cell[189507, 5604, 53, 1, 82, "Print"],
Cell[189563, 5607, 193, 4, 34, "Print"],
Cell[189759, 5613, 59, 1, 34, "Print"],
Cell[189821, 5616, 209, 4, 34, "Print"],
Cell[190033, 5622, 58, 1, 34, "Print"],
Cell[190094, 5625, 242, 3, 298, "Print"],
Cell[190339, 5630, 289, 6, 34, "Print"],
Cell[190631, 5638, 54, 1, 34, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[190734, 5645, 34, 0, 41, "Subsubsection"],
Cell[190771, 5647, 57, 0, 47, "Text"],
Cell[190831, 5649, 68, 1, 42, "Input"],
Cell[190902, 5652, 46, 1, 42, "Input"],
Cell[190951, 5655, 69, 1, 42, "Input"],
Cell[191023, 5658, 57, 1, 42, "Input"],
Cell[191083, 5661, 40, 1, 42, "Input"],
Cell[191126, 5664, 69, 1, 42, "Input"],
Cell[191198, 5667, 57, 1, 42, "Input"],
Cell[191258, 5670, 40, 1, 42, "Input"],
Cell[191301, 5673, 52, 1, 42, "Input"],
Cell[191356, 5676, 151, 2, 68, "Input"],
Cell[191510, 5680, 69, 1, 42, "Input"],
Cell[191582, 5683, 57, 1, 42, "Input"],
Cell[191642, 5686, 64, 1, 42, "Input"],
Cell[191709, 5689, 44, 1, 42, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[191790, 5695, 34, 0, 30, "Subsubsection"],
Cell[191827, 5697, 61, 0, 47, "Text"],
Cell[191891, 5699, 118, 2, 42, "Input"],
Cell[192012, 5703, 74, 1, 35, "Input"],
Cell[192089, 5706, 40, 1, 42, "Input"],
Cell[192132, 5709, 44, 1, 42, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[192213, 5715, 94, 1, 30, "Subsubsection"],
Cell[192310, 5718, 60, 1, 42, "Input"],
Cell[192373, 5721, 46, 1, 42, "Input"],
Cell[192422, 5724, 40, 1, 42, "Input"],
Cell[192465, 5727, 57, 1, 42, "Input"]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{
Cell[192571, 5734, 102, 2, 104, "Section",
  CellTags->"How to define"],
Cell[192676, 5738, 333, 7, 99, "Text"],

Cell[CellGroupData[{
Cell[193034, 5749, 62, 0, 61, "Subsubsection"],
Cell[193099, 5751, 337, 7, 99, "Text"],
Cell[193439, 5760, 70, 1, 42, "Input"],
Cell[193512, 5763, 95, 3, 47, "Text"],
Cell[193610, 5768, 62, 1, 42, "Input"],
Cell[193675, 5771, 38, 0, 47, "Text"],
Cell[193716, 5773, 49, 1, 42, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[193802, 5779, 34, 0, 61, "Subsubsection"],
Cell[193839, 5781, 68, 1, 42, "Input"],
Cell[193910, 5784, 50, 1, 42, "Input"],
Cell[193963, 5787, 145, 3, 68, "Input"],
Cell[194111, 5792, 64, 1, 42, "Input"],
Cell[194178, 5795, 40, 1, 42, "Input"],
Cell[194221, 5798, 56, 1, 42, "Input"],
Cell[194280, 5801, 71, 1, 42, "Input"],
Cell[194354, 5804, 59, 1, 42, "Input"],
Cell[194416, 5807, 69, 1, 42, "Input"],
Cell[194488, 5810, 46, 1, 42, "Input"],
Cell[194537, 5813, 59, 1, 42, "Input"],
Cell[194599, 5816, 40, 1, 42, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[194676, 5822, 34, 0, 38, "Subsubsection"],
Cell[194713, 5824, 61, 0, 47, "Text"],
Cell[194777, 5826, 118, 2, 42, "Input"],
Cell[194898, 5830, 46, 1, 42, "Input"],
Cell[194947, 5833, 69, 1, 42, "Input"],
Cell[195019, 5836, 40, 1, 42, "Input"],
Cell[195062, 5839, 46, 1, 42, "Input"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[195157, 5846, 106, 2, 57, "Section",
  CellTags->"How to define"],

Cell[CellGroupData[{
Cell[195288, 5852, 66, 0, 41, "Subsubsection"],
Cell[195357, 5854, 345, 7, 99, "Text"],
Cell[195705, 5863, 70, 1, 39, "Input"],
Cell[195778, 5866, 95, 3, 47, "Text"],
Cell[195876, 5871, 65, 1, 39, "Input"],
Cell[195944, 5874, 38, 0, 47, "Text"],
Cell[195985, 5876, 53, 1, 39, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[196075, 5882, 34, 0, 41, "Subsubsection"],

Cell[CellGroupData[{
Cell[196134, 5886, 68, 1, 39, "Input"],
Cell[196205, 5889, 114, 2, 34, "Print"],
Cell[196322, 5893, 50, 1, 34, "Print"]
}, Open  ]],
Cell[196387, 5897, 71, 1, 39, "Input"],

Cell[CellGroupData[{
Cell[196483, 5902, 63, 1, 39, "Input"],
Cell[196549, 5905, 84, 1, 34, "Print"],
Cell[196636, 5908, 38, 1, 40, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[196723, 5915, 34, 0, 41, "Subsubsection"],
Cell[196760, 5917, 61, 0, 47, "Text"],

Cell[CellGroupData[{
Cell[196846, 5921, 118, 2, 39, "Input"],
Cell[196967, 5925, 176, 3, 49, "Message"]
}, Open  ]],

Cell[CellGroupData[{
Cell[197180, 5933, 46, 1, 39, "Input"],
Cell[197229, 5936, 128, 2, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[197394, 5943, 76, 1, 39, "Input"],
Cell[197473, 5946, 225, 4, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[197735, 5955, 50, 1, 39, "Input"],
Cell[197788, 5958, 38, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[197863, 5964, 74, 1, 39, "Input"],
Cell[197940, 5967, 221, 4, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[198198, 5976, 40, 1, 39, "Input"],
Cell[198241, 5979, 473, 7, 418, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[198751, 5991, 50, 1, 39, "Input"],
Cell[198804, 5994, 38, 1, 40, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[198893, 6001, 90, 1, 45, "Subsubtitle"]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

