(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[      6899,        294]*)
(*NotebookOutlinePosition[      7733,        323]*)
(*  CellTagsIndexPosition[      7689,        319]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["The \"Normalize\" package", "Title",
  FontColor->RGBColor[1, 0, 0]],

Cell["Patrice Quinton", "Subsubtitle"],

Cell["December 1997, Revised Dec. 27, 2002 and Dec. 29, 2007", "Text"],

Cell["\<\
Irisa, Campus de Beaulieu, 35042 Rennes Cedex, France
quinton@irisa.fr\
\>", "SmallText"],

Cell[CellGroupData[{

Cell["Cr\[EAcute]dits", "Subsubsection",
  FontColor->RGBColor[0, 1, 0]],

Cell["\<\
Zbignew Chamski, Florent de Dinechin, Herv\[EAcute] Le Verge, Doran \
Wilde, Tanguy Risset, Sanjay Rajopadhye\
\>", "SmallText"]
}, Open  ]],

Cell[CellGroupData[{

Cell[" Execute this cell, to redefine the current directory", "Subsubsection",
  FontColor->RGBColor[1, 0, 0]],

Cell[BoxData[
    \(setMMADir[{"\<doc\>", "\<packages\>", "\<Normalization\>"}]\)], "Input",\

  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{

Cell["1. Introduction", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell[TextData[{
  "The ",
  StyleBox["Normalize",
    FontColor->RGBColor[1, 0, 0]],
  " package contains functions to simplify Alpha expressions."
}], "Text",
  TextAlignment->Left,
  TextJustification->1],

Cell[BoxData[
    \(\(?Normalization\)\)], "Input"],

Cell[BoxData[
    \(Directory[]\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["2. normalize and normalize0", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell["The normalization rules are contained in several variables:", "Text"],

Cell[BoxData[
    \(\(?normalizationRules\)\)], "Input"],

Cell[BoxData[
    \(\(?normalizationRules0\)\)], "Input"],

Cell[BoxData[
    \(load["\<Test1.alpha\>"]; ashow[]\)], "Input"],

Cell[BoxData[
    \(normalizeQ[]\)], "Input"],

Cell[BoxData[
    \(substituteInDef["\<Z\>", "\<Z\>"]; ashow[]\)], "Input"],

Cell["The new program is not normalized: ", "Text"],

Cell[BoxData[
    \(normalizeQ[]\)], "Input"],

Cell[BoxData[
    \(normalize0Q[]\)], "Input"],

Cell["Normalize0 does not simplify case expressions:", "Text"],

Cell[BoxData[
    \(normalize0[]; ashow[]\)], "Input"],

Cell[BoxData[
    \(normalize0Q[]\)], "Input"],

Cell[BoxData[
    \(normalizeQ[]\)], "Input"],

Cell["Normalize it: ", "Text"],

Cell[BoxData[
    \(normalize[]; ashow[]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["3. normalizeDef", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell["Again, load test program:", "Text"],

Cell[BoxData[
    \(\(load["\<Test1.alpha\>"];\)\)], "Input"],

Cell["Add a new equation \"Z = U\", and rename old variable Z as U:", "Text"],

Cell[BoxData[
    \(addlocal["\<U\>", "\<Z\>"]; ashow[]\)], "Input"],

Cell["Substitute U in Z twice, and U in U:", "Text"],

Cell[BoxData[{
    \(\(substituteInDef["\<Z\>", "\<U\>"];\)\), "\n", 
    \(\(substituteInDef["\<Z\>", "\<U\>"];\)\), "\n", 
    \(substituteInDef["\<U\>", "\<U\>"]; ashow[]\)}], "Input"],

Cell["Normalize only definition of Z :", "Text"],

Cell[BoxData[{
    \(\(normalizeDef["\<Z\>"];\)\), "\n", 
    \(ashow[]\)}], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["4. Other functions", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell[CellGroupData[{

Cell["checkRestrictionsInCtxt", "Subsection"],

Cell[TextData[{
  "The function ",
  StyleBox["checkrestrictionInCtxt", "Input"],
  " modifies an Alpha expression by removing the constraints which are \
already given by the definition of a domain. "
}], "Text"],

Cell[BoxData[
    \(\(?checkRestrictionsInCtxt\)\)], "Input"],

Cell["Here is an example: ", "Text"],

Cell[BoxData[
    \(\(load["\<Test2.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(ashow[ee = getDefinition["\<Z\>"]]\)], "Input"],

Cell[BoxData[
    \(ashow[
      uu = checkRestrictionsInCtxt[ee, readDom["\<{i|i>0}\>"]]]\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["checkRestrictions", "Subsection"],

Cell["\<\
This function is essentially for internal use by \
normalization.\
\>", "Text"],

Cell[BoxData[
    \(\(?checkRestrictions\)\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["normalizeInCtxt", "Subsection"],

Cell[BoxData[
    \(\(?normalizeInCtxt\)\)], "Input"],

Cell[BoxData[
    \(\(load["\<Test2.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(ee = getDefinition["\<Z\>"]; ashow[ee]\)], "Input"],

Cell[BoxData[
    \(\(ff = normalizeInCtxt[ee, readDom["\<{i|i>=0}\>"]];\)\)], "Input"],

Cell[BoxData[
    \(ashow[ff]\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Functions for Alpha0 programs", "Subsection"],

Cell[BoxData[
    \(\(?FusionneCase\)\)], "Input"],

Cell[BoxData[
    \(\(?FusionneVars\)\)], "Input"],

Cell[BoxData[
    \(\(?normalizeA\)\)], "Input"],

Cell[BoxData[
    \(normalizeA[]; ashow[]\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["simplifyInContext", "Subsection"],

Cell[BoxData[
    \(\(?simplifyInContext\)\)], "Input"],

Cell[BoxData[
    \(load["\<Test3.alpha\>"]; ashow[]\)], "Input"],

Cell[BoxData[
    \(simplifyInContext[]; ashow[]\)], "Input"],

Cell[BoxData[
    \(normalize[]; ashow[]\)], "Input"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["5. Test for normalize", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell[BoxData[
    \(tests["\<Normalization\>"]\)], "Input"]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Macintosh",
ScreenRectangle->{{0, 1440}, {0, 837}},
AutoGeneratedPackage->None,
WindowToolbars->{"RulerBar", "EditBar"},
WindowSize->{1015, 664},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
PrintingCopies->1,
PrintingPageRange->{1, Automatic},
ShowSelection->True,
Magnification->1.5,
StyleDefinitions -> "Default.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 74, 1, 132, "Title"],
Cell[1853, 56, 38, 0, 45, "Subsubtitle"],
Cell[1894, 58, 70, 0, 47, "Text"],
Cell[1967, 60, 99, 3, 62, "SmallText"],

Cell[CellGroupData[{
Cell[2091, 67, 72, 1, 41, "Subsubsection"],
Cell[2166, 70, 138, 3, 40, "SmallText"]
}, Open  ]],

Cell[CellGroupData[{
Cell[2341, 78, 110, 1, 41, "Subsubsection"],
Cell[2454, 81, 122, 3, 42, "Input",
  InitializationCell->True]
}, Closed]],

Cell[CellGroupData[{
Cell[2613, 89, 66, 1, 57, "Section"],
Cell[2682, 92, 206, 7, 47, "Text"],
Cell[2891, 101, 51, 1, 39, "Input"],
Cell[2945, 104, 44, 1, 39, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[3026, 110, 78, 1, 104, "Section"],
Cell[3107, 113, 75, 0, 47, "Text"],
Cell[3185, 115, 56, 1, 39, "Input"],
Cell[3244, 118, 57, 1, 39, "Input"],
Cell[3304, 121, 65, 1, 39, "Input"],
Cell[3372, 124, 45, 1, 39, "Input"],
Cell[3420, 127, 75, 1, 39, "Input"],
Cell[3498, 130, 51, 0, 47, "Text"],
Cell[3552, 132, 45, 1, 39, "Input"],
Cell[3600, 135, 46, 1, 39, "Input"],
Cell[3649, 138, 62, 0, 47, "Text"],
Cell[3714, 140, 54, 1, 39, "Input"],
Cell[3771, 143, 46, 1, 39, "Input"],
Cell[3820, 146, 45, 1, 39, "Input"],
Cell[3868, 149, 30, 0, 47, "Text"],
Cell[3901, 151, 53, 1, 39, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[3991, 157, 66, 1, 57, "Section"],
Cell[4060, 160, 41, 0, 47, "Text"],
Cell[4104, 162, 61, 1, 39, "Input"],
Cell[4168, 165, 77, 0, 47, "Text"],
Cell[4248, 167, 68, 1, 39, "Input"],
Cell[4319, 170, 52, 0, 47, "Text"],
Cell[4374, 172, 187, 3, 108, "Input"],
Cell[4564, 177, 48, 0, 47, "Text"],
Cell[4615, 179, 85, 2, 62, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4737, 186, 69, 1, 104, "Section"],

Cell[CellGroupData[{
Cell[4831, 191, 45, 0, 53, "Subsection"],
Cell[4879, 193, 213, 5, 73, "Text"],
Cell[5095, 200, 61, 1, 39, "Input"],
Cell[5159, 203, 36, 0, 47, "Text"],
Cell[5198, 205, 61, 1, 39, "Input"],
Cell[5262, 208, 67, 1, 39, "Input"],
Cell[5332, 211, 103, 2, 39, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5472, 218, 39, 0, 53, "Subsection"],
Cell[5514, 220, 89, 3, 47, "Text"],
Cell[5606, 225, 55, 1, 39, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5698, 231, 37, 0, 53, "Subsection"],
Cell[5738, 233, 53, 1, 39, "Input"],
Cell[5794, 236, 61, 1, 39, "Input"],
Cell[5858, 239, 71, 1, 39, "Input"],
Cell[5932, 242, 87, 1, 39, "Input"],
Cell[6022, 245, 42, 1, 39, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6101, 251, 51, 0, 53, "Subsection"],
Cell[6155, 253, 50, 1, 39, "Input"],
Cell[6208, 256, 50, 1, 39, "Input"],
Cell[6261, 259, 48, 1, 39, "Input"],
Cell[6312, 262, 54, 1, 39, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6403, 268, 39, 0, 53, "Subsection"],
Cell[6445, 270, 55, 1, 39, "Input"],
Cell[6503, 273, 65, 1, 39, "Input"],
Cell[6571, 276, 61, 1, 39, "Input"],
Cell[6635, 279, 53, 1, 39, "Input"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[6737, 286, 72, 1, 104, "Section"],
Cell[6812, 289, 59, 1, 39, "Input"]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

