(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    319251,       8895]*)
(*NotebookOutlinePosition[    320736,       8941]*)
(*  CellTagsIndexPosition[    320514,       8932]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["The Schedule package", "Title",
  FontColor->RGBColor[1, 0, 0]],

Cell["Patrice Quinton", "Subsubtitle"],

Cell["January 1997 - Revised January 2008", "Text"],

Cell["\<\
Irisa, Campus de Beaulieu, 35042 Rennes Cedex, France
quinton@irisa.fr\
\>", "SmallText"],

Cell["Cr\[EAcute]dits", "SmallText",
  FontColor->RGBColor[0, 1, 0]],

Cell["\<\
      Zbigniew Chamski 
      Catherine Dezan 
      Florent Dupont de Dinechin 
      Patricia Le Mo\[EDoubleDot]nner 
      Herv\[EAcute] Le Verge 
      Christophe Mauras 
      Patrice Quinton 
      Fabien Quillere
      Sanjay Rajopadhye 
      Sophie Robert 
      Tanguy Risset 
      Doran Wilde \
\>", "SmallText"],

Cell[CellGroupData[{

Cell[" This cell is automatically evaluated.", "Subsubsection",
  FontColor->RGBColor[1, 0, 0]],

Cell[CellGroupData[{

Cell[BoxData[
    \(setMMADir[{"\<doc\>", "\<Packages\>", "\<Schedule\>"}]\)], "Input",
  InitializationCell->True],

Cell[BoxData[
    \("/Users/quinton/mmalpha/doc/Packages/Schedule"\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["1. Introduction", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell["\<\
This notebook contains examples of the use of the scheduler.  This is far \
from being complete. Some explanations about the scheduler are available in \
doc/Schedule and also in the Quickstart manual. \
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["2. A simple example (quite strange...)", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell[TextData[{
  "Consider the following Alpha program. We have an input variable, ",
  StyleBox["x",
    FontSlant->"Italic"],
  ", and one output variable, ",
  StyleBox["y",
    FontSlant->"Italic"],
  ". For all i, ",
  Cell[BoxData[
      \(TraditionalForm\`y\_i\)]],
  " = ",
  Cell[BoxData[
      FormBox[
        RowBox[{
          StyleBox[\(\[Sum]\+\(j = 1\)\%i\),
            ScriptLevel->0], 
          StyleBox[" ",
            ScriptLevel->0], \(x\_j\)}], TraditionalForm]]],
  ". To compute this values, we introduce a new variable ",
  Cell[BoxData[
      \(TraditionalForm\`Y\_\(i, j\)\)]],
  ",which is computed iteratively, from ",
  StyleBox["j = 0",
    FontSlant->"Italic"],
  " to ",
  StyleBox["i",
    FontSlant->"Italic"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(load["\<test.alpha\>"]; ashow[]\)], "Input"],

Cell[BoxData[
    \("[testSchedule]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("system testSchedule :{N | 1<=N}\n                    (x : {i | \
1<=i<=N} of integer)\n       returns      (y : {i | 1<=i<=N} of integer);\n\
var\n  Y : {i,j | (1,j)<=i<=N; 0<=j} of integer;\nlet\n  Y[i,j] = \n      \
case\n        { | j=0} : 0[];\n        { | 1<=j} : Y[i,j-1] + x[i];\n      \
esac;\n  y[i] = Y[i,i];\ntel;"\)], "Print"]
}, Open  ]],

Cell["\<\
We may analyze this program (it is a good practice to analyze \
systematically programs, before doing any operation on it: very often, we \
detect flaws in this way.)\
\>", "Text",
  TextJustification->1],

Cell[CellGroupData[{

Cell[BoxData[
    \(analyze[]\)], "Input"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Static Analysis of system \
"\[InvisibleSpace]"testSchedule"\),
      SequenceForm[ "Static Analysis of system ", "testSchedule"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--Checking declaration of variables."\)], "Print"],

Cell[BoxData[
    \("--Checking single assignment rule."\)], "Print"],

Cell[BoxData[
    \("--Checking definitions of output/local variables."\)], "Print"],

Cell[BoxData[
    \("--Checking definition of input variables."\)], "Print"],

Cell[BoxData[
    \("--Checking that input/local variables are used."\)], "Print"],

Cell[BoxData[
    \("--Checking type and domain consistency in the equations:"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("----equation defining "\[InvisibleSpace]"Y"\),
      SequenceForm[ "----equation defining ", "Y"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("----equation defining "\[InvisibleSpace]"y"\),
      SequenceForm[ "----equation defining ", "y"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--Checking regularity of output equations"\)], "Print"],

Cell[BoxData[
    \("----Outputs equations are regular."\)], "Print"],

Cell[BoxData[
    \("\nAnalysis Successful..."\)], "Print"],

Cell[BoxData[
    \(True\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Scheduling this program amounts to finding an order of the variable \
instances, say, x(i), y(i), Y(i,j), such that the definition of each one can \
be computed. Here, we see that ",
  Cell[BoxData[
      \(TraditionalForm\`Y\_\(i, j\)\)]],
  " requires ",
  Cell[BoxData[
      \(TraditionalForm\`Y\_\(i, j - 1\)\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`x\_i\)]],
  " to be computed. "
}], "Text",
  TextJustification->1],

Cell[TextData[{
  "To schedule a program, we associate a timing function to each variable. \
This timing function is constrained to be an affine function of the indexes \
of the variable. So, the timing function for ",
  Cell[BoxData[
      \(TraditionalForm\`Y\_\(i, j\)\)]],
  " has the form t(Y,i,j) = a i + b j + c. For MMA purposes, we shall write "
}], "Text",
  TextJustification->1],

Cell[BoxData[
    \(TYD1\ i\  + \ TYD2\ j\  + \ AY\)], "Input"],

Cell["this function. ", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(scd[addConstraints \[Rule] {TxD1 \[Equal] 1}, 
      additionalObjectiveFunction \[Rule] False, 
      integerSolution \[Rule] True]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\(28\[InvisibleSpace]" constraints, \
"\[InvisibleSpace]12\[InvisibleSpace]" variables \
"\[InvisibleSpace]\(0.047834999999999184`\ Second\)\),
      SequenceForm[ 28, " constraints, ", 12, " variables ", 
        Times[ 0.047834999999999184, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    \({"testSchedule", 12, 27, 0.047834999999999184`\ Second, "Vertex", 
      "Flat"}\)], "Print"],

Cell[BoxData[
    \(scheduleResult[
      "testSchedule", {{"x", {"i", "N"}, sched[{1, 0}, 1]}, {"y", {"i", "N"}, 
          sched[{1, 1}, 2]}, {"Y", {"i", "j", "N"}, 
          sched[{0, 1, 1}, 1]}}, {2, 2}]\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Calling scd returns the number of constraints, variables, the solving time \
ot the corresponding linear problem, and returns the result in a structure. \
The result is also put in the global variable ",
  StyleBox["$schedule", "Input"],
  ". "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \($schedule\)], "Input"],

Cell[BoxData[
    \(scheduleResult[
      "testSchedule", {{"x", {"i", "N"}, sched[{0, 0}, 0]}, {"y", {"i", "N"}, 
          sched[{1, 0}, 1]}, {"Y", {"i", "j", "N"}, 
          sched[{0, 1, 0}, 0]}}, {1, 1}]\)], "Output"]
}, Open  ]],

Cell["It can be viewed using the function showSchedResult.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(showSchedResult[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "i"\)\),
      SequenceForm[ "T_", "x", {"i", "N"}, " = ", 
        Plus[ 1, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2 + "i" + "N"\)\),
      SequenceForm[ "T_", "y", {"i", "N"}, " = ", 
        Plus[ 2, "i", "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"Y"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "j" + "N"\)\),
      SequenceForm[ "T_", "Y", {"i", "j", "N"}, " = ", 
        Plus[ 1, "j", "N"]],
      Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
The schedule will be used later on to produce a parallel architecture. The \
parameters of the schedule are also available (for the scd scheduler only) in \
the global variable $scheduleMDSol[1]. First element is the objective \
function, second is the set of constraints, then the variable, and the \
solution in the form optimum and solution. \
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \($scheduleMDSol[1]\)], "Input"],

Cell[BoxData[
    \({TN, {TxD1 \[Equal] 1, TN - TxD2 \[GreaterEqual] 0, 
        TN - TxD1 - TxD2 \[GreaterEqual] 0, 
        aparam - Ax + TN - TxD1 - TxD2 \[GreaterEqual] 0, 
        TxD2 \[GreaterEqual] 0, TxD1 + TxD2 \[GreaterEqual] 0, 
        TN - TyD2 \[GreaterEqual] 0, TN - TyD1 - TyD2 \[GreaterEqual] 0, 
        aparam - Ay + TN - TyD1 - TyD2 \[GreaterEqual] 0, 
        TyD2 \[GreaterEqual] 0, TyD1 + TyD2 \[GreaterEqual] 0, 
        TYD2 \[GreaterEqual] 1, TN - TYD3 \[GreaterEqual] 0, 
        TN - TYD1 - TYD3 \[GreaterEqual] 0, 
        aparam - AY + TN - TYD1 - TYD3 \[GreaterEqual] 0, 
        TyD2 - TYD3 \[GreaterEqual] 0, 
        TN - TYD1 - TYD2 - TYD3 \[GreaterEqual] 0, 
        aparam - AY + TN - TYD1 - TYD2 - TYD3 \[GreaterEqual] 0, 
        TyD1 - TYD1 + TyD2 - TYD2 - TYD3 \[GreaterEqual] 0, 
        Ay - AY + TyD1 - TYD1 + TyD2 - TYD2 - TYD3 \[GreaterEqual] 1, 
        TYD3 \[GreaterEqual] 0, \(-TxD2\) + TYD3 \[GreaterEqual] 0, 
        TYD1 + TYD3 \[GreaterEqual] 
          0, \(-TxD1\) - TxD2 + TYD1 + TYD3 \[GreaterEqual] 0, 
        TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          0, \(-TxD1\) - TxD2 + TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          0, \(-Ax\) + AY - TxD1 - TxD2 + TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          1}, {aparam, Ax, Ay, AY, TN, TxD1, TxD2, TyD1, TYD1, TyD2, TYD2, 
        TYD3}, {2, {aparam \[Rule] 2, Ax \[Rule] 1, Ay \[Rule] 2, 
          AY \[Rule] 1, TN \[Rule] 2, TxD1 \[Rule] 1, TxD2 \[Rule] 0, 
          TyD1 \[Rule] 1, TYD1 \[Rule] 0, TyD2 \[Rule] 1, TYD2 \[Rule] 1, 
          TYD3 \[Rule] 1}}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(schedule[addConstraints \[Rule] {"\<TxD1==1\>"}]\)], "Input"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(1\[InvisibleSpace]" Recognized added Constraints: "\),
      
      SequenceForm[ 1, " Recognized added Constraints: "],
      Editable->False]], "Print"],

Cell[BoxData[
    \("   -1 + TxD1 = 0,"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]46\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]50\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 46, "      variables, ", 50, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00034066731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408091794, ByteCount \[Rule] 7405}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00034066731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408091794, ByteCount \[Rule] 460}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]\(2\ "N"\)\),
      
      SequenceForm[ "Total execution Time: ", 
        Times[ 2, "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(\(-1\) + "i"\)\),
      SequenceForm[ "T_", "x", {"i", "N"}, " = ", 
        Plus[ -1, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\ "i"\)\),
      SequenceForm[ "T_", "y", {"i", "N"}, " = ", 
        Times[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"Y"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(\(-1\) + "i" + "j"\)\),
      
      SequenceForm[ "T_", "Y", {"i", "j", "N"}, " = ", 
        Plus[ -1, "i", "j"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \(scheduleResult[
      "testSchedule", {{"x", {"i", "N"}, 
          sched[{1, 0}, \(-1\)]}, {"y", {"i", "N"}, 
          sched[{2, 0}, 0]}, {"Y", {"i", "j", "N"}, 
          sched[{1, 1, 0}, \(-1\)]}}, {2, 0}]\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["3. Gathering constraints", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell["\<\
Several functions allows one to gather and modify the scheduling \
constraints. \
\>", "Text"],

Cell[CellGroupData[{

Cell[TextData[{
  "The ",
  StyleBox["scheduleConstraint", "Input"],
  " function"
}], "Subsection"],

Cell["\<\
This function computes the dependence constraints, and the \
positivity constraints, either mono- or multidimensional, with various \
options. \
\>", "Text",
  TextJustification->1],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?scheduleConstraints\)\)], "Input"],

Cell["Alpha`VertexSchedule`scheduleConstraints", "Print",
  CellTags->"Info3407929681-2992127"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {GridBox[{
                {\(scheduleConstraints[
                      Alpha`VertexSchedule`Private`options___Rule] := 
                    scheduleConstraints[$result, {}, 
                      Alpha`VertexSchedule`Private`options]\)},
                {" "},
                {\(scheduleConstraints[
                      Alpha`VertexSchedule`Private`sys_system, 
                      Alpha`VertexSchedule`Private`options___Rule] := 
                    scheduleConstraints[Alpha`VertexSchedule`Private`sys, {}, 
                      Alpha`VertexSchedule`Private`options]\)},
                {" "},
                {\(scheduleConstraints[
                      Alpha`VertexSchedule`Private`variable_String, 
                      Alpha`VertexSchedule`Private`options___Rule] := 
                    scheduleConstraints[$result, 
                      Alpha`VertexSchedule`Private`variable, 
                      Alpha`VertexSchedule`Private`options]\)},
                {" "},
                {\(scheduleConstraints[
                      Alpha`VertexSchedule`Private`sys_system, 
                      Alpha`VertexSchedule`Private`variable : _String | {}, 
                      Alpha`VertexSchedule`Private`options___Rule] := 
                    Catch[Module[{Alpha`VertexSchedule`Private`v, 
                          Alpha`VertexSchedule`Private`lv, 
                          Alpha`VertexSchedule`Private`iv, 
                          Alpha`VertexSchedule`Private`ov, 
                          Alpha`VertexSchedule`Private`declarat, 
                          Alpha`VertexSchedule`Private`dims, 
                          Alpha`VertexSchedule`Private`coeffstau, 
                          Alpha`VertexSchedule`Private`coeffsalpha, 
                          Alpha`VertexSchedule`Private`defs, 
                          Alpha`VertexSchedule`Private`aux, 
                          Alpha`VertexSchedule`Private`result, 
                          Alpha`VertexSchedule`Private`result1, 
                          Alpha`VertexSchedule`Private`pos, 
                          Alpha`VertexSchedule`Private`oploc, 
                          Alpha`VertexSchedule`Private`slp, 
                          Alpha`VertexSchedule`Private`vrb, 
                          Alpha`VertexSchedule`Private`extrac, 
                          Alpha`VertexSchedule`Private`multidim, 
                          Alpha`VertexSchedule`Private`bd, 
                          Alpha`VertexSchedule`Private`vpc, 
                          Alpha`VertexSchedule`Private`dponly, 
                          Alpha`VertexSchedule`Private`dbg, 
                          Alpha`VertexSchedule`Private`depcomp, 
                          Alpha`VertexSchedule`Private`sbs, 
                          Alpha`VertexSchedule`Private`useConstraints, 
                          Alpha`VertexSchedule`Private`df, 
                          Alpha`VertexSchedule`Private`onlyV, 
                          Alpha`VertexSchedule`Private`onlyE, 
                          Alpha`VertexSchedule`Private`onlyD, 
                          Alpha`VertexSchedule`Private`reuseS}, \
$dependencyConstraints = {}; 
                        If[Alpha`VertexSchedule`Private`sys === {}, 
                          Throw[Message[
                              Alpha`VertexSchedule`Private`\
scheduleConstraint::"emptyprog"]]]; 
                        If[\(! normalizeQ[
                              Alpha`VertexSchedule`Private`sys]\), 
                          Throw[Message[
                              Alpha`VertexSchedule`Private`\
scheduleConstraint::"norm"]]]; 
                        Alpha`VertexSchedule`Private`sbs = \(subSystems \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]; 
                        Alpha`VertexSchedule`Private`vpc = \(verticesPositive \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]; 
                        Alpha`VertexSchedule`Private`dponly = \
\(dependenciesOnly /. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\
\) /. \[InvisibleSpace]Options[scheduleConstraints]; 
                        Alpha`VertexSchedule`Private`slp = \(scheduleType \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]; 
                        Alpha`VertexSchedule`Private`vrb = \(verbose \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]; 
                        Alpha`VertexSchedule`Private`df = \(dataFlow \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]; 
                        Alpha`VertexSchedule`Private`multidim = 
                          If[Alpha`VertexSchedule`Private`df, 
                            True, \(multidimensional \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]]; 
                        Alpha`VertexSchedule`Private`dbg = \(debug \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]; 
                        Alpha`VertexSchedule`Private`depcomp = \(depComputed \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]; 
                        Alpha`VertexSchedule`Private`dur = \(durations \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]; 
                        Alpha`VertexSchedule`Private`onlyV = \(onlyVars \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]; 
                        Alpha`VertexSchedule`Private`onlyE = \(onlyEquations \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]; 
                        If[Alpha`VertexSchedule`Private`onlyE =!= all, 
                          Print["Warning: option onlyEquations in not \
implemented"]]; 
                        Alpha`VertexSchedule`Private`onlyD = \(onlyDep \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]; 
                        Alpha`VertexSchedule`Private`reuseS = \(reuseSchedule \
/. \[InvisibleSpace]{Alpha`VertexSchedule`Private`options}\) \
/. \[InvisibleSpace]Options[scheduleConstraints]; 
                        Alpha`VertexSchedule`Private`lv = getLocalVars[]; 
                        If[Alpha`VertexSchedule`Private`onlyV =!= all, 
                          Alpha`VertexSchedule`Private`lv = 
                            Alpha`VertexSchedule`Private`lv \[Intersection] 
                              Alpha`VertexSchedule`Private`onlyV]; 
                        Alpha`VertexSchedule`Private`iv = getInputVars[]; 
                        If[Alpha`VertexSchedule`Private`onlyV =!= all, 
                          Alpha`VertexSchedule`Private`iv = 
                            Alpha`VertexSchedule`Private`iv \[Intersection] 
                              Alpha`VertexSchedule`Private`onlyV]; 
                        Alpha`VertexSchedule`Private`ov = getOutputVars[]; 
                        If[Alpha`VertexSchedule`Private`onlyV =!= all, 
                          Alpha`VertexSchedule`Private`ov = 
                            Alpha`VertexSchedule`Private`ov \[Intersection] 
                              Alpha`VertexSchedule`Private`onlyV]; 
                        Alpha`VertexSchedule`Private`v = 
                          Alpha`VertexSchedule`Private`lv \[Union] 
                            Alpha`VertexSchedule`Private`iv \[Union] 
                            Alpha`VertexSchedule`Private`ov; 
                        If[Alpha`VertexSchedule`Private`variable =!= {}, 
                          If[\(! MemberQ[Alpha`VertexSchedule`Private`v, 
                                Alpha`VertexSchedule`Private`variable]\), 
                            Throw[Message[scheduleConstraints::"wrgvar", 
                                Alpha`VertexSchedule`Private`variable]]]]; 
                        Alpha`VertexSchedule`Private`dv = 
                          Alpha`VertexSchedule`Private`lv \[Union] 
                            Alpha`VertexSchedule`Private`ov; 
                        Alpha`VertexSchedule`Private`declarat = 
                          getDeclaration /@ Alpha`VertexSchedule`Private`v; 
                        Alpha`VertexSchedule`Private`domains = \((#1\
\[LeftDoubleBracket]3\[RightDoubleBracket] &)\) /@ 
                            Alpha`VertexSchedule`Private`declarat; 
                        Alpha`VertexSchedule`Private`dims = \((First[#1] &)\) \
/@ Alpha`VertexSchedule`Private`domains; 
                        Alpha`VertexSchedule`Private`coeffstau = 
                          MapThread[
                            Table[ToExpression[
                                  "T" <> #2 <> "D" <> 
                                    ToString[
                                      Alpha`VertexSchedule`Private`i]], \
{Alpha`VertexSchedule`Private`i, 
                                  1, #1}] &, \
{Alpha`VertexSchedule`Private`dims, Alpha`VertexSchedule`Private`v}]; 
                        If[Alpha`VertexSchedule`Private`slp === 
                            sameLinearPart, 
                          Alpha`VertexSchedule`Private`coeffstau = 
                            MapThread[
                              If[MemberQ[
                                    Alpha`VertexSchedule`Private`lv, #2], 
                                  Table[
                                    ToExpression[
                                      "TD" <> 
                                        ToString[
                                        Alpha`VertexSchedule`Private`i]], \
{Alpha`VertexSchedule`Private`i, 
                                      1, #3}], #1] &, \
{Alpha`VertexSchedule`Private`coeffstau, Alpha`VertexSchedule`Private`v, 
                                Alpha`VertexSchedule`Private`dims}]]; 
                        Alpha`VertexSchedule`Private`coeffsalpha = \
\((ToExpression["A" <> #1] &)\) /@ Alpha`VertexSchedule`Private`v; 
                        If[\(! Alpha`VertexSchedule`Private`depcomp\), 
                          If[Alpha`VertexSchedule`Private`dbg, 
                            Print["Computing dependency Table"]]; \
$scheduleDepTable = Check[dep[Alpha`VertexSchedule`Private`sys, 
                                Alpha`VertexSchedule`Private`options, 
                                parameterConstraints \[Rule] True, 
                                subSystems \[Rule] True], 
                              Throw[Message[scheduleConstraints::"dep"]]], 
                          If[\(! MatchQ[$scheduleDepTable, 
                                dtable[{___depend}, {___dependuse}] | 
                                  dtable[{___depend}]]\), 
                            Throw[Message[
                                scheduleConstraints::"nodeptable"]]]; ]; 
                        If[Alpha`VertexSchedule`Private`onlyD =!= all, 
                          With[{Alpha`VertexSchedule`Private`dept = \
$scheduleDepTable\[LeftDoubleBracket]1\[RightDoubleBracket]}, \
$scheduleDepTable = dtable[Table[
                                  Alpha`VertexSchedule`Private`dept\
\[LeftDoubleBracket]Alpha`VertexSchedule`Private`onlyD\[LeftDoubleBracket]
                                      Alpha`VertexSchedule`Private`i\
\[RightDoubleBracket]\[RightDoubleBracket], {Alpha`VertexSchedule`Private`i, 
                                    1, Length[
                                      Alpha`VertexSchedule`Private`onlyD]}]]]]\
; If[Alpha`VertexSchedule`Private`vrb || Alpha`VertexSchedule`Private`dbg, 
                          Print["Dependencies kept are:"]; 
                          show[$scheduleDepTable]; 
                          Print["----------------------------"]; ]; 
                        If[Alpha`VertexSchedule`Private`onlyV =!= 
                            all, $scheduleDepTable = 
                            dtable[
                              With[{Alpha`VertexSchedule`Private`dept = \
$scheduleDepTable\[LeftDoubleBracket]1\[RightDoubleBracket]}, 
                                Select[Alpha`VertexSchedule`Private`dept, 
                                  Function[{Alpha`VertexSchedule`Private`x}, 
                                    Module[{Alpha`VertexSchedule`Private`v1, 
                                        Alpha`VertexSchedule`Private`v2}, 
                                      Alpha`VertexSchedule`Private`v1 = 
                                        Alpha`VertexSchedule`Private`x\
\[LeftDoubleBracket]2\[RightDoubleBracket]; 
                                      Alpha`VertexSchedule`Private`v2 = 
                                        Alpha`VertexSchedule`Private`x\
\[LeftDoubleBracket]3\[RightDoubleBracket]; 
                                      MemberQ[Alpha`VertexSchedule`Private`v, 
                                        Alpha`VertexSchedule`Private`v1] && 
                                        MemberQ[
                                        Alpha`VertexSchedule`Private`v, 
                                        Alpha`VertexSchedule`Private`v2]]]]]]]\
; If[Alpha`VertexSchedule`Private`dbg, 
                          Print["Dependency Table:\n", 
                            show[$scheduleDepTable, silent \[Rule] True]]]; 
                        If[$scheduleDepTable \[Equal] dtable[{}], 
                          scheduleConstraints::"emptydeptable" = 
                            "dependence table is empty"; 
                          Throw[Message[
                              scheduleConstraints::"emptydeptable"]]]; 
                        If[Alpha`VertexSchedule`Private`dbg, 
                          Print["Computing dependence constraints"]]; 
                        If[Alpha`VertexSchedule`Private`dur === {}, 
                          Alpha`VertexSchedule`Private`dur = 
                            Table[1, {Alpha`VertexSchedule`Private`i, 1, 
                                Length[$scheduleDepTable\[LeftDoubleBracket]1\
\[RightDoubleBracket]]}]]; 
                        If[Length[Alpha`VertexSchedule`Private`dur] =!= 
                            Length[$scheduleDepTable\[LeftDoubleBracket]1\
\[RightDoubleBracket]], Throw[Message[scheduleConstraints::"wrgduration"]]]; 
                        Module[{Alpha`VertexSchedule`Private`listvariables}, 
                          Alpha`VertexSchedule`Private`listvariables = 
                            If[Alpha`VertexSchedule`Private`variable =!= {}, \
{Alpha`VertexSchedule`Private`variable}, Alpha`VertexSchedule`Private`v]; 
                          Alpha`VertexSchedule`Private`result1 = \
\((affineDepConsts[depend[getDeclarationDomain[#1], #1, \*"\<\"\"\>", 
                                      Null, _], 
                                    Alpha`VertexSchedule`Private`v, 0, 
                                    Alpha`VertexSchedule`Private`coeffstau, 
                                    Alpha`VertexSchedule`Private`coeffsalpha, 
                                    Alpha`VertexSchedule`Private`options] \
&)\) /@ Alpha`VertexSchedule`Private`iv; 
                          Alpha`VertexSchedule`Private`result = 
                            MapThread[
                              affineDepConsts[#1, 
                                  Alpha`VertexSchedule`Private`v, #2, 
                                  Alpha`VertexSchedule`Private`coeffstau, 
                                  Alpha`VertexSchedule`Private`coeffsalpha, 
                                  Alpha`VertexSchedule`Private`options] &, \
{$scheduleDepTable\[LeftDoubleBracket]1\[RightDoubleBracket], 
                                Alpha`VertexSchedule`Private`dur}]; 
                          Alpha`VertexSchedule`Private`result = 
                            Join[Alpha`VertexSchedule`Private`result1, 
                              Alpha`VertexSchedule`Private`result]; ]; 
                        Alpha`VertexSchedule`Private`result = \
\((Complement[#1, {True}] &)\) /@ Alpha`VertexSchedule`Private`result; 
                        Alpha`VertexSchedule`Private`pos = \((\(\(Position[
                                      Alpha`VertexSchedule`Private`v, #1]\)\
\[LeftDoubleBracket]1\[RightDoubleBracket]\)\[LeftDoubleBracket]1\
\[RightDoubleBracket] &)\) /@ Alpha`VertexSchedule`Private`v; 
                        Alpha`VertexSchedule`Private`pos = 
                          Flatten[\((Alpha`VertexSchedule`Private`coeffstau\
\[LeftDoubleBracket]#1\[RightDoubleBracket] &)\) /@ 
                                Alpha`VertexSchedule`Private`pos] \[Union] 
                            Flatten[\((Alpha`VertexSchedule`Private`\
coeffsalpha\[LeftDoubleBracket]#1\[RightDoubleBracket] &)\) /@ 
                                Alpha`VertexSchedule`Private`pos]; 
                        If[\(! Alpha`VertexSchedule`Private`multidim\), 
                          Alpha`VertexSchedule`Private`result = 
                            Alpha`VertexSchedule`Private`result \
/. \[InvisibleSpace]"ge"[Alpha`VertexSchedule`Private`x_, 
                                  Alpha`VertexSchedule`Private`y_] \[Rule] 
                                
                                Alpha`VertexSchedule`Private`x \
\[GreaterEqual] 
                                  Alpha`VertexSchedule`Private`y; \
$dependencyConstraints = {Complement[
                                Union @@ 
                                  Alpha`VertexSchedule`Private`result, \
{True}], Alpha`VertexSchedule`Private`pos}; 
                          If[Alpha`VertexSchedule`Private`sbs, 
                            Alpha`VertexSchedule`Private`useConstraints = 
                              If[Length[$scheduleDepTable] =!= 2, 
                                If[Alpha`VertexSchedule`Private`vrb, 
                                  Print[scheduleConstraints::"noUse"]]; {}, 
                                Check[
                                  Alpha`VertexSchedule`Private`useConstraints \
= getUseSchedule[$scheduleDepTable\[LeftDoubleBracket]2\[RightDoubleBracket], 
                                      Alpha`VertexSchedule`Private`options], 
                                  Throw[
                                    "Error while calling \
getUseSchedule..."]]]; $dependencyConstraints = {$dependencyConstraints\
\[LeftDoubleBracket]1\[RightDoubleBracket] \[Union] 
                                  Union @@ \(\((#1\[LeftDoubleBracket]1\
\[RightDoubleBracket] &)\) /@ 
                                      Alpha`VertexSchedule`Private`\
useConstraints\), $dependencyConstraints\[LeftDoubleBracket]2\
\[RightDoubleBracket] \[Union] 
                                  Union @@ \(\((#1\[LeftDoubleBracket]2\
\[RightDoubleBracket] &)\) /@ 
                                      Alpha`VertexSchedule`Private`\
useConstraints\)}]; $dependencyConstraints, 
                          Module[{Alpha`VertexSchedule`Private`aux, 
                              Alpha`VertexSchedule`Private`temp}, 
                            Clear[Alpha`VertexSchedule`Private`aux]; 
                            Alpha`VertexSchedule`Private`aux[
                                Alpha`VertexSchedule`Private`c_, 
                                index : _Integer | Null] := 
                              Module[{Alpha`VertexSchedule`Private`cons1, 
                                  Alpha`VertexSchedule`Private`cons2, 
                                  Alpha`VertexSchedule`Private`vars, 
                                  Alpha`VertexSchedule`Private`cons3, 
                                  Alpha`VertexSchedule`Private`cons}, 
                                Alpha`VertexSchedule`Private`cons1 = 
                                  Cases[Alpha`VertexSchedule`Private`c, 
                                    "ge"[_, Alpha`VertexSchedule`Private`x_] /; 
                                      Alpha`VertexSchedule`Private`x > 0, 2]; 
                                Alpha`VertexSchedule`Private`cons2 = 
                                  Cases[Alpha`VertexSchedule`Private`c, 
                                    "ge"[Alpha`VertexSchedule`Private`x_, 
                                        0] \[Rule] 
                                      Alpha`VertexSchedule`Private`x \
\[GreaterEqual] 0, 2]; If[index =!= Null, 
                                  Alpha`VertexSchedule`Private`cons1 = 
                                    Table[\(Alpha`VertexSchedule`Private`\
cons1\[LeftDoubleBracket]
                                        Alpha`VertexSchedule`Private`i\
\[RightDoubleBracket]\)\[LeftDoubleBracket]1\[RightDoubleBracket] \
\[GreaterEqual] Alpha`VertexSchedule`Private`dur\[LeftDoubleBracket]
                                        index\[RightDoubleBracket]\ \
ToExpression["slack" <> ToString[index]], {Alpha`VertexSchedule`Private`i, 1, 
                                        Length[
                                        Alpha`VertexSchedule`Private`cons1]}];\
 Alpha`VertexSchedule`Private`vars = {ToExpression[
                                        "slack" <> ToString[index]]}; , 
                                  Alpha`VertexSchedule`Private`cons1 = {}; 
                                  Alpha`VertexSchedule`Private`vars = {}; ]; 
                                Alpha`VertexSchedule`Private`cons3 = \((0 \
\[LessEqual] #1 \[LessEqual] 1 &)\) /@ Alpha`VertexSchedule`Private`vars; 
                                Alpha`VertexSchedule`Private`cons = 
                                  Complement[
                                    Alpha`VertexSchedule`Private`cons3 \
\[Union] Alpha`VertexSchedule`Private`cons1 \[Union] 
                                      Alpha`VertexSchedule`Private`cons2, \
{True}]; {Alpha`VertexSchedule`Private`cons, 
                                  Alpha`VertexSchedule`Private`vars}]; 
                            With[{Alpha`VertexSchedule`Private`r1 = 
                                  Take[Alpha`VertexSchedule`Private`result, 
                                    Length[Alpha`VertexSchedule`Private`iv]], 
                                Alpha`VertexSchedule`Private`r2 = 
                                  Drop[Alpha`VertexSchedule`Private`result, 
                                    Length[
                                      Alpha`VertexSchedule`Private`iv]]}, 
                              Alpha`VertexSchedule`Private`temp = 
                                MapThread[
                                  Alpha`VertexSchedule`Private`aux, \
{Alpha`VertexSchedule`Private`r1, Table[
                                      Null, {Alpha`VertexSchedule`Private`i, 
                                        1, Length[
                                        Alpha`VertexSchedule`Private`r1]}]}]; 
                              Alpha`VertexSchedule`Private`temp = 
                                Join[Alpha`VertexSchedule`Private`temp, 
                                  MapThread[
                                    Alpha`VertexSchedule`Private`aux, \
{Alpha`VertexSchedule`Private`r2, Table[
                                        Alpha`VertexSchedule`Private`i, \
{Alpha`VertexSchedule`Private`i, 1, Length[
                                        Alpha`VertexSchedule`Private`r2]}]}]];\
 ]; Alpha`VertexSchedule`Private`pos = 
                              Alpha`VertexSchedule`Private`pos \[Union] 
                                Union @@ \(Last /@ 
                                    Alpha`VertexSchedule`Private`temp\); \
$dependencyConstraints = {First /@ Alpha`VertexSchedule`Private`temp, 
                                Alpha`VertexSchedule`Private`pos}; ]]; 
                        If[Alpha`VertexSchedule`Private`reuseS, 
                          Module[{Alpha`VertexSchedule`Private`rules, 
                              Alpha`VertexSchedule`Private`varNamesloc}, 
                            Clear[Alpha`VertexSchedule`Private`auxFunction]; 
                            Alpha`VertexSchedule`Private`auxFunction[{Alpha`\
VertexSchedule`Private`name_String, {___String}, 
                                  sched[
                                    Alpha`VertexSchedule`Private`values : \
{___Integer}, Alpha`VertexSchedule`Private`constant_Integer]}] := 
                              Module[{Alpha`VertexSchedule`Private`tauCoeffs, 
                                  Alpha`VertexSchedule`Private`alphaCoeff}, 
                                Alpha`VertexSchedule`Private`tauCoeffs = 
                                  Table[
                                    ToExpression[
                                        "T" <> 
                                        Alpha`VertexSchedule`Private`name <> 
                                        "D" <> 
                                        ToString[
                                        Alpha`VertexSchedule`Private`i]] \
\[Rule] Alpha`VertexSchedule`Private`values\[LeftDoubleBracket]
                                        Alpha`VertexSchedule`Private`i\
\[RightDoubleBracket], {Alpha`VertexSchedule`Private`i, 1, 
                                      Length[
                                        Alpha`VertexSchedule`Private`values]}]\
; Alpha`VertexSchedule`Private`alphaCoeff = 
                                  ToExpression[
                                      "A" <> 
                                        Alpha`VertexSchedule`Private`name] \
\[Rule] Alpha`VertexSchedule`Private`constant; 
                                Append[
                                  Alpha`VertexSchedule`Private`tauCoeffs, 
                                  Alpha`VertexSchedule`Private`alphaCoeff]]; 
                            Alpha`VertexSchedule`Private`rules = 
                              Union @@ \
\(Alpha`VertexSchedule`Private`auxFunction /@ $schedule\[LeftDoubleBracket]2\
\[RightDoubleBracket]\); If[Alpha`VertexSchedule`Private`dbg, 
                              Print[Alpha`VertexSchedule`Private`rules]]; 
                            Alpha`VertexSchedule`Private`varNamesloc = 
                              First /@ Alpha`VertexSchedule`Private`rules; 
                            If[Alpha`VertexSchedule`Private`dbg, 
                              Print[Alpha`VertexSchedule`Private`varNamesloc]]\
; $dependencyConstraints = 
                              With[{Alpha`VertexSchedule`Private`r = \
$dependencyConstraints\[LeftDoubleBracket]1\[RightDoubleBracket], 
                                  Alpha`VertexSchedule`Private`v = \
$dependencyConstraints\[LeftDoubleBracket]2\[RightDoubleBracket]}, \
{Complement[
                                    Alpha`VertexSchedule`Private`r \
/. \[InvisibleSpace]Alpha`VertexSchedule`Private`rules, {True}], 
                                  Complement[Alpha`VertexSchedule`Private`v, 
                                    Alpha`VertexSchedule`Private`varNamesloc]}\
]; If[Alpha`VertexSchedule`Private`dbg, 
                              Print[$dependencyConstraints]]]]; \
$dependencyConstraints]]\)},
                {" "},
                {\(scheduleConstraints[___] := 
                    Message[scheduleConstraints::"params"]\)}
                },
              GridBaseline->{Baseline, {1, 1}},
              ColumnWidths->0.999,
              ColumnAlignments->{Left}]},
          {" "},
          {GridBox[{
                {\(Options[
                      scheduleConstraints] = {scheduleType \[Rule] 
                        affineByVar, verbose \[Rule] False, 
                      addConstraints \[Rule] {}, 
                      multidimensional \[Rule] False, 
                      verticesPositive \[Rule] False, 
                      dependenciesOnly \[Rule] False, 
                      onlyEquations \[Rule] all, onlyVars \[Rule] all, 
                      onlyDep \[Rule] all, debug \[Rule] False, 
                      depComputed \[Rule] False, durations \[Rule] {}, 
                      reuseSchedule \[Rule] False, subSystems \[Rule] True, 
                      eliminatesDoubles \[Rule] False, 
                      dataFlow \[Rule] False}\)}
                },
              GridBaseline->{Baseline, {1, 1}},
              ColumnWidths->0.999,
              ColumnAlignments->{Left}]}
          },
        GridBaseline->{Baseline, {1, 1}},
        ColumnAlignments->{Left}],
      Definition[ "scheduleConstraints"],
      Editable->False]], "Print",
  CellTags->"Info3407929681-2992127"]
}, Closed]],

Cell["If we apply this to the previous program, we get: ", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(scheduleConstraints[]\)], "Input"],

Cell[BoxData[
    \({{TxD2 \[GreaterEqual] 0, TxD1 + TxD2 \[GreaterEqual] 0, 
        TyD2 \[GreaterEqual] 0, TyD1 + TyD2 \[GreaterEqual] 0, 
        TYD2 \[GreaterEqual] 1, TyD2 - TYD3 \[GreaterEqual] 0, 
        TyD1 - TYD1 + TyD2 - TYD2 - TYD3 \[GreaterEqual] 0, 
        Ay - AY + TyD1 - TYD1 + TyD2 - TYD2 - TYD3 \[GreaterEqual] 1, 
        TYD3 \[GreaterEqual] 0, \(-TxD2\) + TYD3 \[GreaterEqual] 0, 
        TYD1 + TYD3 \[GreaterEqual] 
          0, \(-TxD1\) - TxD2 + TYD1 + TYD3 \[GreaterEqual] 0, 
        TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          0, \(-TxD1\) - TxD2 + TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          0, \(-Ax\) + AY - TxD1 - TxD2 + TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          1}, {Ax, Ay, AY, TxD1, TxD2, TyD1, TYD1, TyD2, TYD2, 
        TYD3}}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "After the computation, we have a result of the form ",
  StyleBox["{constraints, variables}", "Input"],
  ". The name of the variables are defined as follows: \n- for a variable ",
  StyleBox["x", "Input"],
  ", with dimension 2 (say, with indexes ",
  StyleBox["i", "Input"],
  " and ",
  StyleBox["j", "Input"],
  "), the timing function will have the form ",
  StyleBox["TxD1 i + TxD2 j + Ax", "Input"],
  "."
}], "Text",
  TextJustification->1],

Cell[CellGroupData[{

Cell[BoxData[
    \($optimalityConstraints\)], "Input"],

Cell[BoxData[
    \({{{"x", {TN - TxD2 \[GreaterEqual] 0, 
            TN - TxD1 - TxD2 \[GreaterEqual] 0, 
            aparam - Ax + TN - TxD1 - TxD2 \[GreaterEqual] 
              0}}, {"y", {TN - TyD2 \[GreaterEqual] 0, 
            TN - TyD1 - TyD2 \[GreaterEqual] 0, 
            aparam - Ay + TN - TyD1 - TyD2 \[GreaterEqual] 
              0}}, {"Y", {TN - TYD3 \[GreaterEqual] 0, 
            TN - TYD1 - TYD3 \[GreaterEqual] 0, 
            aparam - AY + TN - TYD1 - TYD3 \[GreaterEqual] 0, 
            TN - TYD1 - TYD2 - TYD3 \[GreaterEqual] 0, 
            aparam - AY + TN - TYD1 - TYD2 - TYD3 \[GreaterEqual] 
              0}}}, {aparam, Ax, Ay, AY, TN, TxD1, TxD2, TyD1, TYD1, TyD2, 
        TYD2, TYD3}}\)], "Output"]
}, Open  ]],

Cell["\<\
There exist various options for this function. Many of these \
options are shared by all scheduling functions.\
\>", "Text",
  TextJustification->1],

Cell[CellGroupData[{

Cell[BoxData[
    \(Options[scheduleConstraints]\)], "Input"],

Cell[BoxData[
    \({scheduleType \[Rule] affineByVar, verbose \[Rule] False, 
      addConstraints \[Rule] {}, multidimensional \[Rule] False, 
      verticesPositive \[Rule] False, dependenciesOnly \[Rule] False, 
      onlyEquations \[Rule] all, onlyVars \[Rule] all, onlyDep \[Rule] all, 
      debug \[Rule] False, depComputed \[Rule] False, durations \[Rule] {}, 
      reuseSchedule \[Rule] False, subSystems \[Rule] True, 
      eliminatesDoubles \[Rule] False, dataFlow \[Rule] False}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  StyleBox["sameLinearPart -> True", "Input"],
  " allows one to constraint variables that have the same number of indexes \
to have the same coefficients."
}], "Text"],

Cell[TextData[{
  StyleBox["verbose -> True", "Input"],
  " provides more information."
}], "Text"],

Cell[TextData[{
  StyleBox["extraConstraints -> {constraints}", "Input"],
  " adds extra constraints."
}], "Text"],

Cell[TextData[{
  StyleBox["multidimensional -> {True}", "Input"],
  " looks for a multidimensional schedule."
}], "Text"],

Cell[TextData[{
  StyleBox["boudedDelay -> True", "Input"],
  " is supposed to look for schedules with bounded delays."
}], "Text"],

Cell[TextData[{
  StyleBox["verticesPositive -> True", "Input"],
  " adds only positivity constraints determined by rays, not vertices."
}], "Text"],

Cell[TextData[{
  StyleBox["dependenciesOnly -> True", "Input"],
  " provides only dependence constraints. "
}], "Text"],

Cell[TextData[{
  StyleBox["debug -> True", "Input"],
  " is the debug option."
}], "Text"],

Cell[TextData[{
  "With the option ",
  StyleBox["dependenciesOnly->True", "Input"],
  ", we get only the dependencies due to causality constaints, not the \
positivity constraints:"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(scheduleConstraints[dependenciesOnly -> True]\)], "Input"],

Cell[BoxData[
    \({{TYD2 \[GreaterEqual] 1, TyD2 - TYD3 \[GreaterEqual] 0, 
        TyD1 - TYD1 + TyD2 - TYD2 - TYD3 \[GreaterEqual] 0, 
        Ay - AY + TyD1 - TYD1 + TyD2 - TYD2 - TYD3 \[GreaterEqual] 
          1, \(-TxD2\) + TYD3 \[GreaterEqual] 
          0, \(-TxD1\) - TxD2 + TYD1 + TYD3 \[GreaterEqual] 
          0, \(-TxD1\) - TxD2 + TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          0, \(-Ax\) + AY - TxD1 - TxD2 + TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          1}, {Ax, Ay, AY, TxD1, TxD2, TyD1, TYD1, TyD2, TYD2, 
        TYD3}}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "The option ",
  StyleBox["samelinearPart -> True", "Input"],
  " constraints the coefficients of variables having the same dimension to be \
the same, only for local variables. The new variables are called TD1, TD2, \
etc. "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(scheduleConstraints[sameLinearPart -> True]\)], "Input"],

Cell[BoxData[
    \({{TxD2 \[GreaterEqual] 0, TxD1 + TxD2 \[GreaterEqual] 0, 
        TyD2 \[GreaterEqual] 0, TyD1 + TyD2 \[GreaterEqual] 0, 
        TYD2 \[GreaterEqual] 1, TyD2 - TYD3 \[GreaterEqual] 0, 
        TyD1 - TYD1 + TyD2 - TYD2 - TYD3 \[GreaterEqual] 0, 
        Ay - AY + TyD1 - TYD1 + TyD2 - TYD2 - TYD3 \[GreaterEqual] 1, 
        TYD3 \[GreaterEqual] 0, \(-TxD2\) + TYD3 \[GreaterEqual] 0, 
        TYD1 + TYD3 \[GreaterEqual] 
          0, \(-TxD1\) - TxD2 + TYD1 + TYD3 \[GreaterEqual] 0, 
        TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          0, \(-TxD1\) - TxD2 + TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          0, \(-Ax\) + AY - TxD1 - TxD2 + TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          1}, {Ax, Ay, AY, TxD1, TxD2, TyD1, TYD1, TyD2, TYD2, 
        TYD3}}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "The timeMinSchedule function builds constraints needed to optimize the \
total computation time of an algorithm. To do so, we model the total time as \
a linear function of the parameters, and we impose that this function be \
greater than the timing functions of all variables. The total function has \
the form ",
  StyleBox["TN n + aparam", "Input"],
  "."
}], "Text",
  TextJustification->1],

Cell[CellGroupData[{

Cell[BoxData[
    \(statScheduleConstraints[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Total time: "\[InvisibleSpace]\(0.02336200000000055`\
\ Second\)\),
      SequenceForm[ "Total time: ", 
        Times[ 0.023362000000000549, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("#variables: "\[InvisibleSpace]10\),
      SequenceForm[ "#variables: ", 10],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("#constraints: "\[InvisibleSpace]15\),
      SequenceForm[ "#constraints: ", 15],
      Editable->False]], "Print"],

Cell[BoxData[
    \({{TxD2 \[GreaterEqual] 0, TxD1 + TxD2 \[GreaterEqual] 0, 
        TyD2 \[GreaterEqual] 0, TyD1 + TyD2 \[GreaterEqual] 0, 
        TYD2 \[GreaterEqual] 1, TyD2 - TYD3 \[GreaterEqual] 0, 
        TyD1 - TYD1 + TyD2 - TYD2 - TYD3 \[GreaterEqual] 0, 
        Ay - AY + TyD1 - TYD1 + TyD2 - TYD2 - TYD3 \[GreaterEqual] 1, 
        TYD3 \[GreaterEqual] 0, \(-TxD2\) + TYD3 \[GreaterEqual] 0, 
        TYD1 + TYD3 \[GreaterEqual] 
          0, \(-TxD1\) - TxD2 + TYD1 + TYD3 \[GreaterEqual] 0, 
        TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          0, \(-TxD1\) - TxD2 + TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          0, \(-Ax\) + AY - TxD1 - TxD2 + TYD1 + TYD2 + TYD3 \[GreaterEqual] 
          1}, {Ax, Ay, AY, TxD1, TxD2, TyD1, TYD1, TyD2, TYD2, 
        TYD3}}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
  "The ",
  StyleBox["timeMinSchedConstraints", "Input"],
  " function"
}], "Subsection"],

Cell["\<\
This function builds a constraint set in order to find out the \
time-minimal schedule for a program. \
\>", "Text",
  TextJustification->1],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?dom2mma\)\)], "Input"],

Cell[BoxData[
    \("dom2mma[d] converts Alpha domain d into a pair {constraints,index} \
(e.g. {{i,j},{i+j>=2,...}}) suitable for Mathematica."\)], "Print",
  CellTags->"Info3407938364-2610364"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?domCompRays\)\)], "Input"],

Cell[BoxData[
    \("domCompRays[dom] recomputes the rays of the Alpha domain dom and \
returns an Alpha domain. domCompRays ignores the ray part of dom, and \
recomputes it. This function allows one to modify a contraint in a domain, \
and to update the domain accordingly."\)], "Print",
  CellTags->"Info3407938399-2759567"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?timeMinSchedConstraints\)\)], "Input"],

Cell[BoxData[
    \("Timeminschedconstraints[] computes the constraints needed to minimize \
the total execution time of $result. timeMinSchedConstraints[var_String] \
computes only the dependence constraints for variable var. Options of \
timeMinSchedConstraints are sameLinearPart (default False), addConstraints \
({}) and verbose (False)."\)], "Print",
  CellTags->"Info3407930501-4698775"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(timeMinSchedConstraints[]\)], "Input"],

Cell[BoxData[
    \({{{"x", {TN - TxD2 \[GreaterEqual] 0, 
            TN - TxD1 - TxD2 \[GreaterEqual] 0, 
            aparam - Ax + TN - TxD1 - TxD2 \[GreaterEqual] 
              0}}, {"y", {TN - TyD2 \[GreaterEqual] 0, 
            TN - TyD1 - TyD2 \[GreaterEqual] 0, 
            aparam - Ay + TN - TyD1 - TyD2 \[GreaterEqual] 
              0}}, {"Y", {TN - TYD3 \[GreaterEqual] 0, 
            TN - TYD1 - TYD3 \[GreaterEqual] 0, 
            aparam - AY + TN - TYD1 - TYD3 \[GreaterEqual] 0, 
            TN - TYD1 - TYD2 - TYD3 \[GreaterEqual] 0, 
            aparam - AY + TN - TYD1 - TYD2 - TYD3 \[GreaterEqual] 
              0}}}, {aparam, Ax, Ay, AY, TN, TxD1, TxD2, TyD1, TYD1, TyD2, 
        TYD2, TYD3}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Options[timeMinSchedConstraints]\)], "Input"],

Cell[BoxData[
    \({scheduleType \[Rule] affineByVar, verbose \[Rule] False, 
      onlyVars \[Rule] all, reuseSchedule \[Rule] False, 
      onlyEquations \[Rule] all, addConstraints \[Rule] {}, 
      verticesPositive \[Rule] False, depComputed \[Rule] False, 
      dataFlow \[Rule] False}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
  "The ",
  StyleBox["summaryScheduleConstraints", "Input"],
  " function"
}], "Subsection"],

Cell["\<\
This function was added in 2007 to project dependencies on the input and \
output variables. \
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(summaryScheduleConstraints[]\)], "Input"],

Cell[BoxData[
    \({{TxD2 \[GreaterEqual] 0, 
        TxD1 + TxD2 \[GreaterEqual] 0, \(-TxD2\) + TyD2 \[GreaterEqual] 
          0, \(-TxD1\) - TxD2 + TyD1 + TyD2 \[GreaterEqual] 
          1, \(-Ax\) + Ay - TxD1 - TxD2 + TyD1 + TyD2 \[GreaterEqual] 
          2}, {Ax, Ay, TxD1, TxD2, TyD1, TyD2}}\)], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["4. A more involved example", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(load["\<testschedule2.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \("[testSchedule2]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \("system testSchedule2 :{N,M | 1<=N; 2<=M}\n                     (x : {i \
| 1<=i<=M} of integer)\n       returns       (y : {i | 1<=i<=M} of integer);\n\
var\n  A : {i | 1<=i<=M} of integer;\n  B : {i,j | 1<=i<=M; 1<=j<=N} of \
integer;\nlet\n  A[i] = \n      case\n        { | i=1} : x[i];\n        { | \
2<=i} : x[i] + B[i-1,N];\n      esac;\n  B[i,j] = \n      case\n        { | \
j=1} : A[i];\n        { | 2<=j} : B[i,j-1] + x[i];\n      esac;\n  y[i] = \
A[i];\ntel;"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(analyze[]\)], "Input"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Static Analysis of system \
"\[InvisibleSpace]"testSchedule2"\),
      SequenceForm[ "Static Analysis of system ", "testSchedule2"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--Checking declaration of variables."\)], "Print"],

Cell[BoxData[
    \("--Checking single assignment rule."\)], "Print"],

Cell[BoxData[
    \("--Checking definitions of output/local variables."\)], "Print"],

Cell[BoxData[
    \("--Checking definition of input variables."\)], "Print"],

Cell[BoxData[
    \("--Checking that input/local variables are used."\)], "Print"],

Cell[BoxData[
    \("--Checking type and domain consistency in the equations:"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("----equation defining "\[InvisibleSpace]"B"\),
      SequenceForm[ "----equation defining ", "B"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("----equation defining "\[InvisibleSpace]"y"\),
      SequenceForm[ "----equation defining ", "y"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--Checking regularity of output equations"\)], "Print"],

Cell[BoxData[
    \("----Outputs equations are regular."\)], "Print"],

Cell[BoxData[
    \("\nAnalysis Successful..."\)], "Print"],

Cell[BoxData[
    \(True\)], "Output"]
}, Closed]],

Cell["This system of equations has no one dimensional schedule. ", "Text"],

Cell[BoxData[
    \(<< Alpha/VertexSchedule.m\)], "Input"],

Cell[BoxData[
    \(scheduleD\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \($scheduleMDSol[1]\)], "Input"],

Cell[BoxData[
    \({\(-AA\) + 3\ AB - 4\ Ax + 2\ Ay + 7\ TBD1 + 7\ TBD2 + 9\ TBD3 + 
        10\ TBD4 + TN - 7\ TxD1 - 8\ TxD2 - 10\ TxD3 + 5\ TyD1 + 4\ TyD2 + 
        8\ TyD3, {TAD2 \[GreaterEqual] 0, TAD3 \[GreaterEqual] 0, 
        TAD1 + TAD3 \[GreaterEqual] 0, TBD2 \[GreaterEqual] 1, 
        TAD2 - TBD2 - TBD3 \[GreaterEqual] 0, 
        TBD3 \[GreaterEqual] 0, \(-TAD2\) + TBD3 \[GreaterEqual] 0, 
        TBD2 + TBD3 \[GreaterEqual] 0, 
        AA - AB + 2\ TAD1 + TAD2 + 2\ TAD3 - TBD1 - TBD2 - TBD3 - 
            2\ TBD4 \[GreaterEqual] 1, TAD3 - TBD4 \[GreaterEqual] 0, 
        TAD1 + TAD3 - TBD1 - TBD4 \[GreaterEqual] 0, 
        TBD4 \[GreaterEqual] 0, \(-TAD3\) + TBD4 \[GreaterEqual] 0, 
        TBD1 + TBD4 \[GreaterEqual] 
          0, \(-TAD1\) - TAD3 + TBD1 + TBD4 \[GreaterEqual] 
          0, \(-AA\) + AB - TAD1 - TAD2 - 2\ TAD3 + TBD1 + TBD2 + TBD3 + 
            2\ TBD4 \[GreaterEqual] 
          1, \(-AA\) + AB - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TBD1 + TBD2 + 
            TBD3 + 2\ TBD4 \[GreaterEqual] 1, \(-TAD3\) + TM \[GreaterEqual] 
          0, \(-TAD1\) - TAD3 + TM \[GreaterEqual] 
          0, \(-TBD4\) + TM \[GreaterEqual] 
          0, \(-TBD1\) - TBD4 + TM \[GreaterEqual] 
          0, \(-TAD2\) + TN \[GreaterEqual] 0, \(-TBD3\) + TN \[GreaterEqual] 
          0, \(-TBD2\) - TBD3 + TN \[GreaterEqual] 
          0, \(-AA\) + aparam - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TM + 
            TN \[GreaterEqual] 
          0, \(-AA\) + aparam - TAD1 - TAD2 - 2\ TAD3 + 2\ TM + 
            TN \[GreaterEqual] 
          0, \(-AB\) + aparam - 2\ TBD1 - TBD2 - TBD3 - 2\ TBD4 + 2\ TM + 
            TN \[GreaterEqual] 
          0, \(-AB\) + aparam - TBD1 - TBD2 - TBD3 - 2\ TBD4 + 2\ TM + 
            TN \[GreaterEqual] 0, TAD2 - TxD2 \[GreaterEqual] 0, 
        TBD3 - TxD2 \[GreaterEqual] 0, TBD2 + TBD3 - TxD2 \[GreaterEqual] 0, 
        TN - TxD2 \[GreaterEqual] 0, TxD2 \[GreaterEqual] 0, 
        AB - Ax + 2\ TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - 2\ TxD1 - 2\ TxD2 - 
            2\ TxD3 \[GreaterEqual] 1, 
        AB - Ax + TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - TxD1 - 2\ TxD2 - 
            2\ TxD3 \[GreaterEqual] 1, 
        AA - Ax + 2\ TAD1 + TAD2 + 2\ TAD3 - 2\ TxD1 - TxD2 - 
            2\ TxD3 \[GreaterEqual] 1, 
        aparam - Ax + 2\ TM + TN - 2\ TxD1 - TxD2 - 2\ TxD3 \[GreaterEqual] 
          0, AA - Ax + TAD1 + TAD2 + 2\ TAD3 - TxD1 - TxD2 - 
            2\ TxD3 \[GreaterEqual] 1, 
        aparam - Ax + 2\ TM + TN - TxD1 - TxD2 - 2\ TxD3 \[GreaterEqual] 0, 
        TAD3 - TxD3 \[GreaterEqual] 0, TBD4 - TxD3 \[GreaterEqual] 0, 
        TM - TxD3 \[GreaterEqual] 0, 
        TAD1 + TAD3 - TxD1 - TxD3 \[GreaterEqual] 0, 
        TBD1 + TBD4 - TxD1 - TxD3 \[GreaterEqual] 0, 
        TM - TxD1 - TxD3 \[GreaterEqual] 0, TxD3 \[GreaterEqual] 0, 
        TxD1 + TxD3 \[GreaterEqual] 0, TN - TyD2 \[GreaterEqual] 0, 
        TyD2 \[GreaterEqual] 0, \(-TAD2\) + TyD2 \[GreaterEqual] 0, 
        aparam - Ay + 2\ TM + TN - 2\ TyD1 - TyD2 - 2\ TyD3 \[GreaterEqual] 
          0, aparam - Ay + 2\ TM + TN - TyD1 - TyD2 - 2\ TyD3 \[GreaterEqual] 
          0, TM - TyD3 \[GreaterEqual] 0, TM - TyD1 - TyD3 \[GreaterEqual] 0, 
        TyD3 \[GreaterEqual] 0, \(-TAD3\) + TyD3 \[GreaterEqual] 0, 
        TyD1 + TyD3 \[GreaterEqual] 
          0, \(-TAD1\) - TAD3 + TyD1 + TyD3 \[GreaterEqual] 
          0, \(-AA\) + Ay - TAD1 - TAD2 - 2\ TAD3 + TyD1 + TyD2 + 
            2\ TyD3 \[GreaterEqual] 
          1, \(-AA\) + Ay - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TyD1 + TyD2 + 
            2\ TyD3 \[GreaterEqual] 1}, {AA, AB, aparam, Ax, Ay, TAD1, TAD2, 
        TAD3, TBD1, TBD2, TBD3, TBD4, TM, TN, TxD1, TxD2, TxD3, TyD1, TyD2, 
        TyD3}, {\[Infinity], {AA \[Rule] Indeterminate, 
          AB \[Rule] Indeterminate, aparam \[Rule] Indeterminate, 
          Ax \[Rule] Indeterminate, Ay \[Rule] Indeterminate, 
          TAD1 \[Rule] Indeterminate, TAD2 \[Rule] Indeterminate, 
          TAD3 \[Rule] Indeterminate, TBD1 \[Rule] Indeterminate, 
          TBD2 \[Rule] Indeterminate, TBD3 \[Rule] Indeterminate, 
          TBD4 \[Rule] Indeterminate, TM \[Rule] Indeterminate, 
          TN \[Rule] Indeterminate, TxD1 \[Rule] Indeterminate, 
          TxD2 \[Rule] Indeterminate, TxD3 \[Rule] Indeterminate, 
          TyD1 \[Rule] Indeterminate, TyD2 \[Rule] Indeterminate, 
          TyD3 \[Rule] Indeterminate}}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(scd[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Additional objective function: \
"\[InvisibleSpace]{TAD2, TAD3, TAD1 + TAD3, \(-1\) + TBD2, 
          TAD2 - TBD2 - TBD3, TBD3, \(-TAD2\) + TBD3, 
          TBD2 + TBD3, \(-1\) + AA - AB + 2\ TAD1 + TAD2 + 2\ TAD3 - TBD1 - 
            TBD2 - TBD3 - 2\ TBD4, TAD3 - TBD4, TAD1 + TAD3 - TBD1 - TBD4, 
          TBD4, \(-TAD3\) + TBD4, 
          TBD1 + TBD4, \(-TAD1\) - TAD3 + TBD1 + TBD4, \(-1\) - AA + AB - 
            TAD1 - TAD2 - 2\ TAD3 + TBD1 + TBD2 + TBD3 + 2\ TBD4, \(-1\) - 
            AA + AB - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TBD1 + TBD2 + TBD3 + 
            2\ TBD4, \(-TAD3\) + TM, \(-TAD1\) - TAD3 + TM, \(-TBD4\) + 
            TM, \(-TBD1\) - TBD4 + TM, \(-TAD2\) + TN, \(-TBD3\) + 
            TN, \(-TBD2\) - TBD3 + TN, \(-AA\) + aparam - 2\ TAD1 - TAD2 - 
            2\ TAD3 + 2\ TM + TN, \(-AA\) + aparam - TAD1 - TAD2 - 2\ TAD3 + 
            2\ TM + TN, \(-AB\) + aparam - 2\ TBD1 - TBD2 - TBD3 - 2\ TBD4 + 
            2\ TM + TN, \(-AB\) + aparam - TBD1 - TBD2 - TBD3 - 2\ TBD4 + 
            2\ TM + TN, TAD2 - TxD2, TBD3 - TxD2, TBD2 + TBD3 - TxD2, 
          TN - TxD2, 
          TxD2, \(-1\) + AB - Ax + 2\ TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - 
            2\ TxD1 - 2\ TxD2 - 2\ TxD3, \(-1\) + AB - Ax + TBD1 + 2\ TBD2 + 
            2\ TBD3 + 2\ TBD4 - TxD1 - 2\ TxD2 - 2\ TxD3, \(-1\) + AA - Ax + 
            2\ TAD1 + TAD2 + 2\ TAD3 - 2\ TxD1 - TxD2 - 2\ TxD3, 
          aparam - Ax + 2\ TM + TN - 2\ TxD1 - TxD2 - 2\ TxD3, \(-1\) + AA - 
            Ax + TAD1 + TAD2 + 2\ TAD3 - TxD1 - TxD2 - 2\ TxD3, 
          aparam - Ax + 2\ TM + TN - TxD1 - TxD2 - 2\ TxD3, TAD3 - TxD3, 
          TBD4 - TxD3, TM - TxD3, TAD1 + TAD3 - TxD1 - TxD3, 
          TBD1 + TBD4 - TxD1 - TxD3, TM - TxD1 - TxD3, TxD3, TxD1 + TxD3, 
          TN - TyD2, TyD2, \(-TAD2\) + TyD2, 
          aparam - Ay + 2\ TM + TN - 2\ TyD1 - TyD2 - 2\ TyD3, 
          aparam - Ay + 2\ TM + TN - TyD1 - TyD2 - 2\ TyD3, TM - TyD3, 
          TM - TyD1 - TyD3, TyD3, \(-TAD3\) + TyD3, 
          TyD1 + TyD3, \(-TAD1\) - TAD3 + TyD1 + TyD3, \(-1\) - AA + Ay - 
            TAD1 - TAD2 - 2\ TAD3 + TyD1 + TyD2 + 2\ TyD3, \(-1\) - AA + Ay - 
            2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TyD1 + TyD2 + 2\ TyD3}\),
      SequenceForm[ "Additional objective function: ", {TAD2, TAD3, 
        Plus[ TAD1, TAD3], 
        Plus[ -1, TBD2], 
        Plus[ TAD2, 
          Times[ -1, TBD2], 
          Times[ -1, TBD3]], TBD3, 
        Plus[ 
          Times[ -1, TAD2], TBD3], 
        Plus[ TBD2, TBD3], 
        Plus[ -1, AA, 
          Times[ -1, AB], 
          Times[ 2, TAD1], TAD2, 
          Times[ 2, TAD3], 
          Times[ -1, TBD1], 
          Times[ -1, TBD2], 
          Times[ -1, TBD3], 
          Times[ -2, TBD4]], 
        Plus[ TAD3, 
          Times[ -1, TBD4]], 
        Plus[ TAD1, TAD3, 
          Times[ -1, TBD1], 
          Times[ -1, TBD4]], TBD4, 
        Plus[ 
          Times[ -1, TAD3], TBD4], 
        Plus[ TBD1, TBD4], 
        Plus[ 
          Times[ -1, TAD1], 
          Times[ -1, TAD3], TBD1, TBD4], 
        Plus[ -1, 
          Times[ -1, AA], AB, 
          Times[ -1, TAD1], 
          Times[ -1, TAD2], 
          Times[ -2, TAD3], TBD1, TBD2, TBD3, 
          Times[ 2, TBD4]], 
        Plus[ -1, 
          Times[ -1, AA], AB, 
          Times[ -2, TAD1], 
          Times[ -1, TAD2], 
          Times[ -2, TAD3], 
          Times[ 2, TBD1], TBD2, TBD3, 
          Times[ 2, TBD4]], 
        Plus[ 
          Times[ -1, TAD3], TM], 
        Plus[ 
          Times[ -1, TAD1], 
          Times[ -1, TAD3], TM], 
        Plus[ 
          Times[ -1, TBD4], TM], 
        Plus[ 
          Times[ -1, TBD1], 
          Times[ -1, TBD4], TM], 
        Plus[ 
          Times[ -1, TAD2], TN], 
        Plus[ 
          Times[ -1, TBD3], TN], 
        Plus[ 
          Times[ -1, TBD2], 
          Times[ -1, TBD3], TN], 
        Plus[ 
          Times[ -1, AA], aparam, 
          Times[ -2, TAD1], 
          Times[ -1, TAD2], 
          Times[ -2, TAD3], 
          Times[ 2, TM], TN], 
        Plus[ 
          Times[ -1, AA], aparam, 
          Times[ -1, TAD1], 
          Times[ -1, TAD2], 
          Times[ -2, TAD3], 
          Times[ 2, TM], TN], 
        Plus[ 
          Times[ -1, AB], aparam, 
          Times[ -2, TBD1], 
          Times[ -1, TBD2], 
          Times[ -1, TBD3], 
          Times[ -2, TBD4], 
          Times[ 2, TM], TN], 
        Plus[ 
          Times[ -1, AB], aparam, 
          Times[ -1, TBD1], 
          Times[ -1, TBD2], 
          Times[ -1, TBD3], 
          Times[ -2, TBD4], 
          Times[ 2, TM], TN], 
        Plus[ TAD2, 
          Times[ -1, TxD2]], 
        Plus[ TBD3, 
          Times[ -1, TxD2]], 
        Plus[ TBD2, TBD3, 
          Times[ -1, TxD2]], 
        Plus[ TN, 
          Times[ -1, TxD2]], TxD2, 
        Plus[ -1, AB, 
          Times[ -1, Ax], 
          Times[ 2, TBD1], 
          Times[ 2, TBD2], 
          Times[ 2, TBD3], 
          Times[ 2, TBD4], 
          Times[ -2, TxD1], 
          Times[ -2, TxD2], 
          Times[ -2, TxD3]], 
        Plus[ -1, AB, 
          Times[ -1, Ax], TBD1, 
          Times[ 2, TBD2], 
          Times[ 2, TBD3], 
          Times[ 2, TBD4], 
          Times[ -1, TxD1], 
          Times[ -2, TxD2], 
          Times[ -2, TxD3]], 
        Plus[ -1, AA, 
          Times[ -1, Ax], 
          Times[ 2, TAD1], TAD2, 
          Times[ 2, TAD3], 
          Times[ -2, TxD1], 
          Times[ -1, TxD2], 
          Times[ -2, TxD3]], 
        Plus[ aparam, 
          Times[ -1, Ax], 
          Times[ 2, TM], TN, 
          Times[ -2, TxD1], 
          Times[ -1, TxD2], 
          Times[ -2, TxD3]], 
        Plus[ -1, AA, 
          Times[ -1, Ax], TAD1, TAD2, 
          Times[ 2, TAD3], 
          Times[ -1, TxD1], 
          Times[ -1, TxD2], 
          Times[ -2, TxD3]], 
        Plus[ aparam, 
          Times[ -1, Ax], 
          Times[ 2, TM], TN, 
          Times[ -1, TxD1], 
          Times[ -1, TxD2], 
          Times[ -2, TxD3]], 
        Plus[ TAD3, 
          Times[ -1, TxD3]], 
        Plus[ TBD4, 
          Times[ -1, TxD3]], 
        Plus[ TM, 
          Times[ -1, TxD3]], 
        Plus[ TAD1, TAD3, 
          Times[ -1, TxD1], 
          Times[ -1, TxD3]], 
        Plus[ TBD1, TBD4, 
          Times[ -1, TxD1], 
          Times[ -1, TxD3]], 
        Plus[ TM, 
          Times[ -1, TxD1], 
          Times[ -1, TxD3]], TxD3, 
        Plus[ TxD1, TxD3], 
        Plus[ TN, 
          Times[ -1, TyD2]], TyD2, 
        Plus[ 
          Times[ -1, TAD2], TyD2], 
        Plus[ aparam, 
          Times[ -1, Ay], 
          Times[ 2, TM], TN, 
          Times[ -2, TyD1], 
          Times[ -1, TyD2], 
          Times[ -2, TyD3]], 
        Plus[ aparam, 
          Times[ -1, Ay], 
          Times[ 2, TM], TN, 
          Times[ -1, TyD1], 
          Times[ -1, TyD2], 
          Times[ -2, TyD3]], 
        Plus[ TM, 
          Times[ -1, TyD3]], 
        Plus[ TM, 
          Times[ -1, TyD1], 
          Times[ -1, TyD3]], TyD3, 
        Plus[ 
          Times[ -1, TAD3], TyD3], 
        Plus[ TyD1, TyD3], 
        Plus[ 
          Times[ -1, TAD1], 
          Times[ -1, TAD3], TyD1, TyD3], 
        Plus[ -1, 
          Times[ -1, AA], Ay, 
          Times[ -1, TAD1], 
          Times[ -1, TAD2], 
          Times[ -2, TAD3], TyD1, TyD2, 
          Times[ 2, TyD3]], 
        Plus[ -1, 
          Times[ -1, AA], Ay, 
          Times[ -2, TAD1], 
          Times[ -1, TAD2], 
          Times[ -2, TAD3], 
          Times[ 2, TyD1], TyD2, 
          Times[ 2, TyD3]]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Additional objective function: \
"\[InvisibleSpace]\(\(-10\) - 3\ AA + AB + 8\ aparam - 6\ Ax - 4\ TAD1 - 
          3\ TAD2 - 6\ TAD3 + 3\ TBD1 + 4\ TBD2 + 5\ TBD3 + 4\ TBD4 + 
          24\ TM + 13\ TN - 11\ TxD1 - 11\ TxD2 - 16\ TxD3 + TyD1 + TyD2 + 
          2\ TyD3\)\),
      SequenceForm[ "Additional objective function: ", 
        Plus[ -10, 
          Times[ -3, AA], AB, 
          Times[ 8, aparam], 
          Times[ -6, Ax], 
          Times[ -4, TAD1], 
          Times[ -3, TAD2], 
          Times[ -6, TAD3], 
          Times[ 3, TBD1], 
          Times[ 4, TBD2], 
          Times[ 5, TBD3], 
          Times[ 4, TBD4], 
          Times[ 24, TM], 
          Times[ 13, TN], 
          Times[ -11, TxD1], 
          Times[ -11, TxD2], 
          Times[ -16, TxD3], TyD1, TyD2, 
          Times[ 2, TyD3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(-10\) - 3\ AA + AB + 8\ aparam - 6\ Ax - 4\ TAD1 - 3\ TAD2 - 
      6\ TAD3 + 3\ TBD1 + 4\ TBD2 + 5\ TBD3 + 4\ TBD4 + 24\ TM + 14\ TN - 
      11\ TxD1 - 11\ TxD2 - 16\ TxD3 + TyD1 + TyD2 + 2\ TyD3\)], "Print"],

Cell[BoxData[
    RowBox[{\(Minimize::"natt"\), \(\(:\)\(\ \)\), "\<\"The minimum is not \
attained at any point satisfying the given constraints. \\!\\(\\*ButtonBox[\\\
\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\\\", \
ButtonFrame->None, ButtonData:>\\\"Minimize::natt\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    InterpretationBox[\(60\[InvisibleSpace]" constraints, \
"\[InvisibleSpace]20\[InvisibleSpace]" variables \
"\[InvisibleSpace]\(0.024214999999999876`\ Second\)\),
      SequenceForm[ 60, " constraints, ", 20, " variables ", 
        Times[ 0.024214999999999876, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    \({"testSchedule2", 20, 60, 0.024214999999999876`\ Second, "Vertex", 
      "Flat"}\)], "Print"],

Cell[BoxData[
    \(scheduleResult[
      "testSchedule2", {{"A", {"i", "N", "M"}, 
          sched[{Indeterminate, Indeterminate, Indeterminate}, 
            Indeterminate]}, {"B", {"i", "j", "N", "M"}, 
          sched[{Indeterminate, Indeterminate, Indeterminate, Indeterminate}, 
            Indeterminate]}, {"x", {"i", "N", "M"}, 
          sched[{Indeterminate, Indeterminate, Indeterminate}, 
            Indeterminate]}, {"y", {"i", "N", "M"}, 
          sched[{Indeterminate, Indeterminate, Indeterminate}, 
            Indeterminate]}}, {Indeterminate, Indeterminate, 
        Indeterminate}]\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(<< Alpha/VertexSchedule.m\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(scd[multidimensional \[Rule] True, verbose \[Rule] True, 
      integerSolution \[Rule] 
        False (*\(,\)\(\ \)\(addConstraints \[Rule] {{TM == 
                0}}\)*) ]\)], "Input"],

Cell[BoxData[
    \("Calling scheduleConstraints"\)], "Print"],

Cell[BoxData[
    \("Dependencies kept are:"\)], "Print"],

Cell[BoxData[
    \("{i,N,M | i=1; 1<=N; 2<=M} : A[i,N,M] -> x[i,N,M]\n{i,N,M | 2<=i<=M; \
1<=N} : A[i,N,M] -> x[i,N,M]\n{i,N,M | 2<=i<=M; 1<=N} : A[i,N,M] -> \
B[i-1,N,N,M]\n{i,j,N,M | 1<=i<=M; j=1; 1<=N; 2<=M} : B[i,j,N,M] -> \
A[i,j+N-1,M]\n{i,j,N,M | 1<=i<=M; 2<=j<=N; 2<=M} : B[i,j,N,M] -> B[i,j-1,N,M]\
\n{i,j,N,M | 1<=i<=M; 2<=j<=N; 2<=M} : B[i,j,N,M] -> x[i,N,M]\n{i,N,M | \
1<=i<=M; 1<=N; 2<=M} : y[i,N,M] -> A[i,N,M]"\)], "Print"],

Cell[BoxData[
    \("----------------------------"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"x"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]\*"\<\"\"\>"\),
      SequenceForm[ "--- LHS variable: ", "x", ", RHS variable: ", ""],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TxD2 \
\[GreaterEqual] 0, TxD3 \[GreaterEqual] 0, TxD1 + TxD3 \[GreaterEqual] 0}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TxD2, 0], 
        GreaterEqual[ TxD3, 0], 
        GreaterEqual[ 
          Plus[ TxD1, TxD3], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"A"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"x"\),
      SequenceForm[ "--- LHS variable: ", "A", ", RHS variable: ", "x"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TAD2 \
\[GreaterEqual] 0, TAD3 \[GreaterEqual] 0, TAD2 - TxD2 \[GreaterEqual] 0, 
          AA - Ax + TAD1 + TAD2 + 2\ TAD3 - TxD1 - TxD2 - 
              2\ TxD3 \[GreaterEqual] 1, TAD3 - TxD3 \[GreaterEqual] 0}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TAD2, 0], 
        GreaterEqual[ TAD3, 0], 
        GreaterEqual[ 
          Plus[ TAD2, 
            Times[ -1, TxD2]], 0], 
        GreaterEqual[ 
          Plus[ AA, 
            Times[ -1, Ax], TAD1, TAD2, 
            Times[ 2, TAD3], 
            Times[ -1, TxD1], 
            Times[ -1, TxD2], 
            Times[ -2, TxD3]], 1], 
        GreaterEqual[ 
          Plus[ TAD3, 
            Times[ -1, TxD3]], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"A"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"x"\),
      SequenceForm[ "--- LHS variable: ", "A", ", RHS variable: ", "x"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TAD2 \
\[GreaterEqual] 0, TAD3 \[GreaterEqual] 0, TAD1 + TAD3 \[GreaterEqual] 0, 
          TAD2 - TxD2 \[GreaterEqual] 0, 
          AA - Ax + 2\ TAD1 + TAD2 + 2\ TAD3 - 2\ TxD1 - TxD2 - 
              2\ TxD3 \[GreaterEqual] 1, TAD3 - TxD3 \[GreaterEqual] 0, 
          TAD1 + TAD3 - TxD1 - TxD3 \[GreaterEqual] 0}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TAD2, 0], 
        GreaterEqual[ TAD3, 0], 
        GreaterEqual[ 
          Plus[ TAD1, TAD3], 0], 
        GreaterEqual[ 
          Plus[ TAD2, 
            Times[ -1, TxD2]], 0], 
        GreaterEqual[ 
          Plus[ AA, 
            Times[ -1, Ax], 
            Times[ 2, TAD1], TAD2, 
            Times[ 2, TAD3], 
            Times[ -2, TxD1], 
            Times[ -1, TxD2], 
            Times[ -2, TxD3]], 1], 
        GreaterEqual[ 
          Plus[ TAD3, 
            Times[ -1, TxD3]], 0], 
        GreaterEqual[ 
          Plus[ TAD1, TAD3, 
            Times[ -1, TxD1], 
            Times[ -1, TxD3]], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"A"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"B"\),
      SequenceForm[ "--- LHS variable: ", "A", ", RHS variable: ", "B"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TAD2 \
\[GreaterEqual] 0, TAD3 \[GreaterEqual] 0, TAD1 + TAD3 \[GreaterEqual] 0, 
          TAD2 - TBD2 - TBD3 \[GreaterEqual] 0, 
          AA - AB + 2\ TAD1 + TAD2 + 2\ TAD3 - TBD1 - TBD2 - TBD3 - 
              2\ TBD4 \[GreaterEqual] 1, TAD3 - TBD4 \[GreaterEqual] 0, 
          TAD1 + TAD3 - TBD1 - TBD4 \[GreaterEqual] 0}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TAD2, 0], 
        GreaterEqual[ TAD3, 0], 
        GreaterEqual[ 
          Plus[ TAD1, TAD3], 0], 
        GreaterEqual[ 
          Plus[ TAD2, 
            Times[ -1, TBD2], 
            Times[ -1, TBD3]], 0], 
        GreaterEqual[ 
          Plus[ AA, 
            Times[ -1, AB], 
            Times[ 2, TAD1], TAD2, 
            Times[ 2, TAD3], 
            Times[ -1, TBD1], 
            Times[ -1, TBD2], 
            Times[ -1, TBD3], 
            Times[ -2, TBD4]], 1], 
        GreaterEqual[ 
          Plus[ TAD3, 
            Times[ -1, TBD4]], 0], 
        GreaterEqual[ 
          Plus[ TAD1, TAD3, 
            Times[ -1, TBD1], 
            Times[ -1, TBD4]], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"B"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"A"\),
      SequenceForm[ "--- LHS variable: ", "B", ", RHS variable: ", "A"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TBD3 \
\[GreaterEqual] 0, \(-TAD2\) + TBD3 \[GreaterEqual] 0, 
          TBD4 \[GreaterEqual] 0, \(-TAD3\) + TBD4 \[GreaterEqual] 0, 
          TBD1 + TBD4 \[GreaterEqual] 
            0, \(-TAD1\) - TAD3 + TBD1 + TBD4 \[GreaterEqual] 
            0, \(-AA\) + AB - TAD1 - TAD2 - 2\ TAD3 + TBD1 + TBD2 + TBD3 + 
              2\ TBD4 \[GreaterEqual] 
            1, \(-AA\) + AB - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TBD1 + TBD2 + 
              TBD3 + 2\ TBD4 \[GreaterEqual] 1}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TBD3, 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD2], TBD3], 0], 
        GreaterEqual[ TBD4, 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD3], TBD4], 0], 
        GreaterEqual[ 
          Plus[ TBD1, TBD4], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD1], 
            Times[ -1, TAD3], TBD1, TBD4], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], AB, 
            Times[ -1, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], TBD1, TBD2, TBD3, 
            Times[ 2, TBD4]], 1], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], AB, 
            Times[ -2, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], 
            Times[ 2, TBD1], TBD2, TBD3, 
            Times[ 2, TBD4]], 1]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"B"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"B"\),
      SequenceForm[ "--- LHS variable: ", "B", ", RHS variable: ", "B"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{True, 
          TBD2 \[GreaterEqual] 1, TBD3 \[GreaterEqual] 0, 
          TBD2 + TBD3 \[GreaterEqual] 0, TBD4 \[GreaterEqual] 0, 
          TBD1 + TBD4 \[GreaterEqual] 0}\),
      SequenceForm[ "    Constraints: ", {True, 
        GreaterEqual[ TBD2, 1], 
        GreaterEqual[ TBD3, 0], 
        GreaterEqual[ 
          Plus[ TBD2, TBD3], 0], 
        GreaterEqual[ TBD4, 0], 
        GreaterEqual[ 
          Plus[ TBD1, TBD4], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"B"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"x"\),
      SequenceForm[ "--- LHS variable: ", "B", ", RHS variable: ", "x"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TBD3 \
\[GreaterEqual] 0, TBD2 + TBD3 \[GreaterEqual] 0, TBD4 \[GreaterEqual] 0, 
          TBD1 + TBD4 \[GreaterEqual] 0, TBD3 - TxD2 \[GreaterEqual] 0, 
          TBD2 + TBD3 - TxD2 \[GreaterEqual] 0, 
          AB - Ax + 2\ TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - 2\ TxD1 - 
              2\ TxD2 - 2\ TxD3 \[GreaterEqual] 1, 
          AB - Ax + TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - TxD1 - 2\ TxD2 - 
              2\ TxD3 \[GreaterEqual] 1, TBD4 - TxD3 \[GreaterEqual] 0, 
          TBD1 + TBD4 - TxD1 - TxD3 \[GreaterEqual] 0}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TBD3, 0], 
        GreaterEqual[ 
          Plus[ TBD2, TBD3], 0], 
        GreaterEqual[ TBD4, 0], 
        GreaterEqual[ 
          Plus[ TBD1, TBD4], 0], 
        GreaterEqual[ 
          Plus[ TBD3, 
            Times[ -1, TxD2]], 0], 
        GreaterEqual[ 
          Plus[ TBD2, TBD3, 
            Times[ -1, TxD2]], 0], 
        GreaterEqual[ 
          Plus[ AB, 
            Times[ -1, Ax], 
            Times[ 2, TBD1], 
            Times[ 2, TBD2], 
            Times[ 2, TBD3], 
            Times[ 2, TBD4], 
            Times[ -2, TxD1], 
            Times[ -2, TxD2], 
            Times[ -2, TxD3]], 1], 
        GreaterEqual[ 
          Plus[ AB, 
            Times[ -1, Ax], TBD1, 
            Times[ 2, TBD2], 
            Times[ 2, TBD3], 
            Times[ 2, TBD4], 
            Times[ -1, TxD1], 
            Times[ -2, TxD2], 
            Times[ -2, TxD3]], 1], 
        GreaterEqual[ 
          Plus[ TBD4, 
            Times[ -1, TxD3]], 0], 
        GreaterEqual[ 
          Plus[ TBD1, TBD4, 
            Times[ -1, TxD1], 
            Times[ -1, TxD3]], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"y"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"A"\),
      SequenceForm[ "--- LHS variable: ", "y", ", RHS variable: ", "A"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TyD2 \
\[GreaterEqual] 0, \(-TAD2\) + TyD2 \[GreaterEqual] 0, 
          TyD3 \[GreaterEqual] 0, \(-TAD3\) + TyD3 \[GreaterEqual] 0, 
          TyD1 + TyD3 \[GreaterEqual] 
            0, \(-TAD1\) - TAD3 + TyD1 + TyD3 \[GreaterEqual] 
            0, \(-AA\) + Ay - TAD1 - TAD2 - 2\ TAD3 + TyD1 + TyD2 + 
              2\ TyD3 \[GreaterEqual] 
            1, \(-AA\) + Ay - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TyD1 + TyD2 + 
              2\ TyD3 \[GreaterEqual] 1}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TyD2, 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD2], TyD2], 0], 
        GreaterEqual[ TyD3, 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD3], TyD3], 0], 
        GreaterEqual[ 
          Plus[ TyD1, TyD3], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD1], 
            Times[ -1, TAD3], TyD1, TyD3], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], Ay, 
            Times[ -1, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], TyD1, TyD2, 
            Times[ 2, TyD3]], 1], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], Ay, 
            Times[ -2, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], 
            Times[ 2, TyD1], TyD2, 
            Times[ 2, TyD3]], 1]}],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Calling timeMinSchedConstraints"\)], "Print"],

Cell[BoxData[
    \("Looking for a multidimensional schedule..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Dependency constraints :\n"\[InvisibleSpace]{TAD2 \
\[GreaterEqual] 0, TAD3 \[GreaterEqual] 0, TAD1 + TAD3 \[GreaterEqual] 0, 
          TBD2 \[GreaterEqual] slack5, TAD2 - TBD2 - TBD3 \[GreaterEqual] 0, 
          TBD3 \[GreaterEqual] 0, \(-TAD2\) + TBD3 \[GreaterEqual] 0, 
          TBD2 + TBD3 \[GreaterEqual] 0, 
          AA - AB + 2\ TAD1 + TAD2 + 2\ TAD3 - TBD1 - TBD2 - TBD3 - 
              2\ TBD4 \[GreaterEqual] slack3, TAD3 - TBD4 \[GreaterEqual] 0, 
          TAD1 + TAD3 - TBD1 - TBD4 \[GreaterEqual] 0, 
          TBD4 \[GreaterEqual] 0, \(-TAD3\) + TBD4 \[GreaterEqual] 0, 
          TBD1 + TBD4 \[GreaterEqual] 
            0, \(-TAD1\) - TAD3 + TBD1 + TBD4 \[GreaterEqual] 
            0, \(-AA\) + AB - TAD1 - TAD2 - 2\ TAD3 + TBD1 + TBD2 + TBD3 + 
              2\ TBD4 \[GreaterEqual] 
            slack4, \(-AA\) + AB - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TBD1 + 
              TBD2 + TBD3 + 2\ TBD4 \[GreaterEqual] slack4, 
          TAD2 - TxD2 \[GreaterEqual] 0, TBD3 - TxD2 \[GreaterEqual] 0, 
          TBD2 + TBD3 - TxD2 \[GreaterEqual] 0, TxD2 \[GreaterEqual] 0, 
          AB - Ax + 2\ TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - 2\ TxD1 - 
              2\ TxD2 - 2\ TxD3 \[GreaterEqual] slack6, 
          AB - Ax + TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - TxD1 - 2\ TxD2 - 
              2\ TxD3 \[GreaterEqual] slack6, 
          AA - Ax + 2\ TAD1 + TAD2 + 2\ TAD3 - 2\ TxD1 - TxD2 - 
              2\ TxD3 \[GreaterEqual] slack2, 
          AA - Ax + TAD1 + TAD2 + 2\ TAD3 - TxD1 - TxD2 - 
              2\ TxD3 \[GreaterEqual] slack1, TAD3 - TxD3 \[GreaterEqual] 0, 
          TBD4 - TxD3 \[GreaterEqual] 0, 
          TAD1 + TAD3 - TxD1 - TxD3 \[GreaterEqual] 0, 
          TBD1 + TBD4 - TxD1 - TxD3 \[GreaterEqual] 0, 
          TxD3 \[GreaterEqual] 0, TxD1 + TxD3 \[GreaterEqual] 0, 
          TyD2 \[GreaterEqual] 0, \(-TAD2\) + TyD2 \[GreaterEqual] 0, 
          TyD3 \[GreaterEqual] 0, \(-TAD3\) + TyD3 \[GreaterEqual] 0, 
          TyD1 + TyD3 \[GreaterEqual] 
            0, \(-TAD1\) - TAD3 + TyD1 + TyD3 \[GreaterEqual] 
            0, \(-AA\) + Ay - TAD1 - TAD2 - 2\ TAD3 + TyD1 + TyD2 + 
              2\ TyD3 \[GreaterEqual] 
            slack7, \(-AA\) + Ay - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TyD1 + 
              TyD2 + 2\ TyD3 \[GreaterEqual] slack7, 
          0 \[LessEqual] slack1 \[LessEqual] 1, 
          0 \[LessEqual] slack2 \[LessEqual] 1, 
          0 \[LessEqual] slack3 \[LessEqual] 1, 
          0 \[LessEqual] slack4 \[LessEqual] 1, 
          0 \[LessEqual] slack5 \[LessEqual] 1, 
          0 \[LessEqual] slack6 \[LessEqual] 1, 
          0 \[LessEqual] slack7 \[LessEqual] 1}\),
      SequenceForm[ "Dependency constraints :\n", {
        GreaterEqual[ TAD2, 0], 
        GreaterEqual[ TAD3, 0], 
        GreaterEqual[ 
          Plus[ TAD1, TAD3], 0], 
        GreaterEqual[ TBD2, slack5], 
        GreaterEqual[ 
          Plus[ TAD2, 
            Times[ -1, TBD2], 
            Times[ -1, TBD3]], 0], 
        GreaterEqual[ TBD3, 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD2], TBD3], 0], 
        GreaterEqual[ 
          Plus[ TBD2, TBD3], 0], 
        GreaterEqual[ 
          Plus[ AA, 
            Times[ -1, AB], 
            Times[ 2, TAD1], TAD2, 
            Times[ 2, TAD3], 
            Times[ -1, TBD1], 
            Times[ -1, TBD2], 
            Times[ -1, TBD3], 
            Times[ -2, TBD4]], slack3], 
        GreaterEqual[ 
          Plus[ TAD3, 
            Times[ -1, TBD4]], 0], 
        GreaterEqual[ 
          Plus[ TAD1, TAD3, 
            Times[ -1, TBD1], 
            Times[ -1, TBD4]], 0], 
        GreaterEqual[ TBD4, 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD3], TBD4], 0], 
        GreaterEqual[ 
          Plus[ TBD1, TBD4], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD1], 
            Times[ -1, TAD3], TBD1, TBD4], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], AB, 
            Times[ -1, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], TBD1, TBD2, TBD3, 
            Times[ 2, TBD4]], slack4], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], AB, 
            Times[ -2, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], 
            Times[ 2, TBD1], TBD2, TBD3, 
            Times[ 2, TBD4]], slack4], 
        GreaterEqual[ 
          Plus[ TAD2, 
            Times[ -1, TxD2]], 0], 
        GreaterEqual[ 
          Plus[ TBD3, 
            Times[ -1, TxD2]], 0], 
        GreaterEqual[ 
          Plus[ TBD2, TBD3, 
            Times[ -1, TxD2]], 0], 
        GreaterEqual[ TxD2, 0], 
        GreaterEqual[ 
          Plus[ AB, 
            Times[ -1, Ax], 
            Times[ 2, TBD1], 
            Times[ 2, TBD2], 
            Times[ 2, TBD3], 
            Times[ 2, TBD4], 
            Times[ -2, TxD1], 
            Times[ -2, TxD2], 
            Times[ -2, TxD3]], slack6], 
        GreaterEqual[ 
          Plus[ AB, 
            Times[ -1, Ax], TBD1, 
            Times[ 2, TBD2], 
            Times[ 2, TBD3], 
            Times[ 2, TBD4], 
            Times[ -1, TxD1], 
            Times[ -2, TxD2], 
            Times[ -2, TxD3]], slack6], 
        GreaterEqual[ 
          Plus[ AA, 
            Times[ -1, Ax], 
            Times[ 2, TAD1], TAD2, 
            Times[ 2, TAD3], 
            Times[ -2, TxD1], 
            Times[ -1, TxD2], 
            Times[ -2, TxD3]], slack2], 
        GreaterEqual[ 
          Plus[ AA, 
            Times[ -1, Ax], TAD1, TAD2, 
            Times[ 2, TAD3], 
            Times[ -1, TxD1], 
            Times[ -1, TxD2], 
            Times[ -2, TxD3]], slack1], 
        GreaterEqual[ 
          Plus[ TAD3, 
            Times[ -1, TxD3]], 0], 
        GreaterEqual[ 
          Plus[ TBD4, 
            Times[ -1, TxD3]], 0], 
        GreaterEqual[ 
          Plus[ TAD1, TAD3, 
            Times[ -1, TxD1], 
            Times[ -1, TxD3]], 0], 
        GreaterEqual[ 
          Plus[ TBD1, TBD4, 
            Times[ -1, TxD1], 
            Times[ -1, TxD3]], 0], 
        GreaterEqual[ TxD3, 0], 
        GreaterEqual[ 
          Plus[ TxD1, TxD3], 0], 
        GreaterEqual[ TyD2, 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD2], TyD2], 0], 
        GreaterEqual[ TyD3, 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD3], TyD3], 0], 
        GreaterEqual[ 
          Plus[ TyD1, TyD3], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD1], 
            Times[ -1, TAD3], TyD1, TyD3], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], Ay, 
            Times[ -1, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], TyD1, TyD2, 
            Times[ 2, TyD3]], slack7], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], Ay, 
            Times[ -2, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], 
            Times[ 2, TyD1], TyD2, 
            Times[ 2, TyD3]], slack7], 
        LessEqual[ 0, slack1, 1], 
        LessEqual[ 0, slack2, 1], 
        LessEqual[ 0, slack3, 1], 
        LessEqual[ 0, slack4, 1], 
        LessEqual[ 0, slack5, 1], 
        LessEqual[ 0, slack6, 1], 
        LessEqual[ 0, slack7, 1]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Optimisation constraints: "\[InvisibleSpace]{\(-TAD3\
\) + TM \[GreaterEqual] 0, \(-TAD1\) - TAD3 + TM \[GreaterEqual] 
            0, \(-TBD4\) + TM \[GreaterEqual] 
            0, \(-TBD1\) - TBD4 + TM \[GreaterEqual] 
            0, \(-TAD2\) + TN \[GreaterEqual] 
            0, \(-TBD3\) + TN \[GreaterEqual] 
            0, \(-TBD2\) - TBD3 + TN \[GreaterEqual] 
            0, \(-AA\) + aparam - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TM + 
              TN \[GreaterEqual] 
            0, \(-AA\) + aparam - TAD1 - TAD2 - 2\ TAD3 + 2\ TM + 
              TN \[GreaterEqual] 
            0, \(-AB\) + aparam - 2\ TBD1 - TBD2 - TBD3 - 2\ TBD4 + 2\ TM + 
              TN \[GreaterEqual] 
            0, \(-AB\) + aparam - TBD1 - TBD2 - TBD3 - 2\ TBD4 + 2\ TM + 
              TN \[GreaterEqual] 0, TN - TxD2 \[GreaterEqual] 0, 
          aparam - Ax + 2\ TM + TN - 2\ TxD1 - TxD2 - 2\ TxD3 \[GreaterEqual] 
            0, aparam - Ax + 2\ TM + TN - TxD1 - TxD2 - 
              2\ TxD3 \[GreaterEqual] 0, TM - TxD3 \[GreaterEqual] 0, 
          TM - TxD1 - TxD3 \[GreaterEqual] 0, TN - TyD2 \[GreaterEqual] 0, 
          aparam - Ay + 2\ TM + TN - 2\ TyD1 - TyD2 - 2\ TyD3 \[GreaterEqual] 
            0, aparam - Ay + 2\ TM + TN - TyD1 - TyD2 - 
              2\ TyD3 \[GreaterEqual] 0, TM - TyD3 \[GreaterEqual] 0, 
          TM - TyD1 - TyD3 \[GreaterEqual] 0}\),
      SequenceForm[ "Optimisation constraints: ", {
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD3], TM], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD1], 
            Times[ -1, TAD3], TM], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TBD4], TM], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TBD1], 
            Times[ -1, TBD4], TM], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD2], TN], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TBD3], TN], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TBD2], 
            Times[ -1, TBD3], TN], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], aparam, 
            Times[ -2, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], 
            Times[ 2, TM], TN], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], aparam, 
            Times[ -1, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], 
            Times[ 2, TM], TN], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AB], aparam, 
            Times[ -2, TBD1], 
            Times[ -1, TBD2], 
            Times[ -1, TBD3], 
            Times[ -2, TBD4], 
            Times[ 2, TM], TN], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AB], aparam, 
            Times[ -1, TBD1], 
            Times[ -1, TBD2], 
            Times[ -1, TBD3], 
            Times[ -2, TBD4], 
            Times[ 2, TM], TN], 0], 
        GreaterEqual[ 
          Plus[ TN, 
            Times[ -1, TxD2]], 0], 
        GreaterEqual[ 
          Plus[ aparam, 
            Times[ -1, Ax], 
            Times[ 2, TM], TN, 
            Times[ -2, TxD1], 
            Times[ -1, TxD2], 
            Times[ -2, TxD3]], 0], 
        GreaterEqual[ 
          Plus[ aparam, 
            Times[ -1, Ax], 
            Times[ 2, TM], TN, 
            Times[ -1, TxD1], 
            Times[ -1, TxD2], 
            Times[ -2, TxD3]], 0], 
        GreaterEqual[ 
          Plus[ TM, 
            Times[ -1, TxD3]], 0], 
        GreaterEqual[ 
          Plus[ TM, 
            Times[ -1, TxD1], 
            Times[ -1, TxD3]], 0], 
        GreaterEqual[ 
          Plus[ TN, 
            Times[ -1, TyD2]], 0], 
        GreaterEqual[ 
          Plus[ aparam, 
            Times[ -1, Ay], 
            Times[ 2, TM], TN, 
            Times[ -2, TyD1], 
            Times[ -1, TyD2], 
            Times[ -2, TyD3]], 0], 
        GreaterEqual[ 
          Plus[ aparam, 
            Times[ -1, Ay], 
            Times[ 2, TM], TN, 
            Times[ -1, TyD1], 
            Times[ -1, TyD2], 
            Times[ -2, TyD3]], 0], 
        GreaterEqual[ 
          Plus[ TM, 
            Times[ -1, TyD3]], 0], 
        GreaterEqual[ 
          Plus[ TM, 
            Times[ -1, TyD1], 
            Times[ -1, TyD3]], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Initial list of unsatisfied edges: \
"\[InvisibleSpace]{1, 2, 3, 4, 5, 6, 7}\),
      SequenceForm[ 
      "Initial list of unsatisfied edges: ", {1, 2, 3, 4, 5, 6, 7}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Slack variables :\n"\[InvisibleSpace]{slack1, 
          slack2, slack3, slack4, slack5, slack6, slack7}\),
      SequenceForm[ 
      "Slack variables :\n", {slack1, slack2, slack3, slack4, slack5, slack6, 
        slack7}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(" Max dimension: "\[InvisibleSpace]2\),
      SequenceForm[ " Max dimension: ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("-----------\n---Dimension : "\[InvisibleSpace]1\),
      SequenceForm[ "-----------\n---Dimension : ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Slack variables: "\[InvisibleSpace]{slack1, slack2, 
          slack3, slack4, slack5, slack6, slack7}\),
      SequenceForm[ 
      "Slack variables: ", {slack1, slack2, slack3, slack4, slack5, slack6, 
        slack7}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Variables : "\[InvisibleSpace]{AA, AB, aparam, Ax, 
          Ay, slack1, slack2, slack3, slack4, slack5, slack6, slack7, TAD1, 
          TAD2, TAD3, TBD1, TBD2, TBD3, TBD4, TM, TN, TxD1, TxD2, TxD3, TyD1, 
          TyD2, TyD3}\),
      SequenceForm[ 
      "Variables : ", {AA, AB, aparam, Ax, Ay, slack1, slack2, slack3, slack4,
         slack5, slack6, slack7, TAD1, TAD2, TAD3, TBD1, TBD2, TBD3, TBD4, TM,
         TN, TxD1, TxD2, TxD3, TyD1, TyD2, TyD3}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Objective Function: "\[InvisibleSpace]\(\(-3\)\ AA + 
          AB + 8\ aparam - 6\ Ax - slack1 - slack2 - slack3 - 2\ slack4 - 
          slack5 - 2\ slack6 - 2\ slack7 - 
          10000\ \((slack1 + slack2 + slack3 + slack4 + slack5 + slack6 + 
                slack7)\) - 4\ TAD1 - 3\ TAD2 - 6\ TAD3 + 3\ TBD1 + 4\ TBD2 + 
          5\ TBD3 + 4\ TBD4 + 24\ TM + 14\ TN - 11\ TxD1 - 11\ TxD2 - 
          16\ TxD3 + TyD1 + TyD2 + 
          2\ TyD3\)\[InvisibleSpace]" Params: "\[InvisibleSpace]{TN, TM}\),
      SequenceForm[ "Objective Function: ", 
        Plus[ 
          Times[ -3, AA], AB, 
          Times[ 8, aparam], 
          Times[ -6, Ax], 
          Times[ -1, slack1], 
          Times[ -1, slack2], 
          Times[ -1, slack3], 
          Times[ -2, slack4], 
          Times[ -1, slack5], 
          Times[ -2, slack6], 
          Times[ -2, slack7], 
          Times[ -10000, 
            Plus[ slack1, slack2, slack3, slack4, slack5, slack6, slack7]], 
          Times[ -4, TAD1], 
          Times[ -3, TAD2], 
          Times[ -6, TAD3], 
          Times[ 3, TBD1], 
          Times[ 4, TBD2], 
          Times[ 5, TBD3], 
          Times[ 4, TBD4], 
          Times[ 24, TM], 
          Times[ 14, TN], 
          Times[ -11, TxD1], 
          Times[ -11, TxD2], 
          Times[ -16, TxD3], TyD1, TyD2, 
          Times[ 2, TyD3]], " Params: ", {TN, TM}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(67\[InvisibleSpace]" constraints, \
"\[InvisibleSpace]27\[InvisibleSpace]" variables \
"\[InvisibleSpace]\(0.057298999999999545`\ Second\)\),
      SequenceForm[ 67, " constraints, ", 27, " variables ", 
        Times[ 0.057298999999999545, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    \({\(-59968\), {AA \[Rule] 0, AB \[Rule] 1, aparam \[Rule] 1, 
        Ax \[Rule] \(-1\), Ay \[Rule] 1, slack1 \[Rule] 1, slack2 \[Rule] 1, 
        slack3 \[Rule] 1, slack4 \[Rule] 1, slack5 \[Rule] 0, 
        slack6 \[Rule] 1, slack7 \[Rule] 1, TAD1 \[Rule] 2, TAD2 \[Rule] 0, 
        TAD3 \[Rule] 0, TBD1 \[Rule] 2, TBD2 \[Rule] 0, TBD3 \[Rule] 0, 
        TBD4 \[Rule] 0, TM \[Rule] 2, TN \[Rule] 0, TxD1 \[Rule] 2, 
        TxD2 \[Rule] 0, TxD3 \[Rule] 0, TyD1 \[Rule] 2, TyD2 \[Rule] 0, 
        TyD3 \[Rule] 0}}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Unsatisfied edges: "\[InvisibleSpace]{1, 2, 3, 4, 5, 
          6, 7}\),
      SequenceForm[ "Unsatisfied edges: ", {1, 2, 3, 4, 5, 6, 7}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("sc: "\[InvisibleSpace]{1, 2, 3, 4, 6, 7}\),
      SequenceForm[ "sc: ", {1, 2, 3, 4, 6, 7}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Unsatisfied edges: "\[InvisibleSpace]{5}\),
      SequenceForm[ "Unsatisfied edges: ", {5}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Unsatisfied dependencies :\n\
"\[InvisibleSpace]"{i,j,N,M | 1<=i<=M; 2<=j<=N; 2<=M} : B[i,j,N,M] -> \
B[i,j-1,N,M]"\),
      SequenceForm[ 
      "Unsatisfied dependencies :\n", 
        "{i,j,N,M | 1<=i<=M; 2<=j<=N; 2<=M} : B[i,j,N,M] -> B[i,j-1,N,M]"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Constraints kept: "\[InvisibleSpace]{{TBD2 \
\[GreaterEqual] slack5, TBD3 \[GreaterEqual] 0, 
            TBD2 + TBD3 \[GreaterEqual] 0, TBD4 \[GreaterEqual] 0, 
            TBD1 + TBD4 \[GreaterEqual] 0, 
            0 \[LessEqual] slack5 \[LessEqual] 1}}\),
      SequenceForm[ "Constraints kept: ", {{
        GreaterEqual[ TBD2, slack5], 
        GreaterEqual[ TBD3, 0], 
        GreaterEqual[ 
          Plus[ TBD2, TBD3], 0], 
        GreaterEqual[ TBD4, 0], 
        GreaterEqual[ 
          Plus[ TBD1, TBD4], 0], 
        LessEqual[ 0, slack5, 1]}}],
      Editable->False]], "Print"],

Cell[BoxData[
    \({slack5}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("-----------\n---Dimension : "\[InvisibleSpace]2\),
      SequenceForm[ "-----------\n---Dimension : ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Slack variables: "\[InvisibleSpace]{slack5}\),
      SequenceForm[ "Slack variables: ", {slack5}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("tempudt: "\[InvisibleSpace]{depend[
            domain[4, {"i", "j", "N", 
                "M"}, {pol[6, 6, 0, 
                  0, {{1, 0, \(-1\), 1, 0, 0}, {1, 0, 1, 0, 0, \(-2\)}, {1, 
                      0, 0, 0, 1, \(-2\)}, {1, 1, 0, 0, 
                      0, \(-1\)}, {1, \(-1\), 0, 0, 1, 0}, {1, 0, 0, 0, 0, 
                      1}}, {{1, 0, 0, 1, 0, 0}, {1, 1, 0, 0, 1, 0}, {1, 0, 0, 
                      0, 1, 0}, {1, 0, 1, 1, 0, 0}, {1, 1, 2, 2, 2, 1}, {1, 
                      2, 2, 2, 2, 1}}]}], "B", "B", 
            matrix[5, 
              5, {"i", "j", "N", 
                "M"}, {{1, 0, 0, 0, 0}, {0, 1, 0, 0, \(-1\)}, {0, 0, 1, 0, 
                  0}, {0, 0, 0, 1, 0}, {0, 0, 0, 0, 1}}], 
            domain[4, {"i", "j", "N", 
                "M"}, {pol[6, 6, 0, 
                  0, {{1, 0, 0, 0, 1, \(-2\)}, {1, 0, \(-1\), 1, 0, 0}, {1, 
                      1, 0, 0, 0, \(-1\)}, {1, \(-1\), 0, 0, 1, 0}, {1, 0, 1, 
                      0, 0, \(-1\)}, {1, 0, 0, 0, 0, 1}}, {{1, 1, 0, 0, 1, 
                      0}, {1, 0, 0, 0, 1, 0}, {1, 0, 0, 1, 0, 0}, {1, 0, 1, 
                      1, 0, 0}, {1, 1, 1, 1, 2, 1}, {1, 2, 1, 1, 2, 
                      1}}]}]]}\),
      SequenceForm[ "tempudt: ", {
        depend[ 
          domain[ 4, {"i", "j", "N", "M"}, {
            pol[ 
            6, 6, 0, 0, {{1, 0, -1, 1, 0, 0}, {1, 0, 1, 0, 0, -2}, {1, 0, 0, 
              0, 1, -2}, {1, 1, 0, 0, 0, -1}, {1, -1, 0, 0, 1, 0}, {1, 0, 0, 
              0, 0, 1}}, {{1, 0, 0, 1, 0, 0}, {1, 1, 0, 0, 1, 0}, {1, 0, 0, 0,
               1, 0}, {1, 0, 1, 1, 0, 0}, {1, 1, 2, 2, 2, 1}, {1, 2, 2, 2, 2, 
              1}}]}], "B", "B", 
          matrix[ 
          5, 5, {"i", "j", "N", "M"}, {{1, 0, 0, 0, 0}, {0, 1, 0, 0, -1}, {0, 
            0, 1, 0, 0}, {0, 0, 0, 1, 0}, {0, 0, 0, 0, 1}}], 
          domain[ 4, {"i", "j", "N", "M"}, {
            pol[ 
            6, 6, 0, 0, {{1, 0, 0, 0, 1, -2}, {1, 0, -1, 1, 0, 0}, {1, 1, 0, 
              0, 0, -1}, {1, -1, 0, 0, 1, 0}, {1, 0, 1, 0, 0, -1}, {1, 0, 0, 
              0, 0, 1}}, {{1, 1, 0, 0, 1, 0}, {1, 0, 0, 0, 1, 0}, {1, 0, 0, 1,
               0, 0}, {1, 0, 1, 1, 0, 0}, {1, 1, 1, 1, 2, 1}, {1, 2, 1, 1, 2, 
              1}}]}]]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\({"B"}\[InvisibleSpace]{}\),
      SequenceForm[ {"B"}, {}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Constraints : "\[InvisibleSpace]{TBD2 \
\[GreaterEqual] slack5, TBD3 \[GreaterEqual] 0, 
          TBD2 + TBD3 \[GreaterEqual] 0, TBD4 \[GreaterEqual] 0, 
          TBD1 + TBD4 \[GreaterEqual] 0, \(-TBD4\) + TM \[GreaterEqual] 
            0, \(-TBD1\) - TBD4 + TM \[GreaterEqual] 
            0, \(-TBD3\) + TN \[GreaterEqual] 
            0, \(-TBD2\) - TBD3 + TN \[GreaterEqual] 
            0, \(-AB\) + aparam - 2\ TBD1 - TBD2 - TBD3 - 2\ TBD4 + 2\ TM + 
              TN \[GreaterEqual] 
            0, \(-AB\) + aparam - TBD1 - TBD2 - TBD3 - 2\ TBD4 + 2\ TM + 
              TN \[GreaterEqual] 0, 0 \[LessEqual] slack5 \[LessEqual] 1}\),
      SequenceForm[ "Constraints : ", {
        GreaterEqual[ TBD2, slack5], 
        GreaterEqual[ TBD3, 0], 
        GreaterEqual[ 
          Plus[ TBD2, TBD3], 0], 
        GreaterEqual[ TBD4, 0], 
        GreaterEqual[ 
          Plus[ TBD1, TBD4], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TBD4], TM], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TBD1], 
            Times[ -1, TBD4], TM], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TBD3], TN], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TBD2], 
            Times[ -1, TBD3], TN], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AB], aparam, 
            Times[ -2, TBD1], 
            Times[ -1, TBD2], 
            Times[ -1, TBD3], 
            Times[ -2, TBD4], 
            Times[ 2, TM], TN], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AB], aparam, 
            Times[ -1, TBD1], 
            Times[ -1, TBD2], 
            Times[ -1, TBD3], 
            Times[ -2, TBD4], 
            Times[ 2, TM], TN], 0], 
        LessEqual[ 0, slack5, 1]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Variables : "\[InvisibleSpace]{AA, AB, aparam, Ax, 
          Ay, slack5, TAD1, TAD2, TAD3, TBD1, TBD2, TBD3, TBD4, TM, TN, TxD1, 
          TxD2, TxD3, TyD1, TyD2, TyD3}\),
      SequenceForm[ 
      "Variables : ", {AA, AB, aparam, Ax, Ay, slack5, TAD1, TAD2, TAD3, TBD1,
         TBD2, TBD3, TBD4, TM, TN, TxD1, TxD2, TxD3, TyD1, TyD2, TyD3}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Objective Function: "\[InvisibleSpace]\(\(-2\)\ AB + 
          2\ aparam - 10001\ slack5 - 3\ TBD1 - TBD2 - 2\ TBD3 - 4\ TBD4 + 
          6\ TM + 5\ TN\)\[InvisibleSpace]" Params: "\[InvisibleSpace]{TN, 
          TM}\),
      SequenceForm[ "Objective Function: ", 
        Plus[ 
          Times[ -2, AB], 
          Times[ 2, aparam], 
          Times[ -10001, slack5], 
          Times[ -3, TBD1], 
          Times[ -1, TBD2], 
          Times[ -2, TBD3], 
          Times[ -4, TBD4], 
          Times[ 6, TM], 
          Times[ 5, TN]], " Params: ", {TN, TM}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(12\[InvisibleSpace]" constraints, \
"\[InvisibleSpace]21\[InvisibleSpace]" variables \
"\[InvisibleSpace]\(0.004554999999999865`\ Second\)\),
      SequenceForm[ 12, " constraints, ", 21, " variables ", 
        Times[ 0.0045549999999998647, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    \({\(-9997\), {AA \[Rule] 0, AB \[Rule] 0, aparam \[Rule] 0, 
        Ax \[Rule] 0, Ay \[Rule] 0, slack5 \[Rule] 1, TAD1 \[Rule] 0, 
        TAD2 \[Rule] 0, TAD3 \[Rule] 0, TBD1 \[Rule] 0, TBD2 \[Rule] 1, 
        TBD3 \[Rule] 0, TBD4 \[Rule] 0, TM \[Rule] 0, TN \[Rule] 1, 
        TxD1 \[Rule] 0, TxD2 \[Rule] 0, TxD3 \[Rule] 0, TyD1 \[Rule] 0, 
        TyD2 \[Rule] 0, TyD3 \[Rule] 0}}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Unsatisfied edges: "\[InvisibleSpace]{5}\),
      SequenceForm[ "Unsatisfied edges: ", {5}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("sc: "\[InvisibleSpace]{5}\),
      SequenceForm[ "sc: ", {5}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Unsatisfied edges: "\[InvisibleSpace]{}\),
      SequenceForm[ "Unsatisfied edges: ", {}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Unsatisfied dependencies :\n\
"\[InvisibleSpace]\*"\<\"\"\>"\),
      SequenceForm[ "Unsatisfied dependencies :\n", ""],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Constraints kept: "\[InvisibleSpace]{}\),
      SequenceForm[ "Constraints kept: ", {}],
      Editable->False]], "Print"],

Cell[BoxData[
    \({}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Schedule is of dimension "\[InvisibleSpace]2\),
      SequenceForm[ "Schedule is of dimension ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    \({scheduleResult[
        "testSchedule2", {{"A", {"i", "N", "M"}, 
            sched[{2, 0, 0}, 0]}, {"B", {"i", "j", "N", "M"}, 
            sched[{2, 0, 0, 0}, 1]}, {"x", {"i", "N", "M"}, 
            sched[{2, 0, 0}, \(-1\)]}, {"y", {"i", "N", "M"}, 
            sched[{2, 0, 0}, 1]}}, {0, 2, 1}], 
      scheduleResult[
        "testSchedule2", {{"A", {"i", "N", "M"}, 
            sched[{0, 0, 0}, 0]}, {"B", {"i", "j", "N", "M"}, 
            sched[{0, 1, 0, 0}, 0]}, {"x", {"i", "N", "M"}, 
            sched[{0, 0, 0}, 0]}, {"y", {"i", "N", "M"}, 
            sched[{0, 0, 0}, 0]}}, {1, 0, 0}]}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(showSchedResult[]\)], "Input"],

Cell[BoxData[
    \("----"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\ "i"\)\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 
        Times[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(\(-1\) + 2\ "i"\)\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 
        Plus[ -1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("----"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    \({Null, Null}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(scd[multidimensional \[Rule] True, 
      addConstraints \[Rule] {{TyD1 \[Equal] 2}, {}}]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("-----------\n---Dimension : "\[InvisibleSpace]1\),
      SequenceForm[ "-----------\n---Dimension : ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(71\[InvisibleSpace]" constraints, \
"\[InvisibleSpace]27\[InvisibleSpace]" variables \
"\[InvisibleSpace]\(0.06134799999999885`\ Second\)\),
      SequenceForm[ 71, " constraints, ", 27, " variables ", 
        Times[ 0.061347999999998848, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("-----------\n---Dimension : "\[InvisibleSpace]2\),
      SequenceForm[ "-----------\n---Dimension : ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(12\[InvisibleSpace]" constraints, \
"\[InvisibleSpace]21\[InvisibleSpace]" variables \
"\[InvisibleSpace]\(0.0040800000000000836`\ Second\)\),
      SequenceForm[ 12, " constraints, ", 21, " variables ", 
        Times[ 0.0040800000000000836, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Schedule is of dimension "\[InvisibleSpace]2\),
      SequenceForm[ "Schedule is of dimension ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    \({scheduleResult[
        "testSchedule2", {{"A", {"i", "N", "M"}, 
            sched[{2, 0, 0}, 1]}, {"B", {"i", "j", "N", "M"}, 
            sched[{2, 0, 0, 0}, 2]}, {"x", {"i", "N", "M"}, 
            sched[{2, 0, 0}, 0]}, {"y", {"i", "N", "M"}, 
            sched[{2, 0, 0}, 2]}}, {0, 2, 2}], 
      scheduleResult[
        "testSchedule2", {{"A", {"i", "N", "M"}, 
            sched[{0, 0, 0}, 0]}, {"B", {"i", "j", "N", "M"}, 
            sched[{0, 1, 0, 0}, 0]}, {"x", {"i", "N", "M"}, 
            sched[{0, 0, 0}, 0]}, {"y", {"i", "N", "M"}, 
            sched[{0, 0, 0}, 0]}}, {1, 0, 0}]}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(showSchedResult[]\)], "Input"],

Cell[BoxData[
    \("----"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2 + 2\ "i"\)\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", 
        Plus[ 2, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\ "i"\)\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 
        Times[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2 + 2\ "i"\)\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 
        Plus[ 2, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("----"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    \({Null, Null}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(load["\<testschedule3.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \("[testSchedule2]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(scheduleConstraints[]\)], "Input"],

Cell[BoxData[
    \({{TAD2 \[GreaterEqual] 0, TAD3 \[GreaterEqual] 0, 
        TAD1 + TAD3 \[GreaterEqual] 0, TBD2 \[GreaterEqual] 1, 
        TAD2 - TBD2 - TBD3 \[GreaterEqual] 0, 
        TBD3 \[GreaterEqual] 0, \(-TAD2\) + TBD3 \[GreaterEqual] 0, 
        TBD2 + TBD3 \[GreaterEqual] 0, 
        AA - AB + 2\ TAD1 + TAD2 + 2\ TAD3 - TBD1 - TBD2 - TBD3 - 
            2\ TBD4 \[GreaterEqual] 1, TAD3 - TBD4 \[GreaterEqual] 0, 
        TAD1 + TAD3 - TBD1 - TBD4 \[GreaterEqual] 0, 
        TBD4 \[GreaterEqual] 0, \(-TAD3\) + TBD4 \[GreaterEqual] 0, 
        TBD1 + TBD4 \[GreaterEqual] 
          0, \(-TAD1\) - TAD3 + TBD1 + TBD4 \[GreaterEqual] 
          0, \(-AA\) + AB - TAD1 - TAD2 - 2\ TAD3 + TBD1 + TBD2 + TBD3 + 
            2\ TBD4 \[GreaterEqual] 
          1, \(-AA\) + AB - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TBD1 + TBD2 + 
            TBD3 + 2\ TBD4 \[GreaterEqual] 1, TAD2 - TxD2 \[GreaterEqual] 0, 
        TBD3 - TxD2 \[GreaterEqual] 0, TBD2 + TBD3 - TxD2 \[GreaterEqual] 0, 
        TxD2 \[GreaterEqual] 0, 
        AB - Ax + 2\ TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - 2\ TxD1 - 2\ TxD2 - 
            2\ TxD3 \[GreaterEqual] 1, 
        AB - Ax + TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - TxD1 - 2\ TxD2 - 
            2\ TxD3 \[GreaterEqual] 1, 
        AA - Ax + 2\ TAD1 + TAD2 + 2\ TAD3 - 2\ TxD1 - TxD2 - 
            2\ TxD3 \[GreaterEqual] 1, 
        AA - Ax + TAD1 + TAD2 + 2\ TAD3 - TxD1 - TxD2 - 
            2\ TxD3 \[GreaterEqual] 1, TAD3 - TxD3 \[GreaterEqual] 0, 
        TBD4 - TxD3 \[GreaterEqual] 0, 
        TAD1 + TAD3 - TxD1 - TxD3 \[GreaterEqual] 0, 
        TBD1 + TBD4 - TxD1 - TxD3 \[GreaterEqual] 0, TxD3 \[GreaterEqual] 0, 
        TxD1 + TxD3 \[GreaterEqual] 0, 
        TyD2 \[GreaterEqual] 0, \(-TAD2\) + TyD2 \[GreaterEqual] 0, 
        TyD3 \[GreaterEqual] 0, \(-TAD3\) + TyD3 \[GreaterEqual] 0, 
        TyD1 + TyD3 \[GreaterEqual] 
          0, \(-TAD1\) - TAD3 + TyD1 + TyD3 \[GreaterEqual] 
          0, \(-AA\) + Ay - TAD1 - TAD2 - 2\ TAD3 + TyD1 + TyD2 + 
            2\ TyD3 \[GreaterEqual] 
          1, \(-AA\) + Ay - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TyD1 + TyD2 + 
            2\ TyD3 \[GreaterEqual] 1}, {AA, AB, Ax, Ay, TAD1, TAD2, TAD3, 
        TBD1, TBD2, TBD3, TBD4, TxD1, TxD2, TxD3, TyD1, TyD2, 
        TyD3}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(timeMinSchedConstraints[]\)], "Input"],

Cell[BoxData[
    \({{{"A", {\(-TAD3\) + TM \[GreaterEqual] 
              0, \(-TAD1\) - TAD3 + TM \[GreaterEqual] 
              0, \(-TAD2\) + TN \[GreaterEqual] 
              0, \(-AA\) + aparam - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TM + 
                TN \[GreaterEqual] 
              0, \(-AA\) + aparam - TAD1 - TAD2 - 2\ TAD3 + 2\ TM + 
                TN \[GreaterEqual] 0}}, {"B", {\(-TBD4\) + TM \[GreaterEqual] 
              0, \(-TBD1\) - TBD4 + TM \[GreaterEqual] 
              0, \(-TBD3\) + TN \[GreaterEqual] 
              0, \(-TBD2\) - TBD3 + TN \[GreaterEqual] 
              0, \(-AB\) + aparam - 2\ TBD1 - TBD2 - TBD3 - 2\ TBD4 + 2\ TM + 
                TN \[GreaterEqual] 
              0, \(-AB\) + aparam - TBD1 - TBD2 - TBD3 - 2\ TBD4 + 2\ TM + 
                TN \[GreaterEqual] 0}}, {"x", {TN - TxD2 \[GreaterEqual] 0, 
            aparam - Ax + 2\ TM + TN - 2\ TxD1 - TxD2 - 
                2\ TxD3 \[GreaterEqual] 0, 
            aparam - Ax + 2\ TM + TN - TxD1 - TxD2 - 2\ TxD3 \[GreaterEqual] 
              0, TM - TxD3 \[GreaterEqual] 0, 
            TM - TxD1 - TxD3 \[GreaterEqual] 
              0}}, {"y", {TN - TyD2 \[GreaterEqual] 0, 
            aparam - Ay + 2\ TM + TN - 2\ TyD1 - TyD2 - 
                2\ TyD3 \[GreaterEqual] 0, 
            aparam - Ay + 2\ TM + TN - TyD1 - TyD2 - 2\ TyD3 \[GreaterEqual] 
              0, TM - TyD3 \[GreaterEqual] 0, 
            TM - TyD1 - TyD3 \[GreaterEqual] 0}}}, {AA, AB, aparam, Ax, Ay, 
        TAD1, TAD2, TAD3, TBD1, TBD2, TBD3, TBD4, TM, TN, TxD1, TxD2, TxD3, 
        TyD1, TyD2, TyD3}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \($optimalityConstraints[2]\)], "Input"],

Cell[BoxData[
    \({{{"A", {\(-TAD3\) + TM \[GreaterEqual] 
                0, \(-TAD1\) - TAD3 + TM \[GreaterEqual] 
                0, \(-TAD2\) + TN \[GreaterEqual] 
                0, \(-AA\) + aparam - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TM + 
                  TN \[GreaterEqual] 
                0, \(-AA\) + aparam - TAD1 - TAD2 - 2\ TAD3 + 2\ TM + 
                  TN \[GreaterEqual] 
                0}}, {"B", {\(-TBD4\) + TM \[GreaterEqual] 
                0, \(-TBD1\) - TBD4 + TM \[GreaterEqual] 
                0, \(-TBD3\) + TN \[GreaterEqual] 
                0, \(-TBD2\) - TBD3 + TN \[GreaterEqual] 
                0, \(-AB\) + aparam - 2\ TBD1 - TBD2 - TBD3 - 2\ TBD4 + 
                  2\ TM + TN \[GreaterEqual] 
                0, \(-AB\) + aparam - TBD1 - TBD2 - TBD3 - 2\ TBD4 + 2\ TM + 
                  TN \[GreaterEqual] 0}}, {"x", {TN - TxD2 \[GreaterEqual] 0, 
              aparam - Ax + 2\ TM + TN - 2\ TxD1 - TxD2 - 
                  2\ TxD3 \[GreaterEqual] 0, 
              aparam - Ax + 2\ TM + TN - TxD1 - TxD2 - 
                  2\ TxD3 \[GreaterEqual] 0, TM - TxD3 \[GreaterEqual] 0, 
              TM - TxD1 - TxD3 \[GreaterEqual] 
                0}}, {"y", {TN - TyD2 \[GreaterEqual] 0, 
              aparam - Ay + 2\ TM + TN - 2\ TyD1 - TyD2 - 
                  2\ TyD3 \[GreaterEqual] 0, 
              aparam - Ay + 2\ TM + TN - TyD1 - TyD2 - 
                  2\ TyD3 \[GreaterEqual] 0, TM - TyD3 \[GreaterEqual] 0, 
              TM - TyD1 - TyD3 \[GreaterEqual] 0}}}, {AA, AB, aparam, Ax, Ay, 
          TAD1, TAD2, TAD3, TBD1, TBD2, TBD3, TBD4, TM, TN, TxD1, TxD2, TxD3, 
          TyD1, TyD2, TyD3}}[2]\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \($scheduleMDSol[2]\)], "Input"],

Cell[BoxData[
    \({\(-slack5\) + 
        TN, {TBD3 \[GreaterEqual] 0, \(-TAD2\) + TBD3 \[GreaterEqual] 0, 
        TBD4 \[GreaterEqual] 0, \(-TAD3\) + TBD4 \[GreaterEqual] 0, 
        TBD1 + TBD4 \[GreaterEqual] 
          0, \(-TAD1\) - TAD3 + TBD1 + TBD4 \[GreaterEqual] 
          0, \(-AA\) + AB - TAD1 - TAD2 - 2\ TAD3 + TBD1 + TBD2 + TBD3 + 
            2\ TBD4 \[GreaterEqual] 
          slack4, \(-AA\) + AB - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TBD1 + TBD2 + 
            TBD3 + 2\ TBD4 \[GreaterEqual] 
          slack4, \(-TBD4\) + TM \[GreaterEqual] 
          0, \(-TBD1\) - TBD4 + TM \[GreaterEqual] 
          0, \(-TBD3\) + TN \[GreaterEqual] 
          0, \(-TBD2\) - TBD3 + TN \[GreaterEqual] 
          0, \(-AB\) + aparam - 2\ TBD1 - TBD2 - TBD3 - 2\ TBD4 + 2\ TM + 
            TN \[GreaterEqual] 
          0, \(-AB\) + aparam - TBD1 - TBD2 - TBD3 - 2\ TBD4 + 2\ TM + 
            TN \[GreaterEqual] 0, 0 \[LessEqual] slack4 \[LessEqual] 1}, {AA, 
        AB, aparam, Ax, Ay, slack5, TAD1, TAD2, TAD3, TBD1, TBD2, TBD3, TBD4, 
        TM, TN, TxD1, TxD2, TxD3, TyD1, TyD2, 
        TyD3}, {\(-\[Infinity]\), {AA \[Rule] Indeterminate, 
          AB \[Rule] Indeterminate, aparam \[Rule] Indeterminate, 
          Ax \[Rule] Indeterminate, Ay \[Rule] Indeterminate, 
          slack5 \[Rule] Indeterminate, TAD1 \[Rule] Indeterminate, 
          TAD2 \[Rule] Indeterminate, TAD3 \[Rule] Indeterminate, 
          TBD1 \[Rule] Indeterminate, TBD2 \[Rule] Indeterminate, 
          TBD3 \[Rule] Indeterminate, TBD4 \[Rule] Indeterminate, 
          TM \[Rule] Indeterminate, TN \[Rule] Indeterminate, 
          TxD1 \[Rule] Indeterminate, TxD2 \[Rule] Indeterminate, 
          TxD3 \[Rule] Indeterminate, TyD1 \[Rule] Indeterminate, 
          TyD2 \[Rule] Indeterminate, 
          TyD3 \[Rule] Indeterminate}}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(multiSched[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Starting multidimensionnal scheduling on program "\
\[InvisibleSpace]"testSchedule2"\),
      SequenceForm[ 
      "Starting multidimensionnal scheduling on program ", "testSchedule2"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Scheduling dimension number "\[InvisibleSpace]1\),
      SequenceForm[ "Scheduling dimension number ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("duration options:"\[InvisibleSpace]{}\),
      SequenceForm[ "duration options:", {}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dep to schedule: "\[InvisibleSpace]{1, 2, 3, 4, 5, 
          6, 7}\),
      SequenceForm[ "dep to schedule: ", {1, 2, 3, 4, 5, 6, 7}],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]102\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]112\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 102, "      variables, ", 112, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00008260341.pip", FileType \[Rule] File, 
      Date \[Rule] 3408084601, ByteCount \[Rule] 33685}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00008260341.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408084601, ByteCount \[Rule] 856}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\ "i"\)\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 
        Times[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Scheduling dimension number "\[InvisibleSpace]2\),
      SequenceForm[ "Scheduling dimension number ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("duration options:"\[InvisibleSpace]{}\),
      SequenceForm[ "duration options:", {}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dep to schedule: "\[InvisibleSpace]{5}\),
      SequenceForm[ "dep to schedule: ", {5}],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]59\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]48\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 59, "      variables, ", 48, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00008560341.pip", FileType \[Rule] File, 
      Date \[Rule] 3408084603, ByteCount \[Rule] 8920}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00008560341.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408084603, ByteCount \[Rule] 553}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Schedule Found has "\[InvisibleSpace]2\
\[InvisibleSpace]" dimension"\),
      SequenceForm[ "Schedule Found has ", 2, " dimension"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("               "\[InvisibleSpace]0\),
      SequenceForm[ "               ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("               "\[InvisibleSpace]0\),
      SequenceForm[ "               ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\ "i"\)\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 
        Times[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("               "\[InvisibleSpace]0\),
      SequenceForm[ "               ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("                  "\[InvisibleSpace]"j"\),
      SequenceForm[ "                  ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    \(scheduleResult[
      "testSchedule2", {{"x", {"i", "N", "M"}, sched[{0, 0, 0}, 0], 
          sched[{0, 0, 0}, 0]}, {"y", {"i", "N", "M"}, sched[{2, 0, 0}, 1], 
          sched[{0, 0, 0}, 0]}, {"A", {"i", "N", "M"}, sched[{2, 0, 0}, 0], 
          sched[{0, 0, 0}, 0]}, {"B", {"i", "j", "N", "M"}, 
          sched[{2, 0, 0, 0}, 1], sched[{0, 1, 0, 0}, 0]}}, {}]\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(scheduleConstraints[multidimensional \[Rule] True, 
      verbose \[Rule] True]\)], "Input"],

Cell[BoxData[
    \("Dependencies kept are:"\)], "Print"],

Cell[BoxData[
    \("{i,N,M | i=1; 1<=N; 2<=M} : A[i,N,M] -> x[i,N,M]\n{i,N,M | 2<=i<=M; \
1<=N} : A[i,N,M] -> x[i,N,M]\n{i,N,M | 2<=i<=M; 1<=N} : A[i,N,M] -> \
B[i-1,N,N,M]\n{i,j,N,M | 1<=i<=M; j=1; 1<=N; 2<=M} : B[i,j,N,M] -> \
A[i,j+N-1,M]\n{i,j,N,M | 1<=i<=M; 2<=j<=N; 2<=M} : B[i,j,N,M] -> B[i,j-1,N,M]\
\n{i,j,N,M | 1<=i<=M; 2<=j<=N; 2<=M} : B[i,j,N,M] -> x[i,N,M]\n{i,N,M | \
1<=i<=M; 1<=N; 2<=M} : y[i,N,M] -> A[i,N,M]"\)], "Print"],

Cell[BoxData[
    \("----------------------------"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"x"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]\*"\<\"\"\>"\),
      SequenceForm[ "--- LHS variable: ", "x", ", RHS variable: ", ""],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TxD2 \
\[GreaterEqual] 0, TxD3 \[GreaterEqual] 0, TxD1 + TxD3 \[GreaterEqual] 0}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TxD2, 0], 
        GreaterEqual[ TxD3, 0], 
        GreaterEqual[ 
          Plus[ TxD1, TxD3], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"A"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"x"\),
      SequenceForm[ "--- LHS variable: ", "A", ", RHS variable: ", "x"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TAD2 \
\[GreaterEqual] 0, TAD3 \[GreaterEqual] 0, TAD2 - TxD2 \[GreaterEqual] 0, 
          AA - Ax + TAD1 + TAD2 + 2\ TAD3 - TxD1 - TxD2 - 
              2\ TxD3 \[GreaterEqual] 1, TAD3 - TxD3 \[GreaterEqual] 0}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TAD2, 0], 
        GreaterEqual[ TAD3, 0], 
        GreaterEqual[ 
          Plus[ TAD2, 
            Times[ -1, TxD2]], 0], 
        GreaterEqual[ 
          Plus[ AA, 
            Times[ -1, Ax], TAD1, TAD2, 
            Times[ 2, TAD3], 
            Times[ -1, TxD1], 
            Times[ -1, TxD2], 
            Times[ -2, TxD3]], 1], 
        GreaterEqual[ 
          Plus[ TAD3, 
            Times[ -1, TxD3]], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"A"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"x"\),
      SequenceForm[ "--- LHS variable: ", "A", ", RHS variable: ", "x"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TAD2 \
\[GreaterEqual] 0, TAD3 \[GreaterEqual] 0, TAD1 + TAD3 \[GreaterEqual] 0, 
          TAD2 - TxD2 \[GreaterEqual] 0, 
          AA - Ax + 2\ TAD1 + TAD2 + 2\ TAD3 - 2\ TxD1 - TxD2 - 
              2\ TxD3 \[GreaterEqual] 1, TAD3 - TxD3 \[GreaterEqual] 0, 
          TAD1 + TAD3 - TxD1 - TxD3 \[GreaterEqual] 0}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TAD2, 0], 
        GreaterEqual[ TAD3, 0], 
        GreaterEqual[ 
          Plus[ TAD1, TAD3], 0], 
        GreaterEqual[ 
          Plus[ TAD2, 
            Times[ -1, TxD2]], 0], 
        GreaterEqual[ 
          Plus[ AA, 
            Times[ -1, Ax], 
            Times[ 2, TAD1], TAD2, 
            Times[ 2, TAD3], 
            Times[ -2, TxD1], 
            Times[ -1, TxD2], 
            Times[ -2, TxD3]], 1], 
        GreaterEqual[ 
          Plus[ TAD3, 
            Times[ -1, TxD3]], 0], 
        GreaterEqual[ 
          Plus[ TAD1, TAD3, 
            Times[ -1, TxD1], 
            Times[ -1, TxD3]], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"A"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"B"\),
      SequenceForm[ "--- LHS variable: ", "A", ", RHS variable: ", "B"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TAD2 \
\[GreaterEqual] 0, TAD3 \[GreaterEqual] 0, TAD1 + TAD3 \[GreaterEqual] 0, 
          TAD2 - TBD2 - TBD3 \[GreaterEqual] 0, 
          AA - AB + 2\ TAD1 + TAD2 + 2\ TAD3 - TBD1 - TBD2 - TBD3 - 
              2\ TBD4 \[GreaterEqual] 1, TAD3 - TBD4 \[GreaterEqual] 0, 
          TAD1 + TAD3 - TBD1 - TBD4 \[GreaterEqual] 0}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TAD2, 0], 
        GreaterEqual[ TAD3, 0], 
        GreaterEqual[ 
          Plus[ TAD1, TAD3], 0], 
        GreaterEqual[ 
          Plus[ TAD2, 
            Times[ -1, TBD2], 
            Times[ -1, TBD3]], 0], 
        GreaterEqual[ 
          Plus[ AA, 
            Times[ -1, AB], 
            Times[ 2, TAD1], TAD2, 
            Times[ 2, TAD3], 
            Times[ -1, TBD1], 
            Times[ -1, TBD2], 
            Times[ -1, TBD3], 
            Times[ -2, TBD4]], 1], 
        GreaterEqual[ 
          Plus[ TAD3, 
            Times[ -1, TBD4]], 0], 
        GreaterEqual[ 
          Plus[ TAD1, TAD3, 
            Times[ -1, TBD1], 
            Times[ -1, TBD4]], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"B"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"A"\),
      SequenceForm[ "--- LHS variable: ", "B", ", RHS variable: ", "A"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TBD3 \
\[GreaterEqual] 0, \(-TAD2\) + TBD3 \[GreaterEqual] 0, 
          TBD4 \[GreaterEqual] 0, \(-TAD3\) + TBD4 \[GreaterEqual] 0, 
          TBD1 + TBD4 \[GreaterEqual] 
            0, \(-TAD1\) - TAD3 + TBD1 + TBD4 \[GreaterEqual] 
            0, \(-AA\) + AB - TAD1 - TAD2 - 2\ TAD3 + TBD1 + TBD2 + TBD3 + 
              2\ TBD4 \[GreaterEqual] 
            1, \(-AA\) + AB - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TBD1 + TBD2 + 
              TBD3 + 2\ TBD4 \[GreaterEqual] 1}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TBD3, 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD2], TBD3], 0], 
        GreaterEqual[ TBD4, 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD3], TBD4], 0], 
        GreaterEqual[ 
          Plus[ TBD1, TBD4], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD1], 
            Times[ -1, TAD3], TBD1, TBD4], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], AB, 
            Times[ -1, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], TBD1, TBD2, TBD3, 
            Times[ 2, TBD4]], 1], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], AB, 
            Times[ -2, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], 
            Times[ 2, TBD1], TBD2, TBD3, 
            Times[ 2, TBD4]], 1]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"B"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"B"\),
      SequenceForm[ "--- LHS variable: ", "B", ", RHS variable: ", "B"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{True, 
          TBD2 \[GreaterEqual] 1, TBD3 \[GreaterEqual] 0, 
          TBD2 + TBD3 \[GreaterEqual] 0, TBD4 \[GreaterEqual] 0, 
          TBD1 + TBD4 \[GreaterEqual] 0}\),
      SequenceForm[ "    Constraints: ", {True, 
        GreaterEqual[ TBD2, 1], 
        GreaterEqual[ TBD3, 0], 
        GreaterEqual[ 
          Plus[ TBD2, TBD3], 0], 
        GreaterEqual[ TBD4, 0], 
        GreaterEqual[ 
          Plus[ TBD1, TBD4], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"B"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"x"\),
      SequenceForm[ "--- LHS variable: ", "B", ", RHS variable: ", "x"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TBD3 \
\[GreaterEqual] 0, TBD2 + TBD3 \[GreaterEqual] 0, TBD4 \[GreaterEqual] 0, 
          TBD1 + TBD4 \[GreaterEqual] 0, TBD3 - TxD2 \[GreaterEqual] 0, 
          TBD2 + TBD3 - TxD2 \[GreaterEqual] 0, 
          AB - Ax + 2\ TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - 2\ TxD1 - 
              2\ TxD2 - 2\ TxD3 \[GreaterEqual] 1, 
          AB - Ax + TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - TxD1 - 2\ TxD2 - 
              2\ TxD3 \[GreaterEqual] 1, TBD4 - TxD3 \[GreaterEqual] 0, 
          TBD1 + TBD4 - TxD1 - TxD3 \[GreaterEqual] 0}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TBD3, 0], 
        GreaterEqual[ 
          Plus[ TBD2, TBD3], 0], 
        GreaterEqual[ TBD4, 0], 
        GreaterEqual[ 
          Plus[ TBD1, TBD4], 0], 
        GreaterEqual[ 
          Plus[ TBD3, 
            Times[ -1, TxD2]], 0], 
        GreaterEqual[ 
          Plus[ TBD2, TBD3, 
            Times[ -1, TxD2]], 0], 
        GreaterEqual[ 
          Plus[ AB, 
            Times[ -1, Ax], 
            Times[ 2, TBD1], 
            Times[ 2, TBD2], 
            Times[ 2, TBD3], 
            Times[ 2, TBD4], 
            Times[ -2, TxD1], 
            Times[ -2, TxD2], 
            Times[ -2, TxD3]], 1], 
        GreaterEqual[ 
          Plus[ AB, 
            Times[ -1, Ax], TBD1, 
            Times[ 2, TBD2], 
            Times[ 2, TBD3], 
            Times[ 2, TBD4], 
            Times[ -1, TxD1], 
            Times[ -2, TxD2], 
            Times[ -2, TxD3]], 1], 
        GreaterEqual[ 
          Plus[ TBD4, 
            Times[ -1, TxD3]], 0], 
        GreaterEqual[ 
          Plus[ TBD1, TBD4, 
            Times[ -1, TxD1], 
            Times[ -1, TxD3]], 0]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("--- LHS variable: "\[InvisibleSpace]"y"\
\[InvisibleSpace]", RHS variable: "\[InvisibleSpace]"A"\),
      SequenceForm[ "--- LHS variable: ", "y", ", RHS variable: ", "A"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("    Constraints: "\[InvisibleSpace]{TyD2 \
\[GreaterEqual] 0, \(-TAD2\) + TyD2 \[GreaterEqual] 0, 
          TyD3 \[GreaterEqual] 0, \(-TAD3\) + TyD3 \[GreaterEqual] 0, 
          TyD1 + TyD3 \[GreaterEqual] 
            0, \(-TAD1\) - TAD3 + TyD1 + TyD3 \[GreaterEqual] 
            0, \(-AA\) + Ay - TAD1 - TAD2 - 2\ TAD3 + TyD1 + TyD2 + 
              2\ TyD3 \[GreaterEqual] 
            1, \(-AA\) + Ay - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TyD1 + TyD2 + 
              2\ TyD3 \[GreaterEqual] 1}\),
      SequenceForm[ "    Constraints: ", {
        GreaterEqual[ TyD2, 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD2], TyD2], 0], 
        GreaterEqual[ TyD3, 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD3], TyD3], 0], 
        GreaterEqual[ 
          Plus[ TyD1, TyD3], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, TAD1], 
            Times[ -1, TAD3], TyD1, TyD3], 0], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], Ay, 
            Times[ -1, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], TyD1, TyD2, 
            Times[ 2, TyD3]], 1], 
        GreaterEqual[ 
          Plus[ 
            Times[ -1, AA], Ay, 
            Times[ -2, TAD1], 
            Times[ -1, TAD2], 
            Times[ -2, TAD3], 
            Times[ 2, TyD1], TyD2, 
            Times[ 2, TyD3]], 1]}],
      Editable->False]], "Print"],

Cell[BoxData[
    \({{{TxD2 \[GreaterEqual] 0, TxD3 \[GreaterEqual] 0, 
          TxD1 + TxD3 \[GreaterEqual] 0}, {TAD2 \[GreaterEqual] 0, 
          TAD3 \[GreaterEqual] 0, TAD2 - TxD2 \[GreaterEqual] 0, 
          AA - Ax + TAD1 + TAD2 + 2\ TAD3 - TxD1 - TxD2 - 
              2\ TxD3 \[GreaterEqual] slack1, TAD3 - TxD3 \[GreaterEqual] 0, 
          0 \[LessEqual] slack1 \[LessEqual] 1}, {TAD2 \[GreaterEqual] 0, 
          TAD3 \[GreaterEqual] 0, TAD1 + TAD3 \[GreaterEqual] 0, 
          TAD2 - TxD2 \[GreaterEqual] 0, 
          AA - Ax + 2\ TAD1 + TAD2 + 2\ TAD3 - 2\ TxD1 - TxD2 - 
              2\ TxD3 \[GreaterEqual] slack2, TAD3 - TxD3 \[GreaterEqual] 0, 
          TAD1 + TAD3 - TxD1 - TxD3 \[GreaterEqual] 0, 
          0 \[LessEqual] slack2 \[LessEqual] 1}, {TAD2 \[GreaterEqual] 0, 
          TAD3 \[GreaterEqual] 0, TAD1 + TAD3 \[GreaterEqual] 0, 
          TAD2 - TBD2 - TBD3 \[GreaterEqual] 0, 
          AA - AB + 2\ TAD1 + TAD2 + 2\ TAD3 - TBD1 - TBD2 - TBD3 - 
              2\ TBD4 \[GreaterEqual] slack3, TAD3 - TBD4 \[GreaterEqual] 0, 
          TAD1 + TAD3 - TBD1 - TBD4 \[GreaterEqual] 0, 
          0 \[LessEqual] slack3 \[LessEqual] 1}, {TBD3 \[GreaterEqual] 
            0, \(-TAD2\) + TBD3 \[GreaterEqual] 0, 
          TBD4 \[GreaterEqual] 0, \(-TAD3\) + TBD4 \[GreaterEqual] 0, 
          TBD1 + TBD4 \[GreaterEqual] 
            0, \(-TAD1\) - TAD3 + TBD1 + TBD4 \[GreaterEqual] 
            0, \(-AA\) + AB - TAD1 - TAD2 - 2\ TAD3 + TBD1 + TBD2 + TBD3 + 
              2\ TBD4 \[GreaterEqual] 
            slack4, \(-AA\) + AB - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TBD1 + 
              TBD2 + TBD3 + 2\ TBD4 \[GreaterEqual] slack4, 
          0 \[LessEqual] slack4 \[LessEqual] 1}, {TBD2 \[GreaterEqual] 
            slack5, TBD3 \[GreaterEqual] 0, TBD2 + TBD3 \[GreaterEqual] 0, 
          TBD4 \[GreaterEqual] 0, TBD1 + TBD4 \[GreaterEqual] 0, 
          0 \[LessEqual] slack5 \[LessEqual] 1}, {TBD3 \[GreaterEqual] 0, 
          TBD2 + TBD3 \[GreaterEqual] 0, TBD4 \[GreaterEqual] 0, 
          TBD1 + TBD4 \[GreaterEqual] 0, TBD3 - TxD2 \[GreaterEqual] 0, 
          TBD2 + TBD3 - TxD2 \[GreaterEqual] 0, 
          AB - Ax + 2\ TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - 2\ TxD1 - 
              2\ TxD2 - 2\ TxD3 \[GreaterEqual] slack6, 
          AB - Ax + TBD1 + 2\ TBD2 + 2\ TBD3 + 2\ TBD4 - TxD1 - 2\ TxD2 - 
              2\ TxD3 \[GreaterEqual] slack6, TBD4 - TxD3 \[GreaterEqual] 0, 
          TBD1 + TBD4 - TxD1 - TxD3 \[GreaterEqual] 0, 
          0 \[LessEqual] slack6 \[LessEqual] 1}, {TyD2 \[GreaterEqual] 
            0, \(-TAD2\) + TyD2 \[GreaterEqual] 0, 
          TyD3 \[GreaterEqual] 0, \(-TAD3\) + TyD3 \[GreaterEqual] 0, 
          TyD1 + TyD3 \[GreaterEqual] 
            0, \(-TAD1\) - TAD3 + TyD1 + TyD3 \[GreaterEqual] 
            0, \(-AA\) + Ay - TAD1 - TAD2 - 2\ TAD3 + TyD1 + TyD2 + 
              2\ TyD3 \[GreaterEqual] 
            slack7, \(-AA\) + Ay - 2\ TAD1 - TAD2 - 2\ TAD3 + 2\ TyD1 + 
              TyD2 + 2\ TyD3 \[GreaterEqual] slack7, 
          0 \[LessEqual] slack7 \[LessEqual] 1}}, {AA, AB, Ax, Ay, slack1, 
        slack2, slack3, slack4, slack5, slack6, slack7, TAD1, TAD2, TAD3, 
        TBD1, TBD2, TBD3, TBD4, TxD1, TxD2, TxD3, TyD1, TyD2, 
        TyD3}}\)], "Output"]
}, Closed]],

Cell[BoxData[
    \(<< Alpha/VertexSchedule.m\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(scd[multidimensional \[Rule] True, verbose \[Rule] False, 
      debug \[Rule] False, integerSolution \[Rule] False, 
      addConstraints \[Rule] {{TyD1 \[Equal] 2}, {}}]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("-----------\n---Dimension : "\[InvisibleSpace]1\),
      SequenceForm[ "-----------\n---Dimension : ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(37\[InvisibleSpace]" constraints, \
"\[InvisibleSpace]15\[InvisibleSpace]" variables \
"\[InvisibleSpace]\(0.009588000000007924`\ Second\)\),
      SequenceForm[ 37, " constraints, ", 15, " variables ", 
        Times[ 0.0095880000000079235, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Schedule is of dimension "\[InvisibleSpace]1\),
      SequenceForm[ "Schedule is of dimension ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    \({scheduleResult[
        "testSchedule2", {{"B", {"i", "N"}, 
            sched[{2, 0}, 0]}, {"x", {"i", "N"}, 
            sched[{1, 0}, 0]}, {"y", {"i", "N"}, sched[{2, 0}, 1]}}, {2, 
          1}]}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(showSchedResult[]\)], "Input"],

Cell[BoxData[
    \("----"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\ "i"\)\),
      SequenceForm[ "T_", "B", {"i", "N"}, " = ", 
        Times[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "x", {"i", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "y", {"i", "N"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \({Null}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ashow[dep[]]\)], "Input"],

Cell[BoxData[
    \("{i,N,M | i=1; 1<=M} : A[i,N,M] -> x[i,N,M]\n{i,N,M | 2<=i<=M} : \
A[i,N,M] -> x[i,N,M]\n{i,N,M | 2<=i<=M; 1<=N} : A[i,N,M] -> B[i-1,N,N,M]\n\
{i,j,N,M | 1<=i<=M; j=1; 1<=N} : B[i,j,N,M] -> A[i,j+N-1,M]\n{i,j,N,M | \
1<=i<=M; 2<=j<=N} : B[i,j,N,M] -> B[i,j-1,N,M]\n{i,j,N,M | 1<=i<=M; 2<=j<=N} \
: B[i,j,N,M] -> x[i,N,M]\n{i,N,M | 1<=i<=M} : y[i,N,M] -> A[i,N,M]"\)], \
"Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(showSchedResult[]\)], "Input"],

Cell[BoxData[
    \("----"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\ "i"\)\),
      SequenceForm[ "T_", "B", {"i", "N"}, " = ", 
        Times[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "x", {"i", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "y", {"i", "N"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \({Null}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(sss2 = $scheduleMDSol[1]\)], "Input"],

Cell[BoxData[
    \({1, {TBD2 \[GreaterEqual] 0, TBD1 + TBD2 \[GreaterEqual] 0, 
        AB - Ax + 2\ TBD1 + 2\ TBD2 - 2\ TxD1 - 2\ TxD2 \[GreaterEqual] 
          slack2, AB - Ax + TBD1 + 2\ TBD2 - TxD1 - 2\ TxD2 \[GreaterEqual] 
          slack1, TBD2 - TxD2 \[GreaterEqual] 0, 
        TBD1 + TBD2 - TxD1 - TxD2 \[GreaterEqual] 0, TxD2 \[GreaterEqual] 0, 
        TxD1 + TxD2 \[GreaterEqual] 0, 
        AB - Ay + 2\ TBD1 + 2\ TBD2 - TyD1 - 2\ TyD2 \[GreaterEqual] slack3, 
        TBD2 - TyD2 \[GreaterEqual] 0, 
        TBD1 + TBD2 - TyD1 - TyD2 \[GreaterEqual] 0, 
        TyD2 \[GreaterEqual] 0, \(-TBD2\) + TyD2 \[GreaterEqual] 0, 
        TyD1 + TyD2 \[GreaterEqual] 
          0, \(-TBD1\) - TBD2 + TyD1 + TyD2 \[GreaterEqual] 
          0, \(-AB\) + Ay - TBD1 - 2\ TBD2 + TyD1 + 2\ TyD2 \[GreaterEqual] 
          slack4, \(-AB\) + Ay - 2\ TBD1 - 2\ TBD2 + 2\ TyD1 + 
            2\ TyD2 \[GreaterEqual] slack4, 
        0 \[LessEqual] slack1 \[LessEqual] 1, 
        0 \[LessEqual] slack2 \[LessEqual] 1, 
        0 \[LessEqual] slack3 \[LessEqual] 1, 
        0 \[LessEqual] slack4 \[LessEqual] 1}, {AB, aparam, Ax, Ay, slack1, 
        slack2, slack3, slack4, TBD1, TBD2, TN, TxD1, TxD2, TyD1, 
        TyD2}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(sss1\)], "Input"],

Cell[BoxData[
    \({{TBD2 \[GreaterEqual] 0, TBD1 + TBD2 \[GreaterEqual] 0, 
        AB - Ax + 2\ TBD1 + 2\ TBD2 - 2\ TxD1 - 2\ TxD2 \[GreaterEqual] 1, 
        AB - Ax + TBD1 + 2\ TBD2 - TxD1 - 2\ TxD2 \[GreaterEqual] 1, 
        TBD2 - TxD2 \[GreaterEqual] 0, 
        TBD1 + TBD2 - TxD1 - TxD2 \[GreaterEqual] 0, TxD2 \[GreaterEqual] 0, 
        TxD1 + TxD2 \[GreaterEqual] 0, 
        AB - Ay + 2\ TBD1 + 2\ TBD2 - TyD1 - 2\ TyD2 \[GreaterEqual] 1, 
        TBD2 - TyD2 \[GreaterEqual] 0, 
        TBD1 + TBD2 - TyD1 - TyD2 \[GreaterEqual] 0, 
        TyD2 \[GreaterEqual] 0, \(-TBD2\) + TyD2 \[GreaterEqual] 0, 
        TyD1 + TyD2 \[GreaterEqual] 
          0, \(-TBD1\) - TBD2 + TyD1 + TyD2 \[GreaterEqual] 
          0, \(-AB\) + Ay - TBD1 - 2\ TBD2 + TyD1 + 2\ TyD2 \[GreaterEqual] 
          1, \(-AB\) + Ay - 2\ TBD1 - 2\ TBD2 + 2\ TyD1 + 
            2\ TyD2 \[GreaterEqual] 1}, {AB, Ax, Ay, TBD1, TBD2, TxD1, TxD2, 
        TyD1, TyD2}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(multiSched[verbose -> True]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Starting multidimensionnal scheduling on program "\
\[InvisibleSpace]"testSchedule2"\),
      SequenceForm[ 
      "Starting multidimensionnal scheduling on program ", "testSchedule2"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Scheduling dimension number "\[InvisibleSpace]1\),
      SequenceForm[ "Scheduling dimension number ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("duration options:"\[InvisibleSpace]{}\),
      SequenceForm[ "duration options:", {}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dep to schedule: "\[InvisibleSpace]{1, 2, 3, 4, 5, 
          6, 7}\),
      SequenceForm[ "dep to schedule: ", {1, 2, 3, 4, 5, 6, 7}],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]102\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]112\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 102, "      variables, ", 112, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00004757841.pip", FileType \[Rule] File, 
      Date \[Rule] 3407938551, ByteCount \[Rule] 33685}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00004757841.pipres", FileType \[Rule] File, 
      Date \[Rule] 3407938551, ByteCount \[Rule] 856}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\ "i"\)\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 
        Times[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Scheduling dimension number "\[InvisibleSpace]2\),
      SequenceForm[ "Scheduling dimension number ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("duration options:"\[InvisibleSpace]{}\),
      SequenceForm[ "duration options:", {}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dep to schedule: "\[InvisibleSpace]{5}\),
      SequenceForm[ "dep to schedule: ", {5}],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]59\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]48\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 59, "      variables, ", 48, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00005057841.pip", FileType \[Rule] File, 
      Date \[Rule] 3407938553, ByteCount \[Rule] 8920}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00005057841.pipres", FileType \[Rule] File, 
      Date \[Rule] 3407938553, ByteCount \[Rule] 553}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Schedule Found has "\[InvisibleSpace]2\
\[InvisibleSpace]" dimension"\),
      SequenceForm[ "Schedule Found has ", 2, " dimension"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("               "\[InvisibleSpace]0\),
      SequenceForm[ "               ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("               "\[InvisibleSpace]0\),
      SequenceForm[ "               ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\ "i"\)\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 
        Times[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("               "\[InvisibleSpace]0\),
      SequenceForm[ "               ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("                  "\[InvisibleSpace]"j"\),
      SequenceForm[ "                  ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    \(scheduleResult[
      "testSchedule2", {{"x", {"i", "N", "M"}, sched[{0, 0, 0}, 0], 
          sched[{0, 0, 0}, 0]}, {"y", {"i", "N", "M"}, sched[{2, 0, 0}, 1], 
          sched[{0, 0, 0}, 0]}, {"A", {"i", "N", "M"}, sched[{2, 0, 0}, 0], 
          sched[{0, 0, 0}, 0]}, {"B", {"i", "j", "N", "M"}, 
          sched[{2, 0, 0, 0}, 1], sched[{0, 1, 0, 0}, 0]}}, {}]\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ashow[$schedule]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("param:"\[InvisibleSpace]scheduleResult[
          "testSchedule2", {{"x", {"i", "N", "M"}, sched[{0, 0, 0}, 0], 
              sched[{0, 0, 0}, 0]}, {"y", {"i", "N", "M"}, 
              sched[{2, 0, 0}, 1], 
              sched[{0, 0, 0}, 0]}, {"A", {"i", "N", "M"}, 
              sched[{2, 0, 0}, 0], 
              sched[{0, 0, 0}, 0]}, {"B", {"i", "j", "N", "M"}, 
              sched[{2, 0, 0, 0}, 1], sched[{0, 1, 0, 0}, 0]}}, {}]\),
      SequenceForm[ "param:", 
        scheduleResult[ "testSchedule2", {{"x", {"i", "N", "M"}, 
          sched[ {0, 0, 0}, 0], 
          sched[ {0, 0, 0}, 0]}, {"y", {"i", "N", "M"}, 
          sched[ {2, 0, 0}, 1], 
          sched[ {0, 0, 0}, 0]}, {"A", {"i", "N", "M"}, 
          sched[ {2, 0, 0}, 0], 
          sched[ {0, 0, 0}, 0]}, {"B", {"i", "j", "N", "M"}, 
          sched[ {2, 0, 0, 0}, 1], 
          sched[ {0, 1, 0, 0}, 0]}}, {}]],
      Editable->False]], "Print"],

Cell[BoxData[
    \(show::"error" \(\(:\)\(\ \)\) 
      "error while showing program or expression"\)], "Message"]
}, Open  ]],

Cell["If one wants to constraint coefficients to have 0-1 values: ", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(multiSched[multidimensional -> True, \ 
      extraConstraints -> {\(-1\) <= TAD1 <= 1, \(-1\) <= TBD1 <= 1, \(-1\) <= 
            TBD2 <= 1}]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Starting multidimensionnal scheduling on program "\
\[InvisibleSpace]"testSchedule2"\),
      SequenceForm[ 
      "Starting multidimensionnal scheduling on program ", "testSchedule2"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Scheduling dimension number "\[InvisibleSpace]1\),
      SequenceForm[ "Scheduling dimension number ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("duration options:"\[InvisibleSpace]{}\),
      SequenceForm[ "duration options:", {}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dep to schedule: "\[InvisibleSpace]{1, 2, 3, 4, 5, 
          6, 7}\),
      SequenceForm[ "dep to schedule: ", {1, 2, 3, 4, 5, 6, 7}],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]102\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]112\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 102, "      variables, ", 112, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00002956491.pip", FileType \[Rule] File, 
      Date \[Rule] 3407930955, ByteCount \[Rule] 33685}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00002956491.pipres", FileType \[Rule] File, 
      Date \[Rule] 3407930955, ByteCount \[Rule] 856}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\ "i"\)\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 
        Times[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Scheduling dimension number "\[InvisibleSpace]2\),
      SequenceForm[ "Scheduling dimension number ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("duration options:"\[InvisibleSpace]{}\),
      SequenceForm[ "duration options:", {}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dep to schedule: "\[InvisibleSpace]{5}\),
      SequenceForm[ "dep to schedule: ", {5}],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]59\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]48\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 59, "      variables, ", 48, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00003256491.pip", FileType \[Rule] File, 
      Date \[Rule] 3407930957, ByteCount \[Rule] 8920}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00003256491.pipres", FileType \[Rule] File, 
      Date \[Rule] 3407930957, ByteCount \[Rule] 553}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Schedule Found has "\[InvisibleSpace]2\
\[InvisibleSpace]" dimension"\),
      SequenceForm[ "Schedule Found has ", 2, " dimension"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("               "\[InvisibleSpace]0\),
      SequenceForm[ "               ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("               "\[InvisibleSpace]0\),
      SequenceForm[ "               ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\ "i"\)\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 
        Times[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("               "\[InvisibleSpace]0\),
      SequenceForm[ "               ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("                  "\[InvisibleSpace]"j"\),
      SequenceForm[ "                  ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    \(scheduleResult[
      "testSchedule2", {{"x", {"i", "N", "M"}, sched[{0, 0, 0}, 0], 
          sched[{0, 0, 0}, 0]}, {"y", {"i", "N", "M"}, sched[{2, 0, 0}, 1], 
          sched[{0, 0, 0}, 0]}, {"A", {"i", "N", "M"}, sched[{2, 0, 0}, 0], 
          sched[{0, 0, 0}, 0]}, {"B", {"i", "j", "N", "M"}, 
          sched[{2, 0, 0, 0}, 1], sched[{0, 1, 0, 0}, 0]}}, {}]\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(showSchedResult[$schedule]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("               "\[InvisibleSpace]0\),
      SequenceForm[ "               ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "y", {"i", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("               "\[InvisibleSpace]0\),
      SequenceForm[ "               ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\ "i"\)\),
      SequenceForm[ "T_", "A", {"i", "N", "M"}, " = ", 
        Times[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("               "\[InvisibleSpace]0\),
      SequenceForm[ "               ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "i"\)\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M"}, " = ", 
        Plus[ 1, 
          Times[ 2, "i"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("                  "\[InvisibleSpace]"j"\),
      SequenceForm[ "                  ", "j"],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \($scheduleSolution\)], "Input"],

Cell[BoxData[
    \({4, {AA \[Rule] \(-1\), AB \[Rule] \(-1\), aparam \[Rule] 0, 
        Ax \[Rule] \(-1\), Ay \[Rule] 0, slack1 \[Rule] 0, slack2 \[Rule] 1, 
        slack3 \[Rule] 0, slack4 \[Rule] 1, slack5 \[Rule] 1, 
        slack6 \[Rule] 1, TAD1 \[Rule] 1, TAD2 \[Rule] 0, TAD3 \[Rule] 0, 
        TBD1 \[Rule] 1, TBD2 \[Rule] 0, TBD3 \[Rule] 0, TBD4 \[Rule] 0, 
        TM \[Rule] 0, TN \[Rule] 1, TxD1 \[Rule] 0, TxD2 \[Rule] 0, 
        TxD3 \[Rule] 0, TyD1 \[Rule] 0, TyD2 \[Rule] 1, TyD3 \[Rule] 0}}\)], 
  "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(solveTimeMinSchedule[multidimensional -> True, \ 
      extraConstraints
         -> {\(-1\) <= TAD1 <= 1, \(-1\) <= TBD1 <= 1, \(-1\) <= TBD2 <= 1}, 
      summary -> True]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Calling "\[InvisibleSpace]timeMinSchedule\),
      SequenceForm[ "Calling ", timeMinSchedule],
      Editable->False]], "Print"],

Cell[BoxData[
    \(dom2al::"typeErr" \( : \ \) 
      "parameters should be {constraint,index}"\)], "Message"],

Cell[BoxData[
    \(summaryScheduleConstraints::"dom2al" \( : \ \) 
      "unexpected error while computing Alpha domain."\)], "Message"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Exemple 1 - Le produit matrice-vecteur", "Section",
  FontColor->RGBColor[0, 1, 0]],

Cell["Load program :", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(load["\<MV1.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \("[prodVect]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"]
}, Open  ]],

Cell["Call scheduler : ", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(scd[]\)], "Input"],

Cell[BoxData[
    \(True\)], "Print"],

Cell[BoxData[
    \(\(-11\) - 8\ Aa - 6\ Ab + 2\ AC + 12\ aparam - 14\ TaD1 - 14\ TaD2 - 
      20\ TaD3 - 11\ TbD1 - 16\ TbD2 + TcD1 + 6\ TCD1 + 2\ TcD2 + 7\ TCD2 + 
      10\ TCD3 + 37\ TN\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(57\[InvisibleSpace]" constraints, \
"\[InvisibleSpace]16\[InvisibleSpace]" variables \
"\[InvisibleSpace]\(0.02779799999999888`\ Second\)\),
      SequenceForm[ 57, " constraints, ", 16, " variables ", 
        Times[ 0.027797999999998879, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    \({"prodVect", 16, 57, 0.02779799999999888`\ Second, "Vertex", 
      "Flat"}\)], "Print"],

Cell[BoxData[
    \(scheduleResult[
      "prodVect", {{"a", {"i", "j", "N"}, 
          sched[{0, 1, 0}, 0]}, {"b", {"i", "N"}, 
          sched[{1, 0}, 0]}, {"c", {"i", "N"}, 
          sched[{0, 1}, 2]}, {"C", {"i", "j", "N"}, 
          sched[{0, 1, 0}, 1]}}, {1, 2}]\)], "Output"]
}, Open  ]],

Cell["Show schedule :", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(showSchedResult[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "a", {"i", "j", "N"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "b", {"i", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2 + "N"\)\),
      SequenceForm[ "T_", "c", {"i", "N"}, " = ", 
        Plus[ 2, "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "j"\)\),
      SequenceForm[ "T_", "C", {"i", "j", "N"}, " = ", 
        Plus[ 1, "j"]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(scd[additionalObjectiveFunction \[Rule] False];\)\)], "Input"],

Cell[BoxData[
    \(False\)], "Print"],

Cell[BoxData[
    \(TN\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(57\[InvisibleSpace]" constraints, \
"\[InvisibleSpace]16\[InvisibleSpace]" variables \
"\[InvisibleSpace]\(0.025125000000000952`\ Second\)\),
      SequenceForm[ 57, " constraints, ", 16, " variables ", 
        Times[ 0.025125000000000952, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    \({"prodVect", 16, 57, 0.025125000000000952`\ Second, "Vertex", 
      "Flat"}\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(showSchedResult[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"i", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "N"\)\),
      SequenceForm[ "T_", "c", {"i", "N"}, " = ", 
        Plus[ 1, "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "C", {"i", "j", "N"}, " = ", "j"],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(schedule[]\)], "Input"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]73\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]64\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 73, "      variables, ", 64, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00004366731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090847, ByteCount \[Rule] 14793}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00004366731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090847, ByteCount \[Rule] 654}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]\(1 + 
          "N"\)\),
      SequenceForm[ "Total execution Time: ", 
        Plus[ 1, "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"i", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "N"\)\),
      SequenceForm[ "T_", "c", {"i", "N"}, " = ", 
        Plus[ 1, "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "C", {"i", "j", "N"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    \(scheduleResult[
      "prodVect", {{"a", {"i", "j", "N"}, 
          sched[{0, 0, 0}, 0]}, {"b", {"i", "N"}, 
          sched[{0, 0}, 0]}, {"c", {"i", "N"}, 
          sched[{0, 1}, 1]}, {"C", {"i", "j", "N"}, 
          sched[{0, 1, 0}, 0]}}, {1, 1}]\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Chargement du programme initial", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(load["\<MV1.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \("[prodVect]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["G\[EAcute]n\[EAcute]ration des contraintes d'ordonnancement", "Section"],

Cell["\<\
Calcul de l'ensemble des contraintes de d\[EAcute]pendance, \
auxquelles on ajoute les contraintes pour optimiser le temps total de calcul, \
suppos\[EAcute] lin\[EAcute]aire :\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ddd\  = \ timeMinSchedConstraints[]\)], "Input"],

Cell[BoxData[
    \({{{"a", {\(-TaD3\) + TN \[GreaterEqual] 
              0, \(-TaD1\) - TaD3 + TN \[GreaterEqual] 
              0, \(-TaD2\) - TaD3 + TN \[GreaterEqual] 
              0, \(-TaD1\) - TaD2 - TaD3 + TN \[GreaterEqual] 
              0, \(-Aa\) + aparam - 2\ TaD1 - 2\ TaD2 - 2\ TaD3 + 
                2\ TN \[GreaterEqual] 
              0, \(-Aa\) + aparam - TaD1 - 2\ TaD2 - 2\ TaD3 + 
                2\ TN \[GreaterEqual] 
              0, \(-Aa\) + aparam - 2\ TaD1 - TaD2 - 2\ TaD3 + 
                2\ TN \[GreaterEqual] 
              0, \(-Aa\) + aparam - TaD1 - TaD2 - 2\ TaD3 + 
                2\ TN \[GreaterEqual] 
              0}}, {"b", {\(-TbD2\) + TN \[GreaterEqual] 
              0, \(-TbD1\) - TbD2 + TN \[GreaterEqual] 
              0, \(-Ab\) + aparam - 2\ TbD1 - 2\ TbD2 + 2\ TN \[GreaterEqual] 
              0, \(-Ab\) + aparam - TbD1 - 2\ TbD2 + 2\ TN \[GreaterEqual] 
              0}}, {"c", {\(-TcD2\) + TN \[GreaterEqual] 
              0, \(-TcD1\) - TcD2 + TN \[GreaterEqual] 
              0, \(-Ac\) + aparam - 2\ TcD1 - 2\ TcD2 + 2\ TN \[GreaterEqual] 
              0, \(-Ac\) + aparam - TcD1 - 2\ TcD2 + 2\ TN \[GreaterEqual] 
              0}}, {"C", {\(-TCD3\) + TN \[GreaterEqual] 
              0, \(-TCD1\) - TCD3 + TN \[GreaterEqual] 
              0, \(-TCD2\) - TCD3 + TN \[GreaterEqual] 
              0, \(-TCD1\) - TCD2 - TCD3 + TN \[GreaterEqual] 
              0, \(-AC\) + aparam - 2\ TCD1 - 2\ TCD3 + 2\ TN \[GreaterEqual] 
              0, \(-AC\) + aparam - TCD1 - 2\ TCD3 + 2\ TN \[GreaterEqual] 
              0, \(-AC\) + aparam - 2\ TCD1 - 2\ TCD2 - 2\ TCD3 + 
                2\ TN \[GreaterEqual] 
              0, \(-AC\) + aparam - TCD1 - 2\ TCD2 - 2\ TCD3 + 
                2\ TN \[GreaterEqual] 0}}}, {Aa, Ab, Ac, AC, aparam, TaD1, 
        TaD2, TaD3, TbD1, TbD2, TcD1, TCD1, TcD2, TCD2, TCD3, 
        TN}}\)], "Output"]
}, Open  ]],

Cell["\<\
Ici, on calcule l'ensemble des contraintes de d\[EAcute]pendance et \
de positivit\[EAcute] pour le syst\[EGrave]me donn\[EAcute] : \
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ccc\  = \ statScheduleConstraints[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Total time: "\[InvisibleSpace]\(0.03323799999999899`\
\ Second\)\),
      SequenceForm[ "Total time: ", 
        Times[ 0.033237999999998991, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("#variables: "\[InvisibleSpace]14\),
      SequenceForm[ "#variables: ", 14],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("#constraints: "\[InvisibleSpace]33\),
      SequenceForm[ "#constraints: ", 33],
      Editable->False]], "Print"],

Cell[BoxData[
    \({{TaD3 \[GreaterEqual] 0, TaD1 + TaD3 \[GreaterEqual] 0, 
        TaD2 + TaD3 \[GreaterEqual] 0, TaD1 + TaD2 + TaD3 \[GreaterEqual] 0, 
        TbD2 \[GreaterEqual] 0, TbD1 + TbD2 \[GreaterEqual] 0, 
        TcD2 \[GreaterEqual] 0, TcD1 + TcD2 \[GreaterEqual] 0, 
        TCD2 \[GreaterEqual] 1, 
        Ac - AC + 2\ TcD1 - 2\ TCD1 + 2\ TcD2 - 2\ TCD2 - 
            2\ TCD3 \[GreaterEqual] 1, 
        Ac - AC + TcD1 - TCD1 + 2\ TcD2 - 2\ TCD2 - 2\ TCD3 \[GreaterEqual] 
          1, TcD2 - TCD2 - TCD3 \[GreaterEqual] 0, 
        TcD1 - TCD1 + TcD2 - TCD2 - TCD3 \[GreaterEqual] 0, 
        TCD3 \[GreaterEqual] 0, \(-TaD3\) + TCD3 \[GreaterEqual] 
          0, \(-TbD2\) + TCD3 \[GreaterEqual] 0, 
        TCD1 + TCD3 \[GreaterEqual] 
          0, \(-TaD1\) - TaD3 + TCD1 + TCD3 \[GreaterEqual] 
          0, \(-TbD2\) + TCD1 + TCD3 \[GreaterEqual] 0, 
        TCD2 + TCD3 \[GreaterEqual] 
          0, \(-TaD2\) - TaD3 + TCD2 + TCD3 \[GreaterEqual] 
          0, \(-TbD1\) - TbD2 + TCD2 + TCD3 \[GreaterEqual] 0, 
        TCD1 + TCD2 + TCD3 \[GreaterEqual] 
          0, \(-TaD1\) - TaD2 - TaD3 + TCD1 + TCD2 + TCD3 \[GreaterEqual] 
          0, \(-TbD1\) - TbD2 + TCD1 + TCD2 + TCD3 \[GreaterEqual] 
          0, \(-Aa\) + AC - TaD1 - TaD2 - 2\ TaD3 + TCD1 + TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Ab\) + AC - TbD1 - 2\ TbD2 + TCD1 + TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Aa\) + AC - 2\ TaD1 - TaD2 - 2\ TaD3 + 2\ TCD1 + TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Ab\) + AC - TbD1 - 2\ TbD2 + 2\ TCD1 + TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Aa\) + AC - TaD1 - 2\ TaD2 - 2\ TaD3 + TCD1 + 2\ TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Ab\) + AC - 2\ TbD1 - 2\ TbD2 + TCD1 + 2\ TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Aa\) + AC - 2\ TaD1 - 2\ TaD2 - 2\ TaD3 + 2\ TCD1 + 2\ TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Ab\) + AC - 2\ TbD1 - 2\ TbD2 + 2\ TCD1 + 2\ TCD2 + 
            2\ TCD3 \[GreaterEqual] 1}, {Aa, Ab, Ac, AC, TaD1, TaD2, TaD3, 
        TbD1, TbD2, TcD1, TCD1, TcD2, TCD2, TCD3}}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Construction du syst\[EGrave]me g\[EAcute]n\[EAcute]rateur de la \
solution", "Section"],

Cell[CellGroupData[{

Cell[BoxData[
    \(scd[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Additional objective function: \
"\[InvisibleSpace]{TaD3, TaD1 + TaD3, TaD2 + TaD3, TaD1 + TaD2 + TaD3, TbD2, 
          TbD1 + TbD2, TcD2, 
          TcD1 + TcD2, \(-1\) + TCD2, \(-1\) + Ac - AC + 2\ TcD1 - 2\ TCD1 + 
            2\ TcD2 - 2\ TCD2 - 2\ TCD3, \(-1\) + Ac - AC + TcD1 - TCD1 + 
            2\ TcD2 - 2\ TCD2 - 2\ TCD3, TcD2 - TCD2 - TCD3, 
          TcD1 - TCD1 + TcD2 - TCD2 - TCD3, 
          TCD3, \(-TaD3\) + TCD3, \(-TbD2\) + TCD3, 
          TCD1 + TCD3, \(-TaD1\) - TaD3 + TCD1 + TCD3, \(-TbD2\) + TCD1 + 
            TCD3, TCD2 + TCD3, \(-TaD2\) - TaD3 + TCD2 + TCD3, \(-TbD1\) - 
            TbD2 + TCD2 + TCD3, 
          TCD1 + TCD2 + TCD3, \(-TaD1\) - TaD2 - TaD3 + TCD1 + TCD2 + 
            TCD3, \(-TbD1\) - TbD2 + TCD1 + TCD2 + TCD3, \(-1\) - Aa + AC - 
            TaD1 - TaD2 - 2\ TaD3 + TCD1 + TCD2 + 2\ TCD3, \(-1\) - Ab + AC - 
            TbD1 - 2\ TbD2 + TCD1 + TCD2 + 2\ TCD3, \(-1\) - Aa + AC - 
            2\ TaD1 - TaD2 - 2\ TaD3 + 2\ TCD1 + TCD2 + 2\ TCD3, \(-1\) - 
            Ab + AC - TbD1 - 2\ TbD2 + 2\ TCD1 + TCD2 + 2\ TCD3, \(-1\) - 
            Aa + AC - TaD1 - 2\ TaD2 - 2\ TaD3 + TCD1 + 2\ TCD2 + 
            2\ TCD3, \(-1\) - Ab + AC - 2\ TbD1 - 2\ TbD2 + TCD1 + 2\ TCD2 + 
            2\ TCD3, \(-1\) - Aa + AC - 2\ TaD1 - 2\ TaD2 - 2\ TaD3 + 
            2\ TCD1 + 2\ TCD2 + 2\ TCD3, \(-1\) - Ab + AC - 2\ TbD1 - 
            2\ TbD2 + 2\ TCD1 + 2\ TCD2 + 2\ TCD3, \(-TaD3\) + 
            TN, \(-TaD1\) - TaD3 + TN, \(-TaD2\) - TaD3 + TN, \(-TaD1\) - 
            TaD2 - TaD3 + TN, \(-TbD2\) + TN, \(-TbD1\) - TbD2 + 
            TN, \(-TcD2\) + TN, \(-TcD1\) - TcD2 + TN, \(-TCD3\) + 
            TN, \(-TCD1\) - TCD3 + TN, \(-TCD2\) - TCD3 + TN, \(-TCD1\) - 
            TCD2 - TCD3 + TN, \(-Aa\) + aparam - 2\ TaD1 - 2\ TaD2 - 
            2\ TaD3 + 2\ TN, \(-Aa\) + aparam - TaD1 - 2\ TaD2 - 2\ TaD3 + 
            2\ TN, \(-Aa\) + aparam - 2\ TaD1 - TaD2 - 2\ TaD3 + 
            2\ TN, \(-Aa\) + aparam - TaD1 - TaD2 - 2\ TaD3 + 
            2\ TN, \(-Ab\) + aparam - 2\ TbD1 - 2\ TbD2 + 2\ TN, \(-Ab\) + 
            aparam - TbD1 - 2\ TbD2 + 2\ TN, \(-Ac\) + aparam - 2\ TcD1 - 
            2\ TcD2 + 2\ TN, \(-Ac\) + aparam - TcD1 - 2\ TcD2 + 
            2\ TN, \(-AC\) + aparam - 2\ TCD1 - 2\ TCD3 + 2\ TN, \(-AC\) + 
            aparam - TCD1 - 2\ TCD3 + 2\ TN, \(-AC\) + aparam - 2\ TCD1 - 
            2\ TCD2 - 2\ TCD3 + 2\ TN, \(-AC\) + aparam - TCD1 - 2\ TCD2 - 
            2\ TCD3 + 2\ TN}\),
      SequenceForm[ "Additional objective function: ", {TaD3, 
        Plus[ TaD1, TaD3], 
        Plus[ TaD2, TaD3], 
        Plus[ TaD1, TaD2, TaD3], TbD2, 
        Plus[ TbD1, TbD2], TcD2, 
        Plus[ TcD1, TcD2], 
        Plus[ -1, TCD2], 
        Plus[ -1, Ac, 
          Times[ -1, AC], 
          Times[ 2, TcD1], 
          Times[ -2, TCD1], 
          Times[ 2, TcD2], 
          Times[ -2, TCD2], 
          Times[ -2, TCD3]], 
        Plus[ -1, Ac, 
          Times[ -1, AC], TcD1, 
          Times[ -1, TCD1], 
          Times[ 2, TcD2], 
          Times[ -2, TCD2], 
          Times[ -2, TCD3]], 
        Plus[ TcD2, 
          Times[ -1, TCD2], 
          Times[ -1, TCD3]], 
        Plus[ TcD1, 
          Times[ -1, TCD1], TcD2, 
          Times[ -1, TCD2], 
          Times[ -1, TCD3]], TCD3, 
        Plus[ 
          Times[ -1, TaD3], TCD3], 
        Plus[ 
          Times[ -1, TbD2], TCD3], 
        Plus[ TCD1, TCD3], 
        Plus[ 
          Times[ -1, TaD1], 
          Times[ -1, TaD3], TCD1, TCD3], 
        Plus[ 
          Times[ -1, TbD2], TCD1, TCD3], 
        Plus[ TCD2, TCD3], 
        Plus[ 
          Times[ -1, TaD2], 
          Times[ -1, TaD3], TCD2, TCD3], 
        Plus[ 
          Times[ -1, TbD1], 
          Times[ -1, TbD2], TCD2, TCD3], 
        Plus[ TCD1, TCD2, TCD3], 
        Plus[ 
          Times[ -1, TaD1], 
          Times[ -1, TaD2], 
          Times[ -1, TaD3], TCD1, TCD2, TCD3], 
        Plus[ 
          Times[ -1, TbD1], 
          Times[ -1, TbD2], TCD1, TCD2, TCD3], 
        Plus[ -1, 
          Times[ -1, Aa], AC, 
          Times[ -1, TaD1], 
          Times[ -1, TaD2], 
          Times[ -2, TaD3], TCD1, TCD2, 
          Times[ 2, TCD3]], 
        Plus[ -1, 
          Times[ -1, Ab], AC, 
          Times[ -1, TbD1], 
          Times[ -2, TbD2], TCD1, TCD2, 
          Times[ 2, TCD3]], 
        Plus[ -1, 
          Times[ -1, Aa], AC, 
          Times[ -2, TaD1], 
          Times[ -1, TaD2], 
          Times[ -2, TaD3], 
          Times[ 2, TCD1], TCD2, 
          Times[ 2, TCD3]], 
        Plus[ -1, 
          Times[ -1, Ab], AC, 
          Times[ -1, TbD1], 
          Times[ -2, TbD2], 
          Times[ 2, TCD1], TCD2, 
          Times[ 2, TCD3]], 
        Plus[ -1, 
          Times[ -1, Aa], AC, 
          Times[ -1, TaD1], 
          Times[ -2, TaD2], 
          Times[ -2, TaD3], TCD1, 
          Times[ 2, TCD2], 
          Times[ 2, TCD3]], 
        Plus[ -1, 
          Times[ -1, Ab], AC, 
          Times[ -2, TbD1], 
          Times[ -2, TbD2], TCD1, 
          Times[ 2, TCD2], 
          Times[ 2, TCD3]], 
        Plus[ -1, 
          Times[ -1, Aa], AC, 
          Times[ -2, TaD1], 
          Times[ -2, TaD2], 
          Times[ -2, TaD3], 
          Times[ 2, TCD1], 
          Times[ 2, TCD2], 
          Times[ 2, TCD3]], 
        Plus[ -1, 
          Times[ -1, Ab], AC, 
          Times[ -2, TbD1], 
          Times[ -2, TbD2], 
          Times[ 2, TCD1], 
          Times[ 2, TCD2], 
          Times[ 2, TCD3]], 
        Plus[ 
          Times[ -1, TaD3], TN], 
        Plus[ 
          Times[ -1, TaD1], 
          Times[ -1, TaD3], TN], 
        Plus[ 
          Times[ -1, TaD2], 
          Times[ -1, TaD3], TN], 
        Plus[ 
          Times[ -1, TaD1], 
          Times[ -1, TaD2], 
          Times[ -1, TaD3], TN], 
        Plus[ 
          Times[ -1, TbD2], TN], 
        Plus[ 
          Times[ -1, TbD1], 
          Times[ -1, TbD2], TN], 
        Plus[ 
          Times[ -1, TcD2], TN], 
        Plus[ 
          Times[ -1, TcD1], 
          Times[ -1, TcD2], TN], 
        Plus[ 
          Times[ -1, TCD3], TN], 
        Plus[ 
          Times[ -1, TCD1], 
          Times[ -1, TCD3], TN], 
        Plus[ 
          Times[ -1, TCD2], 
          Times[ -1, TCD3], TN], 
        Plus[ 
          Times[ -1, TCD1], 
          Times[ -1, TCD2], 
          Times[ -1, TCD3], TN], 
        Plus[ 
          Times[ -1, Aa], aparam, 
          Times[ -2, TaD1], 
          Times[ -2, TaD2], 
          Times[ -2, TaD3], 
          Times[ 2, TN]], 
        Plus[ 
          Times[ -1, Aa], aparam, 
          Times[ -1, TaD1], 
          Times[ -2, TaD2], 
          Times[ -2, TaD3], 
          Times[ 2, TN]], 
        Plus[ 
          Times[ -1, Aa], aparam, 
          Times[ -2, TaD1], 
          Times[ -1, TaD2], 
          Times[ -2, TaD3], 
          Times[ 2, TN]], 
        Plus[ 
          Times[ -1, Aa], aparam, 
          Times[ -1, TaD1], 
          Times[ -1, TaD2], 
          Times[ -2, TaD3], 
          Times[ 2, TN]], 
        Plus[ 
          Times[ -1, Ab], aparam, 
          Times[ -2, TbD1], 
          Times[ -2, TbD2], 
          Times[ 2, TN]], 
        Plus[ 
          Times[ -1, Ab], aparam, 
          Times[ -1, TbD1], 
          Times[ -2, TbD2], 
          Times[ 2, TN]], 
        Plus[ 
          Times[ -1, Ac], aparam, 
          Times[ -2, TcD1], 
          Times[ -2, TcD2], 
          Times[ 2, TN]], 
        Plus[ 
          Times[ -1, Ac], aparam, 
          Times[ -1, TcD1], 
          Times[ -2, TcD2], 
          Times[ 2, TN]], 
        Plus[ 
          Times[ -1, AC], aparam, 
          Times[ -2, TCD1], 
          Times[ -2, TCD3], 
          Times[ 2, TN]], 
        Plus[ 
          Times[ -1, AC], aparam, 
          Times[ -1, TCD1], 
          Times[ -2, TCD3], 
          Times[ 2, TN]], 
        Plus[ 
          Times[ -1, AC], aparam, 
          Times[ -2, TCD1], 
          Times[ -2, TCD2], 
          Times[ -2, TCD3], 
          Times[ 2, TN]], 
        Plus[ 
          Times[ -1, AC], aparam, 
          Times[ -1, TCD1], 
          Times[ -2, TCD2], 
          Times[ -2, TCD3], 
          Times[ 2, TN]]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Additional objective function: \
"\[InvisibleSpace]\(\(-11\) - 8\ Aa - 6\ Ab + 2\ AC + 12\ aparam - 14\ TaD1 - 
          14\ TaD2 - 20\ TaD3 - 11\ TbD1 - 16\ TbD2 + TcD1 + 6\ TCD1 + 
          2\ TcD2 + 7\ TCD2 + 10\ TCD3 + 36\ TN\)\),
      SequenceForm[ "Additional objective function: ", 
        Plus[ -11, 
          Times[ -8, Aa], 
          Times[ -6, Ab], 
          Times[ 2, AC], 
          Times[ 12, aparam], 
          Times[ -14, TaD1], 
          Times[ -14, TaD2], 
          Times[ -20, TaD3], 
          Times[ -11, TbD1], 
          Times[ -16, TbD2], TcD1, 
          Times[ 6, TCD1], 
          Times[ 2, TcD2], 
          Times[ 7, TCD2], 
          Times[ 10, TCD3], 
          Times[ 36, TN]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(-11\) - 8\ Aa - 6\ Ab + 2\ AC + 12\ aparam - 14\ TaD1 - 14\ TaD2 - 
      20\ TaD3 - 11\ TbD1 - 16\ TbD2 + TcD1 + 6\ TCD1 + 2\ TcD2 + 7\ TCD2 + 
      10\ TCD3 + 37\ TN\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(57\[InvisibleSpace]" constraints, \
"\[InvisibleSpace]16\[InvisibleSpace]" variables \
"\[InvisibleSpace]\(0.027407999999999433`\ Second\)\),
      SequenceForm[ 57, " constraints, ", 16, " variables ", 
        Times[ 0.027407999999999433, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    \({"prodVect", 16, 57, 0.027407999999999433`\ Second, "Vertex", 
      "Flat"}\)], "Print"],

Cell[BoxData[
    \(scheduleResult[
      "prodVect", {{"a", {"i", "j", "N"}, 
          sched[{0, 1, 0}, 0]}, {"b", {"i", "N"}, 
          sched[{1, 0}, 0]}, {"c", {"i", "N"}, 
          sched[{0, 1}, 2]}, {"C", {"i", "j", "N"}, 
          sched[{0, 1, 0}, 1]}}, {1, 2}]\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(showSchedResult[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "a", {"i", "j", "N"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "b", {"i", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2 + "N"\)\),
      SequenceForm[ "T_", "c", {"i", "N"}, " = ", 
        Plus[ 2, "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "j"\)\),
      SequenceForm[ "T_", "C", {"i", "j", "N"}, " = ", 
        Plus[ 1, "j"]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(scheduleConstraints[]\)], "Input"],

Cell[BoxData[
    \({{TaD3 \[GreaterEqual] 0, TaD1 + TaD3 \[GreaterEqual] 0, 
        TaD2 + TaD3 \[GreaterEqual] 0, TaD1 + TaD2 + TaD3 \[GreaterEqual] 0, 
        TbD2 \[GreaterEqual] 0, TbD1 + TbD2 \[GreaterEqual] 0, 
        TcD2 \[GreaterEqual] 0, TcD1 + TcD2 \[GreaterEqual] 0, 
        TCD2 \[GreaterEqual] 1, 
        Ac - AC + 2\ TcD1 - 2\ TCD1 + 2\ TcD2 - 2\ TCD2 - 
            2\ TCD3 \[GreaterEqual] 1, 
        Ac - AC + TcD1 - TCD1 + 2\ TcD2 - 2\ TCD2 - 2\ TCD3 \[GreaterEqual] 
          1, TcD2 - TCD2 - TCD3 \[GreaterEqual] 0, 
        TcD1 - TCD1 + TcD2 - TCD2 - TCD3 \[GreaterEqual] 0, 
        TCD3 \[GreaterEqual] 0, \(-TaD3\) + TCD3 \[GreaterEqual] 
          0, \(-TbD2\) + TCD3 \[GreaterEqual] 0, 
        TCD1 + TCD3 \[GreaterEqual] 
          0, \(-TaD1\) - TaD3 + TCD1 + TCD3 \[GreaterEqual] 
          0, \(-TbD2\) + TCD1 + TCD3 \[GreaterEqual] 0, 
        TCD2 + TCD3 \[GreaterEqual] 
          0, \(-TaD2\) - TaD3 + TCD2 + TCD3 \[GreaterEqual] 
          0, \(-TbD1\) - TbD2 + TCD2 + TCD3 \[GreaterEqual] 0, 
        TCD1 + TCD2 + TCD3 \[GreaterEqual] 
          0, \(-TaD1\) - TaD2 - TaD3 + TCD1 + TCD2 + TCD3 \[GreaterEqual] 
          0, \(-TbD1\) - TbD2 + TCD1 + TCD2 + TCD3 \[GreaterEqual] 
          0, \(-Aa\) + AC - TaD1 - TaD2 - 2\ TaD3 + TCD1 + TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Ab\) + AC - TbD1 - 2\ TbD2 + TCD1 + TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Aa\) + AC - 2\ TaD1 - TaD2 - 2\ TaD3 + 2\ TCD1 + TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Ab\) + AC - TbD1 - 2\ TbD2 + 2\ TCD1 + TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Aa\) + AC - TaD1 - 2\ TaD2 - 2\ TaD3 + TCD1 + 2\ TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Ab\) + AC - 2\ TbD1 - 2\ TbD2 + TCD1 + 2\ TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Aa\) + AC - 2\ TaD1 - 2\ TaD2 - 2\ TaD3 + 2\ TCD1 + 2\ TCD2 + 
            2\ TCD3 \[GreaterEqual] 
          1, \(-Ab\) + AC - 2\ TbD1 - 2\ TbD2 + 2\ TCD1 + 2\ TCD2 + 
            2\ TCD3 \[GreaterEqual] 1}, {Aa, Ab, Ac, AC, TaD1, TaD2, TaD3, 
        TbD1, TbD2, TcD1, TCD1, TcD2, TCD2, TCD3}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(summaryScheduleConstraints[]\)], "Input"],

Cell[BoxData[
    \({{TaD3 \[GreaterEqual] 0, TaD1 + TaD3 \[GreaterEqual] 0, 
        TaD2 + TaD3 \[GreaterEqual] 0, TaD1 + TaD2 + TaD3 \[GreaterEqual] 0, 
        TbD2 \[GreaterEqual] 0, 
        TbD1 + TbD2 \[GreaterEqual] 0, \(-TaD3\) + TcD2 \[GreaterEqual] 
          1, \(-TaD2\) - TaD3 + TcD2 \[GreaterEqual] 
          0, \(-TbD2\) + TcD2 \[GreaterEqual] 
          1, \(-TbD1\) - TbD2 + TcD2 \[GreaterEqual] 
          0, \(-TaD1\) - TaD3 + TcD1 + TcD2 \[GreaterEqual] 
          1, \(-TaD1\) - TaD2 - TaD3 + TcD1 + TcD2 \[GreaterEqual] 
          0, \(-TbD2\) + TcD1 + TcD2 \[GreaterEqual] 
          1, \(-TbD1\) - TbD2 + TcD1 + TcD2 \[GreaterEqual] 
          0, \(-Aa\) + Ac - TaD1 - 2\ TaD2 - 2\ TaD3 + TcD1 + 
            2\ TcD2 \[GreaterEqual] 
          2, \(-Aa\) + Ac - TaD1 - TaD2 - 2\ TaD3 + TcD1 + 
            2\ TcD2 \[GreaterEqual] 
          3, \(-Ab\) + Ac - 2\ TbD1 - 2\ TbD2 + TcD1 + 
            2\ TcD2 \[GreaterEqual] 
          2, \(-Ab\) + Ac - TbD1 - 2\ TbD2 + TcD1 + 2\ TcD2 \[GreaterEqual] 
          3, \(-Aa\) + Ac - 2\ TaD1 - 2\ TaD2 - 2\ TaD3 + 2\ TcD1 + 
            2\ TcD2 \[GreaterEqual] 
          2, \(-Aa\) + Ac - 2\ TaD1 - TaD2 - 2\ TaD3 + 2\ TcD1 + 
            2\ TcD2 \[GreaterEqual] 
          3, \(-Ab\) + Ac - 2\ TbD1 - 2\ TbD2 + 2\ TcD1 + 
            2\ TcD2 \[GreaterEqual] 
          2, \(-Ab\) + Ac - TbD1 - 2\ TbD2 + 2\ TcD1 + 
            2\ TcD2 \[GreaterEqual] 3}, {Aa, Ab, Ac, TaD1, TaD2, TaD3, TbD1, 
        TbD2, TcD1, TcD2}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(showSchedResult[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "a", {"i", "j", "N"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "b", {"i", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2 + "N"\)\),
      SequenceForm[ "T_", "c", {"i", "N"}, " = ", 
        Plus[ 2, "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "j"\)\),
      SequenceForm[ "T_", "C", {"i", "j", "N"}, " = ", 
        Plus[ 1, "j"]],
      Editable->False]], "Print"]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Exemple 2 - Multiplication de matrices", "Section",
  FontColor->RGBColor[0, 1, 0]],

Cell[CellGroupData[{

Cell["Chargement du programme initial", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(load["\<MatMat.alpha\>"]; ashow[]\)], "Input"],

Cell[BoxData[
    \("[matmat]? bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad \
char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? \
bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\
\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \r \
(octal=13)\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \
\r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad \
char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? \
bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\
\n"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("system matmat  (a : {i,j | 1<=i<=10; 1<=j<=10} of integer; \n          \
      b : {i,j | 1<=i<=10; 1<=j<=10} of integer)\n       returns (c : {i,j | \
1<=i<=10} of integer);\nvar\n  C : {i,j,k | 1<=i<=10; 1<=j<=10; 0<=k<=10} of \
integer;\n  A : {i,j,k | 1<=i<=10; 0<=j<=10; 1<=k<=10} of integer;\n  B : \
{i,j,k | 0<=i<=10; 1<=j<=10; 1<=k<=10} of integer;\nlet\n  C[i,j,k] = \n      \
case\n        { | k=0} : 0[];\n        { | 0<=k} : C[i,j,k-1] + A[i,j,k] * \
B[i,j,k];\n      esac;\n  A[i,j,k] = \n      case\n        { | j=0} : a[i,k];\
\n        { | 0<=j} : A[i,j-1,k];\n      esac;\n  B[i,j,k] = \n      case\n   \
     { | i=0} : b[k,j];\n        { | 0<=i} : B[i-1,j,k];\n      esac;\n\
tel;"\)], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(load["\<matmat.alpha\>"]; ashow[]\)], "Input"],

Cell[BoxData[
    \("[matmat]? bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad \
char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? \
bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\
\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \r \
(octal=13)\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \
\r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad \
char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? \
bad char: \r (octal=13)\n? bad char: \r (octal=13)\n? bad char: \r (octal=13)\
\n"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("system matmat  (a : {i,j | 1<=i<=10; 1<=j<=10} of integer; \n          \
      b : {i,j | 1<=i<=10; 1<=j<=10} of integer)\n       returns (c : {i,j | \
1<=i<=10} of integer);\nvar\n  C : {i,j,k | 1<=i<=10; 1<=j<=10; 0<=k<=10} of \
integer;\n  A : {i,j,k | 1<=i<=10; 0<=j<=10; 1<=k<=10} of integer;\n  B : \
{i,j,k | 0<=i<=10; 1<=j<=10; 1<=k<=10} of integer;\nlet\n  C[i,j,k] = \n      \
case\n        { | k=0} : 0[];\n        { | 0<=k} : C[i,j,k-1] + A[i,j,k] * \
B[i,j,k];\n      esac;\n  A[i,j,k] = \n      case\n        { | j=0} : a[i,k];\
\n        { | 0<=j} : A[i,j-1,k];\n      esac;\n  B[i,j,k] = \n      case\n   \
     { | i=0} : b[k,j];\n        { | 0<=i} : B[i-1,j,k];\n      esac;\n\
tel;"\)], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(analyze[]\)], "Input"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Static Analysis of system "\[InvisibleSpace]"matmat"\
\),
      SequenceForm[ "Static Analysis of system ", "matmat"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--Checking declaration of variables."\)], "Print"],

Cell[BoxData[
    \("--Checking single assignment rule."\)], "Print"],

Cell[BoxData[
    \("--Checking definitions of output/local variables."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("ERROR: Output/local variable "\[InvisibleSpace]"c"\),
      
      SequenceForm[ "ERROR: Output/local variable ", "c"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("       does not appear at the LHS of any equation.\n"\)], "Print"],

Cell[BoxData[
    \("--Checking definition of input variables."\)], "Print"],

Cell[BoxData[
    \("--Checking that input/local variables are used."\)], "Print"],

Cell[BoxData[
    \("\n*** Analysis failed ***"\)], "Print"],

Cell[BoxData[
    \(False\)], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["G\[EAcute]n\[EAcute]ration des contraintes d'ordonnancement", "Section"],

Cell[CellGroupData[{

Cell[BoxData[
    \(scd[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Additional objective function: \
"\[InvisibleSpace]{\(-Aa\) + aparam - 10\ TaD1 - 10\ TaD2, \(-Aa\) + aparam - 
            TaD1 - 10\ TaD2, \(-Aa\) + aparam - 10\ TaD1 - TaD2, \(-Aa\) + 
            aparam - TaD1 - TaD2, \(-1\) + TAD2, \(-AA\) + aparam - 
            10\ TAD1 - 10\ TAD3, \(-AA\) + aparam - TAD1 - 
            10\ TAD3, \(-AA\) + aparam - 10\ TAD1 - 10\ TAD2 - 
            10\ TAD3, \(-AA\) + aparam - TAD1 - 10\ TAD2 - 
            10\ TAD3, \(-AA\) + aparam - 10\ TAD1 - TAD3, \(-AA\) + aparam - 
            TAD1 - TAD3, \(-AA\) + aparam - 10\ TAD1 - 10\ TAD2 - 
            TAD3, \(-AA\) + aparam - TAD1 - 10\ TAD2 - TAD3, \(-1\) - Aa + 
            AA - TaD1 + TAD1 - TaD2 + TAD3, \(-1\) - Aa + AA - 10\ TaD1 + 
            10\ TAD1 - TaD2 + TAD3, \(-1\) - Aa + AA - TaD1 + TAD1 - 
            10\ TaD2 + 10\ TAD3, \(-1\) - Aa + AA - 10\ TaD1 + 10\ TAD1 - 
            10\ TaD2 + 10\ TAD3, \(-1\) + TBD1, \(-Ab\) + aparam - 10\ TbD1 - 
            10\ TbD2, \(-Ab\) + aparam - TbD1 - 10\ TbD2, \(-Ab\) + aparam - 
            10\ TbD1 - TbD2, \(-Ab\) + aparam - TbD1 - TbD2, \(-AB\) + 
            aparam - 10\ TBD2 - 10\ TBD3, \(-AB\) + aparam - 10\ TBD1 - 
            10\ TBD2 - 10\ TBD3, \(-AB\) + aparam - TBD2 - 
            10\ TBD3, \(-AB\) + aparam - 10\ TBD1 - TBD2 - 
            10\ TBD3, \(-AB\) + aparam - 10\ TBD2 - TBD3, \(-AB\) + aparam - 
            10\ TBD1 - 10\ TBD2 - TBD3, \(-AB\) + aparam - TBD2 - 
            TBD3, \(-AB\) + aparam - 10\ TBD1 - TBD2 - TBD3, \(-1\) - Ab + 
            AB - TbD1 - TbD2 + TBD2 + TBD3, \(-1\) - Ab + AB - TbD1 - 
            10\ TbD2 + 10\ TBD2 + TBD3, \(-1\) - Ab + AB - 10\ TbD1 - TbD2 + 
            TBD2 + 10\ TBD3, \(-1\) - Ab + AB - 10\ TbD1 - 10\ TbD2 + 
            10\ TBD2 + 10\ TBD3, \(-Ac\) + aparam - 10\ TcD1, \(-Ac\) + 
            aparam - TcD1, \(-TcD2\), 
          TcD2, \(-AC\) + aparam - 10\ TCD1 - 10\ TCD2, \(-AC\) + aparam - 
            TCD1 - 10\ TCD2, \(-AC\) + aparam - 10\ TCD1 - TCD2, \(-AC\) + 
            aparam - TCD1 - TCD2, \(-AC\) + aparam - 10\ TCD1 - 10\ TCD2 - 
            10\ TCD3, \(-AC\) + aparam - TCD1 - 10\ TCD2 - 
            10\ TCD3, \(-AC\) + aparam - 10\ TCD1 - TCD2 - 
            10\ TCD3, \(-AC\) + aparam - TCD1 - TCD2 - 10\ TCD3, \(-1\) + 
            TCD3, \(-1\) - AA + AC - TAD1 - TAD2 - TAD3 + TCD1 + TCD2 + 
            TCD3, \(-1\) - AB + AC - TBD1 - TBD2 - TBD3 + TCD1 + TCD2 + 
            TCD3, \(-1\) - AA + AC - 10\ TAD1 - TAD2 - TAD3 + 10\ TCD1 + 
            TCD2 + TCD3, \(-1\) - AB + AC - 10\ TBD1 - TBD2 - TBD3 + 
            10\ TCD1 + TCD2 + TCD3, \(-1\) - AA + AC - TAD1 - 10\ TAD2 - 
            TAD3 + TCD1 + 10\ TCD2 + TCD3, \(-1\) - AB + AC - TBD1 - 
            10\ TBD2 - TBD3 + TCD1 + 10\ TCD2 + TCD3, \(-1\) - AA + AC - 
            10\ TAD1 - 10\ TAD2 - TAD3 + 10\ TCD1 + 10\ TCD2 + TCD3, \(-1\) - 
            AB + AC - 10\ TBD1 - 10\ TBD2 - TBD3 + 10\ TCD1 + 10\ TCD2 + 
            TCD3, \(-1\) - AA + AC - TAD1 - TAD2 - 10\ TAD3 + TCD1 + TCD2 + 
            10\ TCD3, \(-1\) - AB + AC - TBD1 - TBD2 - 10\ TBD3 + TCD1 + 
            TCD2 + 10\ TCD3, \(-1\) - AA + AC - 10\ TAD1 - TAD2 - 10\ TAD3 + 
            10\ TCD1 + TCD2 + 10\ TCD3, \(-1\) - AB + AC - 10\ TBD1 - TBD2 - 
            10\ TBD3 + 10\ TCD1 + TCD2 + 10\ TCD3, \(-1\) - AA + AC - TAD1 - 
            10\ TAD2 - 10\ TAD3 + TCD1 + 10\ TCD2 + 10\ TCD3, \(-1\) - AB + 
            AC - TBD1 - 10\ TBD2 - 10\ TBD3 + TCD1 + 10\ TCD2 + 
            10\ TCD3, \(-1\) - AA + AC - 10\ TAD1 - 10\ TAD2 - 10\ TAD3 + 
            10\ TCD1 + 10\ TCD2 + 10\ TCD3, \(-1\) - AB + AC - 10\ TBD1 - 
            10\ TBD2 - 10\ TBD3 + 10\ TCD1 + 10\ TCD2 + 10\ TCD3}\),
      SequenceForm[ "Additional objective function: ", {
        Plus[ 
          Times[ -1, Aa], aparam, 
          Times[ -10, TaD1], 
          Times[ -10, TaD2]], 
        Plus[ 
          Times[ -1, Aa], aparam, 
          Times[ -1, TaD1], 
          Times[ -10, TaD2]], 
        Plus[ 
          Times[ -1, Aa], aparam, 
          Times[ -10, TaD1], 
          Times[ -1, TaD2]], 
        Plus[ 
          Times[ -1, Aa], aparam, 
          Times[ -1, TaD1], 
          Times[ -1, TaD2]], 
        Plus[ -1, TAD2], 
        Plus[ 
          Times[ -1, AA], aparam, 
          Times[ -10, TAD1], 
          Times[ -10, TAD3]], 
        Plus[ 
          Times[ -1, AA], aparam, 
          Times[ -1, TAD1], 
          Times[ -10, TAD3]], 
        Plus[ 
          Times[ -1, AA], aparam, 
          Times[ -10, TAD1], 
          Times[ -10, TAD2], 
          Times[ -10, TAD3]], 
        Plus[ 
          Times[ -1, AA], aparam, 
          Times[ -1, TAD1], 
          Times[ -10, TAD2], 
          Times[ -10, TAD3]], 
        Plus[ 
          Times[ -1, AA], aparam, 
          Times[ -10, TAD1], 
          Times[ -1, TAD3]], 
        Plus[ 
          Times[ -1, AA], aparam, 
          Times[ -1, TAD1], 
          Times[ -1, TAD3]], 
        Plus[ 
          Times[ -1, AA], aparam, 
          Times[ -10, TAD1], 
          Times[ -10, TAD2], 
          Times[ -1, TAD3]], 
        Plus[ 
          Times[ -1, AA], aparam, 
          Times[ -1, TAD1], 
          Times[ -10, TAD2], 
          Times[ -1, TAD3]], 
        Plus[ -1, 
          Times[ -1, Aa], AA, 
          Times[ -1, TaD1], TAD1, 
          Times[ -1, TaD2], TAD3], 
        Plus[ -1, 
          Times[ -1, Aa], AA, 
          Times[ -10, TaD1], 
          Times[ 10, TAD1], 
          Times[ -1, TaD2], TAD3], 
        Plus[ -1, 
          Times[ -1, Aa], AA, 
          Times[ -1, TaD1], TAD1, 
          Times[ -10, TaD2], 
          Times[ 10, TAD3]], 
        Plus[ -1, 
          Times[ -1, Aa], AA, 
          Times[ -10, TaD1], 
          Times[ 10, TAD1], 
          Times[ -10, TaD2], 
          Times[ 10, TAD3]], 
        Plus[ -1, TBD1], 
        Plus[ 
          Times[ -1, Ab], aparam, 
          Times[ -10, TbD1], 
          Times[ -10, TbD2]], 
        Plus[ 
          Times[ -1, Ab], aparam, 
          Times[ -1, TbD1], 
          Times[ -10, TbD2]], 
        Plus[ 
          Times[ -1, Ab], aparam, 
          Times[ -10, TbD1], 
          Times[ -1, TbD2]], 
        Plus[ 
          Times[ -1, Ab], aparam, 
          Times[ -1, TbD1], 
          Times[ -1, TbD2]], 
        Plus[ 
          Times[ -1, AB], aparam, 
          Times[ -10, TBD2], 
          Times[ -10, TBD3]], 
        Plus[ 
          Times[ -1, AB], aparam, 
          Times[ -10, TBD1], 
          Times[ -10, TBD2], 
          Times[ -10, TBD3]], 
        Plus[ 
          Times[ -1, AB], aparam, 
          Times[ -1, TBD2], 
          Times[ -10, TBD3]], 
        Plus[ 
          Times[ -1, AB], aparam, 
          Times[ -10, TBD1], 
          Times[ -1, TBD2], 
          Times[ -10, TBD3]], 
        Plus[ 
          Times[ -1, AB], aparam, 
          Times[ -10, TBD2], 
          Times[ -1, TBD3]], 
        Plus[ 
          Times[ -1, AB], aparam, 
          Times[ -10, TBD1], 
          Times[ -10, TBD2], 
          Times[ -1, TBD3]], 
        Plus[ 
          Times[ -1, AB], aparam, 
          Times[ -1, TBD2], 
          Times[ -1, TBD3]], 
        Plus[ 
          Times[ -1, AB], aparam, 
          Times[ -10, TBD1], 
          Times[ -1, TBD2], 
          Times[ -1, TBD3]], 
        Plus[ -1, 
          Times[ -1, Ab], AB, 
          Times[ -1, TbD1], 
          Times[ -1, TbD2], TBD2, TBD3], 
        Plus[ -1, 
          Times[ -1, Ab], AB, 
          Times[ -1, TbD1], 
          Times[ -10, TbD2], 
          Times[ 10, TBD2], TBD3], 
        Plus[ -1, 
          Times[ -1, Ab], AB, 
          Times[ -10, TbD1], 
          Times[ -1, TbD2], TBD2, 
          Times[ 10, TBD3]], 
        Plus[ -1, 
          Times[ -1, Ab], AB, 
          Times[ -10, TbD1], 
          Times[ -10, TbD2], 
          Times[ 10, TBD2], 
          Times[ 10, TBD3]], 
        Plus[ 
          Times[ -1, Ac], aparam, 
          Times[ -10, TcD1]], 
        Plus[ 
          Times[ -1, Ac], aparam, 
          Times[ -1, TcD1]], 
        Times[ -1, TcD2], TcD2, 
        Plus[ 
          Times[ -1, AC], aparam, 
          Times[ -10, TCD1], 
          Times[ -10, TCD2]], 
        Plus[ 
          Times[ -1, AC], aparam, 
          Times[ -1, TCD1], 
          Times[ -10, TCD2]], 
        Plus[ 
          Times[ -1, AC], aparam, 
          Times[ -10, TCD1], 
          Times[ -1, TCD2]], 
        Plus[ 
          Times[ -1, AC], aparam, 
          Times[ -1, TCD1], 
          Times[ -1, TCD2]], 
        Plus[ 
          Times[ -1, AC], aparam, 
          Times[ -10, TCD1], 
          Times[ -10, TCD2], 
          Times[ -10, TCD3]], 
        Plus[ 
          Times[ -1, AC], aparam, 
          Times[ -1, TCD1], 
          Times[ -10, TCD2], 
          Times[ -10, TCD3]], 
        Plus[ 
          Times[ -1, AC], aparam, 
          Times[ -10, TCD1], 
          Times[ -1, TCD2], 
          Times[ -10, TCD3]], 
        Plus[ 
          Times[ -1, AC], aparam, 
          Times[ -1, TCD1], 
          Times[ -1, TCD2], 
          Times[ -10, TCD3]], 
        Plus[ -1, TCD3], 
        Plus[ -1, 
          Times[ -1, AA], AC, 
          Times[ -1, TAD1], 
          Times[ -1, TAD2], 
          Times[ -1, TAD3], TCD1, TCD2, TCD3], 
        Plus[ -1, 
          Times[ -1, AB], AC, 
          Times[ -1, TBD1], 
          Times[ -1, TBD2], 
          Times[ -1, TBD3], TCD1, TCD2, TCD3], 
        Plus[ -1, 
          Times[ -1, AA], AC, 
          Times[ -10, TAD1], 
          Times[ -1, TAD2], 
          Times[ -1, TAD3], 
          Times[ 10, TCD1], TCD2, TCD3], 
        Plus[ -1, 
          Times[ -1, AB], AC, 
          Times[ -10, TBD1], 
          Times[ -1, TBD2], 
          Times[ -1, TBD3], 
          Times[ 10, TCD1], TCD2, TCD3], 
        Plus[ -1, 
          Times[ -1, AA], AC, 
          Times[ -1, TAD1], 
          Times[ -10, TAD2], 
          Times[ -1, TAD3], TCD1, 
          Times[ 10, TCD2], TCD3], 
        Plus[ -1, 
          Times[ -1, AB], AC, 
          Times[ -1, TBD1], 
          Times[ -10, TBD2], 
          Times[ -1, TBD3], TCD1, 
          Times[ 10, TCD2], TCD3], 
        Plus[ -1, 
          Times[ -1, AA], AC, 
          Times[ -10, TAD1], 
          Times[ -10, TAD2], 
          Times[ -1, TAD3], 
          Times[ 10, TCD1], 
          Times[ 10, TCD2], TCD3], 
        Plus[ -1, 
          Times[ -1, AB], AC, 
          Times[ -10, TBD1], 
          Times[ -10, TBD2], 
          Times[ -1, TBD3], 
          Times[ 10, TCD1], 
          Times[ 10, TCD2], TCD3], 
        Plus[ -1, 
          Times[ -1, AA], AC, 
          Times[ -1, TAD1], 
          Times[ -1, TAD2], 
          Times[ -10, TAD3], TCD1, TCD2, 
          Times[ 10, TCD3]], 
        Plus[ -1, 
          Times[ -1, AB], AC, 
          Times[ -1, TBD1], 
          Times[ -1, TBD2], 
          Times[ -10, TBD3], TCD1, TCD2, 
          Times[ 10, TCD3]], 
        Plus[ -1, 
          Times[ -1, AA], AC, 
          Times[ -10, TAD1], 
          Times[ -1, TAD2], 
          Times[ -10, TAD3], 
          Times[ 10, TCD1], TCD2, 
          Times[ 10, TCD3]], 
        Plus[ -1, 
          Times[ -1, AB], AC, 
          Times[ -10, TBD1], 
          Times[ -1, TBD2], 
          Times[ -10, TBD3], 
          Times[ 10, TCD1], TCD2, 
          Times[ 10, TCD3]], 
        Plus[ -1, 
          Times[ -1, AA], AC, 
          Times[ -1, TAD1], 
          Times[ -10, TAD2], 
          Times[ -10, TAD3], TCD1, 
          Times[ 10, TCD2], 
          Times[ 10, TCD3]], 
        Plus[ -1, 
          Times[ -1, AB], AC, 
          Times[ -1, TBD1], 
          Times[ -10, TBD2], 
          Times[ -10, TBD3], TCD1, 
          Times[ 10, TCD2], 
          Times[ 10, TCD3]], 
        Plus[ -1, 
          Times[ -1, AA], AC, 
          Times[ -10, TAD1], 
          Times[ -10, TAD2], 
          Times[ -10, TAD3], 
          Times[ 10, TCD1], 
          Times[ 10, TCD2], 
          Times[ 10, TCD3]], 
        Plus[ -1, 
          Times[ -1, AB], AC, 
          Times[ -10, TBD1], 
          Times[ -10, TBD2], 
          Times[ -10, TBD3], 
          Times[ 10, TCD1], 
          Times[ 10, TCD2], 
          Times[ 10, TCD3]]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Additional objective function: \
"\[InvisibleSpace]\(\(-27\) - 8\ Aa - 12\ AA - 8\ Ab - 12\ AB - 2\ Ac + 
          8\ AC + 34\ aparam - 44\ TaD1 - 66\ TAD1 - 44\ TaD2 - 83\ TAD2 - 
          66\ TAD3 - 44\ TbD1 - 83\ TBD1 - 44\ TbD2 - 66\ TBD2 - 66\ TBD3 - 
          11\ TcD1 + 44\ TCD1 + 44\ TCD2 + 49\ TCD3\)\),
      SequenceForm[ "Additional objective function: ", 
        Plus[ -27, 
          Times[ -8, Aa], 
          Times[ -12, AA], 
          Times[ -8, Ab], 
          Times[ -12, AB], 
          Times[ -2, Ac], 
          Times[ 8, AC], 
          Times[ 34, aparam], 
          Times[ -44, TaD1], 
          Times[ -66, TAD1], 
          Times[ -44, TaD2], 
          Times[ -83, TAD2], 
          Times[ -66, TAD3], 
          Times[ -44, TbD1], 
          Times[ -83, TBD1], 
          Times[ -44, TbD2], 
          Times[ -66, TBD2], 
          Times[ -66, TBD3], 
          Times[ -11, TcD1], 
          Times[ 44, TCD1], 
          Times[ 44, TCD2], 
          Times[ 49, TCD3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(-27\) - 8\ Aa - 12\ AA - 8\ Ab - 12\ AB - 2\ Ac + 8\ AC + 
      34\ aparam - 44\ TaD1 - 66\ TAD1 - 44\ TaD2 - 83\ TAD2 - 66\ TAD3 - 
      44\ TbD1 - 83\ TBD1 - 44\ TbD2 - 66\ TBD2 - 66\ TBD3 - 11\ TcD1 + 
      44\ TCD1 + 44\ TCD2 + 49\ TCD3\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(63\[InvisibleSpace]" constraints, \
"\[InvisibleSpace]22\[InvisibleSpace]" variables \
"\[InvisibleSpace]\(0.0787919999999982`\ Second\)\),
      SequenceForm[ 63, " constraints, ", 22, " variables ", 
        Times[ 0.078791999999998197, Second]],
      Editable->False]], "Print"],

Cell[BoxData[
    \({"matmat", 22, 63, 0.0787919999999982`\ Second, "Vertex", 
      "Flat"}\)], "Print"],

Cell[BoxData[
    \(scheduleResult[
      "matmat", {{"a", {"i", "j"}, sched[{0, 1}, 0]}, {"A", {"i", "j", "k"}, 
          sched[{0, 1, 1}, 1]}, {"b", {"i", "j"}, 
          sched[{1, 0}, 0]}, {"B", {"i", "j", "k"}, 
          sched[{1, 0, 1}, 1]}, {"c", {"i", "j"}, 
          sched[{0, 0}, 22]}, {"C", {"i", "j", "k"}, 
          sched[{0, 0, 1}, 12]}}, {22}]\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Options[scd]\)], "Input"],

Cell[BoxData[
    \({verbose \[Rule] False, debug \[Rule] False, onlyEquations \[Rule] all, 
      onlyVars \[Rule] all, resolutionSoft \[Rule] MMA, 
      integerSolution \[Rule] False, addConstraints \[Rule] {}, 
      durations \[Rule] {}, periods \[Rule] {}, durationByEq \[Rule] True, 
      onlyDep \[Rule] all, scheduleType \[Rule] affineByVar, 
      optimizationType \[Rule] time, objFunction \[Rule] lexicographic, 
      verticesPositive \[Rule] False, 
      subSystemSched \[Rule] {scheduleResult[
            "prodVect", {"N"}, {{"a", {"i", "j", "N"}, 
                sched[{0, 0, 0}, 0]}, {"b", {"i", "N"}, 
                sched[{0, 0}, 0]}}, {{"c", {"i", "N"}, sched[{0, 1}, 1]}}]}, 
      multidimensional \[Rule] False, depComputed \[Rule] False, 
      dataFlowVariables \[Rule] all, dataFlowConstantsNull \[Rule] True, 
      dataFlow \[Rule] False, reuseSchedule \[Rule] False, 
      dataFlowPeriod \[Rule] 1, subSystems \[Rule] True, 
      dataFlowObjective \[Rule] Null, eliminatesDoubles \[Rule] False, 
      saveSchedule \[Rule] False}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(schedule[addConstraints -> {TBD3 >= 0}, \n\t
      scheduleType -> sameLinearPart]\)], "Input"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(1\[InvisibleSpace]" Recognized added Constraints: "\),
      
      SequenceForm[ 1, " Recognized added Constraints: "],
      Editable->False]], "Print"],

Cell[BoxData[
    \("   TBD3 >= 0,"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]110\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]118\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 110, "      variables, ", 118, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00001557841.pip", FileType \[Rule] File, 
      Date \[Rule] 3407937782, ByteCount \[Rule] 40420}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00001557841.pipres", FileType \[Rule] File, 
      Date \[Rule] 3407937782, ByteCount \[Rule] 951}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]0\),
      SequenceForm[ "Total execution Time: ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", 
          "j"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "j"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", 
          "j"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"i", "j"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", 
          "j"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "c", {"i", "j"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "k"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "i" + "j" + 
          "k"\)\),
      SequenceForm[ "T_", "C", {"i", "j", "k"}, " = ", 
        Plus[ 1, "i", "j", "k"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "k"}\[InvisibleSpace]" = "\[InvisibleSpace]\("i" + "j" + "k"\)\),
      SequenceForm[ "T_", "A", {"i", "j", "k"}, " = ", 
        Plus[ "i", "j", "k"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "k"}\[InvisibleSpace]" = "\[InvisibleSpace]\("i" + "j" + "k"\)\),
      SequenceForm[ "T_", "B", {"i", "j", "k"}, " = ", 
        Plus[ "i", "j", "k"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \(scheduleResult[
      "matmat", {{"a", {"i", "j"}, sched[{0, 0}, 0]}, {"b", {"i", "j"}, 
          sched[{0, 0}, 0]}, {"c", {"i", "j"}, 
          sched[{0, 0}, 0]}, {"C", {"i", "j", "k"}, 
          sched[{1, 1, 1}, 1]}, {"A", {"i", "j", "k"}, 
          sched[{1, 1, 1}, 0]}, {"B", {"i", "j", "k"}, 
          sched[{1, 1, 1}, 0]}}, {0}]\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(showSchedResult[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", 
          "j"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "j"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", 
          "j"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"i", "j"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", 
          "j"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "c", {"i", "j"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "k"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "i" + "j" + 
          "k"\)\),
      SequenceForm[ "T_", "C", {"i", "j", "k"}, " = ", 
        Plus[ 1, "i", "j", "k"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "k"}\[InvisibleSpace]" = "\[InvisibleSpace]\("i" + "j" + "k"\)\),
      SequenceForm[ "T_", "A", {"i", "j", "k"}, " = ", 
        Plus[ "i", "j", "k"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "k"}\[InvisibleSpace]" = "\[InvisibleSpace]\("i" + "j" + "k"\)\),
      SequenceForm[ "T_", "B", {"i", "j", "k"}, " = ", 
        Plus[ "i", "j", "k"]],
      Editable->False]], "Print"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Testing Schedule", "Section"],

Cell[CellGroupData[{

Cell[BoxData[
    \(tests["\<Schedule\>"]\)], "Input"],

Cell[BoxData[
    \("/Users/quinton/mmalpha/tests/"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Calling Test"\[InvisibleSpace]"Schedule.m"\),
      SequenceForm[ "Calling Test", "Schedule.m"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("tests[\"Schedule\" ] is obsolete, please call tests[\"NewSchedule\"]: \
"\)], "Print"],

Cell[BoxData[
    \("/Users/quinton/mmalpha/tests/"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Calling Test"\[InvisibleSpace]"NewSchedule.m"\),
      SequenceForm[ "Calling Test", "NewSchedule.m"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Test file for the Schedule package\n"\)], "Print"],

Cell[BoxData[
    \("*************************************************"\)], "Print"],

Cell[BoxData[
    \("Part1: Testing the piecewise affine schedule on various example"\)], \
"Print"],

Cell[BoxData[
    \("*************************************************\n"\)], "Print"],

Cell[BoxData[
    \("Test1:3 parameter does nothing"\)], "Print"],

Cell[BoxData[
    \("[Test1]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]26\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]24\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 26, "      variables, ", 24, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00004866731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090930, ByteCount \[Rule] 2168}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00004866731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090931, ByteCount \[Rule] 285}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]0\),
      SequenceForm[ "Total execution Time: ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", "j", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "c", {"i", "j", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Test2: No parameter,  matrix product"\)], "Print"],

Cell[BoxData[
    \("[matvect]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]63\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]58\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 63, "      variables, ", 58, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00005366731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090931, ByteCount \[Rule] 11724}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00005366731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090932, ByteCount \[Rule] 627}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]6\),
      SequenceForm[ "Total execution Time: ", 6],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", 
          "k"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "k"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"j"}\
\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"j"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i"}\
\[InvisibleSpace]" = "\[InvisibleSpace]6\),
      SequenceForm[ "T_", "c", {"i"}, " = ", 6],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", 
          "k"}\[InvisibleSpace]" = "\[InvisibleSpace]1\),
      SequenceForm[ "T_", "A", {"i", "k"}, " = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", 
          "k"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "k"\)\),
      SequenceForm[ "T_", "C", {"i", "k"}, " = ", 
        Plus[ 1, "k"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Test3: assignement of a vector, 1 parameter "\)], "Print"],

Cell[BoxData[
    \("[assign_vect1]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]23\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]24\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 23, "      variables, ", 24, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00005866731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090933, ByteCount \[Rule] 1912}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00005866731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090933, ByteCount \[Rule] 281}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]"N"\),
      SequenceForm[ "Total execution Time: ", "N"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"v"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "v", {"i", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Test4: Testing schedule Type 0"\)], "Print"],

Cell[BoxData[
    \("[test7]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]64\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]64\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 64, "      variables, ", 64, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00006366731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090934, ByteCount \[Rule] 13047}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00006366731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090934, ByteCount \[Rule] 608}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]\(4 + 
          "N"\)\),
      SequenceForm[ "Total execution Time: ", 
        Plus[ 4, "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"k", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"k", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(4 + "j"\)\),
      SequenceForm[ "T_", "b", {"j", "N"}, " = ", 
        Plus[ 4, "j"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A1"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "i"\)\),
      SequenceForm[ "T_", "A1", {"i", "N"}, " = ", 
        Plus[ 1, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A2"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(3 + "i"\)\),
      SequenceForm[ "T_", "A2", {"i", "N"}, " = ", 
        Plus[ 3, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("*************************************************"\)], "Print"],

Cell[BoxData[
    \("Part 2:Testing some Alpha program special forms"\)], "Print"],

Cell[BoxData[
    \("*************************************************"\)], "Print"],

Cell[BoxData[
    \("Testing affine functions"\)], "Print"],

Cell[BoxData[
    \("[matvect]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]102\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]88\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 102, "      variables, ", 88, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00006866731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090936, ByteCount \[Rule] 28338}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00006866731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090936, ByteCount \[Rule] 885}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]"N"\),
      SequenceForm[ "Total execution Time: ", "N"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "k", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "k", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"N"\),
      SequenceForm[ "T_", "c", {"i", "N"}, " = ", "N"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "k", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "A", {"i", "k", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "k", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(\(-1\) + "i"\)\),
      SequenceForm[ "T_", "B", {"i", "k", "N"}, " = ", 
        Plus[ -1, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"k", "i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(\(-1\) + "i"\)\),
      SequenceForm[ "T_", "C", {"k", "i", "N"}, " = ", 
        Plus[ -1, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Testing union of prolyhedra"\)], "Print"],

Cell[BoxData[
    \("[Test9]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]41\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]40\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 41, "      variables, ", 40, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00007366731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090938, ByteCount \[Rule] 5397}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00007366731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090938, ByteCount \[Rule] 456}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]0\),
      SequenceForm[ "Total execution Time: ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i"}\
\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"w"\[InvisibleSpace]{"p"}\
\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "w", {"p"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i"}\
\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "y", {"i"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"Y"\[InvisibleSpace]{"t", 
          "p"}\[InvisibleSpace]" = "\[InvisibleSpace]\(3\ "t"\)\),
      SequenceForm[ "T_", "Y", {"t", "p"}, " = ", 
        Times[ 3, "t"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Testing non computable program"\)], "Print"],

Cell[BoxData[
    \("[Test6]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]96\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]104\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 96, "      variables, ", 104, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00007866731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090940, ByteCount \[Rule] 31234}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00007866731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090940, ByteCount \[Rule] 212}\)], "Print"],

Cell[BoxData[
    \(Alpha`FarkasSchedule`Private`ReadPipResult::"PipOutputError" \(\(:\)\(\ \
\)\) " No schedule found, sorry  ..."\)], "Message"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Testing Unbounded Time program"\)], "Print"],

Cell[BoxData[
    \("[Test9]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]36\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]32\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 36, "      variables, ", 32, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00008366731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090941, ByteCount \[Rule] 3876}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00008366731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090941, ByteCount \[Rule] 210}\)], "Print"],

Cell[BoxData[
    \(Alpha`FarkasSchedule`Private`ReadPipResult::"PipOutputError" \(\(:\)\(\ \
\)\) " No schedule found, sorry  ..."\)], "Message"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("But If you do not specify an objective function:"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]35\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]28\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 35, "      variables, ", 28, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00008666731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090941, ByteCount \[Rule] 3335}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00008666731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090941, ByteCount \[Rule] 418}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i"}\
\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"w"\[InvisibleSpace]{"p"}\
\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "w", {"p"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i"}\
\[InvisibleSpace]" = "\[InvisibleSpace]\(2 + "i"\)\),
      SequenceForm[ "T_", "y", {"i"}, " = ", 
        Plus[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"X"\[InvisibleSpace]{"i"}\
\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "i"\)\),
      SequenceForm[ "T_", "X", {"i"}, " = ", 
        Plus[ 1, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("*************************************************"\)], "Print"],

Cell[BoxData[
    \("Part 3: Testing the options"\)], "Print"],

Cell[BoxData[
    \("*************************************************\n\n"\)], "Print"],

Cell[BoxData[
    \("***Debug option"\)], "Print"],

Cell[BoxData[
    \("[Test1]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("optimization Type: "\[InvisibleSpace]time\),
      SequenceForm[ "optimization Type: ", time],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Number of Variables: "\)], "Print"],

Cell[BoxData[
    \({{"c", 1}}\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Dependencies (numbered in this order):"\)], "Print"],

Cell[BoxData[
    \(\*"\<\"\"\>"\)], "Print"],

Cell[BoxData[
    \("Duration : 1 for each equation "\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]26\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]24\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 26, "      variables, ", 24, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("   MuTcD1 - MuTcD5 + TcD1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -MuTcD2 + MuTcD3 + TcD2 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -MuTcD1 + TcD3 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -MuTcD4 + TcD4 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -MuTcD3 + TcD5 = 0,"\)], "Print"],

Cell[BoxData[
    \("   Cc + MuTcD2 + MuTcD4 + MuTcD5 - MuTcD6 - MuTcD7 = 0,"\)], "Print"],

Cell[BoxData[
    \("   RhoTcD1 - RhoTcD5 - TcD1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -RhoTcD2 + RhoTcD3 - TcD2 = 0,"\)], "Print"],

Cell[BoxData[
    \("   Phi1 - RhoTcD1 - TcD3 = 0,"\)], "Print"],

Cell[BoxData[
    \("   Phi2 - RhoTcD4 - TcD4 = 0,"\)], "Print"],

Cell[BoxData[
    \("   Phi3 - RhoTcD3 - TcD5 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -Cc + Phi0 + RhoTcD2 + RhoTcD4 + RhoTcD5 - RhoTcD6 - RhoTcD7 = \
0,"\)], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("temporary file: \
"\[InvisibleSpace]"/tmp/m00009266731"\),
      SequenceForm[ "temporary file: ", "/tmp/m00009266731"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Opening: "\[InvisibleSpace]"/tmp/m00009266731.pip"\),
      
      SequenceForm[ "Opening: ", "/tmp/m00009266731.pip"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Opening "\[InvisibleSpace]"/tmp/m00009366731"\),
      SequenceForm[ "Opening ", "/tmp/m00009366731"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Running sed process on \
"\[InvisibleSpace]"/tmp/m00009366731"\[InvisibleSpace]" to \
"\[InvisibleSpace]"/tmp/m00009466731"\),
      SequenceForm[ 
      "Running sed process on ", "/tmp/m00009366731", " to ", 
        "/tmp/m00009466731"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Appending "\[InvisibleSpace]"/tmp/m00009466731"\
\[InvisibleSpace]" the end of "\[InvisibleSpace]"/tmp/m00009266731.pip"\),
      SequenceForm[ 
      "Appending ", "/tmp/m00009466731", " the end of ", 
        "/tmp/m00009266731.pip"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("/tmp/m00009366731"\[InvisibleSpace]"/tmp/\
m00009466731"\[InvisibleSpace]" generated"\),
      SequenceForm[ "/tmp/m00009366731", "/tmp/m00009466731", " generated"],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00009266731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090943, ByteCount \[Rule] 2168}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00009266731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090943, ByteCount \[Rule] 285}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Shift coef: "\[InvisibleSpace]0\),
      SequenceForm[ "Shift coef: ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    \("File /tmp/m00009266731.pip,/tmp/m000092667312.pip, \
/tmp/m00009266731.pipres generated"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]0\),
      SequenceForm[ "Total execution Time: ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", "j", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "c", {"i", "j", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("\n***verbose (False) on the same program"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]26\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]24\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 26, "      variables, ", 24, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00009566731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090943, ByteCount \[Rule] 2168}\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00009566731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090943, ByteCount \[Rule] 285}\)], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("\n***default value for $result"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]26\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]24\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 26, "      variables, ", 24, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00009866731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090943, ByteCount \[Rule] 2168}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00009866731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090944, ByteCount \[Rule] 285}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]0\),
      SequenceForm[ "Total execution Time: ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", "j", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "c", {"i", "j", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("\n***scheduleType option"\)], "Print"],

Cell[BoxData[
    \("[assign_vect1]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("affine by variable"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]41\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]42\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 41, "      variables, ", 42, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00010366731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090944, ByteCount \[Rule] 5635}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00010366731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090944, ByteCount \[Rule] 438}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]"N"\),
      SequenceForm[ "Total execution Time: ", "N"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"v"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "i"\)\),
      SequenceForm[ "T_", "v", {"i", "N"}, " = ", 
        Plus[ 1, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"vv"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "vv", {"i", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"vvv"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "vvv", {"i", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("affine of unique linear Part"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]41\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]46\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 41, "      variables, ", 46, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00010666731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090946, ByteCount \[Rule] 6147}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00010666731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090946, ByteCount \[Rule] 438}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]"N"\),
      SequenceForm[ "Total execution Time: ", "N"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"v"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "i"\)\),
      SequenceForm[ "T_", "v", {"i", "N"}, " = ", 
        Plus[ 1, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"vv"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "vv", {"i", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"vvv"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "vvv", {"i", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("affine of unique linear part except on parameters "\)], "Print"],

Cell[BoxData[
    \("[assign_vect1]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]52\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]58\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 52, "      variables, ", 58, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00011166731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090947, ByteCount \[Rule] 9762}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00011166731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090947, ByteCount \[Rule] 536}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]"N"\),
      SequenceForm[ "Total execution Time: ", "N"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"v"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "i"\)\),
      SequenceForm[ "T_", "v", {"i", "N"}, " = ", 
        Plus[ 1, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"vv"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "vv", {"i", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"vvv"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "vvv", {"i", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"vvvv"\[InvisibleSpace]{"i", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\("i" + "N"\)\),
      SequenceForm[ "T_", "vvvv", {"i", "N"}, " = ", 
        Plus[ "i", "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("\n** optimizationType option"\)], "Print"],

Cell[BoxData[
    \("[Test9]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("No objectibve function on unbounded time"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]35\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]28\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 35, "      variables, ", 28, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00011666731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090948, ByteCount \[Rule] 3335}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00011666731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090948, ByteCount \[Rule] 418}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"x"\[InvisibleSpace]{"i"}\
\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "x", {"i"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"w"\[InvisibleSpace]{"p"}\
\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "w", {"p"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"y"\[InvisibleSpace]{"i"}\
\[InvisibleSpace]" = "\[InvisibleSpace]\(2 + "i"\)\),
      SequenceForm[ "T_", "y", {"i"}, " = ", 
        Plus[ 2, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"X"\[InvisibleSpace]{"i"}\
\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "i"\)\),
      SequenceForm[ "T_", "X", {"i"}, " = ", 
        Plus[ 1, "i"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("And Failing if the total time is minimized"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]36\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]32\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 36, "      variables, ", 32, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00011966731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090949, ByteCount \[Rule] 3876}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00011966731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090949, ByteCount \[Rule] 210}\)], "Print"],

Cell[BoxData[
    \(Alpha`FarkasSchedule`Private`ReadPipResult::"PipOutputError" \(\(:\)\(\ \
\)\) " No schedule found, sorry  ..."\)], "Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(Alpha`FarkasSchedule`Private`ReadPipResult :: \\\"PipOutputError\\\"\\) \
will be suppressed during this calculation. \\!\\(\\*ButtonBox[\\\"More\
\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("\n***Testing Rationnal resolution for Pip"\)], "Print"],

Cell[BoxData[
    \("[test6]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("First in Rationnal"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]28\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]24\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 28, "      variables, ", 24, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00012466731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090950, ByteCount \[Rule] 2350}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00012466731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090950, ByteCount \[Rule] 348}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]0\),
      SequenceForm[ "Total execution Time: ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A1"\[InvisibleSpace]{"j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\/3\),
      SequenceForm[ "T_", "A1", {"j", "N"}, " = ", 
        Times[ 
          Rational[ 1, 3], "j"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A2"\[InvisibleSpace]{"j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2\/3 + "j"\/3\)\),
      SequenceForm[ "T_", "A2", {"j", "N"}, " = ", 
        Plus[ 
          Rational[ 2, 3], 
          Times[ 
            Rational[ 1, 3], "j"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Then in integer"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]28\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]24\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 28, "      variables, ", 24, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00012766731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090951, ByteCount \[Rule] 2350}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00012766731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090951, ByteCount \[Rule] 336}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]0\),
      SequenceForm[ "Total execution Time: ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A1"\[InvisibleSpace]{"j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "A1", {"j", "N"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A2"\[InvisibleSpace]{"j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "A2", {"j", "N"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("\n***Testing added constraints"\)], "Print"],

Cell[BoxData[
    \("[prodVect]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(9\[InvisibleSpace]" Recognized added Constraints: "\),
      
      SequenceForm[ 9, " Recognized added Constraints: "],
      Editable->False]], "Print"],

Cell[BoxData[
    \("   -2 + TaD1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -3 + TaD2 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -3 + TaD3 >= 0,"\)], "Print"],

Cell[BoxData[
    \("   -5 + Ca >= 0,"\)], "Print"],

Cell[BoxData[
    \("   -10 + TAD1 + TCD1 + TCD2 >= 0,"\)], "Print"],

Cell[BoxData[
    \("   -3 + TcD1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   TcD2 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -10 + TcD3 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -4 + Cc = 0,"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]154\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]158\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 154, "      variables, ", 158, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00013266731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090953, ByteCount \[Rule] 74473}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00013266731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090954, ByteCount \[Rule] 1229}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]\(4 + 
          13\ "N"\)\),
      SequenceForm[ "Total execution Time: ", 
        Plus[ 4, 
          Times[ 13, "N"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(5 + 2\ "i" + 3\ "j" + 
          3\ "N"\)\),
      SequenceForm[ "T_", "a", {"i", "j", "N"}, " = ", 
        Plus[ 5, 
          Times[ 2, "i"], 
          Times[ 3, "j"], 
          Times[ 3, "N"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(4 + 3\ "i" + 
          10\ "N"\)\),
      SequenceForm[ "T_", "c", {"i", "j", "N"}, " = ", 
        Plus[ 4, 
          Times[ 3, "i"], 
          Times[ 10, "N"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "B", {"i", "j", "k", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(4 + 3\ "i" + 
          "j" + 3\ "k" + 3\ "N"\)\),
      SequenceForm[ "T_", "A", {"i", "j", "k", "N"}, " = ", 
        Plus[ 4, 
          Times[ 3, "i"], "j", 
          Times[ 3, "k"], 
          Times[ 3, "N"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2 + 6\ "i" + 
          "j" + 3\ "k" + 3\ "N"\)\),
      SequenceForm[ "T_", "C", {"i", "j", "k", "N"}, " = ", 
        Plus[ 2, 
          Times[ 6, "i"], "j", 
          Times[ 3, "k"], 
          Times[ 3, "N"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Testing durations Options"\)], "Print"],

Cell[BoxData[
    \("[prodVect]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]154\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]138\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 154, "      variables, ", 138, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00013766731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090956, ByteCount \[Rule] 66363}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00013766731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090956, ByteCount \[Rule] 1219}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]\(3 + 
          "N"\)\),
      SequenceForm[ "Total execution Time: ", 
        Plus[ 3, "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(3 + "j"\)\),
      SequenceForm[ "T_", "c", {"i", "j", "N"}, " = ", 
        Plus[ 3, "j"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]1\),
      SequenceForm[ "T_", "B", {"i", "j", "k", "N"}, " = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]2\),
      SequenceForm[ "T_", "A", {"i", "j", "k", "N"}, " = ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2 + "k"\)\),
      SequenceForm[ "T_", "C", {"i", "j", "k", "N"}, " = ", 
        Plus[ 2, "k"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Duration not null for :"\[InvisibleSpace]{"a", "b", 
          "c", "B", "A", "C"}\),
      SequenceForm[ 
      "Duration not null for :", {"a", "b", "c", "B", "A", "C"}],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]154\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]138\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 154, "      variables, ", 138, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00014066731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090959, ByteCount \[Rule] 66391}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00014066731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090959, ByteCount \[Rule] 1268}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]\(113 + 
          1000\ "N"\)\),
      SequenceForm[ "Total execution Time: ", 
        Plus[ 113, 
          Times[ 1000, "N"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(113 + 1000\ "j"\)\),
      SequenceForm[ "T_", "c", {"i", "j", "N"}, " = ", 
        Plus[ 113, 
          Times[ 1000, "j"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]10\),
      SequenceForm[ "T_", "B", {"i", "j", "k", "N"}, " = ", 10],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]110\),
      SequenceForm[ "T_", "A", {"i", "j", "k", "N"}, " = ", 110],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(110 + 
          1000\ "k"\)\),
      SequenceForm[ "T_", "C", {"i", "j", "k", "N"}, " = ", 
        Plus[ 110, 
          Times[ 1000, "k"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]154\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]138\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 154, "      variables, ", 138, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00014366731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090962, ByteCount \[Rule] 66363}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00014366731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090962, ByteCount \[Rule] 1219}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]\(3 + 
          "N"\)\),
      SequenceForm[ "Total execution Time: ", 
        Plus[ 3, "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(3 + "j"\)\),
      SequenceForm[ "T_", "c", {"i", "j", "N"}, " = ", 
        Plus[ 3, "j"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]1\),
      SequenceForm[ "T_", "B", {"i", "j", "k", "N"}, " = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]2\),
      SequenceForm[ "T_", "A", {"i", "j", "k", "N"}, " = ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2 + "k"\)\),
      SequenceForm[ "T_", "C", {"i", "j", "k", "N"}, " = ", 
        Plus[ 2, "k"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Duration not null for :"\[InvisibleSpace]{"a", "b", 
          "c", "B", "A", "C"}\),
      SequenceForm[ 
      "Duration not null for :", {"a", "b", "c", "B", "A", "C"}],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]154\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]138\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 154, "      variables, ", 138, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00014666731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090964, ByteCount \[Rule] 66379}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00014666731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090965, ByteCount \[Rule] 1248}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]\(18 + 
          100\ "N"\)\),
      SequenceForm[ "Total execution Time: ", 
        Plus[ 18, 
          Times[ 100, "N"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(18 + 100\ "j"\)\),
      SequenceForm[ "T_", "c", {"i", "j", "N"}, " = ", 
        Plus[ 18, 
          Times[ 100, "j"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]5\),
      SequenceForm[ "T_", "B", {"i", "j", "k", "N"}, " = ", 5],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]15\),
      SequenceForm[ "T_", "A", {"i", "j", "k", "N"}, " = ", 15],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(15 + 100\ "k"\)\),
      
      SequenceForm[ "T_", "C", {"i", "j", "k", "N"}, " = ", 
        Plus[ 15, 
          Times[ 100, "k"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("*** outputForm option"\)], "Print"],

Cell[BoxData[
    \("[Test1]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]26\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]24\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 26, "      variables, ", 24, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("{Phi1, Phi2, Phi3, Phi0, TcD1, TcD2, TcD3, TcD4, TcD5, Cc, MuTcD1, \
MuTcD2, MuTcD3, MuTcD4, MuTcD5, MuTcD6, MuTcD7, RhoTcD1, RhoTcD2, RhoTcD3, \
RhoTcD4, RhoTcD5, RhoTcD6, RhoTcD7|   MuTcD1 - MuTcD5 + TcD1 = 0;   -MuTcD2 + \
MuTcD3 + TcD2 = 0;   -MuTcD1 + TcD3 = 0;   -MuTcD4 + TcD4 = 0;   -MuTcD3 + \
TcD5 = 0;   Cc + MuTcD2 + MuTcD4 + MuTcD5 - MuTcD6 - MuTcD7 = 0;   RhoTcD1 - \
RhoTcD5 - TcD1 = 0;   -RhoTcD2 + RhoTcD3 - TcD2 = 0;   Phi1 - RhoTcD1 - TcD3 \
= 0;   Phi2 - RhoTcD4 - TcD4 = 0;   Phi3 - RhoTcD3 - TcD5 = 0;   -Cc + Phi0 + \
RhoTcD2 + RhoTcD4 + RhoTcD5 - RhoTcD6 - RhoTcD7 = 0;Phi1>=0; Phi2>=0; \
Phi3>=0; Phi0>=0; MuTcD1>=0; MuTcD2>=0; MuTcD3>=0; MuTcD4>=0; MuTcD5>=0; \
MuTcD6>=0; MuTcD7>=0; RhoTcD1>=0; RhoTcD2>=0; RhoTcD3>=0; RhoTcD4>=0; \
RhoTcD5>=0; RhoTcD6>=0; RhoTcD7>=0}"\)], "Print"],

Cell[BoxData[
    \("{Phi1, Phi2, Phi3, Phi0, TcD1, TcD2, TcD3, TcD4, TcD5, Cc, MuTcD1, \
MuTcD2, MuTcD3, MuTcD4, MuTcD5, MuTcD6, MuTcD7, RhoTcD1, RhoTcD2, RhoTcD3, \
RhoTcD4, RhoTcD5, RhoTcD6, RhoTcD7|   MuTcD1 - MuTcD5 + TcD1 = 0;   -MuTcD2 + \
MuTcD3 + TcD2 = 0;   -MuTcD1 + TcD3 = 0;   -MuTcD4 + TcD4 = 0;   -MuTcD3 + \
TcD5 = 0;   Cc + MuTcD2 + MuTcD4 + MuTcD5 - MuTcD6 - MuTcD7 = 0;   RhoTcD1 - \
RhoTcD5 - TcD1 = 0;   -RhoTcD2 + RhoTcD3 - TcD2 = 0;   Phi1 - RhoTcD1 - TcD3 \
= 0;   Phi2 - RhoTcD4 - TcD4 = 0;   Phi3 - RhoTcD3 - TcD5 = 0;   -Cc + Phi0 + \
RhoTcD2 + RhoTcD4 + RhoTcD5 - RhoTcD6 - RhoTcD7 = 0;Phi1>=0; Phi2>=0; \
Phi3>=0; Phi0>=0; MuTcD1>=0; MuTcD2>=0; MuTcD3>=0; MuTcD4>=0; MuTcD5>=0; \
MuTcD6>=0; MuTcD7>=0; RhoTcD1>=0; RhoTcD2>=0; RhoTcD3>=0; RhoTcD4>=0; \
RhoTcD5>=0; RhoTcD6>=0; RhoTcD7>=0}"\)], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]26\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]24\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 26, "      variables, ", 24, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("LP SOLVE FORMAT, WARNING, some Variable may be negative "\)], "Print"],

Cell[BoxData[
    \("min:Phi1;\n   MuTcD1 - MuTcD5 + TcD1 = 0;   -MuTcD2 + MuTcD3 + TcD2 = \
0;   -MuTcD1 + TcD3 = 0;   -MuTcD4 + TcD4 = 0;   -MuTcD3 + TcD5 = 0;   Cc + \
MuTcD2 + MuTcD4 + MuTcD5 - MuTcD6 - MuTcD7 = 0;   RhoTcD1 - RhoTcD5 - TcD1 = \
0;   -RhoTcD2 + RhoTcD3 - TcD2 = 0;   Phi1 - RhoTcD1 - TcD3 = 0;   Phi2 - \
RhoTcD4 - TcD4 = 0;   Phi3 - RhoTcD3 - TcD5 = 0;   -Cc + Phi0 + RhoTcD2 + \
RhoTcD4 + RhoTcD5 - RhoTcD6 - RhoTcD7 = 0;int :Phi1, Phi2, Phi3, Phi0, TcD1, \
TcD2, TcD3, TcD4, TcD5, Cc, MuTcD1, MuTcD2, MuTcD3, MuTcD4, MuTcD5, MuTcD6, \
MuTcD7, RhoTcD1, RhoTcD2, RhoTcD3, RhoTcD4, RhoTcD5, RhoTcD6, RhoTcD7;"\)], \
"Print"],

Cell[BoxData[
    InterpretationBox[\(\("Phi1"\ Alpha`FarkasSchedule`Private`g\)\
\[InvisibleSpace]";"\),
      SequenceForm[ 
        Times[ "Phi1", Alpha`FarkasSchedule`Private`g], ";"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("   MuTcD1 - MuTcD5 + TcD1 = 0;"\)], "Print"],

Cell[BoxData[
    \("   -MuTcD2 + MuTcD3 + TcD2 = 0;"\)], "Print"],

Cell[BoxData[
    \("   -MuTcD1 + TcD3 = 0;"\)], "Print"],

Cell[BoxData[
    \("   -MuTcD4 + TcD4 = 0;"\)], "Print"],

Cell[BoxData[
    \("   -MuTcD3 + TcD5 = 0;"\)], "Print"],

Cell[BoxData[
    \("   Cc + MuTcD2 + MuTcD4 + MuTcD5 - MuTcD6 - MuTcD7 = 0;"\)], "Print"],

Cell[BoxData[
    \("   RhoTcD1 - RhoTcD5 - TcD1 = 0;"\)], "Print"],

Cell[BoxData[
    \("   -RhoTcD2 + RhoTcD3 - TcD2 = 0;"\)], "Print"],

Cell[BoxData[
    \("   Phi1 - RhoTcD1 - TcD3 = 0;"\)], "Print"],

Cell[BoxData[
    \("   Phi2 - RhoTcD4 - TcD4 = 0;"\)], "Print"],

Cell[BoxData[
    \("   Phi3 - RhoTcD3 - TcD5 = 0;"\)], "Print"],

Cell[BoxData[
    \("   -Cc + Phi0 + RhoTcD2 + RhoTcD4 + RhoTcD5 - RhoTcD6 - RhoTcD7 = \
0;"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("int :"\[InvisibleSpace]"Phi1, Phi2, Phi3, Phi0, \
TcD1, TcD2, TcD3, TcD4, TcD5, Cc, MuTcD1, MuTcD2, MuTcD3, MuTcD4, MuTcD5, \
MuTcD6, MuTcD7, RhoTcD1, RhoTcD2, RhoTcD3, RhoTcD4, RhoTcD5, RhoTcD6, \
RhoTcD7"\[InvisibleSpace]";"\),
      SequenceForm[ 
      "int :", "Phi1, Phi2, Phi3, Phi0, TcD1, TcD2, TcD3, TcD4, TcD5, Cc, \
MuTcD1, MuTcD2, MuTcD3, MuTcD4, MuTcD5, MuTcD6, MuTcD7, RhoTcD1, RhoTcD2, \
RhoTcD3, RhoTcD4, RhoTcD5, RhoTcD6, RhoTcD7", ";"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("*************************************************"\)], "Print"],

Cell[BoxData[
    \("Part 4: Testing schedule Tools"\)], "Print"],

Cell[BoxData[
    \("*************************************************\n\n"\)], "Print"],

Cell[BoxData[
    \("Testing appSched "\)], "Print"],

Cell[BoxData[
    \("[prodVect]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]154\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]140\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 154, "      variables, ", 140, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00015766731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090969, ByteCount \[Rule] 67307}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00015766731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090969, ByteCount \[Rule] 1221}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]\(1 + 
          2\ "N"\)\),
      SequenceForm[ "Total execution Time: ", 
        Plus[ 1, 
          Times[ 2, "N"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 2\ "N"\)\),
      SequenceForm[ "T_", "c", {"i", "j", "N"}, " = ", 
        Plus[ 1, 
          Times[ 2, "N"]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"i"\),
      SequenceForm[ "T_", "B", {"i", "j", "k", "N"}, " = ", "i"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "A", {"i", "j", "k", "N"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\("k" + "N"\)\),
      SequenceForm[ "T_", "C", {"i", "j", "k", "N"}, " = ", 
        Plus[ "k", "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Testing isScheduledQ "\)], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Testing appSched with projVector and projMatrix options "\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]154\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]156\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 154, "      variables, ", 156, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00016066731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408090978, ByteCount \[Rule] 74939}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00016066731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408090979, ByteCount \[Rule] 1221}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]\(3\ "N"\)\),
      
      SequenceForm[ "Total execution Time: ", 
        Times[ 3, "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"i", "j", "N"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", "j", 
          "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\("i" + "j" + "N"\)\),
      SequenceForm[ "T_", "c", {"i", "j", "N"}, " = ", 
        Plus[ "i", "j", "N"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(\(-2\) + "i" + 
          "j" + "k"\)\),
      SequenceForm[ "T_", "B", {"i", "j", "k", "N"}, " = ", 
        Plus[ -2, "i", "j", "k"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(\(-2\) + "i" + 
          "j" + "k"\)\),
      SequenceForm[ "T_", "A", {"i", "j", "k", "N"}, " = ", 
        Plus[ -2, "i", "j", "k"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "k", "N"}\[InvisibleSpace]" = "\[InvisibleSpace]\(\(-1\) + "i" + 
          "j" + "k"\)\),
      SequenceForm[ "T_", "C", {"i", "j", "k", "N"}, " = ", 
        Plus[ -1, "i", "j", "k"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Testing equationOrderQ and reorder equation "\)], "Print"],

Cell[BoxData[
    \("[firr]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Warning, the identity schedule is not valid (probably due to I/O)    "\
\)], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("known bug : equationOrderQ corrupted   "\)], "Print"],

Cell[BoxData[
    \("known bug : reorderEquations corrupted   "\)], "Print"],

Cell[BoxData[
    \("Testing rename Indices "\)], "Print"],

Cell[BoxData[
    \("[prodVect]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("[matvect1][matvect2][matvect3][oneStage][PIC]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Warning2 : "\[InvisibleSpace]"dhatNew"\
\[InvisibleSpace]" appears in the output list of a use :"\),
      SequenceForm[ 
      "Warning2 : ", "dhatNew", " appears in the output list of a use :"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("  Adding the auxilliary variable \
"\[InvisibleSpace]"oldXdhatNew"\),
      SequenceForm[ "  Adding the auxilliary variable ", "oldXdhatNew"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Warning, performing COB on I/O of system "\
\[InvisibleSpace]"oneStage"\[InvisibleSpace]": changing its semantic"\),
      SequenceForm[ 
      "Warning, performing COB on I/O of system ", "oneStage", 
        ": changing its semantic"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("oneStage"\[InvisibleSpace]" replaced in library."\),
      SequenceForm[ "oneStage", " replaced in library."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Warning, performing COB on I/O of system "\
\[InvisibleSpace]"oneStage"\[InvisibleSpace]": changing its semantic"\),
      SequenceForm[ 
      "Warning, performing COB on I/O of system ", "oneStage", 
        ": changing its semantic"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("oneStage"\[InvisibleSpace]" replaced in library."\),
      SequenceForm[ "oneStage", " replaced in library."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Warning, performing COB on I/O of system "\
\[InvisibleSpace]"oneStage"\[InvisibleSpace]": changing its semantic"\),
      SequenceForm[ 
      "Warning, performing COB on I/O of system ", "oneStage", 
        ": changing its semantic"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("oneStage"\[InvisibleSpace]" replaced in library."\),
      SequenceForm[ "oneStage", " replaced in library."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Warning, performing COB on I/O of system "\
\[InvisibleSpace]"oneStage"\[InvisibleSpace]": changing its semantic"\),
      SequenceForm[ 
      "Warning, performing COB on I/O of system ", "oneStage", 
        ": changing its semantic"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("oneStage"\[InvisibleSpace]" replaced in library."\),
      SequenceForm[ "oneStage", " replaced in library."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Warning, performing COB on I/O of system "\
\[InvisibleSpace]"oneStage"\[InvisibleSpace]": changing its semantic"\),
      SequenceForm[ 
      "Warning, performing COB on I/O of system ", "oneStage", 
        ": changing its semantic"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Warning, performing COB on I/O of system "\
\[InvisibleSpace]"matvect3"\[InvisibleSpace]": changing its semantic"\),
      SequenceForm[ 
      "Warning, performing COB on I/O of system ", "matvect3", 
        ": changing its semantic"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("matvect3"\[InvisibleSpace]" replaced in library."\),
      SequenceForm[ "matvect3", " replaced in library."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("oneStage"\[InvisibleSpace]" replaced in library."\),
      SequenceForm[ "oneStage", " replaced in library."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Warning, performing COB on I/O of system "\
\[InvisibleSpace]"oneStage"\[InvisibleSpace]": changing its semantic"\),
      SequenceForm[ 
      "Warning, performing COB on I/O of system ", "oneStage", 
        ": changing its semantic"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Warning, performing COB on I/O of system "\
\[InvisibleSpace]"matvect2"\[InvisibleSpace]": changing its semantic"\),
      SequenceForm[ 
      "Warning, performing COB on I/O of system ", "matvect2", 
        ": changing its semantic"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("matvect2"\[InvisibleSpace]" replaced in library."\),
      SequenceForm[ "matvect2", " replaced in library."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("oneStage"\[InvisibleSpace]" replaced in library."\),
      SequenceForm[ "oneStage", " replaced in library."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Warning, performing COB on I/O of system "\
\[InvisibleSpace]"oneStage"\[InvisibleSpace]": changing its semantic"\),
      SequenceForm[ 
      "Warning, performing COB on I/O of system ", "oneStage", 
        ": changing its semantic"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Warning, performing COB on I/O of system "\
\[InvisibleSpace]"matvect1"\[InvisibleSpace]": changing its semantic"\),
      SequenceForm[ 
      "Warning, performing COB on I/O of system ", "matvect1", 
        ": changing its semantic"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("matvect1"\[InvisibleSpace]" replaced in library."\),
      SequenceForm[ "matvect1", " replaced in library."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("oneStage"\[InvisibleSpace]" replaced in library."\),
      SequenceForm[ "oneStage", " replaced in library."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Warning, performing COB on I/O of system "\
\[InvisibleSpace]"oneStage"\[InvisibleSpace]": changing its semantic"\),
      SequenceForm[ 
      "Warning, performing COB on I/O of system ", "oneStage", 
        ": changing its semantic"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("oneStage"\[InvisibleSpace]" replaced in library."\),
      SequenceForm[ "oneStage", " replaced in library."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("*************************************************"\)], "Print"],

Cell[BoxData[
    \("Part5: Testing the structured scheduling "\)], "Print"],

Cell[BoxData[
    \("*************************************************\n"\)], "Print"],

Cell[BoxData[
    \("***Testing Structured scheduling"\)], "Print"],

Cell[BoxData[
    \("[MatVect][MatMat]"\)], "Print"],

Cell[BoxData[
    \(" Library Loaded"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]87\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]82\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 87, "      variables, ", 82, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00031666731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408091012, ByteCount \[Rule] 22486}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00031666731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408091012, ByteCount \[Rule] 738}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]\(1 + 
          "M"\)\),
      SequenceForm[ "Total execution Time: ", 
        Plus[ 1, "M"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"a"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "a", {"i", "j", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"b"\[InvisibleSpace]{"j", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "b", {"j", "N", "M"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"c"\[InvisibleSpace]{"i", "N", 
          "M"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + "M"\)\),
      SequenceForm[ "T_", "c", {"i", "N", "M"}, " = ", 
        Plus[ 1, "M"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "j", 
          "N", "M"}\[InvisibleSpace]" = "\[InvisibleSpace]"j"\),
      SequenceForm[ "T_", "C", {"i", "j", "N", "M"}, " = ", "j"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(14\[InvisibleSpace]" Recognized added Constraints: \
"\),
      SequenceForm[ 14, " Recognized added Constraints: "],
      Editable->False]], "Print"],

Cell[BoxData[
    \("   TAbisD1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   TAbisD2 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -CAbis + CVectIN = 0,"\)], "Print"],

Cell[BoxData[
    \("   TVectIND1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD4 + TVectIND3 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD5 + TVectIND4 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD6 + TVectIND5 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -1 - CAbis + CVectOUT = 0,"\)], "Print"],

Cell[BoxData[
    \("   TVectOUTD1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD4 + TVectOUTD3 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -1 - TAbisD5 + TVectOUTD4 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD6 + TVectOUTD5 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD3 + TVectIND2 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD3 + TVectOUTD2 = 0,"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]137\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]152\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 137, "      variables, ", 152, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00031966731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408091014, ByteCount \[Rule] 64909}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00031966731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408091014, ByteCount \[Rule] 1083}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Total execution Time: "\[InvisibleSpace]\(3 + 
          "M"\)\),
      SequenceForm[ "Total execution Time: ", 
        Plus[ 3, "M"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "A", {"i", "j", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "k", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]\(3 + "M"\)\),
      SequenceForm[ "T_", "C", {"i", "k", "N", "M", "P"}, " = ", 
        Plus[ 3, "M"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"Abis"\[InvisibleSpace]{"i", 
          "j", "k", "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]1\),
      SequenceForm[ "T_", "Abis", {"i", "j", "k", "N", "M", "P"}, " = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"VectIN"\[InvisibleSpace]{"i", 
          "k", "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]1\),
      SequenceForm[ "T_", "VectIN", {"i", "k", "N", "M", "P"}, " = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"VectOUT"\[InvisibleSpace]{"j", 
          "k", "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]\(2 + 
          "M"\)\),
      SequenceForm[ "T_", "VectOUT", {"j", "k", "N", "M", "P"}, " = ", 
        Plus[ 2, "M"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("OK"\)], "Print"],

Cell[BoxData[
    \("Checking options..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Starting multidimensionnal scheduling on program "\
\[InvisibleSpace]"MatMat"\),
      SequenceForm[ 
      "Starting multidimensionnal scheduling on program ", "MatMat"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Scheduling dimension number "\[InvisibleSpace]1\),
      SequenceForm[ "Scheduling dimension number ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("duration options:"\[InvisibleSpace]{}\),
      SequenceForm[ "duration options:", {}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dep to schedule: "\[InvisibleSpace]{1, 2, 3}\),
      SequenceForm[ "dep to schedule: ", {1, 2, 3}],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(14\[InvisibleSpace]" Recognized added Constraints: \
"\),
      SequenceForm[ 14, " Recognized added Constraints: "],
      Editable->False]], "Print"],

Cell[BoxData[
    \("   TAbisD1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   TAbisD2 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -CAbis + CVectIN = 0,"\)], "Print"],

Cell[BoxData[
    \("   TVectIND1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD3 + TVectIND2 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD4 + TVectIND3 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD5 + TVectIND4 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD6 + TVectIND5 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -CAbis + CVectOUT = 0,"\)], "Print"],

Cell[BoxData[
    \("   TVectOUTD1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD3 + TVectOUTD2 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD4 + TVectOUTD3 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD5 + TVectOUTD4 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD6 + TVectOUTD5 = 0,"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]137\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]148\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 137, "      variables, ", 148, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00032266731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408091017, ByteCount \[Rule] 61930}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00032266731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408091017, ByteCount \[Rule] 1086}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "A", {"i", "j", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "k", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]2\),
      SequenceForm[ "T_", "C", {"i", "k", "N", "M", "P"}, " = ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"Abis"\[InvisibleSpace]{"i", 
          "j", "k", "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]1\),
      SequenceForm[ "T_", "Abis", {"i", "j", "k", "N", "M", "P"}, " = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"VectIN"\[InvisibleSpace]{"i", 
          "k", "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]1\),
      SequenceForm[ "T_", "VectIN", {"i", "k", "N", "M", "P"}, " = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"VectOUT"\[InvisibleSpace]{"j", 
          "k", "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]1\),
      SequenceForm[ "T_", "VectOUT", {"j", "k", "N", "M", "P"}, " = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Scheduling dimension number "\[InvisibleSpace]2\),
      SequenceForm[ "Scheduling dimension number ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("duration options:"\[InvisibleSpace]{}\),
      SequenceForm[ "duration options:", {}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dep to schedule: "\[InvisibleSpace]{}\),
      SequenceForm[ "dep to schedule: ", {}],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Dependence analysis..."\)], "Print"],

Cell[BoxData[
    \("Building LP..."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\(12\[InvisibleSpace]" Recognized added Constraints: \
"\),
      SequenceForm[ 12, " Recognized added Constraints: "],
      Editable->False]], "Print"],

Cell[BoxData[
    \("   TAbisD1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   TAbisD2 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -CAbis + CVectIN = 0,"\)], "Print"],

Cell[BoxData[
    \("   TVectIND1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD4 + TVectIND3 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD5 + TVectIND4 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD6 + TVectIND5 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -1 - CAbis + CVectOUT = 0,"\)], "Print"],

Cell[BoxData[
    \("   TVectOUTD1 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD4 + TVectOUTD3 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -1 - TAbisD5 + TVectOUTD4 = 0,"\)], "Print"],

Cell[BoxData[
    \("   -TAbisD6 + TVectOUTD5 = 0,"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("LP: "\[InvisibleSpace]105\[InvisibleSpace]"      \
variables, "\[InvisibleSpace]100\[InvisibleSpace]" Constraints"\),
      SequenceForm[ "LP: ", 105, "      variables, ", 100, " Constraints"],
      Editable->False]], "Print"],

Cell[BoxData[
    \("Writing file for PIP...."\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("Writing line "\[InvisibleSpace]100\),
      SequenceForm[ "Writing line ", 100],
      Editable->False]], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00032566731.pip", FileType \[Rule] File, 
      Date \[Rule] 3408091018, ByteCount \[Rule] 32973}\)], "Print"],

Cell[BoxData[
    \("Solving the LP..."\)], "Print"],

Cell[BoxData[
    \({File \[Rule] "/tmp/m00032566731.pipres", FileType \[Rule] File, 
      Date \[Rule] 3408091018, ByteCount \[Rule] 874}\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "A", {"i", "j", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "k", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "C", {"i", "k", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"Abis"\[InvisibleSpace]{"i", 
          "j", "k", "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "Abis", {"i", "j", "k", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"VectIN"\[InvisibleSpace]{"i", 
          "k", "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "VectIN", {"i", "k", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"VectOUT"\[InvisibleSpace]{"j", 
          "k", "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]\(1 + 
          "M"\)\),
      SequenceForm[ "T_", "VectOUT", {"j", "k", "N", "M", "P"}, " = ", 
        Plus[ 1, "M"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Schedule Found has "\[InvisibleSpace]2\
\[InvisibleSpace]" dimension"\),
      SequenceForm[ "Schedule Found has ", 2, " dimension"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"A"\[InvisibleSpace]{"i", "j", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "A", {"i", "j", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("                     "\[InvisibleSpace]0\),
      SequenceForm[ "                     ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"B"\[InvisibleSpace]{"i", "j", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]0\),
      SequenceForm[ "T_", "B", {"i", "j", "N", "M", "P"}, " = ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("                     "\[InvisibleSpace]0\),
      SequenceForm[ "                     ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"C"\[InvisibleSpace]{"i", "k", 
          "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]2\),
      SequenceForm[ "T_", "C", {"i", "k", "N", "M", "P"}, " = ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("                     "\[InvisibleSpace]0\),
      SequenceForm[ "                     ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"Abis"\[InvisibleSpace]{"i", 
          "j", "k", "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]1\),
      SequenceForm[ "T_", "Abis", {"i", "j", "k", "N", "M", "P"}, " = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("                           "\[InvisibleSpace]0\),
      SequenceForm[ "                           ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"VectIN"\[InvisibleSpace]{"i", 
          "k", "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]1\),
      SequenceForm[ "T_", "VectIN", {"i", "k", "N", "M", "P"}, " = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("                          "\[InvisibleSpace]0\),
      SequenceForm[ "                          ", 0],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("T_"\[InvisibleSpace]"VectOUT"\[InvisibleSpace]{"j", 
          "k", "N", "M", "P"}\[InvisibleSpace]" = "\[InvisibleSpace]1\),
      SequenceForm[ "T_", "VectOUT", {"j", "k", "N", "M", "P"}, " = ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("                           "\[InvisibleSpace]\(1 + 
          "M"\)\),
      SequenceForm[ "                           ", 
        Plus[ 1, "M"]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("******************* Test Schedule Result summary ******"\)], "Print"],

Cell[BoxData[
    \(" *******   No Problem occured *****"\)], "Print"],

Cell[BoxData[
    \("*******************************************************"\)], "Print"],

Cell[BoxData[
    InterpretationBox[\("------------- > Test for \
"\[InvisibleSpace]"NewSchedule"\[InvisibleSpace]" was passed successfully."\),
      
      SequenceForm[ 
      "------------- > Test for ", "NewSchedule", 
        " was passed successfully."],
      Editable->False]], "Print"]
}, Closed]]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Macintosh",
ScreenRectangle->{{0, 1440}, {0, 834}},
AutoGeneratedPackage->None,
WindowToolbars->{"RulerBar", "EditBar"},
WindowSize->{1144, 759},
WindowMargins->{{0, Automatic}, {26, Automatic}},
ShowSelection->True,
Magnification->1.5
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "Info3407929681-2992127"->{
    Cell[14808, 486, 95, 1, 33, "Print",
      CellTags->"Info3407929681-2992127"],
    Cell[14906, 489, 28757, 466, 5680, "Print",
      CellTags->"Info3407929681-2992127"]},
  "Info3407938364-2610364"->{
    Cell[51969, 1201, 195, 3, 58, "Print",
      CellTags->"Info3407938364-2610364"]},
  "Info3407938399-2759567"->{
    Cell[52253, 1212, 326, 5, 82, "Print",
      CellTags->"Info3407938399-2759567"]},
  "Info3407930501-4698775"->{
    Cell[52680, 1225, 394, 6, 106, "Print",
      CellTags->"Info3407930501-4698775"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"Info3407929681-2992127", 319961, 8914},
  {"Info3407938364-2610364", 320167, 8919},
  {"Info3407938399-2759567", 320284, 8922},
  {"Info3407930501-4698775", 320401, 8925}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 69, 1, 132, "Title"],
Cell[1848, 56, 38, 0, 45, "Subsubtitle"],
Cell[1889, 58, 51, 0, 47, "Text"],
Cell[1943, 60, 99, 3, 62, "SmallText"],
Cell[2045, 65, 68, 1, 40, "SmallText"],
Cell[2116, 68, 334, 13, 282, "SmallText"],

Cell[CellGroupData[{
Cell[2475, 85, 95, 1, 41, "Subsubsection"],

Cell[CellGroupData[{
Cell[2595, 90, 115, 2, 39, "Input",
  InitializationCell->True],
Cell[2713, 94, 80, 1, 70, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[2842, 101, 66, 1, 57, "Section"],
Cell[2911, 104, 226, 4, 73, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[3174, 113, 89, 1, 104, "Section"],
Cell[3266, 116, 767, 28, 107, "Text"],

Cell[CellGroupData[{
Cell[4058, 148, 64, 1, 39, "Input"],
Cell[4125, 151, 49, 1, 34, "Print"],
Cell[4177, 154, 50, 1, 34, "Print"],
Cell[4230, 157, 359, 5, 322, "Print"]
}, Open  ]],
Cell[4604, 165, 214, 5, 73, "Text"],

Cell[CellGroupData[{
Cell[4843, 174, 42, 1, 39, "Input"],
Cell[4888, 177, 36, 1, 34, "Print"],
Cell[4927, 180, 203, 4, 34, "Print"],
Cell[5133, 186, 71, 1, 34, "Print"],
Cell[5207, 189, 69, 1, 34, "Print"],
Cell[5279, 192, 84, 1, 34, "Print"],
Cell[5366, 195, 76, 1, 34, "Print"],
Cell[5445, 198, 82, 1, 34, "Print"],
Cell[5530, 201, 91, 1, 34, "Print"],
Cell[5624, 204, 171, 3, 34, "Print"],
Cell[5798, 209, 171, 3, 34, "Print"],
Cell[5972, 214, 76, 1, 34, "Print"],
Cell[6051, 217, 69, 1, 34, "Print"],
Cell[6123, 220, 59, 1, 58, "Print"],
Cell[6185, 223, 38, 1, 40, "Output"]
}, Closed]],
Cell[6238, 227, 457, 14, 66, "Text"],
Cell[6698, 243, 390, 8, 73, "Text"],
Cell[7091, 253, 63, 1, 39, "Input"],
Cell[7157, 256, 31, 0, 47, "Text"],

Cell[CellGroupData[{
Cell[7213, 260, 165, 3, 62, "Input"],
Cell[7381, 265, 322, 6, 34, "Print"],
Cell[7706, 273, 113, 2, 34, "Print"],
Cell[7822, 277, 222, 4, 64, "Output"]
}, Open  ]],
Cell[8059, 284, 275, 6, 73, "Text"],

Cell[CellGroupData[{
Cell[8359, 294, 42, 1, 39, "Input"],
Cell[8404, 297, 222, 4, 64, "Output"]
}, Open  ]],
Cell[8641, 304, 68, 0, 47, "Text"],

Cell[CellGroupData[{
Cell[8734, 308, 50, 1, 39, "Input"],
Cell[8787, 311, 262, 5, 34, "Print"],
Cell[9052, 318, 273, 5, 34, "Print"],
Cell[9328, 325, 283, 5, 34, "Print"]
}, Open  ]],
Cell[9626, 333, 369, 6, 99, "Text"],

Cell[CellGroupData[{
Cell[10020, 343, 50, 1, 39, "Input"],
Cell[10073, 346, 1584, 26, 256, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[11694, 377, 81, 1, 39, "Input"],
Cell[11778, 380, 54, 1, 34, "Print"],
Cell[11835, 383, 57, 1, 34, "Print"],
Cell[11895, 386, 49, 1, 34, "Print"],
Cell[11947, 389, 192, 4, 34, "Print"],
Cell[12142, 395, 52, 1, 34, "Print"],
Cell[12197, 398, 262, 4, 34, "Print"],
Cell[12462, 404, 59, 1, 34, "Print"],
Cell[12524, 407, 150, 2, 34, "Print"],
Cell[12677, 411, 52, 1, 34, "Print"],
Cell[12732, 414, 152, 2, 34, "Print"],
Cell[12887, 418, 205, 5, 34, "Print"],
Cell[13095, 425, 268, 5, 34, "Print"],
Cell[13366, 432, 262, 5, 34, "Print"],
Cell[13631, 439, 296, 6, 34, "Print"],
Cell[13930, 447, 243, 5, 64, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[14222, 458, 75, 1, 104, "Section"],
Cell[14300, 461, 104, 3, 47, "Text"],

Cell[CellGroupData[{
Cell[14429, 468, 100, 4, 53, "Subsection"],
Cell[14532, 474, 191, 5, 47, "Text"],

Cell[CellGroupData[{
Cell[14748, 483, 57, 1, 39, "Input"],
Cell[14808, 486, 95, 1, 33, "Print",
  CellTags->"Info3407929681-2992127"],
Cell[14906, 489, 28757, 466, 5680, "Print",
  CellTags->"Info3407929681-2992127"]
}, Closed]],
Cell[43678, 958, 66, 0, 40, "Text"],

Cell[CellGroupData[{
Cell[43769, 962, 54, 1, 39, "Input"],
Cell[43826, 965, 790, 13, 136, "Output"]
}, Open  ]],
Cell[44631, 981, 467, 13, 73, "Text"],

Cell[CellGroupData[{
Cell[45123, 998, 55, 1, 39, "Input"],
Cell[45181, 1001, 734, 13, 136, "Output"]
}, Open  ]],
Cell[45930, 1017, 158, 4, 47, "Text"],

Cell[CellGroupData[{
Cell[46113, 1025, 61, 1, 39, "Input"],
Cell[46177, 1028, 507, 7, 112, "Output"]
}, Open  ]],
Cell[46699, 1038, 184, 4, 47, "Text"],
Cell[46886, 1044, 99, 3, 47, "Text"],
Cell[46988, 1049, 114, 3, 47, "Text"],
Cell[47105, 1054, 122, 3, 47, "Text"],
Cell[47230, 1059, 131, 3, 47, "Text"],
Cell[47364, 1064, 148, 3, 47, "Text"],
Cell[47515, 1069, 120, 3, 47, "Text"],
Cell[47638, 1074, 91, 3, 47, "Text"],
Cell[47732, 1079, 194, 5, 47, "Text"],

Cell[CellGroupData[{
Cell[47951, 1088, 78, 1, 39, "Input"],
Cell[48032, 1091, 553, 9, 112, "Output"]
}, Open  ]],
Cell[48600, 1103, 255, 6, 73, "Text"],

Cell[CellGroupData[{
Cell[48880, 1113, 76, 1, 39, "Input"],
Cell[48959, 1116, 790, 13, 136, "Output"]
}, Open  ]],
Cell[49764, 1132, 414, 9, 99, "Text"],

Cell[CellGroupData[{
Cell[50203, 1145, 58, 1, 32, "Input"],
Cell[50264, 1148, 224, 5, 34, "Print"],
Cell[50491, 1155, 149, 3, 34, "Print"],
Cell[50643, 1160, 153, 3, 34, "Print"],
Cell[50799, 1165, 790, 13, 136, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[51638, 1184, 105, 4, 39, "Subsection"],
Cell[51746, 1190, 150, 4, 47, "Text"],

Cell[CellGroupData[{
Cell[51921, 1198, 45, 1, 39, "Input"],
Cell[51969, 1201, 195, 3, 58, "Print",
  CellTags->"Info3407938364-2610364"]
}, Open  ]],

Cell[CellGroupData[{
Cell[52201, 1209, 49, 1, 39, "Input"],
Cell[52253, 1212, 326, 5, 82, "Print",
  CellTags->"Info3407938399-2759567"]
}, Open  ]],

Cell[CellGroupData[{
Cell[52616, 1222, 61, 1, 39, "Input"],
Cell[52680, 1225, 394, 6, 106, "Print",
  CellTags->"Info3407930501-4698775"]
}, Open  ]],

Cell[CellGroupData[{
Cell[53111, 1236, 58, 1, 39, "Input"],
Cell[53172, 1239, 734, 13, 136, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[53943, 1257, 65, 1, 39, "Input"],
Cell[54011, 1260, 307, 5, 88, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[54367, 1271, 108, 4, 39, "Subsection"],
Cell[54478, 1277, 117, 3, 47, "Text"],

Cell[CellGroupData[{
Cell[54620, 1284, 61, 1, 39, "Input"],
Cell[54684, 1287, 313, 5, 64, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[55058, 1299, 77, 1, 57, "Section"],

Cell[CellGroupData[{
Cell[55160, 1304, 69, 1, 39, "Input"],
Cell[55232, 1307, 50, 1, 34, "Print"],
Cell[55285, 1310, 50, 1, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[55372, 1316, 40, 1, 39, "Input"],
Cell[55415, 1319, 505, 7, 466, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[55957, 1331, 42, 1, 39, "Input"],
Cell[56002, 1334, 36, 1, 34, "Print"],
Cell[56041, 1337, 205, 4, 34, "Print"],
Cell[56249, 1343, 71, 1, 34, "Print"],
Cell[56323, 1346, 69, 1, 34, "Print"],
Cell[56395, 1349, 84, 1, 34, "Print"],
Cell[56482, 1352, 76, 1, 34, "Print"],
Cell[56561, 1355, 82, 1, 34, "Print"],
Cell[56646, 1358, 91, 1, 34, "Print"],
Cell[56740, 1361, 171, 3, 34, "Print"],
Cell[56914, 1366, 171, 3, 34, "Print"],
Cell[57088, 1371, 76, 1, 34, "Print"],
Cell[57167, 1374, 69, 1, 34, "Print"],
Cell[57239, 1377, 59, 1, 58, "Print"],
Cell[57301, 1380, 38, 1, 40, "Output"]
}, Closed]],
Cell[57354, 1384, 74, 0, 40, "Text"],
Cell[57431, 1386, 58, 1, 39, "Input"],
Cell[57492, 1389, 42, 1, 39, "Input"],

Cell[CellGroupData[{
Cell[57559, 1394, 50, 1, 39, "Input"],
Cell[57612, 1397, 4386, 71, 784, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[62035, 1473, 38, 1, 32, "Input"],
Cell[62076, 1476, 7618, 224, 514, "Print"],
Cell[69697, 1702, 897, 24, 58, "Print"],
Cell[70597, 1728, 232, 3, 58, "Print"],
Cell[70832, 1733, 311, 4, 30, "Message"],
Cell[71146, 1739, 322, 6, 34, "Print"],
Cell[71471, 1747, 114, 2, 34, "Print"],
Cell[71588, 1751, 618, 11, 184, "Output"]
}, Open  ]],
Cell[72221, 1765, 58, 1, 39, "Input"],

Cell[CellGroupData[{
Cell[72304, 1770, 208, 4, 62, "Input"],
Cell[72515, 1776, 62, 1, 34, "Print"],
Cell[72580, 1779, 57, 1, 34, "Print"],
Cell[72640, 1782, 440, 6, 178, "Print"],
Cell[73083, 1790, 63, 1, 34, "Print"],
Cell[73149, 1793, 253, 4, 34, "Print"],
Cell[73405, 1799, 359, 8, 34, "Print"],
Cell[73767, 1809, 245, 4, 34, "Print"],
Cell[74015, 1815, 811, 21, 58, "Print"],
Cell[74829, 1838, 245, 4, 34, "Print"],
Cell[75077, 1844, 1115, 30, 82, "Print"],
Cell[76195, 1876, 245, 4, 34, "Print"],
Cell[76443, 1882, 1188, 32, 82, "Print"],
Cell[77634, 1916, 245, 4, 34, "Print"],
Cell[77882, 1922, 1497, 40, 82, "Print"],
Cell[79382, 1964, 245, 4, 34, "Print"],
Cell[79630, 1970, 547, 13, 34, "Print"],
Cell[80180, 1985, 245, 4, 34, "Print"],
Cell[80428, 1991, 1794, 49, 106, "Print"],
Cell[82225, 2042, 245, 4, 34, "Print"],
Cell[82473, 2048, 1471, 40, 82, "Print"],
Cell[83947, 2090, 66, 1, 34, "Print"],
Cell[84016, 2093, 77, 1, 34, "Print"],
Cell[84096, 2096, 7356, 199, 418, "Print"],
Cell[91455, 2297, 4402, 127, 202, "Print"],
Cell[95860, 2426, 242, 5, 34, "Print"],
Cell[96105, 2433, 298, 6, 58, "Print"],
Cell[96406, 2441, 155, 3, 34, "Print"],
Cell[96564, 2446, 179, 3, 58, "Print"],
Cell[96746, 2451, 294, 6, 34, "Print"],
Cell[97043, 2459, 535, 9, 58, "Print"],
Cell[97581, 2470, 1457, 36, 106, "Print"],
Cell[99041, 2508, 322, 6, 34, "Print"],
Cell[99366, 2516, 551, 8, 82, "Print"],
Cell[99920, 2526, 212, 4, 34, "Print"],
Cell[100135, 2532, 165, 3, 34, "Print"],
Cell[100303, 2537, 165, 3, 34, "Print"],
Cell[100471, 2542, 327, 7, 58, "Print"],
Cell[100801, 2551, 632, 15, 58, "Print"],
Cell[101436, 2568, 41, 1, 34, "Print"],
Cell[101480, 2571, 179, 3, 58, "Print"],
Cell[101662, 2576, 171, 3, 34, "Print"],
Cell[101836, 2581, 2307, 41, 250, "Print"],
Cell[104146, 2624, 131, 3, 34, "Print"],
Cell[104280, 2629, 1869, 51, 106, "Print"],
Cell[106152, 2682, 419, 7, 58, "Print"],
Cell[106574, 2691, 639, 16, 58, "Print"],
Cell[107216, 2709, 323, 6, 34, "Print"],
Cell[107542, 2717, 419, 6, 82, "Print"],
Cell[107964, 2725, 165, 3, 34, "Print"],
Cell[108132, 2730, 135, 3, 34, "Print"],
Cell[108270, 2735, 163, 3, 34, "Print"],
Cell[108436, 2740, 193, 4, 58, "Print"],
Cell[108632, 2746, 161, 3, 34, "Print"],
Cell[108796, 2751, 35, 1, 34, "Print"],
Cell[108834, 2754, 173, 3, 34, "Print"],
Cell[109010, 2759, 636, 12, 160, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[109683, 2776, 50, 1, 32, "Input"],
Cell[109736, 2779, 39, 1, 34, "Print"],
Cell[109778, 2782, 272, 5, 34, "Print"],
Cell[110053, 2789, 307, 6, 34, "Print"],
Cell[110363, 2797, 303, 6, 34, "Print"],
Cell[110669, 2805, 297, 6, 34, "Print"],
Cell[110969, 2813, 39, 1, 34, "Print"],
Cell[111011, 2816, 241, 4, 34, "Print"],
Cell[111255, 2822, 255, 4, 34, "Print"],
Cell[111513, 2828, 241, 4, 34, "Print"],
Cell[111757, 2834, 241, 4, 34, "Print"],
Cell[112001, 2840, 46, 1, 40, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[112084, 2846, 122, 2, 32, "Input"],
Cell[112209, 2850, 179, 3, 58, "Print"],
Cell[112391, 2855, 321, 6, 34, "Print"],
Cell[112715, 2863, 179, 3, 58, "Print"],
Cell[112897, 2868, 324, 6, 34, "Print"],
Cell[113224, 2876, 173, 3, 34, "Print"],
Cell[113400, 2881, 631, 12, 160, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[114068, 2898, 50, 1, 39, "Input"],
Cell[114121, 2901, 39, 1, 34, "Print"],
Cell[114163, 2904, 297, 6, 34, "Print"],
Cell[114463, 2912, 307, 6, 34, "Print"],
Cell[114773, 2920, 272, 5, 34, "Print"],
Cell[115048, 2927, 297, 6, 34, "Print"],
Cell[115348, 2935, 39, 1, 34, "Print"],
Cell[115390, 2938, 241, 4, 34, "Print"],
Cell[115634, 2944, 255, 4, 34, "Print"],
Cell[115892, 2950, 241, 4, 34, "Print"],
Cell[116136, 2956, 241, 4, 34, "Print"],
Cell[116380, 2962, 46, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[116463, 2968, 69, 1, 39, "Input"],
Cell[116535, 2971, 50, 1, 34, "Print"],
Cell[116588, 2974, 50, 1, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[116675, 2980, 54, 1, 39, "Input"],
Cell[116732, 2983, 2278, 39, 376, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[119047, 3027, 58, 1, 32, "Input"],
Cell[119108, 3030, 1599, 27, 304, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[120744, 3062, 58, 1, 32, "Input"],
Cell[120805, 3065, 1675, 28, 304, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[122517, 3098, 50, 1, 32, "Input"],
Cell[122570, 3101, 1844, 31, 352, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[124451, 3137, 45, 1, 32, "Input"],
Cell[124499, 3140, 258, 5, 34, "Print"],
Cell[124760, 3147, 179, 3, 34, "Print"],
Cell[124942, 3152, 159, 3, 34, "Print"],
Cell[125104, 3157, 208, 4, 34, "Print"],
Cell[125315, 3163, 57, 1, 34, "Print"],
Cell[125375, 3166, 49, 1, 34, "Print"],
Cell[125427, 3169, 266, 4, 34, "Print"],
Cell[125696, 3175, 59, 1, 34, "Print"],
Cell[125758, 3178, 153, 3, 34, "Print"],
Cell[125914, 3183, 151, 2, 34, "Print"],
Cell[126068, 3187, 52, 1, 34, "Print"],
Cell[126123, 3190, 152, 2, 34, "Print"],
Cell[126278, 3194, 241, 4, 34, "Print"],
Cell[126522, 3200, 297, 6, 34, "Print"],
Cell[126822, 3208, 272, 5, 34, "Print"],
Cell[127097, 3215, 307, 6, 34, "Print"],
Cell[127407, 3223, 179, 3, 34, "Print"],
Cell[127589, 3228, 159, 3, 34, "Print"],
Cell[127751, 3233, 161, 3, 34, "Print"],
Cell[127915, 3238, 57, 1, 34, "Print"],
Cell[127975, 3241, 49, 1, 34, "Print"],
Cell[128027, 3244, 262, 4, 34, "Print"],
Cell[128292, 3250, 59, 1, 34, "Print"],
Cell[128354, 3253, 150, 2, 34, "Print"],
Cell[128507, 3257, 52, 1, 34, "Print"],
Cell[128562, 3260, 152, 2, 34, "Print"],
Cell[128717, 3264, 241, 4, 34, "Print"],
Cell[128961, 3270, 241, 4, 34, "Print"],
Cell[129205, 3276, 241, 4, 34, "Print"],
Cell[129449, 3282, 255, 4, 34, "Print"],
Cell[129707, 3288, 206, 4, 34, "Print"],
Cell[129916, 3294, 241, 4, 34, "Print"],
Cell[130160, 3300, 153, 3, 34, "Print"],
Cell[130316, 3305, 297, 6, 34, "Print"],
Cell[130616, 3313, 153, 3, 34, "Print"],
Cell[130772, 3318, 272, 5, 34, "Print"],
Cell[131047, 3325, 153, 3, 34, "Print"],
Cell[131203, 3330, 307, 6, 34, "Print"],
Cell[131513, 3338, 163, 3, 34, "Print"],
Cell[131679, 3343, 397, 6, 112, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[132113, 3354, 112, 2, 32, "Input"],
Cell[132228, 3358, 57, 1, 34, "Print"],
Cell[132288, 3361, 440, 6, 178, "Print"],
Cell[132731, 3369, 63, 1, 34, "Print"],
Cell[132797, 3372, 253, 4, 34, "Print"],
Cell[133053, 3378, 359, 8, 34, "Print"],
Cell[133415, 3388, 245, 4, 34, "Print"],
Cell[133663, 3394, 811, 21, 58, "Print"],
Cell[134477, 3417, 245, 4, 34, "Print"],
Cell[134725, 3423, 1115, 30, 82, "Print"],
Cell[135843, 3455, 245, 4, 34, "Print"],
Cell[136091, 3461, 1188, 32, 82, "Print"],
Cell[137282, 3495, 245, 4, 34, "Print"],
Cell[137530, 3501, 1497, 40, 82, "Print"],
Cell[139030, 3543, 245, 4, 34, "Print"],
Cell[139278, 3549, 547, 13, 34, "Print"],
Cell[139828, 3564, 245, 4, 34, "Print"],
Cell[140076, 3570, 1794, 49, 106, "Print"],
Cell[141873, 3621, 245, 4, 34, "Print"],
Cell[142121, 3627, 1471, 40, 82, "Print"],
Cell[143595, 3669, 3252, 51, 496, "Output"]
}, Closed]],
Cell[146862, 3723, 58, 1, 32, "Input"],

Cell[CellGroupData[{
Cell[146945, 3728, 204, 3, 62, "Input"],
Cell[147152, 3733, 179, 3, 58, "Print"],
Cell[147334, 3738, 323, 6, 34, "Print"],
Cell[147660, 3746, 173, 3, 34, "Print"],
Cell[147836, 3751, 234, 5, 64, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[148107, 3761, 50, 1, 39, "Input"],
Cell[148160, 3764, 39, 1, 34, "Print"],
Cell[148202, 3767, 262, 5, 34, "Print"],
Cell[148467, 3774, 235, 4, 34, "Print"],
Cell[148705, 3780, 287, 6, 34, "Print"],
Cell[148995, 3788, 40, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[149072, 3794, 45, 1, 39, "Input"],
Cell[149120, 3797, 398, 6, 178, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[149555, 3808, 50, 1, 39, "Input"],
Cell[149608, 3811, 39, 1, 34, "Print"],
Cell[149650, 3814, 262, 5, 34, "Print"],
Cell[149915, 3821, 235, 4, 34, "Print"],
Cell[150153, 3827, 287, 6, 34, "Print"],
Cell[150443, 3835, 40, 1, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[150520, 3841, 57, 1, 39, "Input"],
Cell[150580, 3844, 1237, 21, 208, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[151854, 3870, 37, 1, 39, "Input"],
Cell[151894, 3873, 964, 16, 160, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[152895, 3894, 60, 1, 39, "Input"],
Cell[152958, 3897, 258, 5, 34, "Print"],
Cell[153219, 3904, 179, 3, 34, "Print"],
Cell[153401, 3909, 159, 3, 34, "Print"],
Cell[153563, 3914, 208, 4, 34, "Print"],
Cell[153774, 3920, 57, 1, 34, "Print"],
Cell[153834, 3923, 49, 1, 34, "Print"],
Cell[153886, 3926, 266, 4, 34, "Print"],
Cell[154155, 3932, 59, 1, 34, "Print"],
Cell[154217, 3935, 153, 3, 34, "Print"],
Cell[154373, 3940, 151, 2, 34, "Print"],
Cell[154527, 3944, 52, 1, 34, "Print"],
Cell[154582, 3947, 152, 2, 34, "Print"],
Cell[154737, 3951, 241, 4, 34, "Print"],
Cell[154981, 3957, 297, 6, 34, "Print"],
Cell[155281, 3965, 272, 5, 34, "Print"],
Cell[155556, 3972, 307, 6, 34, "Print"],
Cell[155866, 3980, 179, 3, 34, "Print"],
Cell[156048, 3985, 159, 3, 34, "Print"],
Cell[156210, 3990, 161, 3, 34, "Print"],
Cell[156374, 3995, 57, 1, 34, "Print"],
Cell[156434, 3998, 49, 1, 34, "Print"],
Cell[156486, 4001, 262, 4, 34, "Print"],
Cell[156751, 4007, 59, 1, 34, "Print"],
Cell[156813, 4010, 150, 2, 34, "Print"],
Cell[156966, 4014, 52, 1, 34, "Print"],
Cell[157021, 4017, 152, 2, 34, "Print"],
Cell[157176, 4021, 241, 4, 34, "Print"],
Cell[157420, 4027, 241, 4, 34, "Print"],
Cell[157664, 4033, 241, 4, 34, "Print"],
Cell[157908, 4039, 255, 4, 34, "Print"],
Cell[158166, 4045, 206, 4, 34, "Print"],
Cell[158375, 4051, 241, 4, 34, "Print"],
Cell[158619, 4057, 153, 3, 34, "Print"],
Cell[158775, 4062, 297, 6, 34, "Print"],
Cell[159075, 4070, 153, 3, 34, "Print"],
Cell[159231, 4075, 272, 5, 34, "Print"],
Cell[159506, 4082, 153, 3, 34, "Print"],
Cell[159662, 4087, 307, 6, 34, "Print"],
Cell[159972, 4095, 163, 3, 34, "Print"],
Cell[160138, 4100, 397, 6, 112, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[160572, 4111, 49, 1, 39, "Input"],
Cell[160624, 4114, 964, 19, 106, "Print"],
Cell[161591, 4135, 115, 2, 30, "Message"]
}, Open  ]],
Cell[161721, 4140, 76, 0, 47, "Text"],

Cell[CellGroupData[{
Cell[161822, 4144, 176, 3, 62, "Input"],
Cell[162001, 4149, 258, 5, 34, "Print"],
Cell[162262, 4156, 179, 3, 34, "Print"],
Cell[162444, 4161, 159, 3, 34, "Print"],
Cell[162606, 4166, 208, 4, 34, "Print"],
Cell[162817, 4172, 57, 1, 34, "Print"],
Cell[162877, 4175, 49, 1, 34, "Print"],
Cell[162929, 4178, 266, 4, 34, "Print"],
Cell[163198, 4184, 59, 1, 34, "Print"],
Cell[163260, 4187, 153, 3, 34, "Print"],
Cell[163416, 4192, 151, 2, 34, "Print"],
Cell[163570, 4196, 52, 1, 34, "Print"],
Cell[163625, 4199, 152, 2, 34, "Print"],
Cell[163780, 4203, 241, 4, 34, "Print"],
Cell[164024, 4209, 297, 6, 34, "Print"],
Cell[164324, 4217, 272, 5, 34, "Print"],
Cell[164599, 4224, 307, 6, 34, "Print"],
Cell[164909, 4232, 179, 3, 34, "Print"],
Cell[165091, 4237, 159, 3, 34, "Print"],
Cell[165253, 4242, 161, 3, 34, "Print"],
Cell[165417, 4247, 57, 1, 34, "Print"],
Cell[165477, 4250, 49, 1, 34, "Print"],
Cell[165529, 4253, 262, 4, 34, "Print"],
Cell[165794, 4259, 59, 1, 34, "Print"],
Cell[165856, 4262, 150, 2, 34, "Print"],
Cell[166009, 4266, 52, 1, 34, "Print"],
Cell[166064, 4269, 152, 2, 34, "Print"],
Cell[166219, 4273, 241, 4, 34, "Print"],
Cell[166463, 4279, 241, 4, 34, "Print"],
Cell[166707, 4285, 241, 4, 34, "Print"],
Cell[166951, 4291, 255, 4, 34, "Print"],
Cell[167209, 4297, 206, 4, 34, "Print"],
Cell[167418, 4303, 241, 4, 34, "Print"],
Cell[167662, 4309, 153, 3, 34, "Print"],
Cell[167818, 4314, 297, 6, 34, "Print"],
Cell[168118, 4322, 153, 3, 34, "Print"],
Cell[168274, 4327, 272, 5, 34, "Print"],
Cell[168549, 4334, 153, 3, 34, "Print"],
Cell[168705, 4339, 307, 6, 34, "Print"],
Cell[169015, 4347, 163, 3, 34, "Print"],
Cell[169181, 4352, 397, 6, 112, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[169615, 4363, 59, 1, 39, "Input"],
Cell[169677, 4366, 241, 4, 34, "Print"],
Cell[169921, 4372, 153, 3, 34, "Print"],
Cell[170077, 4377, 297, 6, 34, "Print"],
Cell[170377, 4385, 153, 3, 34, "Print"],
Cell[170533, 4390, 272, 5, 34, "Print"],
Cell[170808, 4397, 153, 3, 34, "Print"],
Cell[170964, 4402, 307, 6, 34, "Print"],
Cell[171274, 4410, 163, 3, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[171474, 4418, 50, 1, 39, "Input"],
Cell[171527, 4421, 529, 8, 88, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[172093, 4434, 206, 4, 62, "Input"],
Cell[172302, 4440, 167, 3, 34, "Print"],
Cell[172472, 4445, 111, 2, 30, "Message"],
Cell[172586, 4449, 137, 2, 30, "Message"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[172772, 4457, 89, 1, 57, "Section"],
Cell[172864, 4460, 30, 0, 47, "Text"],

Cell[CellGroupData[{
Cell[172919, 4464, 59, 1, 39, "Input"],
Cell[172981, 4467, 45, 1, 34, "Print"],
Cell[173029, 4470, 50, 1, 34, "Print"]
}, Open  ]],
Cell[173094, 4474, 33, 0, 47, "Text"],

Cell[CellGroupData[{
Cell[173152, 4478, 38, 1, 39, "Input"],
Cell[173193, 4481, 37, 1, 34, "Print"],
Cell[173233, 4484, 203, 3, 58, "Print"],
Cell[173439, 4489, 321, 6, 34, "Print"],
Cell[173763, 4497, 108, 2, 34, "Print"],
Cell[173874, 4501, 285, 6, 64, "Output"]
}, Open  ]],
Cell[174174, 4510, 31, 0, 47, "Text"],

Cell[CellGroupData[{
Cell[174230, 4514, 50, 1, 39, "Input"],
Cell[174283, 4517, 245, 4, 34, "Print"],
Cell[174531, 4523, 235, 4, 34, "Print"],
Cell[174769, 4529, 262, 5, 34, "Print"],
Cell[175034, 4536, 272, 5, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[175343, 4546, 84, 1, 39, "Input"],
Cell[175430, 4549, 38, 1, 34, "Print"],
Cell[175471, 4552, 35, 1, 34, "Print"],
Cell[175509, 4555, 322, 6, 34, "Print"],
Cell[175834, 4563, 109, 2, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[175980, 4570, 50, 1, 39, "Input"],
Cell[176033, 4573, 241, 4, 34, "Print"],
Cell[176277, 4579, 231, 4, 34, "Print"],
Cell[176511, 4585, 262, 5, 34, "Print"],
Cell[176776, 4592, 245, 4, 34, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[177058, 4601, 43, 1, 39, "Input"],
Cell[177104, 4604, 54, 1, 34, "Print"],
Cell[177161, 4607, 57, 1, 34, "Print"],
Cell[177221, 4610, 49, 1, 34, "Print"],
Cell[177273, 4613, 262, 4, 34, "Print"],
Cell[177538, 4619, 59, 1, 34, "Print"],
Cell[177600, 4622, 151, 2, 34, "Print"],
Cell[177754, 4626, 52, 1, 34, "Print"],
Cell[177809, 4629, 152, 2, 34, "Print"],
Cell[177964, 4633, 209, 5, 34, "Print"],
Cell[178176, 4640, 241, 4, 34, "Print"],
Cell[178420, 4646, 231, 4, 34, "Print"],
Cell[178654, 4652, 262, 5, 34, "Print"],
Cell[178919, 4659, 245, 4, 34, "Print"],
Cell[179167, 4665, 285, 6, 64, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[179489, 4676, 53, 0, 53, "Subsection"],

Cell[CellGroupData[{
Cell[179567, 4680, 59, 1, 39, "Input"],
Cell[179629, 4683, 45, 1, 34, "Print"],
Cell[179677, 4686, 50, 1, 34, "Print"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[179788, 4694, 78, 0, 57, "Section"],
Cell[179869, 4696, 200, 4, 73, "Text"],

Cell[CellGroupData[{
Cell[180094, 4704, 68, 1, 39, "Input"],
Cell[180165, 4707, 1909, 32, 280, "Output"]
}, Open  ]],
Cell[182089, 4742, 156, 3, 47, "Text"],

Cell[CellGroupData[{
Cell[182270, 4749, 68, 1, 39, "Input"],
Cell[182341, 4752, 224, 5, 34, "Print"],
Cell[182568, 4759, 149, 3, 34, "Print"],
Cell[182720, 4764, 153, 3, 34, "Print"],
Cell[182876, 4769, 2175, 38, 400, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[185100, 4813, 94, 1, 57, "Section"],

Cell[CellGroupData[{
Cell[185219, 4818, 38, 1, 39, "Input"],
Cell[185260, 4821, 8236, 242, 538, "Print"],
Cell[193499, 5065, 793, 21, 58, "Print"],
Cell[194295, 5088, 203, 3, 58, "Print"],
Cell[194501, 5093, 322, 6, 34, "Print"],
Cell[194826, 5101, 109, 2, 34, "Print"],
Cell[194938, 5105, 285, 6, 64, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[195260, 5116, 50, 1, 32, "Input"],
Cell[195313, 5119, 245, 4, 34, "Print"],
Cell[195561, 5125, 235, 4, 34, "Print"],
Cell[195799, 5131, 262, 5, 34, "Print"],
Cell[196064, 5138, 272, 5, 34, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[196373, 5148, 54, 1, 32, "Input"],
Cell[196430, 5151, 2175, 38, 400, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[198642, 5194, 61, 1, 32, "Input"],
Cell[198706, 5197, 1507, 27, 232, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[200250, 5229, 50, 1, 32, "Input"],
Cell[200303, 5232, 245, 4, 34, "Print"],
Cell[200551, 5238, 235, 4, 34, "Print"],
Cell[200789, 5244, 262, 5, 34, "Print"],
Cell[201054, 5251, 272, 5, 34, "Print"]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{
Cell[201375, 5262, 89, 1, 104, "Section"],

Cell[CellGroupData[{
Cell[201489, 5267, 53, 0, 53, "Subsection"],

Cell[CellGroupData[{
Cell[201567, 5271, 66, 1, 39, "Input"],
Cell[201636, 5274, 653, 9, 1090, "Print"],
Cell[202292, 5285, 50, 1, 34, "Print"],
Cell[202345, 5288, 735, 10, 586, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[203117, 5303, 66, 1, 32, "Input"],
Cell[203186, 5306, 653, 9, 1090, "Print"],
Cell[203842, 5317, 50, 1, 34, "Print"],
Cell[203895, 5320, 735, 10, 586, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[204667, 5335, 42, 1, 32, "Input"],
Cell[204712, 5338, 36, 1, 34, "Print"],
Cell[204751, 5341, 191, 4, 34, "Print"],
Cell[204945, 5347, 71, 1, 34, "Print"],
Cell[205019, 5350, 69, 1, 34, "Print"],
Cell[205091, 5353, 84, 1, 34, "Print"],
Cell[205178, 5356, 192, 4, 34, "Print"],
Cell[205373, 5362, 87, 1, 58, "Print"],
Cell[205463, 5365, 76, 1, 34, "Print"],
Cell[205542, 5368, 82, 1, 34, "Print"],
Cell[205627, 5371, 60, 1, 58, "Print"],
Cell[205690, 5374, 39, 1, 40, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[205790, 5382, 78, 0, 57, "Section"],

Cell[CellGroupData[{
Cell[205893, 5386, 38, 1, 39, "Input"],
Cell[205934, 5389, 12201, 348, 850, "Print"],
Cell[218138, 5739, 1076, 29, 82, "Print"],
Cell[219217, 5770, 278, 4, 58, "Print"],
Cell[219498, 5776, 320, 6, 34, "Print"],
Cell[219821, 5784, 105, 2, 34, "Print"],
Cell[219929, 5788, 376, 7, 88, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[220342, 5800, 45, 1, 39, "Input"],
Cell[220390, 5803, 1079, 17, 256, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[221506, 5825, 115, 2, 62, "Input"],
Cell[221624, 5829, 54, 1, 34, "Print"],
Cell[221681, 5832, 57, 1, 34, "Print"],
Cell[221741, 5835, 49, 1, 34, "Print"],
Cell[221793, 5838, 192, 4, 34, "Print"],
Cell[221988, 5844, 48, 1, 34, "Print"],
Cell[222039, 5847, 266, 4, 34, "Print"],
Cell[222308, 5853, 59, 1, 34, "Print"],
Cell[222370, 5856, 153, 3, 34, "Print"],
Cell[222526, 5861, 151, 2, 34, "Print"],
Cell[222680, 5865, 52, 1, 34, "Print"],
Cell[222735, 5868, 152, 2, 34, "Print"],
Cell[222890, 5872, 167, 3, 34, "Print"],
Cell[223060, 5877, 231, 4, 34, "Print"],
Cell[223294, 5883, 231, 4, 34, "Print"],
Cell[223528, 5889, 231, 4, 34, "Print"],
Cell[223762, 5895, 305, 6, 34, "Print"],
Cell[224070, 5903, 287, 5, 34, "Print"],
Cell[224360, 5910, 287, 5, 34, "Print"],
Cell[224650, 5917, 373, 7, 88, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[225060, 5929, 50, 1, 32, "Input"],
Cell[225113, 5932, 231, 4, 34, "Print"],
Cell[225347, 5938, 231, 4, 34, "Print"],
Cell[225581, 5944, 231, 4, 34, "Print"],
Cell[225815, 5950, 305, 6, 34, "Print"],
Cell[226123, 5958, 287, 5, 34, "Print"],
Cell[226413, 5965, 287, 5, 34, "Print"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[226749, 5976, 35, 0, 57, "Section"],

Cell[CellGroupData[{
Cell[226809, 5980, 54, 1, 39, "Input"],
Cell[226866, 5983, 64, 1, 34, "Print"],
Cell[226933, 5986, 169, 3, 34, "Print"],
Cell[227105, 5991, 107, 2, 34, "Print"],
Cell[227215, 5995, 64, 1, 34, "Print"],
Cell[227282, 5998, 175, 3, 34, "Print"],
Cell[227460, 6003, 71, 1, 58, "Print"],
Cell[227534, 6006, 84, 1, 34, "Print"],
Cell[227621, 6009, 100, 2, 34, "Print"],
Cell[227724, 6013, 86, 1, 58, "Print"],
Cell[227813, 6016, 65, 1, 34, "Print"],
Cell[227881, 6019, 42, 1, 34, "Print"],
Cell[227926, 6022, 50, 1, 34, "Print"],
Cell[227979, 6025, 54, 1, 34, "Print"],
Cell[228036, 6028, 57, 1, 34, "Print"],
Cell[228096, 6031, 49, 1, 34, "Print"],
Cell[228148, 6034, 262, 4, 34, "Print"],
Cell[228413, 6040, 59, 1, 34, "Print"],
Cell[228475, 6043, 150, 2, 34, "Print"],
Cell[228628, 6047, 52, 1, 34, "Print"],
Cell[228683, 6050, 152, 2, 34, "Print"],
Cell[228838, 6054, 167, 3, 34, "Print"],
Cell[229008, 6059, 261, 4, 34, "Print"],
Cell[229272, 6065, 37, 1, 34, "Print"],
Cell[229312, 6068, 71, 1, 34, "Print"],
Cell[229386, 6071, 44, 1, 34, "Print"],
Cell[229433, 6074, 50, 1, 34, "Print"],
Cell[229486, 6077, 54, 1, 34, "Print"],
Cell[229543, 6080, 57, 1, 34, "Print"],
Cell[229603, 6083, 49, 1, 34, "Print"],
Cell[229655, 6086, 262, 4, 34, "Print"],
Cell[229920, 6092, 59, 1, 34, "Print"],
Cell[229982, 6095, 151, 2, 34, "Print"],
Cell[230136, 6099, 52, 1, 34, "Print"],
Cell[230191, 6102, 152, 2, 34, "Print"],
Cell[230346, 6106, 167, 3, 34, "Print"],
Cell[230516, 6111, 231, 4, 34, "Print"],
Cell[230750, 6117, 212, 4, 34, "Print"],
Cell[230965, 6123, 212, 4, 34, "Print"],
Cell[231180, 6129, 231, 4, 34, "Print"],
Cell[231414, 6135, 262, 5, 34, "Print"],
Cell[231679, 6142, 37, 1, 34, "Print"],
Cell[231719, 6145, 79, 1, 34, "Print"],
Cell[231801, 6148, 49, 1, 34, "Print"],
Cell[231853, 6151, 50, 1, 34, "Print"],
Cell[231906, 6154, 54, 1, 34, "Print"],
Cell[231963, 6157, 57, 1, 34, "Print"],
Cell[232023, 6160, 49, 1, 34, "Print"],
Cell[232075, 6163, 262, 4, 34, "Print"],
Cell[232340, 6169, 59, 1, 34, "Print"],
Cell[232402, 6172, 150, 2, 34, "Print"],
Cell[232555, 6176, 52, 1, 34, "Print"],
Cell[232610, 6179, 152, 2, 34, "Print"],
Cell[232765, 6183, 171, 3, 34, "Print"],
Cell[232939, 6188, 235, 4, 34, "Print"],
Cell[233177, 6194, 37, 1, 34, "Print"],
Cell[233217, 6197, 65, 1, 34, "Print"],
Cell[233285, 6200, 42, 1, 34, "Print"],
Cell[233330, 6203, 50, 1, 34, "Print"],
Cell[233383, 6206, 54, 1, 34, "Print"],
Cell[233440, 6209, 57, 1, 34, "Print"],
Cell[233500, 6212, 49, 1, 34, "Print"],
Cell[233552, 6215, 262, 4, 34, "Print"],
Cell[233817, 6221, 59, 1, 34, "Print"],
Cell[233879, 6224, 151, 2, 34, "Print"],
Cell[234033, 6228, 52, 1, 34, "Print"],
Cell[234088, 6231, 152, 2, 34, "Print"],
Cell[234243, 6235, 209, 5, 34, "Print"],
Cell[234455, 6242, 231, 4, 34, "Print"],
Cell[234689, 6248, 262, 5, 34, "Print"],
Cell[234954, 6255, 264, 5, 34, "Print"],
Cell[235221, 6262, 264, 5, 34, "Print"],
Cell[235488, 6269, 37, 1, 34, "Print"],
Cell[235528, 6272, 84, 1, 34, "Print"],
Cell[235615, 6275, 82, 1, 34, "Print"],
Cell[235700, 6278, 84, 1, 34, "Print"],
Cell[235787, 6281, 59, 1, 34, "Print"],
Cell[235849, 6284, 44, 1, 34, "Print"],
Cell[235896, 6287, 50, 1, 34, "Print"],
Cell[235949, 6290, 54, 1, 34, "Print"],
Cell[236006, 6293, 57, 1, 34, "Print"],
Cell[236066, 6296, 49, 1, 34, "Print"],
Cell[236118, 6299, 264, 4, 34, "Print"],
Cell[236385, 6305, 59, 1, 34, "Print"],
Cell[236447, 6308, 151, 2, 34, "Print"],
Cell[236601, 6312, 52, 1, 34, "Print"],
Cell[236656, 6315, 152, 2, 34, "Print"],
Cell[236811, 6319, 171, 3, 34, "Print"],
Cell[236985, 6324, 241, 4, 34, "Print"],
Cell[237229, 6330, 231, 4, 34, "Print"],
Cell[237463, 6336, 235, 4, 34, "Print"],
Cell[237701, 6342, 245, 4, 34, "Print"],
Cell[237949, 6348, 278, 5, 34, "Print"],
Cell[238230, 6355, 278, 5, 34, "Print"],
Cell[238511, 6362, 37, 1, 34, "Print"],
Cell[238551, 6365, 62, 1, 34, "Print"],
Cell[238616, 6368, 42, 1, 34, "Print"],
Cell[238661, 6371, 50, 1, 34, "Print"],
Cell[238714, 6374, 54, 1, 34, "Print"],
Cell[238771, 6377, 57, 1, 34, "Print"],
Cell[238831, 6380, 49, 1, 34, "Print"],
Cell[238883, 6383, 262, 4, 34, "Print"],
Cell[239148, 6389, 59, 1, 34, "Print"],
Cell[239210, 6392, 150, 2, 34, "Print"],
Cell[239363, 6396, 52, 1, 34, "Print"],
Cell[239418, 6399, 152, 2, 34, "Print"],
Cell[239573, 6403, 167, 3, 34, "Print"],
Cell[239743, 6408, 212, 4, 34, "Print"],
Cell[239958, 6414, 212, 4, 34, "Print"],
Cell[240173, 6420, 212, 4, 34, "Print"],
Cell[240388, 6426, 262, 5, 34, "Print"],
Cell[240653, 6433, 37, 1, 34, "Print"],
Cell[240693, 6436, 65, 1, 34, "Print"],
Cell[240761, 6439, 42, 1, 34, "Print"],
Cell[240806, 6442, 50, 1, 34, "Print"],
Cell[240859, 6445, 54, 1, 34, "Print"],
Cell[240916, 6448, 57, 1, 34, "Print"],
Cell[240976, 6451, 49, 1, 34, "Print"],
Cell[241028, 6454, 264, 4, 34, "Print"],
Cell[241295, 6460, 59, 1, 34, "Print"],
Cell[241357, 6463, 153, 3, 34, "Print"],
Cell[241513, 6468, 151, 2, 34, "Print"],
Cell[241667, 6472, 52, 1, 34, "Print"],
Cell[241722, 6475, 152, 2, 34, "Print"],
Cell[241877, 6479, 146, 2, 30, "Message"],
Cell[242026, 6483, 37, 1, 34, "Print"],
Cell[242066, 6486, 65, 1, 34, "Print"],
Cell[242134, 6489, 42, 1, 34, "Print"],
Cell[242179, 6492, 50, 1, 34, "Print"],
Cell[242232, 6495, 54, 1, 34, "Print"],
Cell[242289, 6498, 57, 1, 34, "Print"],
Cell[242349, 6501, 49, 1, 34, "Print"],
Cell[242401, 6504, 262, 4, 34, "Print"],
Cell[242666, 6510, 59, 1, 34, "Print"],
Cell[242728, 6513, 150, 2, 34, "Print"],
Cell[242881, 6517, 52, 1, 34, "Print"],
Cell[242936, 6520, 152, 2, 34, "Print"],
Cell[243091, 6524, 146, 2, 30, "Message"],
Cell[243240, 6528, 37, 1, 34, "Print"],
Cell[243280, 6531, 83, 1, 34, "Print"],
Cell[243366, 6534, 54, 1, 34, "Print"],
Cell[243423, 6537, 57, 1, 34, "Print"],
Cell[243483, 6540, 49, 1, 34, "Print"],
Cell[243535, 6543, 262, 4, 34, "Print"],
Cell[243800, 6549, 59, 1, 34, "Print"],
Cell[243862, 6552, 150, 2, 34, "Print"],
Cell[244015, 6556, 52, 1, 34, "Print"],
Cell[244070, 6559, 152, 2, 34, "Print"],
Cell[244225, 6563, 212, 4, 34, "Print"],
Cell[244440, 6569, 212, 4, 34, "Print"],
Cell[244655, 6575, 243, 5, 34, "Print"],
Cell[244901, 6582, 243, 5, 34, "Print"],
Cell[245147, 6589, 37, 1, 34, "Print"],
Cell[245187, 6592, 84, 1, 34, "Print"],
Cell[245274, 6595, 62, 1, 34, "Print"],
Cell[245339, 6598, 88, 1, 82, "Print"],
Cell[245430, 6601, 50, 1, 34, "Print"],
Cell[245483, 6604, 42, 1, 34, "Print"],
Cell[245528, 6607, 50, 1, 34, "Print"],
Cell[245581, 6610, 54, 1, 34, "Print"],
Cell[245638, 6613, 167, 3, 34, "Print"],
Cell[245808, 6618, 56, 1, 34, "Print"],
Cell[245867, 6621, 43, 1, 34, "Print"],
Cell[245913, 6624, 57, 1, 34, "Print"],
Cell[245973, 6627, 73, 1, 34, "Print"],
Cell[246049, 6630, 45, 1, 34, "Print"],
Cell[246097, 6633, 66, 1, 34, "Print"],
Cell[246166, 6636, 49, 1, 34, "Print"],
Cell[246218, 6639, 262, 4, 34, "Print"],
Cell[246483, 6645, 65, 1, 34, "Print"],
Cell[246551, 6648, 66, 1, 34, "Print"],
Cell[246620, 6651, 57, 1, 34, "Print"],
Cell[246680, 6654, 57, 1, 34, "Print"],
Cell[246740, 6657, 57, 1, 34, "Print"],
Cell[246800, 6660, 90, 1, 34, "Print"],
Cell[246893, 6663, 67, 1, 34, "Print"],
Cell[246963, 6666, 68, 1, 34, "Print"],
Cell[247034, 6669, 64, 1, 34, "Print"],
Cell[247101, 6672, 64, 1, 34, "Print"],
Cell[247168, 6675, 64, 1, 34, "Print"],
Cell[247235, 6678, 105, 2, 34, "Print"],
Cell[247343, 6682, 59, 1, 34, "Print"],
Cell[247405, 6685, 193, 4, 34, "Print"],
Cell[247601, 6691, 192, 4, 34, "Print"],
Cell[247796, 6697, 175, 3, 34, "Print"],
Cell[247974, 6702, 313, 7, 34, "Print"],
Cell[248290, 6711, 309, 6, 34, "Print"],
Cell[248602, 6719, 238, 4, 34, "Print"],
Cell[248843, 6725, 150, 2, 34, "Print"],
Cell[248996, 6729, 52, 1, 34, "Print"],
Cell[249051, 6732, 152, 2, 34, "Print"],
Cell[249206, 6736, 147, 3, 34, "Print"],
Cell[249356, 6741, 122, 2, 34, "Print"],
Cell[249481, 6745, 167, 3, 34, "Print"],
Cell[249651, 6750, 261, 4, 34, "Print"],
Cell[249915, 6756, 37, 1, 34, "Print"],
Cell[249955, 6759, 75, 1, 58, "Print"],
Cell[250033, 6762, 262, 4, 34, "Print"],
Cell[250298, 6768, 150, 2, 34, "Print"],
Cell[250451, 6772, 152, 2, 34, "Print"],
Cell[250606, 6776, 37, 1, 34, "Print"],
Cell[250646, 6779, 65, 1, 58, "Print"],
Cell[250714, 6782, 54, 1, 34, "Print"],
Cell[250771, 6785, 57, 1, 34, "Print"],
Cell[250831, 6788, 49, 1, 34, "Print"],
Cell[250883, 6791, 262, 4, 34, "Print"],
Cell[251148, 6797, 59, 1, 34, "Print"],
Cell[251210, 6800, 150, 2, 34, "Print"],
Cell[251363, 6804, 52, 1, 34, "Print"],
Cell[251418, 6807, 152, 2, 34, "Print"],
Cell[251573, 6811, 167, 3, 34, "Print"],
Cell[251743, 6816, 261, 4, 34, "Print"],
Cell[252007, 6822, 37, 1, 34, "Print"],
Cell[252047, 6825, 59, 1, 58, "Print"],
Cell[252109, 6828, 49, 1, 34, "Print"],
Cell[252161, 6831, 50, 1, 34, "Print"],
Cell[252214, 6834, 53, 1, 34, "Print"],
Cell[252270, 6837, 54, 1, 34, "Print"],
Cell[252327, 6840, 57, 1, 34, "Print"],
Cell[252387, 6843, 49, 1, 34, "Print"],
Cell[252439, 6846, 262, 4, 34, "Print"],
Cell[252704, 6852, 59, 1, 34, "Print"],
Cell[252766, 6855, 150, 2, 34, "Print"],
Cell[252919, 6859, 52, 1, 34, "Print"],
Cell[252974, 6862, 152, 2, 34, "Print"],
Cell[253129, 6866, 171, 3, 34, "Print"],
Cell[253303, 6871, 262, 5, 34, "Print"],
Cell[253568, 6878, 237, 4, 34, "Print"],
Cell[253808, 6884, 235, 4, 34, "Print"],
Cell[254046, 6890, 37, 1, 34, "Print"],
Cell[254086, 6893, 63, 1, 34, "Print"],
Cell[254152, 6896, 54, 1, 34, "Print"],
Cell[254209, 6899, 57, 1, 34, "Print"],
Cell[254269, 6902, 49, 1, 34, "Print"],
Cell[254321, 6905, 262, 4, 34, "Print"],
Cell[254586, 6911, 59, 1, 34, "Print"],
Cell[254648, 6914, 150, 2, 34, "Print"],
Cell[254801, 6918, 52, 1, 34, "Print"],
Cell[254856, 6921, 152, 2, 34, "Print"],
Cell[255011, 6925, 171, 3, 34, "Print"],
Cell[255185, 6930, 262, 5, 34, "Print"],
Cell[255450, 6937, 237, 4, 34, "Print"],
Cell[255690, 6943, 239, 4, 34, "Print"],
Cell[255932, 6949, 37, 1, 34, "Print"],
Cell[255972, 6952, 85, 1, 34, "Print"],
Cell[256060, 6955, 49, 1, 34, "Print"],
Cell[256112, 6958, 50, 1, 34, "Print"],
Cell[256165, 6961, 54, 1, 34, "Print"],
Cell[256222, 6964, 57, 1, 34, "Print"],
Cell[256282, 6967, 49, 1, 34, "Print"],
Cell[256334, 6970, 262, 4, 34, "Print"],
Cell[256599, 6976, 59, 1, 34, "Print"],
Cell[256661, 6979, 150, 2, 34, "Print"],
Cell[256814, 6983, 52, 1, 34, "Print"],
Cell[256869, 6986, 152, 2, 34, "Print"],
Cell[257024, 6990, 171, 3, 34, "Print"],
Cell[257198, 6995, 262, 5, 34, "Print"],
Cell[257463, 7002, 237, 4, 34, "Print"],
Cell[257703, 7008, 239, 4, 34, "Print"],
Cell[257945, 7014, 272, 5, 34, "Print"],
Cell[258220, 7021, 37, 1, 34, "Print"],
Cell[258260, 7024, 63, 1, 58, "Print"],
Cell[258326, 7027, 42, 1, 34, "Print"],
Cell[258371, 7030, 50, 1, 34, "Print"],
Cell[258424, 7033, 75, 1, 34, "Print"],
Cell[258502, 7036, 54, 1, 34, "Print"],
Cell[258559, 7039, 57, 1, 34, "Print"],
Cell[258619, 7042, 49, 1, 34, "Print"],
Cell[258671, 7045, 262, 4, 34, "Print"],
Cell[258936, 7051, 59, 1, 34, "Print"],
Cell[258998, 7054, 150, 2, 34, "Print"],
Cell[259151, 7058, 52, 1, 34, "Print"],
Cell[259206, 7061, 152, 2, 34, "Print"],
Cell[259361, 7065, 212, 4, 34, "Print"],
Cell[259576, 7071, 212, 4, 34, "Print"],
Cell[259791, 7077, 243, 5, 34, "Print"],
Cell[260037, 7084, 243, 5, 34, "Print"],
Cell[260283, 7091, 37, 1, 34, "Print"],
Cell[260323, 7094, 77, 1, 34, "Print"],
Cell[260403, 7097, 54, 1, 34, "Print"],
Cell[260460, 7100, 57, 1, 34, "Print"],
Cell[260520, 7103, 49, 1, 34, "Print"],
Cell[260572, 7106, 262, 4, 34, "Print"],
Cell[260837, 7112, 59, 1, 34, "Print"],
Cell[260899, 7115, 150, 2, 34, "Print"],
Cell[261052, 7119, 52, 1, 34, "Print"],
Cell[261107, 7122, 152, 2, 34, "Print"],
Cell[261262, 7126, 146, 2, 30, "Message"],
Cell[261411, 7130, 376, 5, 49, "Message"],
Cell[261790, 7137, 37, 1, 34, "Print"],
Cell[261830, 7140, 76, 1, 58, "Print"],
Cell[261909, 7143, 42, 1, 34, "Print"],
Cell[261954, 7146, 50, 1, 34, "Print"],
Cell[262007, 7149, 53, 1, 34, "Print"],
Cell[262063, 7152, 54, 1, 34, "Print"],
Cell[262120, 7155, 57, 1, 34, "Print"],
Cell[262180, 7158, 49, 1, 34, "Print"],
Cell[262232, 7161, 262, 4, 34, "Print"],
Cell[262497, 7167, 59, 1, 34, "Print"],
Cell[262559, 7170, 150, 2, 34, "Print"],
Cell[262712, 7174, 52, 1, 34, "Print"],
Cell[262767, 7177, 152, 2, 34, "Print"],
Cell[262922, 7181, 167, 3, 34, "Print"],
Cell[263092, 7186, 285, 6, 52, "Print"],
Cell[263380, 7194, 344, 8, 52, "Print"],
Cell[263727, 7204, 37, 1, 34, "Print"],
Cell[263767, 7207, 50, 1, 34, "Print"],
Cell[263820, 7210, 54, 1, 34, "Print"],
Cell[263877, 7213, 57, 1, 34, "Print"],
Cell[263937, 7216, 49, 1, 34, "Print"],
Cell[263989, 7219, 262, 4, 34, "Print"],
Cell[264254, 7225, 59, 1, 34, "Print"],
Cell[264316, 7228, 150, 2, 34, "Print"],
Cell[264469, 7232, 52, 1, 34, "Print"],
Cell[264524, 7235, 152, 2, 34, "Print"],
Cell[264679, 7239, 167, 3, 34, "Print"],
Cell[264849, 7244, 237, 4, 34, "Print"],
Cell[265089, 7250, 237, 4, 34, "Print"],
Cell[265329, 7256, 37, 1, 34, "Print"],
Cell[265369, 7259, 65, 1, 58, "Print"],
Cell[265437, 7262, 45, 1, 34, "Print"],
Cell[265485, 7265, 50, 1, 34, "Print"],
Cell[265538, 7268, 54, 1, 34, "Print"],
Cell[265595, 7271, 57, 1, 34, "Print"],
Cell[265655, 7274, 49, 1, 34, "Print"],
Cell[265707, 7277, 192, 4, 34, "Print"],
Cell[265902, 7283, 52, 1, 34, "Print"],
Cell[265957, 7286, 52, 1, 34, "Print"],
Cell[266012, 7289, 53, 1, 34, "Print"],
Cell[266068, 7292, 51, 1, 34, "Print"],
Cell[266122, 7295, 68, 1, 34, "Print"],
Cell[266193, 7298, 52, 1, 34, "Print"],
Cell[266248, 7301, 47, 1, 34, "Print"],
Cell[266298, 7304, 53, 1, 34, "Print"],
Cell[266354, 7307, 50, 1, 34, "Print"],
Cell[266407, 7310, 266, 4, 34, "Print"],
Cell[266676, 7316, 59, 1, 34, "Print"],
Cell[266738, 7319, 153, 3, 34, "Print"],
Cell[266894, 7324, 151, 2, 34, "Print"],
Cell[267048, 7328, 52, 1, 34, "Print"],
Cell[267103, 7331, 153, 2, 34, "Print"],
Cell[267259, 7335, 236, 6, 34, "Print"],
Cell[267498, 7343, 380, 9, 34, "Print"],
Cell[267881, 7354, 241, 4, 34, "Print"],
Cell[268125, 7360, 346, 8, 34, "Print"],
Cell[268474, 7370, 255, 4, 34, "Print"],
Cell[268732, 7376, 401, 9, 34, "Print"],
Cell[269136, 7387, 401, 9, 34, "Print"],
Cell[269540, 7398, 37, 1, 34, "Print"],
Cell[269580, 7401, 60, 1, 34, "Print"],
Cell[269643, 7404, 45, 1, 34, "Print"],
Cell[269691, 7407, 50, 1, 34, "Print"],
Cell[269744, 7410, 54, 1, 34, "Print"],
Cell[269801, 7413, 57, 1, 34, "Print"],
Cell[269861, 7416, 49, 1, 34, "Print"],
Cell[269913, 7419, 266, 4, 34, "Print"],
Cell[270182, 7425, 59, 1, 34, "Print"],
Cell[270244, 7428, 153, 3, 34, "Print"],
Cell[270400, 7433, 151, 2, 34, "Print"],
Cell[270554, 7437, 52, 1, 34, "Print"],
Cell[270609, 7440, 153, 2, 34, "Print"],
Cell[270765, 7444, 209, 5, 34, "Print"],
Cell[270977, 7451, 241, 4, 34, "Print"],
Cell[271221, 7457, 241, 4, 34, "Print"],
Cell[271465, 7463, 272, 5, 34, "Print"],
Cell[271740, 7470, 251, 4, 34, "Print"],
Cell[271994, 7476, 251, 4, 34, "Print"],
Cell[272248, 7482, 282, 5, 34, "Print"],
Cell[272533, 7489, 37, 1, 34, "Print"],
Cell[272573, 7492, 54, 1, 34, "Print"],
Cell[272630, 7495, 57, 1, 34, "Print"],
Cell[272690, 7498, 245, 5, 34, "Print"],
Cell[272938, 7505, 49, 1, 34, "Print"],
Cell[272990, 7508, 266, 4, 34, "Print"],
Cell[273259, 7514, 59, 1, 34, "Print"],
Cell[273321, 7517, 153, 3, 34, "Print"],
Cell[273477, 7522, 151, 2, 34, "Print"],
Cell[273631, 7526, 52, 1, 34, "Print"],
Cell[273686, 7529, 153, 2, 34, "Print"],
Cell[273842, 7533, 244, 6, 34, "Print"],
Cell[274089, 7541, 241, 4, 34, "Print"],
Cell[274333, 7547, 241, 4, 34, "Print"],
Cell[274577, 7553, 307, 6, 34, "Print"],
Cell[274887, 7561, 253, 4, 34, "Print"],
Cell[275143, 7567, 255, 4, 34, "Print"],
Cell[275401, 7573, 328, 7, 34, "Print"],
Cell[275732, 7582, 37, 1, 34, "Print"],
Cell[275772, 7585, 54, 1, 34, "Print"],
Cell[275829, 7588, 57, 1, 34, "Print"],
Cell[275889, 7591, 49, 1, 34, "Print"],
Cell[275941, 7594, 266, 4, 34, "Print"],
Cell[276210, 7600, 59, 1, 34, "Print"],
Cell[276272, 7603, 153, 3, 34, "Print"],
Cell[276428, 7608, 151, 2, 34, "Print"],
Cell[276582, 7612, 52, 1, 34, "Print"],
Cell[276637, 7615, 153, 2, 34, "Print"],
Cell[276793, 7619, 209, 5, 34, "Print"],
Cell[277005, 7626, 241, 4, 34, "Print"],
Cell[277249, 7632, 241, 4, 34, "Print"],
Cell[277493, 7638, 272, 5, 34, "Print"],
Cell[277768, 7645, 251, 4, 34, "Print"],
Cell[278022, 7651, 251, 4, 34, "Print"],
Cell[278276, 7657, 282, 5, 34, "Print"],
Cell[278561, 7664, 37, 1, 34, "Print"],
Cell[278601, 7667, 54, 1, 34, "Print"],
Cell[278658, 7670, 57, 1, 34, "Print"],
Cell[278718, 7673, 245, 5, 34, "Print"],
Cell[278966, 7680, 49, 1, 34, "Print"],
Cell[279018, 7683, 266, 4, 34, "Print"],
Cell[279287, 7689, 59, 1, 34, "Print"],
Cell[279349, 7692, 153, 3, 34, "Print"],
Cell[279505, 7697, 151, 2, 34, "Print"],
Cell[279659, 7701, 52, 1, 34, "Print"],
Cell[279714, 7704, 153, 2, 34, "Print"],
Cell[279870, 7708, 240, 6, 34, "Print"],
Cell[280113, 7716, 241, 4, 34, "Print"],
Cell[280357, 7722, 241, 4, 34, "Print"],
Cell[280601, 7728, 303, 6, 34, "Print"],
Cell[280907, 7736, 251, 4, 34, "Print"],
Cell[281161, 7742, 253, 4, 34, "Print"],
Cell[281417, 7748, 320, 7, 34, "Print"],
Cell[281740, 7757, 37, 1, 34, "Print"],
Cell[281780, 7760, 56, 1, 34, "Print"],
Cell[281839, 7763, 42, 1, 34, "Print"],
Cell[281884, 7766, 50, 1, 34, "Print"],
Cell[281937, 7769, 54, 1, 34, "Print"],
Cell[281994, 7772, 57, 1, 34, "Print"],
Cell[282054, 7775, 49, 1, 34, "Print"],
Cell[282106, 7778, 262, 4, 34, "Print"],
Cell[282371, 7784, 830, 11, 250, "Print"],
Cell[283204, 7797, 830, 11, 250, "Print"],
Cell[284037, 7810, 37, 1, 34, "Print"],
Cell[284077, 7813, 54, 1, 34, "Print"],
Cell[284134, 7816, 57, 1, 34, "Print"],
Cell[284194, 7819, 49, 1, 34, "Print"],
Cell[284246, 7822, 262, 4, 34, "Print"],
Cell[284511, 7828, 91, 1, 34, "Print"],
Cell[284605, 7831, 645, 9, 202, "Print"],
Cell[285253, 7842, 222, 5, 34, "Print"],
Cell[285478, 7849, 65, 1, 34, "Print"],
Cell[285546, 7852, 66, 1, 34, "Print"],
Cell[285615, 7855, 57, 1, 34, "Print"],
Cell[285675, 7858, 57, 1, 34, "Print"],
Cell[285735, 7861, 57, 1, 34, "Print"],
Cell[285795, 7864, 90, 1, 34, "Print"],
Cell[285888, 7867, 67, 1, 34, "Print"],
Cell[285958, 7870, 68, 1, 34, "Print"],
Cell[286029, 7873, 64, 1, 34, "Print"],
Cell[286096, 7876, 64, 1, 34, "Print"],
Cell[286163, 7879, 64, 1, 34, "Print"],
Cell[286230, 7882, 105, 2, 34, "Print"],
Cell[286338, 7886, 527, 9, 82, "Print"],
Cell[286868, 7897, 37, 1, 34, "Print"],
Cell[286908, 7900, 84, 1, 34, "Print"],
Cell[286995, 7903, 65, 1, 34, "Print"],
Cell[287063, 7906, 88, 1, 82, "Print"],
Cell[287154, 7909, 52, 1, 34, "Print"],
Cell[287209, 7912, 45, 1, 34, "Print"],
Cell[287257, 7915, 50, 1, 34, "Print"],
Cell[287310, 7918, 54, 1, 34, "Print"],
Cell[287367, 7921, 57, 1, 34, "Print"],
Cell[287427, 7924, 49, 1, 34, "Print"],
Cell[287479, 7927, 266, 4, 34, "Print"],
Cell[287748, 7933, 59, 1, 34, "Print"],
Cell[287810, 7936, 153, 3, 34, "Print"],
Cell[287966, 7941, 151, 2, 34, "Print"],
Cell[288120, 7945, 52, 1, 34, "Print"],
Cell[288175, 7948, 153, 2, 34, "Print"],
Cell[288331, 7952, 234, 6, 34, "Print"],
Cell[288568, 7960, 241, 4, 34, "Print"],
Cell[288812, 7966, 241, 4, 34, "Print"],
Cell[289056, 7972, 297, 6, 34, "Print"],
Cell[289356, 7980, 255, 4, 34, "Print"],
Cell[289614, 7986, 255, 4, 34, "Print"],
Cell[289872, 7992, 286, 5, 34, "Print"],
Cell[290161, 7999, 37, 1, 34, "Print"],
Cell[290201, 8002, 56, 1, 34, "Print"],
Cell[290260, 8005, 37, 1, 34, "Print"],
Cell[290300, 8008, 37, 1, 34, "Print"],
Cell[290340, 8011, 37, 1, 34, "Print"],
Cell[290380, 8014, 91, 1, 34, "Print"],
Cell[290474, 8017, 54, 1, 34, "Print"],
Cell[290531, 8020, 57, 1, 34, "Print"],
Cell[290591, 8023, 49, 1, 34, "Print"],
Cell[290643, 8026, 266, 4, 34, "Print"],
Cell[290912, 8032, 59, 1, 34, "Print"],
Cell[290974, 8035, 153, 3, 34, "Print"],
Cell[291130, 8040, 151, 2, 34, "Print"],
Cell[291284, 8044, 52, 1, 34, "Print"],
Cell[291339, 8047, 153, 2, 34, "Print"],
Cell[291495, 8051, 205, 5, 34, "Print"],
Cell[291703, 8058, 241, 4, 34, "Print"],
Cell[291947, 8064, 241, 4, 34, "Print"],
Cell[292191, 8070, 287, 5, 34, "Print"],
Cell[292481, 8077, 321, 6, 34, "Print"],
Cell[292805, 8085, 321, 6, 34, "Print"],
Cell[293129, 8093, 321, 6, 34, "Print"],
Cell[293453, 8101, 37, 1, 34, "Print"],
Cell[293493, 8104, 37, 1, 34, "Print"],
Cell[293533, 8107, 37, 1, 34, "Print"],
Cell[293573, 8110, 37, 1, 34, "Print"],
Cell[293613, 8113, 37, 1, 34, "Print"],
Cell[293653, 8116, 37, 1, 34, "Print"],
Cell[293693, 8119, 79, 1, 34, "Print"],
Cell[293775, 8122, 41, 1, 34, "Print"],
Cell[293819, 8125, 50, 1, 34, "Print"],
Cell[293872, 8128, 106, 2, 34, "Print"],
Cell[293981, 8132, 37, 1, 34, "Print"],
Cell[294021, 8135, 74, 1, 34, "Print"],
Cell[294098, 8138, 76, 1, 34, "Print"],
Cell[294177, 8141, 58, 1, 34, "Print"],
Cell[294238, 8144, 45, 1, 34, "Print"],
Cell[294286, 8147, 50, 1, 34, "Print"],
Cell[294339, 8150, 37, 1, 34, "Print"],
Cell[294379, 8153, 80, 1, 34, "Print"],
Cell[294462, 8156, 50, 1, 34, "Print"],
Cell[294515, 8159, 269, 5, 34, "Print"],
Cell[294787, 8166, 215, 4, 34, "Print"],
Cell[295005, 8172, 37, 1, 34, "Print"],
Cell[295045, 8175, 310, 6, 34, "Print"],
Cell[295358, 8183, 183, 3, 34, "Print"],
Cell[295544, 8188, 310, 6, 34, "Print"],
Cell[295857, 8196, 183, 3, 34, "Print"],
Cell[296043, 8201, 310, 6, 34, "Print"],
Cell[296356, 8209, 183, 3, 34, "Print"],
Cell[296542, 8214, 310, 6, 34, "Print"],
Cell[296855, 8222, 183, 3, 34, "Print"],
Cell[297041, 8227, 310, 6, 34, "Print"],
Cell[297354, 8235, 310, 6, 34, "Print"],
Cell[297667, 8243, 183, 3, 34, "Print"],
Cell[297853, 8248, 183, 3, 34, "Print"],
Cell[298039, 8253, 310, 6, 34, "Print"],
Cell[298352, 8261, 310, 6, 34, "Print"],
Cell[298665, 8269, 183, 3, 34, "Print"],
Cell[298851, 8274, 183, 3, 34, "Print"],
Cell[299037, 8279, 310, 6, 34, "Print"],
Cell[299350, 8287, 310, 6, 34, "Print"],
Cell[299663, 8295, 183, 3, 34, "Print"],
Cell[299849, 8300, 183, 3, 34, "Print"],
Cell[300035, 8305, 310, 6, 34, "Print"],
Cell[300348, 8313, 183, 3, 34, "Print"],
Cell[300534, 8318, 37, 1, 34, "Print"],
Cell[300574, 8321, 84, 1, 34, "Print"],
Cell[300661, 8324, 76, 1, 34, "Print"],
Cell[300740, 8327, 86, 1, 58, "Print"],
Cell[300829, 8330, 67, 1, 34, "Print"],
Cell[300899, 8333, 52, 1, 34, "Print"],
Cell[300954, 8336, 50, 1, 34, "Print"],
Cell[301007, 8339, 54, 1, 34, "Print"],
Cell[301064, 8342, 57, 1, 34, "Print"],
Cell[301124, 8345, 49, 1, 34, "Print"],
Cell[301176, 8348, 262, 4, 34, "Print"],
Cell[301441, 8354, 59, 1, 34, "Print"],
Cell[301503, 8357, 151, 2, 34, "Print"],
Cell[301657, 8361, 52, 1, 34, "Print"],
Cell[301712, 8364, 152, 2, 34, "Print"],
Cell[301867, 8368, 209, 5, 34, "Print"],
Cell[302079, 8375, 251, 4, 34, "Print"],
Cell[302333, 8381, 241, 4, 34, "Print"],
Cell[302577, 8387, 272, 5, 34, "Print"],
Cell[302852, 8394, 255, 4, 34, "Print"],
Cell[303110, 8400, 37, 1, 34, "Print"],
Cell[303150, 8403, 54, 1, 34, "Print"],
Cell[303207, 8406, 57, 1, 34, "Print"],
Cell[303267, 8409, 49, 1, 34, "Print"],
Cell[303319, 8412, 189, 4, 34, "Print"],
Cell[303511, 8418, 50, 1, 34, "Print"],
Cell[303564, 8421, 50, 1, 34, "Print"],
Cell[303617, 8424, 59, 1, 34, "Print"],
Cell[303679, 8427, 52, 1, 34, "Print"],
Cell[303734, 8430, 63, 1, 34, "Print"],
Cell[303800, 8433, 63, 1, 34, "Print"],
Cell[303866, 8436, 63, 1, 34, "Print"],
Cell[303932, 8439, 64, 1, 34, "Print"],
Cell[303999, 8442, 53, 1, 34, "Print"],
Cell[304055, 8445, 64, 1, 34, "Print"],
Cell[304122, 8448, 68, 1, 34, "Print"],
Cell[304193, 8451, 64, 1, 34, "Print"],
Cell[304260, 8454, 63, 1, 34, "Print"],
Cell[304326, 8457, 64, 1, 34, "Print"],
Cell[304393, 8460, 266, 4, 34, "Print"],
Cell[304662, 8466, 59, 1, 34, "Print"],
Cell[304724, 8469, 153, 3, 34, "Print"],
Cell[304880, 8474, 151, 2, 34, "Print"],
Cell[305034, 8478, 52, 1, 34, "Print"],
Cell[305089, 8481, 153, 2, 34, "Print"],
Cell[305245, 8485, 209, 5, 34, "Print"],
Cell[305457, 8492, 261, 4, 34, "Print"],
Cell[305721, 8498, 261, 4, 34, "Print"],
Cell[305985, 8504, 292, 5, 34, "Print"],
Cell[306280, 8511, 277, 4, 34, "Print"],
Cell[306560, 8517, 271, 4, 34, "Print"],
Cell[306834, 8523, 315, 6, 34, "Print"],
Cell[307152, 8531, 37, 1, 34, "Print"],
Cell[307192, 8534, 54, 1, 34, "Print"],
Cell[307249, 8537, 244, 5, 34, "Print"],
Cell[307496, 8544, 179, 3, 34, "Print"],
Cell[307678, 8549, 159, 3, 34, "Print"],
Cell[307840, 8554, 173, 3, 34, "Print"],
Cell[308016, 8559, 57, 1, 34, "Print"],
Cell[308076, 8562, 49, 1, 34, "Print"],
Cell[308128, 8565, 189, 4, 34, "Print"],
Cell[308320, 8571, 50, 1, 34, "Print"],
Cell[308373, 8574, 50, 1, 34, "Print"],
Cell[308426, 8577, 59, 1, 34, "Print"],
Cell[308488, 8580, 52, 1, 34, "Print"],
Cell[308543, 8583, 63, 1, 34, "Print"],
Cell[308609, 8586, 63, 1, 34, "Print"],
Cell[308675, 8589, 63, 1, 34, "Print"],
Cell[308741, 8592, 63, 1, 34, "Print"],
Cell[308807, 8595, 60, 1, 34, "Print"],
Cell[308870, 8598, 53, 1, 34, "Print"],
Cell[308926, 8601, 64, 1, 34, "Print"],
Cell[308993, 8604, 64, 1, 34, "Print"],
Cell[309060, 8607, 64, 1, 34, "Print"],
Cell[309127, 8610, 64, 1, 34, "Print"],
Cell[309194, 8613, 266, 4, 34, "Print"],
Cell[309463, 8619, 59, 1, 34, "Print"],
Cell[309525, 8622, 153, 3, 34, "Print"],
Cell[309681, 8627, 151, 2, 34, "Print"],
Cell[309835, 8631, 52, 1, 34, "Print"],
Cell[309890, 8634, 153, 2, 34, "Print"],
Cell[310046, 8638, 261, 4, 34, "Print"],
Cell[310310, 8644, 261, 4, 34, "Print"],
Cell[310574, 8650, 261, 4, 34, "Print"],
Cell[310838, 8656, 277, 4, 34, "Print"],
Cell[311118, 8662, 271, 4, 34, "Print"],
Cell[311392, 8668, 273, 4, 34, "Print"],
Cell[311668, 8674, 179, 3, 34, "Print"],
Cell[311850, 8679, 159, 3, 34, "Print"],
Cell[312012, 8684, 159, 3, 34, "Print"],
Cell[312174, 8689, 57, 1, 34, "Print"],
Cell[312234, 8692, 49, 1, 34, "Print"],
Cell[312286, 8695, 189, 4, 34, "Print"],
Cell[312478, 8701, 50, 1, 34, "Print"],
Cell[312531, 8704, 50, 1, 34, "Print"],
Cell[312584, 8707, 59, 1, 34, "Print"],
Cell[312646, 8710, 52, 1, 34, "Print"],
Cell[312701, 8713, 63, 1, 34, "Print"],
Cell[312767, 8716, 63, 1, 34, "Print"],
Cell[312833, 8719, 63, 1, 34, "Print"],
Cell[312899, 8722, 64, 1, 34, "Print"],
Cell[312966, 8725, 53, 1, 34, "Print"],
Cell[313022, 8728, 64, 1, 34, "Print"],
Cell[313089, 8731, 68, 1, 34, "Print"],
Cell[313160, 8734, 64, 1, 34, "Print"],
Cell[313227, 8737, 266, 4, 34, "Print"],
Cell[313496, 8743, 59, 1, 34, "Print"],
Cell[313558, 8746, 153, 3, 34, "Print"],
Cell[313714, 8751, 151, 2, 34, "Print"],
Cell[313868, 8755, 52, 1, 34, "Print"],
Cell[313923, 8758, 152, 2, 34, "Print"],
Cell[314078, 8762, 261, 4, 34, "Print"],
Cell[314342, 8768, 261, 4, 34, "Print"],
Cell[314606, 8774, 261, 4, 34, "Print"],
Cell[314870, 8780, 277, 4, 34, "Print"],
Cell[315150, 8786, 271, 4, 34, "Print"],
Cell[315424, 8792, 315, 6, 34, "Print"],
Cell[315742, 8800, 206, 4, 34, "Print"],
Cell[315951, 8806, 261, 4, 34, "Print"],
Cell[316215, 8812, 165, 3, 34, "Print"],
Cell[316383, 8817, 261, 4, 34, "Print"],
Cell[316647, 8823, 165, 3, 34, "Print"],
Cell[316815, 8828, 261, 4, 34, "Print"],
Cell[317079, 8834, 165, 3, 34, "Print"],
Cell[317247, 8839, 277, 4, 34, "Print"],
Cell[317527, 8845, 177, 3, 34, "Print"],
Cell[317707, 8850, 271, 4, 34, "Print"],
Cell[317981, 8856, 175, 3, 34, "Print"],
Cell[318159, 8861, 273, 4, 34, "Print"],
Cell[318435, 8867, 219, 5, 34, "Print"],
Cell[318657, 8874, 90, 1, 34, "Print"],
Cell[318750, 8877, 70, 1, 34, "Print"],
Cell[318823, 8880, 90, 1, 34, "Print"],
Cell[318916, 8883, 295, 7, 34, "Print"]
}, Closed]]
}, Open  ]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

