(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[      5669,        239]*)
(*NotebookOutlinePosition[      7004,        283]*)
(*  CellTagsIndexPosition[      6960,        279]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["\<\
The ToAlpha0v2 package
and the Alphard package\
\>", "Title",
  FontColor->RGBColor[1, 0, 0]],

Cell["Patrice Quinton", "Subsubtitle"],

Cell["July 2002", "Text"],

Cell["\<\
Irisa, Campus de Beaulieu, 35042 Rennes Cedex, France
alpha@irisa.fr\
\>", "SmallText"],

Cell[CellGroupData[{

Cell[" Execute this cell, to redefine the current directory", "Subsubsection",
  FontColor->RGBColor[1, 0, 0]],

Cell[BoxData[
    \(setMMADir[{"\<doc\>", "\<packages\>", "\<ToAlpha0v2\>"}]\)], "Input",
  InitializationCell->True]
}, Open  ]],

Cell[CellGroupData[{

Cell["Introduction", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell[TextData[{
  "The ",
  StyleBox["Semantics",
    FontColor->RGBColor[1, 0, 0]],
  " package contains functions to analyze the type of Alpha expressions. "
}], "Text",
  TextAlignment->Left,
  TextJustification->1],

Cell[BoxData[
    \(SymbolsInContext["\<Alpha`ToAlpha0v2`\>"]\)], "Input"],

Cell[BoxData[
    \(SymbolsInContext["\<Alpha`Alphard`\>"]\)], "Input"],

Cell[BoxData[
    \(\(?buildInterface\)\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["toAlpha0v2", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell[BoxData[
    \(\(load["\<delayline.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(analyze[recurse \[Rule] True]\)], "Input"],

Cell[BoxData[
    \(\(inlineAll[];\)\)], "Input"],

Cell[BoxData[
    \(\(removeIdEqus[norm \[Rule] True];\)\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \(\(simplifySystem[];\)\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \(\(asave["\<delaylineClean.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(<< Alpha/ToAlpha0v2.m\)], "Input"],

Cell[BoxData[
    \(\(load["\<delaylineClean.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(analyze[]\)], "Input"],

Cell[BoxData[
    \(\(toAlpha0v2[verbose \[Rule] True];\)\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \(\(alpha0ToAlphard[];\)\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \(systemNames[]\)], "Input"],

Cell[BoxData[
    \(getSystem["\<ControldelaylineModule\>"]; ashow[]\)], "Input"],

Cell[BoxData[
    \(getSystem["\<delaylineModule\>"]; ashow[]\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["buildInterface", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell[BoxData[
    \(\(?buildInterface\)\)], "Input"],

Cell[BoxData[
    \(\(load["\<delayline0.alpha\>"];\)\)], "Input"],

Cell["\<\
This system has one input, in, one output, out. The input has one input \
mirror equation, in_mirr1, and an output equation.\
\>", "Text"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell["\<\
The interface contains the input (in) and output (out) of the initial system, \
and a call to a module which takes the in_mirr1 input, and the outXregister1 \
output. \
\>", "Text"],

Cell[BoxData[
    \(buildInterface[]; ashow[]\)], "Input"],

Cell["The module contains all equations of the initial system. ", "Text"],

Cell[BoxData[
    \(getSystem["\<delaylineModule\>"]; ashow[]\)], "Input"],

Cell[BoxData[
    \(\(save["\<delaylineModule.alpha\>"];\)\)], "Input"],

Cell["Another example.", "Text"],

Cell[BoxData[
    \(\(load["\<QR0.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \(buildInterface[]; ashow[]\)], "Input"],

Cell[BoxData[
    \(systemNames[]\)], "Input"],

Cell[BoxData[
    \(\(getSystem["\<QRModule\>"];\)\)], "Input"],

Cell[BoxData[
    \(\(getArrayDomains[];\)\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["getArrayDomains", "Section",
  FontColor->RGBColor[0, 0, 1]],

Cell[BoxData[
    \(FileNames[]\)], "Input"],

Cell[BoxData[
    \(\(load["\<QR0.alpha\>"];\)\)], "Input"],

Cell[BoxData[
    \(ashow[]\)], "Input"],

Cell[BoxData[
    \(getArrayDomains[debug \[Rule] False, 
      mergeDomains \[Rule] True]\)], "Input"]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Macintosh",
ScreenRectangle->{{0, 1440}, {0, 834}},
AutoGeneratedPackage->None,
WindowToolbars->{"RulerBar", "EditBar"},
WindowSize->{1016, 664},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
PrintingCopies->1,
PrintingPageRange->{1, 9999},
PageHeaders->{{Cell[ 
        TextData[ {
          CounterBox[ "Page"]}], "PageNumber"], Inherited, Cell[ 
        TextData[ {
          ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
        TextData[ {
          ValueBox[ "FileName"]}], "Header"], Inherited, Cell[ 
        TextData[ {
          CounterBox[ "Page"]}], "PageNumber"]}},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"FirstPageHeader"->False,
"FirstPageFooter"->False,
"FacingPages"->True,
"Magnification"->1},
ShowSelection->True,
Magnification->1.5,
StyleDefinitions -> "Default.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 103, 4, 204, "Title"],
Cell[1882, 59, 38, 0, 45, "Subsubtitle"],
Cell[1923, 61, 25, 0, 47, "Text"],
Cell[1951, 63, 97, 3, 62, "SmallText"],

Cell[CellGroupData[{
Cell[2073, 70, 110, 1, 41, "Subsubsection"],
Cell[2186, 73, 117, 2, 39, "Input",
  InitializationCell->True]
}, Open  ]],

Cell[CellGroupData[{
Cell[2340, 80, 63, 1, 104, "Section"],
Cell[2406, 83, 218, 7, 47, "Text"],
Cell[2627, 92, 74, 1, 39, "Input"],
Cell[2704, 95, 71, 1, 39, "Input"],
Cell[2778, 98, 52, 1, 39, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[2867, 104, 61, 1, 104, "Section"],
Cell[2931, 107, 65, 1, 39, "Input"],
Cell[2999, 110, 62, 1, 39, "Input"],
Cell[3064, 113, 49, 1, 39, "Input"],
Cell[3116, 116, 69, 1, 39, "Input"],
Cell[3188, 119, 40, 1, 39, "Input"],
Cell[3231, 122, 54, 1, 39, "Input"],
Cell[3288, 125, 40, 1, 39, "Input"],
Cell[3331, 128, 71, 1, 39, "Input"],
Cell[3405, 131, 54, 1, 39, "Input"],
Cell[3462, 134, 70, 1, 39, "Input"],
Cell[3535, 137, 42, 1, 39, "Input"],
Cell[3580, 140, 70, 1, 39, "Input"],
Cell[3653, 143, 40, 1, 39, "Input"],
Cell[3696, 146, 55, 1, 39, "Input"],
Cell[3754, 149, 40, 1, 39, "Input"],
Cell[3797, 152, 46, 1, 39, "Input"],
Cell[3846, 155, 81, 1, 39, "Input"],
Cell[3930, 158, 74, 1, 39, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4041, 164, 65, 1, 104, "Section"],
Cell[4109, 167, 52, 1, 39, "Input"],
Cell[4164, 170, 66, 1, 39, "Input"],
Cell[4233, 173, 148, 3, 47, "Text"],
Cell[4384, 178, 40, 1, 39, "Input"],
Cell[4427, 181, 191, 4, 73, "Text"],
Cell[4621, 187, 58, 1, 39, "Input"],
Cell[4682, 190, 73, 0, 47, "Text"],
Cell[4758, 192, 74, 1, 39, "Input"],
Cell[4835, 195, 71, 1, 39, "Input"],
Cell[4909, 198, 32, 0, 47, "Text"],
Cell[4944, 200, 59, 1, 39, "Input"],
Cell[5006, 203, 40, 1, 39, "Input"],
Cell[5049, 206, 58, 1, 39, "Input"],
Cell[5110, 209, 46, 1, 39, "Input"],
Cell[5159, 212, 63, 1, 39, "Input"],
Cell[5225, 215, 55, 1, 39, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5317, 221, 66, 1, 104, "Section"],
Cell[5386, 224, 44, 1, 39, "Input"],
Cell[5433, 227, 59, 1, 39, "Input"],
Cell[5495, 230, 40, 1, 39, "Input"],
Cell[5538, 233, 103, 2, 32, "Input"]
}, Open  ]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

